package com.trifork.sosigw.inttest;

import com.trifork.sosigw.util.Util;
import com.trifork.sosigw.ws.IdCardTestUtils;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.pki.InMemoryIntermediateCertificateCache;
import dk.sosi.seal.pki.SOSIFederation;
import dk.sosi.seal.vault.FileBasedCredentialVault;
import dk.sosi.seal.xml.Base64;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.StringRequestEntity;
import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.io.*;
import java.net.HttpURLConnection;
import java.net.URL;
import java.security.*;
import java.security.cert.CertificateException;
import java.util.Properties;
import java.util.UUID;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;


/**
 * This test class is dependent on a niab running with sosigw.mode=test, also sosigw console must have been configured according to the "Golden image" documentation.
 */
public class ProxyIdCardTest {
    private static final Logger logger = Logger.getLogger(ProxyIdCardTest.class);
    private static final String HTTP_STS_SECURITY_TOKEN_SERVICE = "http://test1-cnsp.ekstern-test.nspop.dk:8080/sts/services/SecurityTokenService";
    private String signonUrl = "http://localhost:8080/sosigw/service/sosigw";
    private String proxyUrl = "http://localhost:8080/sosigw/proxy/soap-request";
    private int statusCode;

    private static final String MOCES_VAULT = "validMocesVault.jks";

    @Before
    public void init() throws Exception {
        statusCode = 200;
    }

    @Test
    public void getMedicinCardWithSignedRequestThroughSosiProxy() throws Exception {
        String fullIdCardSoapResponse = createLevel4IdCard();
        String idCardHeader = fullIdCardSoapResponse.substring(0, fullIdCardSoapResponse.length() - 51); // remove "</soapenv:Header><soapenv:Body/></soapenv:Envelope>"

        // add SOAP action
        idCardHeader = idCardHeader.replace("<soapenv:Header>", "<soapenv:Header><Action xmlns=\"http://schemas.xmlsoap.org/ws/2004/08/addressing\">http://www.dkma.dk/medicinecard/xml.schema/2015/01/01/E1#GetMedicineCard</Action>");

        final String signedIdCardRequest = idCardHeader + loadFileContent("getMedicineCard_request.xml");
        System.out.println("\n----request:\n" + signedIdCardRequest + "\n\n");

        final String requestIdCardResp = callServer(signedIdCardRequest, "", proxyUrl);
        if (statusCode != 200) {
            Matcher matcher = Pattern.compile("<(.*:)?BrowserUrl>([^<]*)</(.*:)?BrowserUrl>").matcher(requestIdCardResp);
            if (matcher.find()) {
                String browserUrl = matcher.group(2).trim().replaceAll("&amp;", "&");
                System.out.println("BrowserUrl: " + browserUrl);
            }
            fail("Fault response from server");
        }

        System.out.println("Server response to proxy request: \n" + requestIdCardResp);
    }

    @Test
    public void getMedicinCardWithUnSignedRequestThroughSosiProxy() throws Exception {
        logger.info("Creating level1 idcard");
        String idCardAssertion = createLevel1IdCard();

        String idCardSignatureValue = getSignatureValue(idCardAssertion);
        signIdCardAndInsertCardInGatewayCache(idCardAssertion, idCardSignatureValue);

        logger.info("Preparing gatewayProxyMedecineCardRequest from callGatewayProxyMedecineCardRequest-template.xml");
        final String callGatewayProxyMedecineCardRequestTemplate = loadFileContent("callGatewayProxyMedecineCardRequest-template.xml");
        String callGatewayProxyMedecineCardRequest = callGatewayProxyMedecineCardRequestTemplate.replace("#IDCARD_ASSERTION#", idCardAssertion).replace("#MESSAGE_ID#", UUID.randomUUID().toString());

        logger.info("Calling server " + proxyUrl + " with XML\n" + callGatewayProxyMedecineCardRequest);
        String callGatewayProxyMedecineCardResponse = callServer(callGatewayProxyMedecineCardRequest, "http://www.dkma.dk/medicinecard/xml.schema/2015/01/01/E1#GetMedicineCard", proxyUrl);
        logger.debug("callGatewayProxyMedecineCardResponse: \n" + callGatewayProxyMedecineCardResponse + "\n");

        System.out.println("Server response to proxy request: \n" + callGatewayProxyMedecineCardResponse);
    }

    private String loadFileContent(String filename) throws IOException {
        return Util.readAllFromStreamAsString(new FileInputStream(filename)); // this.getClass().getResourceAsStream("getMedicineCard_request.xml"));
    }


    private void signIdCardAndInsertCardInGatewayCache(String idCardAssertion, String idCardSignatureValue) throws IOException {
        logger.debug("Preparing signIdCardRequest from callGatewaySignIdCard-request-template.xml");
        String signIdCardRequestTemplate = loadFileContent("callGatewaySignIdCard-request-template.xml"); // this.getClass().getResourceAsStream("callGatewaySignIdCard-request-template.xml"));
        String signIdCardRequest = signIdCardRequestTemplate.replace("#SIGNATURE_VALUE#", idCardSignatureValue).replace("#IDCARD_ASSERTION#", idCardAssertion);
        logger.debug("Calling server " + signonUrl + " with XML\n" + signIdCardRequest);
        String signIdCardResponse = callServer(signIdCardRequest, "", signonUrl); //this will put the idcard in the gateway cache
        logger.debug("signIdCardResponse: \n" + signIdCardResponse + "\n");
    }

    private String getSignatureValue(String idCardAssertion) throws Exception {
        logger.debug("Preparing requestSignatureXml from callGatewayRequestSignature-template.xml");
        String requestSignatureXmlTemplate = loadFileContent("callGatewayRequestSignature-template.xml");
        String requestSignatureXml = requestSignatureXmlTemplate.replaceAll("#IDCARD_ASSERTION#", idCardAssertion);
        logger.debug("Calling server " + signonUrl + " with XML\n" + requestSignatureXml);
        String requestIdCardResp = callServer(requestSignatureXml, "", signonUrl);
        logger.debug("Parsing DigestValue from response");
        Matcher matcher = Pattern.compile("<(.*:)?DigestValue>([^<]*)</(.*:)?DigestValue>").matcher(requestIdCardResp);
        assertTrue(requestIdCardResp + " contains DigestValue", matcher.find());
        final String digestValue = matcher.group(2).trim();
        final String signatureValue = sign(digestValue);
        if (statusCode != 200) {
            Matcher browserUrlMatcher = Pattern.compile("<(.*:)?BrowserUrl>([^<]*)</(.*:)?BrowserUrl>").matcher(requestIdCardResp);
            if (browserUrlMatcher.find()) {
                String browserUrl = browserUrlMatcher.group(2).trim().replaceAll("&amp;", "&");
                logger.error("BrowserUrl: " + browserUrl);
            }
            fail("Fault response from server");
        }
        return signatureValue;
    }

    private String sign(String digestValue) throws NoSuchAlgorithmException, KeyStoreException, IOException, CertificateException, UnrecoverableKeyException, InvalidKeyException, SignatureException {
        String password = "Test1234";
        KeyStore keystore = KeyStore.getInstance("JKS");
        keystore.load(new FileInputStream(MOCES_VAULT), password.toCharArray());
        // signIdCardAndInsertCardInGatewayCache-request-template must contain the certificate from the keystore loaded above

        PrivateKey key = (PrivateKey) keystore.getKey("SOSI:ALIAS_SYSTEM", password.toCharArray());

        byte[] signatureValue = IdCardTestUtils.signDigest(Base64.decode(digestValue), key);

        return Base64.encode(signatureValue);
    }

    private String callServer(String requestIdCardRequest, String soapAction, String serverUrl) throws IOException {
        HttpClient httpClient = makeHttpClient();
        PostMethod postMethod = makePostMethod(serverUrl);
        postMethod.addRequestHeader("SOAPAction", "\"" + soapAction + "\"");
        postMethod.setRequestEntity(new StringRequestEntity(requestIdCardRequest, "text/xml", "utf-8"));
        httpClient.executeMethod(postMethod);
        String response = readResponse(postMethod);

        if (postMethod.getStatusCode() != 200) {
            statusCode = postMethod.getStatusCode();
            fail("Status code: " + postMethod.getStatusCode() + " - status text: " + postMethod.getStatusText() + " - XML response:\n" + response);
        }

        return response;
    }

    private HttpClient makeHttpClient() {
        final HttpClient httpClient = new HttpClient();
        return httpClient;
    }

    private PostMethod makePostMethod(String serverUrl) {
        PostMethod postMethod = new PostMethod(serverUrl);
        return postMethod;
    }

    private String readResponse(PostMethod postMethod) throws IOException {
        final InputStream responseStream = postMethod.getResponseBodyAsStream();

        return Util.readAllFromStreamAsString(responseStream);
    }

    private String createLevel1IdCard() throws IOException {
        String cpr = "0501792275";
        CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "20921897", "test");

        Properties props = SignatureUtil.setupCryptoProviderForJVM();
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE_ENHANCED, "false");
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, "false");

        SOSIFederation federation = new SOSIFederation(props, new InMemoryIntermediateCertificateCache());

        FileBasedCredentialVault vault = new FileBasedCredentialVault(props, new File(MOCES_VAULT), "Test1234");
        SOSIFactory factory = new SOSIFactory(federation, vault, props);

        SecurityTokenRequest stReq = factory.createNewSecurityTokenRequest();


        UserIDCard card = factory.createNewUserIDCard("FMK-testclient", cpr, "Lars", "Larsen", "ll@trifork.com",
                "developer", "Doctor", careProvider, "J0184", AuthenticationLevel.NO_AUTHENTICATION,
                vault.getSystemCredentialPair().getCertificate(), "Test");
        stReq.setIDCard(card);
        Element idCardNode = stReq.getIDCard().serialize2DOMDocument(factory, stReq.serialize2DOMDocument());
        String xml = XmlUtil.node2String(idCardNode, false, false);

        return xml;
    }

    private String createLevel4IdCard() throws NoSuchAlgorithmException, NoSuchProviderException, InvalidKeyException, SignatureException, IOException {
        String cpr = "0501792275";
        Properties props = SignatureUtil.setupCryptoProviderForJVM();
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE_ENHANCED, "false");
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, "false");

        SOSIFederation federation = new SOSIFederation(props, new InMemoryIntermediateCertificateCache());
        FileBasedCredentialVault vault = new FileBasedCredentialVault(props, new File(MOCES_VAULT), "Test1234");
        SOSIFactory factory = new SOSIFactory(federation, vault, props);

        CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "20921897", "test");
        SecurityTokenRequest stReq = factory.createNewSecurityTokenRequest();

        UserIDCard card = factory.createNewUserIDCard("SOSITEST", cpr, "Lars", "Larsen", "ll@trifork.com",
                "developer", "Doctor", careProvider, "J0184", AuthenticationLevel.MOCES_TRUSTED_USER, vault
                        .getSystemCredentialPair().getCertificate(), "Test");


        stReq.setIDCard(card);
        byte[] bytesForSigning = stReq.getIDCard().getBytesForSigning(stReq.serialize2DOMDocument());

        Signature jceSign = Signature.getInstance("SHA1withRSA", SignatureUtil.getCryptoProvider(props,
                SOSIFactory.PROPERTYNAME_SOSI_CRYPTOPROVIDER_SHA1WITHRSA));
        PrivateKey key = vault.getSystemCredentialPair().getPrivateKey();
        jceSign.initSign(key);
        jceSign.update(bytesForSigning);
        String signature = XmlUtil.toBase64(jceSign.sign());

        stReq.getIDCard().injectSignature(signature, vault.getSystemCredentialPair().getCertificate());

        String xml = XmlUtil.node2String(stReq.serialize2DOMDocument(), false, true);
        String stsUrl = HTTP_STS_SECURITY_TOKEN_SERVICE;
        String stResXml = sendRequest(stsUrl, "", xml, false);

        SecurityTokenRequest stRes = factory.deserializeSecurityTokenRequest(stResXml);
        IDCard idCard = stRes.getIDCard();

        Request sreq = factory.createNewRequest(false, "flowId");
        sreq.setIDCard(idCard);
        Document reqdoc = sreq.serialize2DOMDocument();

        System.out.println(XmlUtil.node2String(reqdoc, true, true));

        return XmlUtil.node2String(reqdoc, false, true);
    }

    public static String sendRequest(String url, String action, String docXml, boolean failOnError) throws IOException {
        URL u = new URL(url);
        HttpURLConnection uc = (HttpURLConnection) u.openConnection();
        uc.setDoOutput(true);
        uc.setDoInput(true);
        uc.setRequestMethod("POST");
        uc.setRequestProperty("SOAPAction", "\"" + action + "\"");
        uc.setRequestProperty("Content-Type", "text/xml; encoding=utf-8");
        OutputStream os = uc.getOutputStream();

        IOUtils.write(docXml, os);
        os.flush();

        InputStream is;
        if (uc.getResponseCode() != 200) {
            is = uc.getErrorStream();
        } else {
            is = uc.getInputStream();
        }
        String res = IOUtils.toString(is);

        is.close();
        if (uc.getResponseCode() != 200 && (uc.getResponseCode() != 500 || failOnError)) {
            throw new IOException(res);
        }
        os.close();
        uc.disconnect();

        return res;
    }
}
