/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw;

import com.trifork.sosigw.api.IDCardCacheEntry;
import com.trifork.sosigw.api.Intercom;
import com.trifork.sosigw.auditlog.*;
import com.trifork.sosigw.cluster.intercom.IntercomImpl;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.StaticConfigApi;
import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.proxy.ProxyRequestHandler;
import com.trifork.sosigw.util.Util;
import com.trifork.sosigw.util.XMLWatchdog;
import com.trifork.sosigw.util.timing.Log4jTimeSink;
import com.trifork.sosigw.util.timing.Timer;
import dk.sdsd.nsp.slalog.api.SLALogConfig;
import dk.sdsd.nsp.slalog.impl.BackendInstantiator;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.apache.ws.security.util.UUIDGenerator;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.Properties;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class Bootstrap {
    private static final String SLA_LOG_APP_NAME = "NSP GW";
    private static final String SLA_LOG_APP_SHORT_NAME = "sosigw";

    private static final String ORG_APACHE_XERCES_IMPL_VERSION = "org.apache.xerces.impl.Version";

    private static final Logger log = Logger.getLogger(Bootstrap.class);
    
    private static XMLWatchdog xdog;

    private static final Intercom mockIntercom = new Intercom() {
        public void configUpdatedLocally() {}
        public void entryUpdatedLocally(IDCardCacheEntry peerEntry) {}
        public String getUID() {
            return UUIDGenerator.getUUID();
        }
        public void kill() {}
        public void requestConfig(long timestampOfKnownConfig) {}
        public boolean isBootstrapped() {
            return true;
        }
        public boolean isSingleMember() {
            return true;
        }
    };

    public static void start() {
        Bootstrap.initLogging();
        log.info("Bootstrapping SOSI-GW");
        verifyDependencies();
        StaticConfigImpl.initSettings();
        RuntimeConfig.getRuntimeConfigState();

        initSlaLogging();

        // Make JAX-WS less noisy. We do use faults as regular replies, so don't make a fuss about them.
        System.setProperty("com.sun.xml.ws.fault.SOAPFaultBuilder.disableCaptureStackTrace", "false");
        java.util.logging.Logger.getLogger("com.sun.xml.ws.server.sei.EndpointMethodHandler").setLevel(java.util.logging.Level.WARNING);

        StaticConfigApi configApi = StaticConfigImpl.getInstance();
        if (configApi.isBootstrapRestrictedMode()) {
            log.warn("SOSI-GW running in BootstrapRestrictedMode. Please configure and restart in normal mode");
            Factory.setIntercom(mockIntercom);
        } else if (configApi.isDeveloperRestrictedMode()) {
            String message = "SOSI-GW running in DeveloperRestrictedMode. USE FOR SINGLE-USER TESTING ONLY!";
            log.warn(message);
            System.err.println(message);
            setLevel("com.trifork.sosigw", Level.DEBUG);
            Factory.setIntercom(mockIntercom);
        } else {
            if (configApi.isProductionMode()) {
                log.info("SOSI-GW running in ProductionMode");
            } else if (configApi.isTestMode()) {
                log.info("SOSI-GW running in TestMode");
            }
            
            try {
                Factory.setIntercom(new IntercomImpl());
            } catch (IOException e) {
                throw new RuntimeException(e);
            }

            Timer.setTimer(new Timer(new Log4jTimeSink(), 32));

            initAuditLogging(configApi);
            
            log.info("Basic Bootstrap of SOSI-GW completed.");
        }        

        BootStrapRejectorFilter.start();
     }

    private static void initAuditLogging(StaticConfigApi configApi) {
        log.info("Enabling audit logging to files");
        AuditLog.addLogger(new Audit2Log4j());
        if (configApi.isAuditLogToDbEnabled()) {
            log.info("Enabling audit logging to database");
            AuditLog.addLogger(AuditLog2DB.getInstance());
        } else {
            log.info("Disabling audit logging to database");
        }

        AuditLog.start();
        if (StaticConfigImpl.getInstance().isGlobalAuditLogEnabled()) {
            AuditLogGlobal.start();
            AuditLog.addSearcher(AuditLogGlobal.getInstance());
        }

        String centralAuditLogTablePrefix = configApi.getCentralAuditLogTablePrefix();
        if (centralAuditLogTablePrefix != null) {
            log.info("Enabling audit log db centralizer");
            AuditLogCentralizerMysql.start(centralAuditLogTablePrefix);
        } else {
            log.info("Audit log db centralizer disabled");
        }
    }

    private static void initSlaLogging() {
        SLALogConfig slaLogConfig;
        if (StaticConfigImpl.getInstance().isSLALogEnabled()) {
            slaLogConfig = new SLALogConfig(SLA_LOG_APP_NAME, SLA_LOG_APP_SHORT_NAME);
        } else {
            // disable SLA logging
            final Properties properties = new Properties();
            properties.put(BackendInstantiator.NSPSLALOG_LOG4J_CONFIGURE_KEY, false);
            slaLogConfig = new SLALogConfig(properties);
        }
        Util.setSLALogger(slaLogConfig.getSLALogger());
    }

    @SuppressWarnings({ "unchecked", "rawtypes" })
    private static void verifyDependencies() {
        if (org.apache.xalan.Version.getMajorVersionNum() < 2
                || (org.apache.xalan.Version.getMajorVersionNum() == 2 && org.apache.xalan.Version.getReleaseVersionNum() < 7)) {
            logAndPrintError("Apache xalan must be at least version 2.7.0. See installation guide for more information. " 
                + "Version is " + org.apache.xalan.Version.getVersion(), null);
        }

        try {
            Class xerces = Class.forName(ORG_APACHE_XERCES_IMPL_VERSION);
            if (xerces != null) {
                Method m = xerces.getMethod("getVersion", new Class[0]);
                String versionString = (String)m.invoke(null, new Object[0]);
                if (versionString.compareTo("Xerces-J 2.6.0") < 0) {
                    logAndPrintError("Apache xerces must be at least version 2.6.0. See installation guide for more information. " 
                            + "Version is " + versionString, null);
                }
            }
        } catch (Exception e) {
            logAndPrintError("Failed to load class " + ORG_APACHE_XERCES_IMPL_VERSION, e);
        }
    }

    private static void logAndPrintError(String msg, Throwable t) {
        if (t != null) {
            log.error(msg, t);
        } else {
            log.error(msg);
        }
        System.err.println(msg);
    }

    public static void stop() {
        BootStrapRejectorFilter.kill();
        AuditLogGlobal.stop();
        Factory.getIntercom().kill();
        Factory.setIntercom(null);
        Factory.setTheIDCardCache(null);
        if (xdog != null) {
            xdog.kill();
            xdog = null;
        }
        ProxyRequestHandler.kill();
        AuditLog.stop();
        org.apache.commons.logging.LogFactory.releaseAll();
        LogManager.shutdown();
        LogManager.resetConfiguration();
        LogManager.shutdown();
    }

    public static void initLogging() {
        String logFile = Util.getConfigFiledirName() + "/log4j-sosigw.xml";
        if (new File(logFile).canRead()) {
            // permspace leaking code in log4j: DOMConfigurator.configureAndWatch(logFile, 30 * 1000);
            xdog = new XMLWatchdog(logFile, 30 * 1000);
            xdog.start();
            log.info("Log4j configured from " + logFile);
        } else {
            String msg = "Unable to read file " + logFile + " - falling back to default logging settings.";
            System.err.println(msg);
            BasicConfigurator.configure();
            Logger logger = Logger.getRootLogger();
            logger.setLevel(Level.INFO);
            logger.setAdditivity(true);

            setLevel("org.apache.xml.security.signature.Reference", Level.WARN);
            logger.warn(msg);
        }
    }

    public static void setLevel(String packageName, Level level) {
        Logger logger = Logger.getLogger(packageName);
        logger.setLevel(level);
        logger.setAdditivity(true);
    }
}
