/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.auditlog;

import com.trifork.sosigw.api.IDCardCacheEntry;
import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.console.datamodel.AuditLogEntry;
import com.trifork.sosigw.console.datamodel.AuditLogQuery;
import dk.sdsd.nsp.slalog.impl.NSPAwareRequestWrapper;
import dk.sosi.seal.model.UserIDCard;
import org.apache.log4j.Logger;

import javax.servlet.http.HttpServletRequest;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

/**
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public final class AuditLog {
    private static final int QUEUE_CAPACITY = 1000;

    private static final Logger log = Logger.getLogger(AuditLog.class); 
    
    // Used by multiple threads concurrently, but expected to by read-only after bootstrap.
    private static final List<AuditLogger> loggers = new ArrayList<AuditLogger>();
    private static final List<AuditLogSearch> searchers = new ArrayList<AuditLogSearch>();
    
    // Queue of items to be logged. added to by many threads, read by a single thread.
    private static final BlockingQueue<AuditLogItem> todo = new ArrayBlockingQueue<AuditLogItem>(QUEUE_CAPACITY, false);

    // We only need to place log items on the shared queue if we are logging to the database, file based logging works fine for concurrent threads, and
    // there's no noticeable delay
    private static final boolean useSharedQueue = StaticConfigImpl.getInstance().isAuditLogToDbEnabled();

    private static Worker worker;

    public static void addLogger(AuditLogger instance) {
        loggers.add(instance);
    }

    public static void addSearcher(AuditLogSearch searcher) {
        searchers.add(searcher);
    }
    
    public static void start() {
        if (useSharedQueue) {
            worker = new Worker();
            worker.setDaemon(true);
            worker.setName("Loq queue processor");
            worker.start();
        }
    }
    
    public static boolean isWorking() {
        return worker != null && !worker.stopped;
    }
    
    public static void stop() {
        if (worker == null) return;
        
        worker.killed = true;
        worker.interrupt();
    }
    
    public static void logIdcardChange(IDCardCacheEntry changedEntry, HttpServletRequest requestForClientIP) {
        logIdcardChange(changedEntry, new NSPAwareRequestWrapper(requestForClientIP).getRemoteAddr());
    }

    protected static void logIdcardChange(IDCardCacheEntry changedEntry, String clientIP) {
        UserIDCard idCard = changedEntry.getIdCard();
        if (idCard != null && idCard.getUserInfo() != null && idCard.getUserInfo().getCPR() != null) {
            AuditLogItem item = new AuditLogItem(changedEntry.getState().toString(), idCard.getUserInfo().getCPR(), System.currentTimeMillis(), null, null, clientIP);
            queueItem(item);
        }
    }


    public static void logIdcardUse(IDCardCacheEntry changedEntry, HttpServletRequest req, String wsaTo, String wsaAction) {
        UserIDCard idCard = changedEntry.getIdCard();
        AuditLogItem item = new AuditLogItem(changedEntry.getState().toString(), idCard.getUserInfo().getCPR(), System.currentTimeMillis(), wsaTo, wsaAction, new NSPAwareRequestWrapper(req).getRemoteAddr());
        queueItem(item);
    }
    
    public static void logIdcardUse(IDCardCacheEntry changedEntry, String req, String wsaTo, String wsaAction) {
        UserIDCard idCard = changedEntry.getIdCard();
        AuditLogItem item = new AuditLogItem(changedEntry.getState().toString(), idCard.getUserInfo().getCPR(), System.currentTimeMillis(), wsaTo, wsaAction, req);
        queueItem(item);
    }
    
    public static List<AuditLogEntry> search(AuditLogQuery query) {
        for (AuditLogSearch searcher : searchers) {
            return searcher.search(query);
        }
        return Collections.emptyList();
    }
    
    public static int getQueueSize() {
        return todo.size();
    }

    
    private static void queueItem(AuditLogItem item) {
        if (useSharedQueue) {
            if (!todo.offer(item)) {
                String msg = "Unable to audit-log due to overflow in queue";
                log.error(msg);
                AlertSender.sendAlert(msg);
            }
        } else {
            logItem(item);
        }
    }

    // TODO: Optimize. Send a chunk, every second or so, as an array to each logger, so that the DB may use batched inserts.
    // Currently spends approx 10ms per item when inserting in db. It also takes approx 10ms to insert 10 items in one batch...
    private static class Worker extends Thread {
        public boolean killed;
        public boolean stopped = false;
        public void run() {
            while (!killed) {
                try {
                    logItem(todo.take());
                } catch (InterruptedException e) {
                    continue;
                } catch (Exception any) {
                    String msg = "Unable to audit-log, problem fetching from queue ";
                    log.error(msg, any);
                    AlertSender.sendAlert(msg);
                }
            }
            while (!todo.isEmpty()) {
                try {
                    logItem(todo.take());
                } catch (InterruptedException e) {
                    // ignore.
                }
            }
            stopped = true;
        }
    }
    
    private static void logItem(AuditLogItem item) {
        for (AuditLogger instance : loggers) {
            try {
                instance.log(item);
            } catch (Exception any) {
                log.error("Unable to audit-log ", any);
                AlertSender.sendAlert("Unable to audit-log");
            }
        }
    }
}
