/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.auditlog;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import org.apache.log4j.Logger;

import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.StaticConfigImpl;

/**
 * Implements centralizing a SOSI GW Audit Log from a local SOSI GW to a centralized
 * Audit Log at a central Audit Log database. Reuses some parts from AuditLogGlobalMysql.
 * It transfers all available records up until a max number.
 * <p>
 * The centralizer will normalize the Audit Log tables
 * to ensure proper normalization at the destination. Once the entries are
 * transferred to the remote database, they are removed locally.
 * <p>
 * Usage:
 * <pre>
 *
 * AuditLogCentralizerMysql.start(tablePrefix);
 * </pre>
 * <p>
 * Copied some of the MySQLSLALogCentralizer code from NSP SLA Log implementation - version 0.9.
 * 
 * @see AuditLogGlobalMysql
 * @author Christian Callsen
 */
public class AuditLogCentralizerMysql {
    
    private static final int MAX_LOG_RECORDS_TO_PROCESS = 500;
    private final String destinationDataSourceName = "java:jdbc/sosigwcentral";
    private final String destinationTablePrefix;

    private final String hostName;
    private final String applicationName;
    
    public AuditLogCentralizerMysql(String destinationTablePrefix) {
        super();
        if (destinationTablePrefix.indexOf('.') < 0) {
            throw new IllegalArgumentException("Destination table prefix must contain a '.'");
        }
        this.destinationTablePrefix = destinationTablePrefix;
        try {
            this.hostName = InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException e) {
            throw new RuntimeException("Cannot determine local host name: "+e.getMessage(),e);
        }
        this.applicationName = "SOSIGW";
    }

    // copy & paste from AuditLogGlobalMysql.java
    
    public static final Logger log = Logger.getLogger(AuditLogCentralizerMysql.class); 

    private static AuditLogCentralizerMysql auditLogCentralizerMysql = null;
    private static Thread worker = null;
    private String sourceDataSourceName = StaticConfigImpl.getInstance().getDataSourceName();        
    private String sourceTablePrefix = "sosigwlog";

    public static void start(String destinationTablePrefix) {
        auditLogCentralizerMysql = new AuditLogCentralizerMysql(destinationTablePrefix);
        worker = new Thread() {

            public void run() {
                while (worker == this) {
                    try {
                        long interval = RuntimeConfig.getGlobalDatabaseJDBCInterval();
                        if (interval <= 0) {
                            log.warn("Unreasonably low sleep interval, using 10 seconds");
                            interval = 10000;
                        }
                        Thread.sleep(interval);
                        if (worker == this && auditLogCentralizerMysql != null) {
                            auditLogCentralizerMysql.execute();
                        }
                    } catch (java.lang.InterruptedException e) {
                        log.debug("Interrupted.");
                        // Fine, probably got killed.
                    } catch (Exception e) {
                        log.error("Move to central db failed", e);
                    }
                }
                if (auditLogCentralizerMysql != null) {
                    auditLogCentralizerMysql.shutdown();
                    auditLogCentralizerMysql = null;
                }
            }
        };
        worker.setName("SOSIGW-AuditLogCentralizer");
        worker.setDaemon(true);
        worker.start();
    }
    
    public static AuditLogCentralizerMysql getInstance() {
        return auditLogCentralizerMysql;
    }
    
    public static void stop() {
        if (worker != null) {
            Thread oldWorker = worker;
            worker = null;
            oldWorker.interrupt();
        }
    }
    
    protected Connection getLocalDBConnectionFromPool() throws Exception {
        DataSource ds = lookupDataSource(sourceDataSourceName);
        
        Connection con = ds.getConnection();
        DatabaseMetaData dbmd = con.getMetaData(); //get MetaData to confirm connection
        log.debug("Connection to Local database " + dbmd.getURL() + " as user " + dbmd.getUserName() + " (" + dbmd.getDatabaseProductName() + " " + dbmd.getDatabaseProductVersion() + ") successful.");
        return con;
    }

    private DataSource lookupDataSource(String name) throws NamingException {
        Context ctx = new InitialContext();

        DataSource ds = (DataSource)ctx.lookup(name);
        if (ds == null) {
            throw new IllegalStateException("Unable to lookup Datasource in jndi from '" + name + "'");
        }
        return ds;
    }
    
    // end copy & paste
    
    protected Connection getGlobalDBConnection() throws Exception {
        DataSource ds = lookupDataSource(destinationDataSourceName);
        Connection con = ds.getConnection();
        DatabaseMetaData dbmd = con.getMetaData(); //get MetaData to confirm connection
        log.debug("Connection to Central database " + dbmd.getURL() + " as user " + dbmd.getUserName() + " (" + dbmd.getDatabaseProductName() + " " + dbmd.getDatabaseProductVersion() + ") successful.");
        return con;
    }
    
    /**
     * Builds local SQL statements given a table prefix
     */
    static class LocalSQLBuilder {        
        // local SOSI GW log SQLs
        private final String SELECT_ALL_LOG_RECORDS_PREFIX = "SELECT lpk, systemid, operation, cpr, moment, wsato, wsaaction, clientip FROM ";
        private final String SELECT_ALL_LOG_RECORDS_POSTFIX = ".SOSIGWLOG ORDER BY lpk";        
        private final String DELETE_LOG_RECORDS_PREFIX = "DELETE FROM ";
        private final String DELETE_LOG_RECORDS_POSTFIX = ".SOSIGWLOG"+" WHERE lpk <= ?";
        // computed final SQLs containing proper table names        
        private final String selectAllLogRecords;
        private final String deleteLogRecords;
        
        public LocalSQLBuilder(final String tablePrefix) {
            super();
            selectAllLogRecords = SELECT_ALL_LOG_RECORDS_PREFIX + tablePrefix + SELECT_ALL_LOG_RECORDS_POSTFIX;
            deleteLogRecords = DELETE_LOG_RECORDS_PREFIX + tablePrefix + DELETE_LOG_RECORDS_POSTFIX;
        }
        public String getSelectAllLogRecords() {
            return selectAllLogRecords;
        }
        public String getDeleteLogRecords() {
            return deleteLogRecords;
        }
    }

    /**
     * Builds central log SQL statements given a table prefix
     */
    static class RemoteSQLBuilder {        
        // centralized SOSI GW log SQLs
        // these assume that tablePrefix contains a '.' where appropriate
        private final String SELECT_ALL_WSAKEYS_PREFIX = "SELECT pk, wsato, wsaaction FROM ";
        private final String SELECT_ALL_WSAKEYS_POSTFIX = "wsa ";
        private final String SELECT_ALL_CLIENTIP_PREFIX = "SELECT pk, clientip FROM ";
        private final String SELECT_ALL_CLIENTIP_POSTFIX = "clientip ";
        private final String SELECT_ALL_OPER_PREFIX = "SELECT pk, operation FROM ";
        private final String SELECT_ALL_OPER_POSTFIX = "operation ";
        private final String SELECT_ALL_SYSTEMID_PREFIX = "SELECT pk, systemid FROM ";
        private final String SELECT_ALL_SYSTEMID_POSTFIX = "systemid ";

        private final String INSERT_WSAKEY_PREFIX = "INSERT INTO ";
        private final String INSERT_WSAKEY_POSTFIX = "wsa (wsato,wsaaction) VALUES (?,?) ";
        private final String INSERT_CLIENTIP_PREFIX = "INSERT INTO ";
        private final String INSERT_CLIENTIP_POSTFIX = "clientip (clientip) VALUES (?) ";
        private final String INSERT_OPERATION_PREFIX = "INSERT INTO ";
        private final String INSERT_OPERATION_POSTFIX = "operation (operation) VALUES (?) ";
        private final String INSERT_SYSTEMID_PREFIX = "INSERT INTO ";
        private final String INSERT_SYSTEMID_POSTFIX = "systemid (systemid) VALUES (?) ";
        
        private final String INSERT_LOG_RECORDS_PREFIX = "INSERT INTO ";
        private final String INSERT_LOG_RECORDS_POSTFIX = "sosigwcentrallog (systemid,operation,cpr,moment,wsa,clientip,createdTime,createdByHost,createdByApplication) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)";
        
        // computed final SQLs containing proper table names
        private final String selectAllWSAKeys;
        private final String selectAllClientIPs;
        private final String selectAllOperations;
        private final String selectAllSystemIDs;
        private final String insertIntoWSAKeys;
        private final String insertIntoClientIPs;
        private final String insertIntoOperations;
        private final String insertIntoSystemIDs;
        
        private final String insertIntoLogRecords;
        
        public RemoteSQLBuilder(final String tablePrefix) {
            super();
            selectAllWSAKeys = SELECT_ALL_WSAKEYS_PREFIX + tablePrefix + SELECT_ALL_WSAKEYS_POSTFIX;
            selectAllClientIPs = SELECT_ALL_CLIENTIP_PREFIX + tablePrefix + SELECT_ALL_CLIENTIP_POSTFIX;
            selectAllOperations = SELECT_ALL_OPER_PREFIX + tablePrefix + SELECT_ALL_OPER_POSTFIX;
            selectAllSystemIDs = SELECT_ALL_SYSTEMID_PREFIX + tablePrefix + SELECT_ALL_SYSTEMID_POSTFIX;

            insertIntoWSAKeys = INSERT_WSAKEY_PREFIX + tablePrefix + INSERT_WSAKEY_POSTFIX;
            insertIntoClientIPs = INSERT_CLIENTIP_PREFIX + tablePrefix + INSERT_CLIENTIP_POSTFIX;
            insertIntoOperations = INSERT_OPERATION_PREFIX + tablePrefix + INSERT_OPERATION_POSTFIX;
            insertIntoSystemIDs = INSERT_SYSTEMID_PREFIX + tablePrefix + INSERT_SYSTEMID_POSTFIX;

            insertIntoLogRecords = INSERT_LOG_RECORDS_PREFIX + tablePrefix + INSERT_LOG_RECORDS_POSTFIX;
        }

        public String getSelectAllWSAKeys() {
            return selectAllWSAKeys;
        }

        public String getSelectAllClientIPs() {
            return selectAllClientIPs;
        }

        public String getSelectAllOperations() {
            return selectAllOperations;
        }

        public String getSelectAllSystemIDs() {
            return selectAllSystemIDs;
        }

        public String getInsertIntoWSAKeys() {
            return insertIntoWSAKeys;
        }

        public String getInsertIntoClientIPs() {
            return insertIntoClientIPs;
        }

        public String getInsertIntoOperations() {
            return insertIntoOperations;
        }

        public String getInsertIntoSystemIDs() {
            return insertIntoSystemIDs;
        }

        public String getInsertIntoLogRecords() {
            return insertIntoLogRecords;
        }
    }
    
    /**
     * Caches normalized WSAKeys, ClientIPs, Operations and SystemIDs from the remote 
     * central log server. Can also add missing entries (given a list) and reload
     * the local caches.
     */
    static class NormalizationCache {
        static class WSAKey {
            public final int    wsaKey; // not part of identity
            public final String wsaTo;
            public final String wsaAction;

            public WSAKey(int wsaKey, String wsaTo, String wsaAction) {
                this.wsaKey = wsaKey;
                this.wsaTo = wsaTo;
                this.wsaAction = wsaAction;
            }


            public int hashCode() {
                final int prime = 31;
                int result = 1;
                result = prime * result + ((wsaAction == null) ? 0 : wsaAction.hashCode());
                result = prime * result + ((wsaTo == null) ? 0 : wsaTo.hashCode());
                return result;
            }


            public boolean equals(Object obj) {
                if (this == obj)
                    return true;
                if (obj == null)
                    return false;
                if (getClass() != obj.getClass())
                    return false;
                WSAKey other = (WSAKey)obj;
                if (wsaAction == null) {
                    if (other.wsaAction != null)
                        return false;
                } else if (!wsaAction.equals(other.wsaAction))
                    return false;
                if (wsaTo == null) {
                    if (other.wsaTo != null)
                        return false;
                } else if (!wsaTo.equals(other.wsaTo))
                    return false;
                return true;
            }
        }
        static class SingleValue {
            public final int pk; // not part of identity
            public final String value;
            public SingleValue(int pk, String clientIP) {
                super();
                this.pk = pk;
                this.value = clientIP;
            }

            public int hashCode() {
                final int prime = 31;
                int result = 1;
                result = prime * result + ((value == null) ? 0 : value.hashCode());
                return result;
            }

            public boolean equals(Object obj) {
                if (this == obj)
                    return true;
                if (obj == null)
                    return false;
                if (getClass() != obj.getClass())
                    return false;
                SingleValue other = (SingleValue)obj;
                if (value == null) {
                    if (other.value != null)
                        return false;
                } else if (!value.equals(other.value))
                    return false;
                return true;
            }
        }
        
        private final Map<WSAKey, WSAKey> wsaMap = new HashMap<WSAKey, WSAKey>();
        private final Map<SingleValue, SingleValue> clientIPMap = new HashMap<SingleValue, SingleValue>();
        private final Map<SingleValue, SingleValue> operationMap = new HashMap<SingleValue, SingleValue>();
        private final Map<SingleValue, SingleValue> systemIDMap = new HashMap<SingleValue, SingleValue>();
        
        private final Connection globalDB;
        private final RemoteSQLBuilder sqlBuilder;

        public NormalizationCache(Connection conn, RemoteSQLBuilder sqlBuilder) {
            super();
            this.globalDB = conn;
            this.sqlBuilder = sqlBuilder;
        }

        public void cacheAll() throws SQLException {
            cacheWSAKeys();
            cacheClientIPs();
            cacheOperations();
            cacheSystemIDs();
            log.debug("NormalizationCache: cached "+wsaMap.size()+" WSAKeys, "+clientIPMap.size()+" ClientIPs, "+
                    operationMap.size()+" Operations, and "+systemIDMap.size()+" SystemIDs");
        }
        
        private void cacheWSAKeys() throws SQLException {
            PreparedStatement selectStmt = null;
            ResultSet rs = null;
            try {
                selectStmt = globalDB.prepareStatement(sqlBuilder.getSelectAllWSAKeys());
                if (selectStmt.execute()) {
                    rs = selectStmt.getResultSet();
                    while (rs.next()) {
                        int pk = rs.getInt(1);
                        String wsaTo = rs.getString(2);
                        String wsaAction = rs.getString(3);
                        WSAKey wsaKey = new WSAKey(pk, wsaTo, wsaAction);
                        wsaMap.put(wsaKey, wsaKey);
                    }
                }
                selectStmt.close();
            } catch (SQLException e) {
                log.error("cacheWSAKeys(): Caught exception caching all WSAKeys using "+sqlBuilder.getSelectAllWSAKeys(),e);
                throw e;
            } finally {
                if (rs != null) { rs.close(); }
                if (selectStmt != null) { selectStmt.close(); }
            }
        }
        private void cacheClientIPs() throws SQLException {
            cacheOneSingleValueCategory("ClientIPs", sqlBuilder.getSelectAllClientIPs(), clientIPMap);
        }
        private void cacheOperations() throws SQLException {
            cacheOneSingleValueCategory("Operations", sqlBuilder.getSelectAllOperations(), operationMap);
        }
        private void cacheSystemIDs() throws SQLException {
            cacheOneSingleValueCategory("SystemIDs", sqlBuilder.getSelectAllSystemIDs(), systemIDMap);
        }
        private void cacheOneSingleValueCategory(final String categoryName, final String sqlStmt, final Map<SingleValue, SingleValue> map) throws SQLException {
            PreparedStatement selectStmt = null;
            ResultSet rs = null;
            try {
                selectStmt = globalDB.prepareStatement(sqlStmt);
                if (selectStmt.execute()) {
                    rs = selectStmt.getResultSet();
                    while (rs.next()) {
                        int pk = rs.getInt(1);
                        String value = rs.getString(2);
                        SingleValue entry = new SingleValue(pk, value);
                        map.put(entry, entry);
                    }
                }
                selectStmt.close();
            } catch (SQLException e) {
                log.error("cacheOneSingleValueCategory(): Caught exception caching all "+ categoryName +" using "+sqlStmt,e);
                throw e;
            } finally {
                if (rs != null) { rs.close(); }
                if (selectStmt != null) { selectStmt.close(); }
            }            
        }

        public Map<WSAKey, WSAKey> getWsaMap() {
            return wsaMap;
        }

        public Map<SingleValue, SingleValue> getClientIPMap() {
            return clientIPMap;
        }

        public Map<SingleValue, SingleValue> getOperationMap() {
            return operationMap;
        }

        public Map<SingleValue, SingleValue> getSystemIDMap() {
            return systemIDMap;
        }

        public void addMissingWSAKeys(List<WSAKey> missingWSAKeys) throws SQLException {
            log.debug("addMissingWSAKeys: adding "+missingWSAKeys.size()+" missing WSA Keys");
            if (missingWSAKeys.size() == 0) {
                return;
            }
            PreparedStatement insertStmt = globalDB.prepareStatement(sqlBuilder.getInsertIntoWSAKeys());
            try {
                for (WSAKey thing : missingWSAKeys) {
                    insertStmt.setString(1, thing.wsaTo);
                    if (thing.wsaAction != null) {
                        insertStmt.setString(2, thing.wsaAction);
                    } else {
                        insertStmt.setNull(2, Types.VARCHAR);
                    }
                    insertStmt.addBatch();
                }
                insertStmt.executeBatch();
            } finally {
                insertStmt.close();
            }
            wsaMap.clear();
            cacheWSAKeys();
        }

        public void addMissingClientIPs(List<String> missingClientIPs) throws SQLException {
            log.debug("addMissingClientIPs: adding "+missingClientIPs.size()+" missing ClientIPs");
            if (missingClientIPs.size() == 0) {
                return;
            }
            addMissing(missingClientIPs,sqlBuilder.getInsertIntoClientIPs());
            clientIPMap.clear();
            cacheClientIPs();
        }

        public void addMissingOperations(List<String> missingOperations) throws SQLException {
            log.debug("addMissingClientIPs: adding "+missingOperations.size()+" missing Operations");
            if (missingOperations.size() == 0) {
                return;
            }
            addMissing(missingOperations,sqlBuilder.getInsertIntoOperations());
            operationMap.clear();
            cacheOperations();
        }

        public void addMissingSystemIDs(List<String> missingSystemIDs) throws SQLException {
            log.debug("addMissingSystemIDs: adding "+missingSystemIDs.size()+" missing SystemIDs");
            if (missingSystemIDs.size() == 0) {
                return;
            }
            addMissing(missingSystemIDs,sqlBuilder.getInsertIntoSystemIDs());
            systemIDMap.clear();
            cacheSystemIDs();
        }
        private void addMissing(List<String> missing, String insertStmtSql) throws SQLException {
            PreparedStatement insertStmt = globalDB.prepareStatement(insertStmtSql);
            try {
                for (String thing : missing) {
                    insertStmt.setString(1, thing);
                    insertStmt.addBatch();
                }
                insertStmt.executeBatch();
            } finally {
                insertStmt.close();
            }
        }
    }
    
    protected void shutdown() {
        log.info("Shutting down");
    }
    
    /**
     * Overall processing, which obtains connections, processes the log, and commits (or rolls back).
     */
    protected void execute() {
        Connection fromConnection = null;
        Connection toConnection = null;

        try {
            log.debug("execute(): Processing begins");
            long beginTime = System.nanoTime();
            fromConnection = getLocalDBConnectionFromPool();
            fromConnection.setAutoCommit(false);
            toConnection = getGlobalDBConnection();
            toConnection.setAutoCommit(false);
            
            int items = processLog(fromConnection,toConnection);
            log.debug("execute(): Committing remote!");
            toConnection.commit();
            fromConnection.commit();
            long endTime = System.nanoTime();
            long processTime = (long)((endTime-beginTime)/1000);
            if (items > 0) {
                log.info("Processing took "+processTime+" mikros to create "+items+" records or "+(long)processTime/items + " mikros each");
            } else {
                log.debug("execute(): Processing ends; took "+processTime+" mikros to check for records (none moved, status="+items+")");             
            }
            toConnection.setAutoCommit(true);
            fromConnection.setAutoCommit(true);

        } catch (Exception e) {
            log.error("Failed establishing connections for transfer of audit log to central database", e);
            try {
                toConnection.rollback();
            } catch (SQLException e1) {
            }
            try {
                fromConnection.rollback();
            } catch (SQLException e1) {
            }
            return;
        } finally {
            if (toConnection != null) { 
                try {
                    toConnection.close();
                } catch (SQLException e) {
                } 
            }
            if (fromConnection != null) { 
                try {
                    fromConnection.close();
                } catch (SQLException e) {
                } 
            }
        }
    }

    static class Record extends AuditLogItem {
        public final String systemID;
        public final int pk;
        
        public Record(int pk, String operation, String cpr, long timestamp, String wsaTo, String wsaAction, String clientIP, String systemID) {
            super(operation, cpr, timestamp, wsaTo, wsaAction, clientIP);
            this.systemID = systemID;
            this.pk = pk;
        }
        
    }
    
    /**
     * Processes the local log into the central log including normalization, given connections to local and central.
     * @param fromConnection database connection containing local audit log
     * @param toConnection database connection for central log destination
     * @return number of items transferred
     * @throws SQLException
     */
    private int processLog(Connection fromConnection, Connection toConnection) throws SQLException {
        // algorithm
        // 1. cache remote normalization (if any)
        // 2. query all local records
        // 3.   collect all in memory
        // 4.   determine missing normalisation
        // 5. add missing normalization to remote cache
        // 6. add all records from memory to remote using normalization maps
        // 7. purge local records
        
        java.sql.Timestamp createdTime = new java.sql.Timestamp(System.currentTimeMillis());
                
        int numberOfRecordsProcessed = 0;
        
        LocalSQLBuilder localSQLBuilder = new LocalSQLBuilder(sourceTablePrefix);
        RemoteSQLBuilder remoteSQLBuilder = new RemoteSQLBuilder(destinationTablePrefix);

        log.debug("processLog: caching all remote");
        NormalizationCache remoteCache = new NormalizationCache(toConnection, remoteSQLBuilder); // 4 remote queries
        remoteCache.cacheAll();
 
        Map<NormalizationCache.WSAKey, NormalizationCache.WSAKey> knownWSAs = remoteCache.getWsaMap();
        Map<NormalizationCache.SingleValue, NormalizationCache.SingleValue> knownClientIPs = remoteCache.getClientIPMap();
        Map<NormalizationCache.SingleValue, NormalizationCache.SingleValue> knownOperations = remoteCache.getOperationMap();
        Map<NormalizationCache.SingleValue, NormalizationCache.SingleValue> knownSystemIDs = remoteCache.getSystemIDMap();
        
        List<NormalizationCache.WSAKey> missingWSAKeys = new LinkedList<NormalizationCache.WSAKey>();
        List<String> missingClientIPs = new LinkedList<String>();
        List<String> missingOperations = new LinkedList<String>();
        List<String> missingSystemIDs = new LinkedList<String>();

        List<Record> records = new LinkedList<Record>();
        
        log.debug("processLog: computing missing keys from log contents");
        int numberOfLogRecords = 0;
        {
            PreparedStatement stmt = null;
            ResultSet rs = null;
            try {
                log.debug("processLog: perusing log contents using "+localSQLBuilder.getSelectAllLogRecords());
                // 1 local query
                stmt = fromConnection.prepareStatement(localSQLBuilder.getSelectAllLogRecords(), ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY);
                rs = stmt.executeQuery();
                while (rs.next() && numberOfLogRecords < MAX_LOG_RECORDS_TO_PROCESS) {
                    String systemid = rs.getString(2);
                    if (!knownSystemIDs.containsKey(new NormalizationCache.SingleValue(-1, systemid))) {
                        missingSystemIDs.add(systemid);
                    }
                    String oper = rs.getString(3);
                    if (!knownOperations.containsKey(new NormalizationCache.SingleValue(-1, oper))) {
                        missingOperations.add(oper);
                    }
                    String clientIP = rs.getString(8);
                    if (!knownClientIPs.containsKey(new NormalizationCache.SingleValue(-1, clientIP))) {
                        missingClientIPs.add(clientIP);
                    }
                    String wsaTo = rs.getString(6);
                    String wsaAction = rs.getString(7);
                    NormalizationCache.WSAKey wsaKey = new NormalizationCache.WSAKey(-1, wsaTo, wsaAction);
                    if (!knownWSAs.containsKey(wsaKey)) {
                        missingWSAKeys.add(wsaKey);
                    }
                    int pk = rs.getInt(1);
                    String cpr = rs.getString(4);
                    Timestamp timestamp = rs.getTimestamp(5);

                    records.add(new Record(pk,oper,cpr,timestamp.getTime(),wsaTo,wsaAction,clientIP,systemid));
                    
                    numberOfLogRecords++;
                }

            } catch (Exception e) {
                log.error("Caught exception trying to query local records",e);
                return -1;
            } finally {
                if (rs != null) { rs.close(); }            
                if (stmt != null) { stmt.close(); }                            
            }
        }
        
        log.debug("processLog: saw "+numberOfLogRecords+" records in log contents");
        log.debug("processLog: handling missing keys from log contents");
        
        try {
            // insert and re-cache keys - max 4 remote batch inserts
            remoteCache.addMissingWSAKeys(missingWSAKeys);
            remoteCache.addMissingClientIPs(missingClientIPs);
            remoteCache.addMissingOperations(missingOperations);
            remoteCache.addMissingSystemIDs(missingSystemIDs);
        } catch (Exception e) {
            log.error("Caught exception trying to add missing normalization",e);
            return -1;
        }
        
        int maxPK = 0;
        {
            PreparedStatement insertStmt = null;
            ResultSet rs = null;
            try {
                log.debug("processLog: inserting log contents using "+remoteSQLBuilder.getInsertIntoLogRecords());
                insertStmt = toConnection.prepareStatement(remoteSQLBuilder.getInsertIntoLogRecords());    
                for (Record record : records) {
                    String systemid = record.systemID;
                    String oper = record.operation;
                    String cpr = record.cpr;
                    Timestamp timestamp = new Timestamp(record.timestamp);
                    String wsaTo = record.wsaTo;
                    String wsaAction = record.wsaAction;
                    String clientIP = record.clientIP;

                    NormalizationCache.SingleValue systemIdKey = new NormalizationCache.SingleValue(-1, systemid);
                    if (!knownSystemIDs.containsKey(systemIdKey)) {
                        // missing normalization, just break, and next run will fix this
                        log.debug("processLog: systemid "+systemid+" unknown");
                        break;
                    }
                    NormalizationCache.SingleValue operationKey = new NormalizationCache.SingleValue(-1, oper);
                    if (!knownOperations.containsKey(operationKey)) {
                        // missing normalization, just break, and next run will fix this
                        log.debug("processLog: operation "+oper+" unknown");
                        break;
                    }
                    NormalizationCache.SingleValue clientipKey = new NormalizationCache.SingleValue(-1, clientIP);
                    if (!knownClientIPs.containsKey(clientipKey)) {
                        // missing normalization, just break, and next run will fix this
                        log.debug("processLog: clientip "+clientIP+" unknown");
                        break;
                    }
                    NormalizationCache.WSAKey wsaKey = new NormalizationCache.WSAKey(0, wsaTo, wsaAction);
                    if (wsaTo != null) {
                        if (!knownWSAs.containsKey(wsaKey)) {
                            log.debug("processLog: WSAKey "+wsaTo+" unknown");
                            // missing normalization, just break, and next run will fix this
                            break;
                        }                
                    }

                    // map pks
                    Integer systemIDKey = knownSystemIDs.get(systemIdKey).pk;
                    Integer clientIPKey = knownClientIPs.get(clientipKey).pk;
                    Integer operKey = knownOperations.get(operationKey).pk;
                    Integer wsaKeyKey;
                    if (wsaTo != null) {
                        wsaKeyKey = knownWSAs.get(wsaKey).wsaKey;
                    } else {
                        wsaKeyKey = wsaKey.wsaKey;
                    }

                    // add
                    insertStmt.setInt(1, systemIDKey.intValue());
                    insertStmt.setInt(2, operKey.intValue());
                    insertStmt.setString(3, cpr);
                    insertStmt.setTimestamp(4, timestamp);
                    insertStmt.setInt(5, wsaKeyKey.intValue());
                    insertStmt.setInt(6, clientIPKey);
                    insertStmt.setTimestamp(7, createdTime);
                    insertStmt.setString(8, hostName);
                    insertStmt.setString(9, applicationName);

                    insertStmt.addBatch();
                    numberOfRecordsProcessed++;
                    if (record.pk > maxPK) {
                        maxPK = record.pk;
                    }

                }
                int newRows[] = insertStmt.executeBatch(); // 1 remote batch insert
                if (newRows.length != numberOfRecordsProcessed) {
                    log.warn("processLog(): Inserting "+numberOfRecordsProcessed+" items results in "+newRows.length+" rows being inserted in DB");
                }
            } catch (Exception e) {
                log.error("Caught exception trying to process local records",e);
                return -1;
            } finally {
                if (rs != null) { rs.close(); }
                if (insertStmt != null) { insertStmt.close(); }
            }
            log.debug("processLog: processed "+numberOfRecordsProcessed+" records from log contents");
        }
        
        {
            PreparedStatement updateStmt = null;
            try {
                updateStmt = fromConnection.prepareStatement(localSQLBuilder.getDeleteLogRecords());
                updateStmt.setInt(1, maxPK);
                updateStmt.executeUpdate();
            } catch (Exception e) {
                log.error("Caught exception trying to purge local records",e);
                return -1;            
            } finally {
                updateStmt.close();
            }
        }
        
        return numberOfRecordsProcessed;
    }
}
