/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.auditlog;

import java.sql.BatchUpdateException;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.sql.DataSource;

import org.apache.log4j.Logger;

import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.console.datamodel.AuditLogEntry;
import com.trifork.sosigw.console.datamodel.AuditLogQuery;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public abstract class AuditLogGlobal implements AuditLogSearch {
    public static final Logger log = Logger.getLogger(AuditLogGlobal.class); 

    private static final String SOSIGWLOCAL_PREFIX = "sosigwlog";
    protected static final String SOSIGWLOCALLOG_SOSIGWLOG = SOSIGWLOCAL_PREFIX + ".SOSIGWLOG";
    
    protected static final String SOSIGWGLOBAL_PREFIX = "sosigwgloballog";
    protected static final String SOSIGWGLOBALLOG_SOSIGWGLOBALLOG = SOSIGWGLOBAL_PREFIX + ".SOSIGWGLOBALLOG";
    protected static final String SOSIGWGLOBALLOG_SYSTEMID = SOSIGWGLOBAL_PREFIX + ".systemid";
    protected static final String SOSIGWGLOBALLOG_OPERATION = SOSIGWGLOBAL_PREFIX + ".operation";
    protected static final String SOSIGWGLOBALLOG_WSA = SOSIGWGLOBAL_PREFIX + ".wsa";
    protected static final String SOSIGWGLOBALLOG_CLIENTIP = SOSIGWGLOBAL_PREFIX + ".clientip";


    private static AuditLogGlobal auditLogGlobal = null;
    private static Thread worker = null;
    
    public static void start() {

        String jdbcDriver = StaticConfigImpl.getInstance().getJdbcDriver().toLowerCase();
        if (jdbcDriver.contains("mysql")) {
            auditLogGlobal = new AuditLogGlobalMysql();
        } else if (jdbcDriver.contains("postgresql")) {
            auditLogGlobal = new AuditLogGlobalPostgresql();
        } else if (jdbcDriver.contains("hsqldb")) {
            auditLogGlobal = new AuditLogGlobalHsqldb();
        } else {
            log.fatal("Could not determine type of database support from jdbc driver \"" + StaticConfigImpl.getInstance().getJdbcDriver() + "\"");
        }

        worker = new Thread() {

            public void run() {
                while (worker == this) {
                    try {
                        long interval = RuntimeConfig.getGlobalDatabaseJDBCInterval();
                        if (interval <= 0) {
                            log.warn("Unreasonably low sleep interval, using 10 seconds");
                            interval = 10000;
                        }
                        Thread.sleep(interval);
                        if (worker == this && auditLogGlobal != null) {
                            auditLogGlobal.execute();
                        }
                    } catch (java.lang.InterruptedException e) {
                        log.debug("Interrupted.");
                        // Fine, probably got killed.
                    } catch (Exception e) {
                        log.error("Move to global db failed", e);
                    }
                }
                if (auditLogGlobal != null) {
                    auditLogGlobal.shutdown();
                    auditLogGlobal = null;
                }
            }
        };
        worker.setName("SOSIGW-GlobalLogTransfer");
        worker.setDaemon(true);
        worker.start();
    }
    
    public static AuditLogGlobal getInstance() {
        return auditLogGlobal;
    }
    
    public static void stop() {
        if (worker != null) {
            Thread oldWorker = worker;
            worker = null;
            oldWorker.interrupt();
        }
    }
    
    protected Connection getLocalDBConnectionFromPool() throws Exception {
        String name = StaticConfigImpl.getInstance().getDataSourceName();
        Context ctx = new InitialContext();

        DataSource ds = (DataSource)ctx.lookup(name);
        if (ds == null) {
            throw new IllegalStateException("Unable to lookup Datasource in jndi from '" + name + "'");
        }
        
        Connection con = ds.getConnection();
        DatabaseMetaData dbmd = con.getMetaData(); //get MetaData to confirm connection
        log.info("Connection to Local database " + dbmd.getURL() + " as user " + dbmd.getUserName() + " (" + dbmd.getDatabaseProductName() + " " + dbmd.getDatabaseProductVersion() + ") successful.");
        
        return con;
    }
    

    // Not from pool, as it is configurable  through console
    private Connection getGlobalDBConnection() throws Exception {
        Class.forName(StaticConfigImpl.getInstance().getJdbcDriver());
        String databaseUrl = RuntimeConfig.getGlobalDatabaseJDBCUrl();
        String username = RuntimeConfig.getGlobalDatabaseJDBCUsername();
        String password = RuntimeConfig.getGlobalDatabaseJDBCPassword();
        Connection db = DriverManager.getConnection(databaseUrl, username, password);
        return db;
    }

    private static class WSAKey {
        public final String wsaTo;
        public final String wsaAction;

        public WSAKey(String wsaTo, String wsaAction) {
            this.wsaTo = wsaTo;
            this.wsaAction = wsaAction;
        }
        

        public boolean equals(Object obj) {
            WSAKey other = (WSAKey)obj;
            if (other.wsaTo == null) {
                if (wsaTo != null) {
                    return false;
                }
            } else {
                if (!other.wsaTo.equals(wsaTo)) {
                    return false;
                }
            }
            if (other.wsaAction == null) {
                if (wsaAction != null) {
                    return false;
                }
            } else {
                if (!other.wsaAction.equals(wsaAction)) {
                    return false;
                }
                
            }
            return true;
        }
        

        public int hashCode() {
            return (wsaTo == null ? 0 : wsaTo.hashCode()) + (wsaAction == null ? 0 : wsaAction.hashCode());
        }
    }
    
    
    protected Map<String, Integer> systemidStates = new HashMap<String, Integer>();
    private Map<String, Integer> operation2PK = new HashMap<String, Integer>();
    private Map<String, Integer> clientIP2PK = new HashMap<String, Integer>();
    private Map<WSAKey, Integer> wsa2PKcache = new HashMap<WSAKey, Integer>();
    
    protected Connection localDB = null;
    protected Connection globalDB = null;
    
    protected void shutdown() {
        log.info("Shutting down");
        closeAll();
        systemidStates = null;
        operation2PK = null;
        clientIP2PK = null;
        wsa2PKcache = null;
    }
    
    protected void execute() {
        try {
            if (localDB == null) {
                localDB = getLocalDBConnectionFromPool();
                localDB.setAutoCommit(false);
            }
            if (globalDB == null) {
                globalDB = getGlobalDBConnection();
                globalDB.setAutoCommit(false);
                DatabaseMetaData dbmd = globalDB.getMetaData(); //get MetaData to confirm connection
                log.info("Connection to Global database " + dbmd.getURL() + " as user " + dbmd.getUserName() + " (" + dbmd.getDatabaseProductName() + " " + dbmd.getDatabaseProductVersion() + ") successful.");
            }
        } catch (Exception e) {
            log.error("Failed establishing connections for transfer of audit log to global database", e);
            return;
        }
        
        boolean commit = false;
        try {
            Set<String> systemids = findUnknownLocalSystemIdSet(localDB);
            if (!systemids.isEmpty()) {
                updateServeruidGlobalMapping(globalDB, systemids);
                for (String uid: systemids) {
                    if (!systemidStates.containsKey(uid)) {
                        addServeruid2GlobalState(uid);
                    }
                }
                updateServeruidGlobalMapping(globalDB, systemids);
            }
            transferLogItems();
            commit = true;
        } catch(BatchUpdateException buex) {
            SQLException e = buex;
            while (e != null) {
                log.error("Transfer of audit log to global database failed", e);
                e = e.getNextException();
            }
        } catch (SQLException e) {
            log.error("Transfer of audit log to global database failed", e);
        } finally {
            boolean close = false;
            if (commit) {
                try {
                    globalDB.commit();
                } catch (SQLException e) {
                    log.debug("Global commit failed", e);
                    commit = false;
                    close = true;
                }
            }
            if (commit) {
                try {
                    localDB.commit();
                } catch (SQLException e) {
                    log.error("Local commit failed after global commit: Log entries will get duplicated!", e);
                    close = true;
                }
            }            
            if (!commit) {
                try {
                    globalDB.rollback();
                } catch (SQLException e1) {
                    log.error("Glocal rollback failed!", e1);
                    close = true;
                }
                try {
                    localDB.rollback();
                } catch (SQLException e1) {
                    log.error("Local rollback failed!", e1);
                    close = true;
                }
            }

            if (close) {
                closeAll();
            }
        }
    }


    private void closeAll() {
        if (localDB != null) {
            try {
                localDB.close();
            } catch (SQLException e) {
                // Ignore
            }
            localDB = null;
        }
        if (globalDB != null) {
            try {
                globalDB.close();
            } catch (SQLException e) {
                // Ignore
            }
            globalDB = null;
        }
    }

    protected abstract String getInsertIntoGlobalLogSql();
    protected abstract String getInsertIntoWsaSql();
    protected abstract String getInsertIntoOperation();
    protected abstract String getInsertIntoClientIpSql();
    protected abstract String getInsertIntoSystemIdSql();

    protected String getSelectFromWsaWithActionSql() {
        return "SELECT pk FROM " + SOSIGWGLOBALLOG_WSA + " WHERE wsato=?" + " and wsaaction=?";
    }

    protected String getSelectFromWsaSql() {
        return "SELECT pk FROM " + SOSIGWGLOBALLOG_WSA + " WHERE wsato=?";
    }

    protected void transferLogItems() throws SQLException {
        PreparedStatement insertStmt = globalDB.prepareStatement(getInsertIntoGlobalLogSql());
        Statement stmt = localDB.createStatement(ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_UPDATABLE);
        if (stmt.execute("SELECT * FROM " + SOSIGWLOCALLOG_SOSIGWLOG + " FOR UPDATE")) {
            ResultSet rs = stmt.getResultSet();
            while (rs.next()) {
                //long lpk = rs.getLong(1);
                String uid = rs.getString(2);
                String oper = rs.getString(3);
                String cpr = rs.getString(4);
                Timestamp timestamp = rs.getTimestamp(5);
                String wsaTo = rs.getString(6);
                String wsaAction = rs.getString(7);
                String clientIP = rs.getString(8);
                
                Integer systemIdState = systemidStates.get(uid);
                if (systemIdState == null) {
                    continue; // unknown systemid, we will find it next time round.
                }
                Integer operId = ensureOperation(oper);
                Integer wsaId = ensureWSA(wsaTo, wsaAction);
                Integer clienID = ensureClientIP(clientIP);

                insertStmt.setInt(1, systemIdState.intValue());
                insertStmt.setInt(2, operId.intValue());
                insertStmt.setString(3, cpr);
                insertStmt.setTimestamp(4, timestamp);
                if (wsaId != null) {
                    insertStmt.setInt(5, wsaId.intValue());
                } else {
                    insertStmt.setNull(5, Types.INTEGER);
                }
                insertStmt.setInt(6, clienID);
                insertStmt.addBatch();
                rs.deleteRow();
            }
            try { rs.close(); } catch (SQLException e) {}
        }
        insertStmt.executeBatch();
        insertStmt.close();
    }


    protected Integer ensureWSA(String wsaTo, String wsaAction) throws SQLException {
        if (wsaTo == null && wsaAction == null) {
            return null;
        }
        WSAKey wsaKey = new WSAKey(wsaTo, wsaAction);
        Integer wsaId = wsa2PKcache.get(wsaKey);
        if (wsaId == null) {
            wsaId = lookupWsa(wsaKey);
        }
        if (wsaId == null) {
            PreparedStatement insertStmt = globalDB.prepareStatement(getInsertIntoWsaSql());
            insertStmt.setString(1, wsaTo);
            if (wsaAction != null) {
                insertStmt.setString(2, wsaAction);
            } else {
                insertStmt.setNull(2, Types.VARCHAR);
            }
            insertStmt.executeUpdate();
            insertStmt.close();
            wsaId = lookupWsa(wsaKey);
        }
        
        return wsaId;
    }


    private Integer lookupWsa(WSAKey wsaKey) throws SQLException {
        PreparedStatement selectStmt;
        if (wsaKey != null) {
            selectStmt = globalDB.prepareStatement(getSelectFromWsaWithActionSql());
        } else {
            selectStmt = globalDB.prepareStatement(getSelectFromWsaSql());
        }
        selectStmt.setString(1, wsaKey.wsaTo);
        if (wsaKey.wsaAction != null) {
            selectStmt.setString(2, wsaKey.wsaAction);
        }
        if (selectStmt.execute()) {
            ResultSet rs = selectStmt.getResultSet();
            try {
                if (rs.next()) {
                    int pk = rs.getInt(1);
                    Integer wsaId = new Integer(pk);
                    wsa2PKcache.put(wsaKey, wsaId);
                    return wsaId;
                }
            } finally {
                if (rs != null) rs.close();
            }
        }
        selectStmt.close();
        return null;
    }


    protected Integer ensureOperation(String oper) throws SQLException {
        Integer operId = operation2PK.get(oper);
        if (operId == null) {
            loadAllOperation();
            operId = operation2PK.get(oper);
            if (operId == null) {
                addOperation2GlobalState(oper);
                loadAllOperation();
                operId = operation2PK.get(oper);
            }
        }
        return operId;
    }


    private void loadAllOperation() throws SQLException {
        Statement stmt = globalDB.createStatement();
        if (stmt.execute("SELECT * FROM " + SOSIGWGLOBALLOG_OPERATION)) {
            ResultSet rs = stmt.getResultSet();
            while (rs.next()) {
                int pk = rs.getInt(1);
                String oper = rs.getString(2);
                operation2PK.put(oper, new Integer(pk));
            }
            try { rs.close(); } catch (SQLException e) {}
        }
        stmt.close();
    }

    private void addOperation2GlobalState(String oper) throws SQLException {
        PreparedStatement stmt = globalDB.prepareStatement(getInsertIntoOperation());
        stmt.setString(1, oper);
        stmt.executeUpdate();
        stmt.close();
    }


    protected Integer ensureClientIP(String clientIP) throws SQLException {
        Integer id = clientIP2PK.get(clientIP);
        if (id == null) {
            loadClientIP(clientIP);
            id = clientIP2PK.get(clientIP);
            if (id == null) {
                addClientIP2GlobalState(clientIP);
                loadClientIP(clientIP);
                id = clientIP2PK.get(clientIP);
            }
        }
        return id;
    }

    private void loadClientIP(String clientIP) throws SQLException {
        PreparedStatement stmt = globalDB.prepareStatement("SELECT pk FROM " + SOSIGWGLOBALLOG_CLIENTIP + " WHERE clientip=?");
        stmt.setString(1, clientIP);
        if (stmt.execute()) {
            ResultSet rs = stmt.getResultSet();
            if (rs.next()) {
                int pk = rs.getInt(1);
                clientIP2PK.put(clientIP, new Integer(pk));
            }
            rs.close();
        }
        stmt.close();
    }

    private void addClientIP2GlobalState(String clientIP) throws SQLException {
        PreparedStatement stmt = globalDB.prepareStatement(getInsertIntoClientIpSql());
        stmt.setString(1, clientIP);
        stmt.executeUpdate();
        stmt.close();
    }

    private void updateServeruidGlobalMapping(Connection globalDB, Set<String> systemIds) throws SQLException {
        int count = systemIds.size();
        StringBuffer sb = new StringBuffer();
        sb.append("SELECT pk, systemid FROM " + SOSIGWGLOBALLOG_SYSTEMID + " WHERE systemid in (");
        for (int i = 0; i < count; i++) {
            if (i > 0) {
                sb.append(",");
            }
            sb.append("?");
        }
        sb.append(")");
        PreparedStatement stmt = globalDB.prepareStatement(sb.toString());
        int i = 0;
        for (String systemId : systemIds) {
            stmt.setString(++i, systemId);
        }
        if (stmt.execute()) {
            ResultSet rs = stmt.getResultSet();
            while (rs.next()) {
                int pk = rs.getInt(1);
                String uid = rs.getString(2);
                systemidStates.put(uid, new Integer(pk));
            }
            try { rs.close(); } catch (SQLException e) {}
        }
        stmt.close();
    }

    private void addServeruid2GlobalState(String uid) throws SQLException {
        //Statement uidStmt = globalDB.createStatement();
//        uidStmt.executeUpdate(getInsertStmt5(String uid));
//        uidStmt.close();
        PreparedStatement stmt = globalDB.prepareStatement(getInsertIntoSystemIdSql());
        stmt.setString(1, uid);
        stmt.executeUpdate();
        stmt.close();
    }

    private Set<String> findUnknownLocalSystemIdSet(Connection localDB) throws SQLException {
        Set<String> newServeruids = new HashSet<String>();
        Statement stmt = localDB.createStatement();
        if (stmt.execute("SELECT DISTINCT systemid FROM " + SOSIGWLOCALLOG_SOSIGWLOG)) {
            ResultSet resultSet = stmt.getResultSet();
            while (resultSet.next()) {
                String uid = resultSet.getString(1);
                if (!systemidStates.containsKey(uid)) {
                    newServeruids.add(uid);
                }
            }
            try { resultSet.close(); } catch (SQLException e) {}
        }
        stmt.close();
        return newServeruids;
    }


    private static final String SELECT_PART =
        "SELECT log.cpr, log.moment, systemid.systemid, wsa.wsato, wsa.wsaaction, operation.operation, clientip.clientip"
         + " FROM " + SOSIGWGLOBALLOG_SOSIGWGLOBALLOG + " log"
         + " LEFT OUTER JOIN " + SOSIGWGLOBALLOG_SYSTEMID + " systemid ON log.systemid=systemid.pk"
         + " LEFT OUTER JOIN " + SOSIGWGLOBALLOG_WSA + " wsa ON log.wsa=wsa.pk"
         + " LEFT OUTER JOIN " + SOSIGWGLOBALLOG_OPERATION + " operation ON log.operation=operation.pk"
         + " LEFT OUTER JOIN " + SOSIGWGLOBALLOG_CLIENTIP + " clientip ON log.clientip=clientip.pk";

    
    private void addClause(StringBuffer where, List<Object> values, String field, String op, Object value) {
        if (value != null) {
            if (where.length() > 0) {
                where.append(" AND ");
            }
            where.append(field).append(" ").append(op).append(" ").append("?");
            values.add(value);
        }
    }
    
    public List<AuditLogEntry> search(AuditLogQuery query) {
        StringBuffer where = new StringBuffer();
        List<Object> values = new ArrayList<Object>();
        addClause(where, values, "CPR", "=", query.getCpr());
        if (query.getMinTimeStamp() != null) {
            addClause(where, values, "MOMENT", ">=", new Timestamp(query.getMinTimeStamp()));
        }
        if (query.getMaxTimeStamp() != null) {
            addClause(where, values, "MOMENT", "<=", new Timestamp(query.getMaxTimeStamp()));
        }
        StringBuffer sql = new StringBuffer();
        sql.append(SELECT_PART);
        if (where.length() > 0) {
            sql.append(" WHERE ").append(where.toString());
        }
        
        Connection con = null;
        PreparedStatement stmt = null;
        ResultSet resultSet = null;
        try {
            con = getGlobalDBConnection();
            stmt = con.prepareStatement(sql.toString());          
            assignValuesToQuery(stmt, values);
            if (stmt.execute()) {
                List<AuditLogEntry> result = new ArrayList<AuditLogEntry>();
                resultSet = stmt.getResultSet();
                while (resultSet.next()) {
                    AuditLogEntry entry = new AuditLogEntry();
                    entry.setCpr(resultSet.getString(1));
                    entry.setTimeStamp(resultSet.getTimestamp(2).getTime());
                    entry.setSystemID(resultSet.getString(3));
                    entry.setWsaTo(resultSet.getString(4));
                    entry.setWsaAction(resultSet.getString(5));
                    entry.setOperation(resultSet.getString(6));
                    entry.setClientIP(resultSet.getString(7));
                    result.add(entry);
                }
                return result;
            }
        } catch (Exception e) {
            log.error("Unable to connect to global DB", e);
        } finally {
            try {
                if (resultSet != null) {
                    resultSet.close();
                }
            } catch (SQLException e) {
                log.warn("Unable to close ResultSet", e);
            }
            try {
                if (stmt != null) {
                    stmt.close();
                }
            } catch (SQLException e) {
                log.warn("Unable to close stmt", e);
            }
            try {
                if (con != null) {
                    con.close();
                }
            } catch (SQLException e) {
                log.warn("Unable to close con", e);
            }
        }
        return null;
    }

    private void assignValuesToQuery(PreparedStatement stmt, List<Object> values) throws SQLException {
        for (int i = 0; i < values.size(); i++) {
            Object value = values.get(i);
            if (value instanceof Timestamp) {
                stmt.setTimestamp(i + 1, (Timestamp)value);
            } else if (value instanceof String) {
                stmt.setString(i + 1, (String)value);
            } else {
                throw new IllegalStateException("Unhandled type in querybuilder");
            }
        }
    }
}
