/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.browser;

import com.sun.org.apache.xerces.internal.dom.DOMInputImpl;
import com.trifork.sosigw.util.DOMUtil;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.ls.LSParser;

import java.io.*;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class BrowserSigningResponse {
    private static final Logger log = Logger.getLogger(BrowserSigningResponse.class);
    private static final String SHA1_DIGESTMETHOD_NS = "http://www.w3.org/2000/09/xmldsig#sha1";
    private static final String SHA1_SIGNATUREMETHOD_NS = "http://www.w3.org/2000/09/xmldsig#rsa-sha1";

    private byte[] message;
    private byte[] signatureValue;
    private byte[] certBytes;

    public BrowserSigningResponse(byte[] message) {
        this.message = message;
        parseMessage();
    }

    public byte[] getSignatureValue() {
        return signatureValue;
    }

    public byte[] getCertBytes() {
        return certBytes;
    }

    private void parseMessage() {
        if (log.isDebugEnabled()) {
            log.debug("Browser signing response received: " + asString(message));
        }
        final ByteArrayInputStream bais = new ByteArrayInputStream(message);

        Document doc = parseToDom(bais);

        verifyDigestMethod(doc);
        verifySignatureMethod(doc);

        final Node signatureValueNode = findNode(doc, "ds:SignatureValue", true);
        this.signatureValue = XmlUtil.fromBase64(signatureValueNode.getTextContent());

        final String certificate = findCertificateAtEndOfCertificateChain(doc);
        this.certBytes = XmlUtil.fromBase64(certificate);
    }

    private String findCertificateAtEndOfCertificateChain(Document doc) {
        final NodeList nodes = doc.getElementsByTagName("ds:X509Certificate");
        if (nodes.getLength() < 1) {
            throw new IllegalArgumentException("ds:X509Certificate" + " not found in sign response from browser");
        }

        Map<String, String> certsBase64BySubjectDN = new HashMap(nodes.getLength());
        List<X509Certificate> certs = new ArrayList<X509Certificate>(nodes.getLength());
        for (int i = 0; i < nodes.getLength(); i++) {
            final Node certNode = nodes.item(i);
            X509Certificate cert;
            try {
                cert = (X509Certificate)CertificateFactory.getInstance("X.509").generateCertificate(new ByteArrayInputStream(XmlUtil.fromBase64(certNode.getTextContent())));
            } catch (CertificateException e) {
                throw new RuntimeException(e);
            }

            log.debug("Putting cert with DN " + cert.getSubjectDN().getName() + " issued by " + cert.getIssuerDN().getName());
            certsBase64BySubjectDN.put(cert.getSubjectDN().getName(), certNode.getTextContent());
            certs.add(cert);
        }

        // we are looking for the end user cert
        // we know the certs are orderes as this: root <- issuing ca <- end user cert, with the root also pointing at "root"
        // so when we find a cert which noone "points to" (ie. names as issuer) we have the end user cert
        for (X509Certificate x509Certificate : certs) {
            final String issuerDN = x509Certificate.getIssuerDN().getName();
            certsBase64BySubjectDN.remove(issuerDN);
        }

        // now we expect the certs map to contain only one cert
        if (certsBase64BySubjectDN.size() > 1) {
            throw new IllegalArgumentException("Found more than one end user certificate in browser signing response: " + certsBase64BySubjectDN.size());
        }
        if (certsBase64BySubjectDN.size() < 1) {
            throw new IllegalArgumentException("Did not find end user certificate in browser signing response: " + certsBase64BySubjectDN.size());
        }

        return certsBase64BySubjectDN.values().iterator().next();
    }


    private void verifySignatureMethod(Document doc) {
        final Node foundNode = findNode(doc, "ds:SignatureMethod", false); // due to a peculiarity in the patched opensign code, the sign response contains two identical SignedInfo nodes each with a SignatureMethod
        final String algorithm = foundNode.getAttributes().getNamedItem("Algorithm").getTextContent();
        if (!SHA1_SIGNATUREMETHOD_NS.equals(algorithm)) {
            throw new IllegalDigestMethodException("Signature method must be " + SHA1_SIGNATUREMETHOD_NS + " but is " + algorithm);
        }
    }

    private void verifyDigestMethod(Document doc) {
        final Node foundNode = findNode(doc, "ds:DigestMethod", false); // due to a peculiarity in the patched opensign code, the sign response contains two identical SignedInfo nodes each with a DigestMethod
        final String algorithm = foundNode.getAttributes().getNamedItem("Algorithm").getTextContent();
        if (!SHA1_DIGESTMETHOD_NS.equals(algorithm)) {
            throw new IllegalDigestMethodException("Digest method must be " + SHA1_DIGESTMETHOD_NS + " but is " + algorithm);
        }
    }

    private Node findNode(Document doc, String qualifiedNodeName, boolean requireUnique) {
        final NodeList nodes = doc.getElementsByTagName(qualifiedNodeName);
        if (nodes.getLength() < 1) {
            throw new IllegalArgumentException(qualifiedNodeName + " not found in sign response from browser");
        }
        if (requireUnique && nodes.getLength() > 1) {
            throw new IllegalArgumentException(qualifiedNodeName + " found more than once in sign response from browser (found " + nodes.getLength() + " times). Expected it to be unique");
        }

        return nodes.item(0);
    }

    private Document parseToDom(final ByteArrayInputStream bais) {
        LSParser builder = DOMUtil.getDomLSParser();
        return builder.parse(new DOMInputImpl() {

            public InputStream getByteStream() {
                return bais;
            }
        });
    }

    private String asString(byte[] signedBytes) {
        try {
            return new String(signedBytes, "utf-8");
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException("Most surprisingly, utf-8 is unsupported", e);
        }
    }
}
