/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.browser;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.*;
import com.trifork.sosigw.auditlog.AuditLog;
import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.util.Util;
import com.trifork.sosigw.ws.IdCardSignerImpl;
import dk.sosi.gw._2007_09.DGWSFault;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.UserInfo;
import org.apache.log4j.Logger;
import dk.sosi.seal.xml.XmlUtil;

import javax.servlet.RequestDispatcher;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.Map;
import java.util.UUID;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class BrowserSigningServlet extends HttpServlet {
    private static final Logger log = Logger.getLogger(BrowserSigningServlet.class);
    private static final String APPLET_PAGE_URL = "/jsp/browsersignapplet.jsp";
    private static final String SIGNING_DONE_PAGE_URL = "/jsp/browsersigndone.jsp";
    private IdCardSigner signer;

    public static final int extraIEPayload = 1024;


    public void init() throws ServletException {
        setIdCardSigner(new IdCardSignerImpl());
    }
    

    protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        doPost(req, resp);
    }


    protected void doPost(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        try {
            execute(req, resp);
        } catch (Exception e) {
            String message = e.getMessage();
            String messageID = Util.getMessageID();
            if (messageID != null) {
                message = messageID + "\r\n" + message;
            }
            log.error(message, e);
            failed(resp, e.getMessage());
        }
    }

    private void execute(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
                //Generating a random SOSIGW uuid for use in the slalog when calling the STS later in the flow
        String messageId = "SOSIGW" + UUID.randomUUID().toString();
        Util.setActiveHttpRequest(req);
        Util.setMessageID(messageId);

        Map<String, String[]> parameterMap = getParameterMap(req);
        IDCardClusterCacheEntry entry = BrowserBasedSigning.findAndCheckIDCardCacheEntry(parameterMap);
        if (entry == null) {
            failed(resp, "ID-kort findes ikke");
            return;
        }
        //Now we have the right idcard for the user issuing the request.

        if (log.isDebugEnabled()) {
            debugDumpParameterMap(parameterMap);
        }

        switch (entry.getState()) {
            case AWAITING_SIGNED_BYTES:
                // The user wants to start a browser-signing session.
                startBrowserSigning(req, resp, entry);
                return;
            case AWAITING_BROWSER_RESPONSE:
                try {
                    receiveBrowserSigningResponse(req, resp, parameterMap, entry);    
                } catch (DGWSFault e) {
                    log.debug("Browser response handling failed: " + e.getMessage(), e);
                    req.getSession().setAttribute("exception", e);
                    failed(resp, e.getMessage());
                    return;
                }
                AuditLog.logIdcardChange(entry, req);
                return;
            case SIGNED:
                doneSigningIDCard(entry, req, resp);
                return;
            case AWAITING_STS_RESPONSE:
                Exception e = (Exception)req.getSession().getAttribute("exception");
                failed(resp, e != null ? e.getMessage() : null);
                return;
        }
        
        if (log.isInfoEnabled()) {
            log.info("Got request from browser for a card in an unhandled state. id=" + entry.getCacheKey() + " state=" + entry.getState());
        }
        failed(resp, "Internal error");
    }

    private void doneSigningIDCard(IDCardCacheEntry entry, HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        if (ConsoleLogin.isConsoleLogin(req, resp)) {
            ConsoleLogin.userLoggedIn(req, resp, entry);
        } else {
            RequestDispatcher requestDispatcher = req.getRequestDispatcher(SIGNING_DONE_PAGE_URL);
            requestDispatcher.forward(req, resp);
        }
    }

    private void debugDumpParameterMap(Map<String, String[]> parameterMap) {
        for (String key : parameterMap.keySet()) {
            String[] vals = parameterMap.get(key);
            log.debug("Parameter-key: " + key);
            for (String string : vals) {
                if (key.equals("message")) {
                    byte[] bytes = XmlUtil.fromBase64(string);
                    try {
                        log.debug("Parameter-value-for-message: " + new String(bytes, "UTF-8"));
                    } catch (UnsupportedEncodingException ignore) {
                        // ignore.
                    }
                } else {
                    log.debug("Parameter-value: " + string);
                }
            }
        }
    }

    @SuppressWarnings("unchecked")
    private Map<String, String[]> getParameterMap(ServletRequest req) {
        Map<String, String[]> parameterMap = req.getParameterMap();
        return parameterMap;
    }

    private void startBrowserSigning(HttpServletRequest req, ServletResponse resp, IDCardClusterCacheEntry entry) throws ServletException, IOException {
        byte[] bytesForSigning = entry.getBytesForSigning();
        String encodedDigest = XmlUtil.toBase64(bytesForSigning);
        req.setAttribute("digest.to.sign", encodedDigest);
        String textForUser = makeDisplayTextForUser(entry.getIdCard());
        req.setAttribute("text.for.user", XmlUtil.toBase64(textForUser.getBytes(Constants.UTF_8)));
        final IDCardCache idCardCache = Factory.getIDCardCache();
        idCardCache.updateEntryState(entry, IDCardState.AWAITING_BROWSER_RESPONSE);
        idCardCache.entryUpdatedLocally(entry);
        AuditLog.logIdcardChange(entry, req);
        RequestDispatcher requestDispatcher = req.getRequestDispatcher(APPLET_PAGE_URL);
        requestDispatcher.forward(req, resp);
    }

    private String makeDisplayTextForUser(UserIDCard idCard) {
        StringBuffer sb = new StringBuffer();
        UserInfo userInfo = idCard.getUserInfo();
        sb.append("IDKort for ").append(userInfo.getGivenName()).append(' ').append(userInfo.getSurName()).append("\n");
        sb.append("CPR: ").append(userInfo.getCPR()).append("\n");
        sb.append("Rolle: ").append(userInfo.getRole()).append("\n");
        sb.append("Stilling: ").append(userInfo.getOccupation()).append("\n");
        sb.append("Autorisation: ").append(userInfo.getAuthorizationCode()).append("\n");
        sb.append("E-mail: ").append(userInfo.getEmail()).append("\n");
        return sb.toString();
    }

    private void receiveBrowserSigningResponse(HttpServletRequest req, HttpServletResponse resp, Map<String, String[]> parameterMap, IDCardCacheEntry entry) throws ServletException, IOException, DGWSFault {
        String[] messages = parameterMap.get("message");
        if (messages == null) {
            throw new IllegalArgumentException("Must include message parameter");
        }
        byte[] message = XmlUtil.fromBase64(messages[0]);
        BrowserSigningResponse browserSigningResponse = new BrowserSigningResponse(message);
        byte[] signatureValue = browserSigningResponse.getSignatureValue();
        byte[] certBytes = browserSigningResponse.getCertBytes();
        signer.signIdCard(entry.getCacheKey(), signatureValue, certBytes);

        log.debug("Entry signed by browser: " + entry.toString());
        AuditLog.logIdcardChange(entry, req);
        doneSigningIDCard(entry, req, resp);
    }


    private void failed(HttpServletResponse resp, String reason) throws IOException {
        resp.setStatus(500);
        PrintWriter pw = resp.getWriter();
        pw.write("<html><head><title>Fejl</title></head><body>");
        pw.write("Der er sket en fejl. ");
        if (reason != null) {
            pw.write("<br />Fejlen var: " + reason);
        }
        String messageID = Util.getMessageID();
        if (messageID != null) {
            pw.write("<br /><br />Reference ID: " + messageID);
        }
        // To work around an issue found in IE; returning error code with "small" payloads
        // will make IE render a standard default page instead. To prevent this, 1k of spaces is
        // added to the payload.
        for(int i=extraIEPayload; i > 0; i--) {
            pw.write(' ');
        }
        pw.write("</body></html>");
    }
    
    public void setIdCardSigner(IdCardSigner signer) {
        this.signer = signer;
    }

}
