/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.browser;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.Constants;
import com.trifork.sosigw.api.IDCardCache;
import com.trifork.sosigw.api.IDCardCacheEntry;
import com.trifork.sosigw.api.IDCardState;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.util.SosiSeal;
import com.trifork.sosigw.util.Util;
import com.trifork.sosigw.ws.SosiGWFacadeBusiness;
import dk.sosi.gw._2007_09.DGWSFault;
import dk.sosi.gw._2007_09.RequestIdCardDigestForSigningResponse;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import org.apache.log4j.Logger;

import javax.servlet.*;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class ConsoleLogin extends HttpServlet implements Filter {
    private static final String CONSOLE = "CONSOLE_";

    private static final Logger log = Logger.getLogger(ConsoleLogin.class);
    
    public static final String CPR_PARAM = "CPR";
    public  static final String LOGIN = "LOGIN";
    //public static final String PURPOSE_PARAM = "PURPOSE";
    public static final String CONSOLE_COOKIE = "SOSIGWConsole";
    //public static final String CONSOLE_USER_NAMEID = "ConsoleUserNameID";
    public static final String CONSOLE_LOGIN_URL = "/jsp/consolelogin.jsp";
    private static final int SECRET_BITLENGTH = 256; 

    public void init(FilterConfig filterconfig) throws ServletException {
    }

    public void destroy() {
    }

    public void doFilter(ServletRequest servletrequest, ServletResponse servletresponse, FilterChain filterchain) throws IOException, ServletException {
        HttpServletRequest req = (HttpServletRequest)servletrequest;
        HttpServletResponse resp = (HttpServletResponse)servletresponse;

        boolean restrictedMode = StaticConfigImpl.getInstance().isBootstrapRestrictedMode() || StaticConfigImpl.getInstance().isDeveloperRestrictedMode();
        if (ConsoleLogin.isUserLoggedIn(req) || restrictedMode) {
            // User is okay.
            Util.setActiveHttpRequest(req);
            filterchain.doFilter(servletrequest, servletresponse);
            Util.setActiveHttpRequest(null);
        } else {
            // Initiate browserbased login.
            resp.sendRedirect(req.getContextPath() + CONSOLE_LOGIN_URL);
        }
    }
    
    protected static String getBrowserLoginUrl(String cpr, String secret) throws DGWSFault {
        UserInfo userInfo = RuntimeConfig.getConsoleUserInfo(cpr);
        if (userInfo == null) {
            return null;
        }
        String cvr = RuntimeConfig.getConsoleCVRNumber();
        String cvrOrgName = RuntimeConfig.getConsoleCVRName();
        String type = SubjectIdentifierTypeValues.CVR_NUMBER;

        String itSystemName = SosiSeal.getSosigwITSystemName();
        UserIDCard userIDCard = new UserIDCard(StaticConfigImpl.getInstance().getSupportedDGWSVersion(),
                AuthenticationLevel.MOCES_TRUSTED_USER, "SOSI-GW", new SystemInfo(new CareProvider(type, cvr,
                        cvrOrgName), itSystemName), userInfo, null, null, null, null);
        RequestIdCardDigestForSigningResponse signingResponse = SosiGWFacadeBusiness.requestIdCardDigestForSigningImpl(userIDCard, secret);
        return signingResponse.getBrowserUrl();
    }
    
    public static boolean isUserLoggedIn(HttpServletRequest req) {
        final IDCardCache idCardCache = Factory.getIDCardCache();
        final String cacheKey = getSosigwCookie(req);
        final IDCardCacheEntry entry = idCardCache.findIDCardOrNull(cacheKey);
        return entry != null && entry.getState().equals(IDCardState.SIGNED);
    }
    
    private static String getSosigwCookie(HttpServletRequest req) {
        Cookie[] cookies = req.getCookies();
        if (cookies != null) {
            for (Cookie cookie : cookies) {
                if (CONSOLE_COOKIE.equals(cookie.getName())) {
                    return cookie.getValue();
                }
            }
        }
        return null;
    }
    
    private static void setSosigwCookie(HttpServletResponse resp, String value) {
        Cookie cookie = new Cookie(CONSOLE_COOKIE, value);
        cookie.setPath("/");
        resp.addCookie(cookie);
    }
    

    protected void doPost(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        try {
            Util.setActiveHttpRequest(req);
            String cpr = req.getParameter(CPR_PARAM);
            if (cpr != null) {
                cpr = cpr.replaceAll("\\'", "");
                try {
                    String secret = CONSOLE + Util.makeSecretKey(SECRET_BITLENGTH);
                    String browserLoginUrl = getBrowserLoginUrl(cpr, secret);
                    if (browserLoginUrl != null) {
                        setSosigwCookie(resp, secret);
                        resp.sendRedirect(browserLoginUrl);
                        return;
                    }
                } catch (DGWSFault e) {
                    log.error("Console login problem", e);
                    // ignore, fall through to retry.
                }
            }

        } finally {
            Util.setActiveHttpRequest(null);
        }

        // Not logged in. Redirect to login page.
        resp.sendRedirect(req.getContextPath() + CONSOLE_LOGIN_URL);
    }

    public static boolean isConsoleLogin(HttpServletRequest req, HttpServletResponse resp) {
        final String cacheKey = getSosigwCookie(req);
        final IDCardCache idCardCache = Factory.getIDCardCache();
        return cacheKey != null && cacheKey.startsWith(CONSOLE) && idCardCache.findIDCardOrNull(cacheKey) != null;
    }

    public static void userLoggedIn(HttpServletRequest req, HttpServletResponse resp, IDCardCacheEntry entry) throws ServletException, IOException {
        resp.sendRedirect(req.getContextPath() + Constants.CONSOLE_LOGIN_DONE_PAGE_URL);
    }
}
