/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster;

import com.sun.org.apache.xerces.internal.dom.DOMInputImpl;
import com.trifork.sosigw.api.Constants;
import com.trifork.sosigw.api.IDCardCacheEntry;
import com.trifork.sosigw.api.IDCardState;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.util.DOMUtil;
import com.trifork.sosigw.util.SosiSeal;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.SecurityTokenRequest;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.modelbuilders.IDCardModelBuilder;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.ls.LSParser;

import java.io.*;
import java.util.Comparator;

/**
 * Keeps track of the life cycle of an IDCard. Cluster-aware.
 * 
 * @author Morten Grouleff <mgr@trifork.com>
 *
 */
public class IDCardClusterCacheEntry implements Serializable, IDCardCacheEntry {
    private static final long serialVersionUID = 3L;
    private static final Logger log = Logger.getLogger(IDCardClusterCacheEntry.class);

    private final String cacheKey;
    private final int hashCode;
    private transient long stateHash;
    private byte[] idCardAsXmlBytes;
    private transient UserIDCard idCard;
    private IDCardState state = IDCardState.UNKNOWN;
    private String browserSigningSecret;
    private byte[] bytesForSigning;
    private long lastmodified;
    
    
    public String toString() {
        return "IDCardCacheEntry("
        + "cacheKey=" + cacheKey
        + (idCard != null ? (" cpr=" + idCard.getUserInfo().getCPR()) : "")
        + ": state=" + state 
        + ")";
    }
    
    public IDCardClusterCacheEntry(String cacheKey) {
        this.cacheKey = cacheKey;
        this.hashCode = calcHashCode();
        stateChanged();
    }

    public UserIDCard getIdCard() {
        return idCard;
    }
    
    public synchronized void setIdCard(UserIDCard idCard) {
        if (idCard != null) {
            // Sanity check
            if (!AuthenticationLevel.MOCES_TRUSTED_USER.equals(idCard.getAuthenticationLevel())) {
                throw new IllegalArgumentException("IDcard in cache must always be MOCES_TRUSTED_USER, also when unsigned");
            }
        }
        this.idCard = idCard;
        idCardChanged();
    }
    
    private void idCardChanged() {
        if (idCard != null) {
            Node domElementForIDCard = calcDomElementForIDCard();
            String nodeAsString = XmlUtil.node2String(domElementForIDCard);
            try {
                this.idCardAsXmlBytes = nodeAsString.getBytes(Constants.UTF_8);
            } catch (UnsupportedEncodingException e) {
                log.error("UTF-8 is missing", e);
                throw new IllegalStateException(e);
            }
        } else {
            this.idCardAsXmlBytes = null;
        }
    }

    protected void stateChanged() {
        if (state.equals(IDCardState.SIGNED)) {
            // We have been signed. Get rid of temp info needed while signing to save space-
            this.browserSigningSecret = null;
            this.bytesForSigning = null;
        } else if (state.equals(IDCardState.REVOKED)) {
            idCard = null;
        }
        stateHash = (long)Math.pow(10, state.ordinal() * 2);
    }

    private Node calcDomElementForIDCard() {
        SecurityTokenRequest securityTokenRequest = SosiSeal.getSosiSealFactory().createNewSecurityTokenRequest();
        securityTokenRequest.setIDCard(idCard);
        Document doc = securityTokenRequest.serialize2DOMDocument();

        Element element = DOMUtil.getIDCardElement(doc);
        return element;
    }

    private int calcHashCode() {
        return cacheKey != null ? cacheKey.hashCode() : 0;
    }

    /*package*/ synchronized void setState(IDCardState state) {
        if (log.isDebugEnabled()) {
            log.debug("setState(" + cacheKey + ", " + state + ")");
        }
        this.state = state;
        stateChanged();
    }
    
    public IDCardState getState() {
        return state;
    }


    public boolean equals(Object obj) {
        if (obj instanceof IDCardClusterCacheEntry) {
            IDCardClusterCacheEntry other = (IDCardClusterCacheEntry)obj;
            if (cacheKey == null) {
                return other.cacheKey == null;
            }
            return cacheKey.equals(other.cacheKey);
        }
        return false;
    }
    
    public long stateHashCode() {
        return stateHash;
    }

    public int hashCode() {
        return this.hashCode;
    }

    public synchronized void setSecret(String secret) {
        this.browserSigningSecret = secret;
    }
    
    public boolean validateSecret(String secret) {
        return this.browserSigningSecret != null && this.browserSigningSecret.equals(secret);
    }

    public String getCacheKey() {
        return cacheKey;
    }

    public synchronized void setBytesForSigning(byte[] bytesForSigning) {
        this.bytesForSigning = bytesForSigning;
    }

    public byte[] getBytesForSigning() {
        return bytesForSigning;
    }
    
    public Node getDOMelement() {
        Document doc = parseBytesAsDom();
        Element element = DOMUtil.getIDCardElement(doc);
        return element.cloneNode(true);

    }

    private Document parseBytesAsDom() {
        LSParser domLSParser = DOMUtil.getDomLSParser();
        DOMInputImpl inputImpl = new DOMInputImpl() {

            public InputStream getByteStream() {
                return new ByteArrayInputStream(idCardAsXmlBytes);
            }
        };
        inputImpl.setEncoding(Constants.UTF_8);
        Document doc = domLSParser.parse(inputImpl);
        return doc;
    }

    public synchronized void setLastModified(long now) {
        this.lastmodified = now;
    }
    public long getLastmodified() {
        return lastmodified;
    }
    
    public static class IDCardAgeComparator implements Comparator<IDCardClusterCacheEntry> {
        public int compare(IDCardClusterCacheEntry o1, IDCardClusterCacheEntry o2) {
            long l1;
            synchronized (o1) {
                l1 = o1.getLastmodified();   
            }
            long l2;
            synchronized (o2) {
                l2 = o2.getLastmodified();
            }
            long l = l1 - l2;
            return l < 0 ? -1: (l > 0 ? 1 : 0);
        }
    }
    
    private void readObject(ObjectInputStream ois) throws IOException, ClassNotFoundException {
        ois.defaultReadObject();
        if (idCardAsXmlBytes != null && idCardAsXmlBytes.length > 0) {
            Document doc = parseBytesAsDom();
            idCard = (UserIDCard)new IDCardModelBuilder().buildModel(doc);
        } else {
            idCard = null;
            idCardAsXmlBytes = null;
        }
        stateHash = (long)Math.pow(10, state.ordinal() * 2); // Transient to keep backward compatability
    }

    public boolean isTimedOutWaitingForSigning() {
        return (state.equals(IDCardState.AWAITING_BROWSER_RESPONSE)
                || state.equals(IDCardState.AWAITING_SIGNED_BYTES)
                || state.equals(IDCardState.AWAITING_STS_RESPONSE))
                && RuntimeConfig.getIdCardSigningTimeout() < System.currentTimeMillis() - lastmodified;
    }
}
