/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster;

import com.trifork.sosigw.api.*;
import com.trifork.sosigw.cluster.intercom.TimeSource;
import org.apache.log4j.Logger;

import java.util.*;
import java.util.concurrent.atomic.AtomicLong;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class IDCardClusterCacheImpl implements IDCardCache, IDCardClusterCache {
    private static final Logger log = Logger.getLogger(IDCardClusterCacheImpl.class);

    private static final long OSOLETION_AGE = Constants.MAX_IDCARD_LIFETIME_HOURS  * 60 * 60 * 1000; // in miliseconds. 

    private Map<String, IDCardClusterCacheEntry> map = new HashMap<String, IDCardClusterCacheEntry>();
    private AtomicLong hash = new AtomicLong();
    private PriorityQueue<IDCardClusterCacheEntry> obsoletionQueue = new PriorityQueue<IDCardClusterCacheEntry>(10, new IDCardClusterCacheEntry.IDCardAgeComparator());

    private IDCardCacheNotification notificationListener;

    private final AtomicLong arrivedFromPeer = new AtomicLong();
    private final AtomicLong addedFromPeer = new AtomicLong();
    private final AtomicLong replacedFromPeer = new AtomicLong();
    private final AtomicLong ignoredFromPeer = new AtomicLong();
    
    public IDCardClusterCacheImpl(IDCardCacheNotification notificationListener) {
        this.notificationListener = notificationListener;
    }
    
    public IDCardClusterCacheEntry findIDCardOrNull(String cacheKey) {
        if (cacheKey == null) {
            return null;
        }
        IDCardClusterCacheEntry entry = findIDCardOrNull_0(cacheKey);
        if (log.isDebugEnabled()) {
            log.debug("findIDCardOrNull(" + cacheKey + "): " + entry);
        }
        return entry;
    }

    private synchronized IDCardClusterCacheEntry findIDCardOrNull_0(String cacheKey) {
        IDCardClusterCacheEntry entry = map.get(cacheKey);
        return entry;
    }

    public synchronized IDCardClusterCacheEntry ensureIDCard(String cacheKey) {
        IDCardClusterCacheEntry entry = findIDCardOrNull_0(cacheKey);
        if (entry == null) {
            if (log.isDebugEnabled()) {
                log.debug("Adding new entry for " + cacheKey);
            }
            entry = new IDCardClusterCacheEntry(cacheKey);
            entry.setLastModified(TimeSource.currentTimeMillis());
            addNewEntry(entry);
        }
        return entry;
    }

    private synchronized void addNewEntry(IDCardClusterCacheEntry entry) {
        // Call here, as this keeps the cache-size down, and we are inside login here
        // Which has no response-time limits.
        removeObsoleteEntries(false);
        if (entry != null) {
            IDCardClusterCacheEntry oldEntry = map.put(entry.getCacheKey(), entry);
            obsoletionQueue.offer(entry);
            if (oldEntry != null) {
                hash.addAndGet(-1 * oldEntry.stateHashCode());
                hash.addAndGet(-1 * oldEntry.hashCode());
            }
            hash.addAndGet(entry.stateHashCode());
            hash.addAndGet(entry.hashCode());
        }
    }

    private synchronized void replaceEntry(IDCardClusterCacheEntry newEntry) {
        String cacheKey = newEntry.getCacheKey();
        IDCardClusterCacheEntry oldEntry = map.get(cacheKey);
        if (oldEntry != null) {
            obsoletionQueue.remove(oldEntry);
        }
        addNewEntry(newEntry);
    }

    public synchronized void arrivedFromPeer(IDCardClusterCacheEntry peerEntry, int senderTimeOffset) {
        arrivedFromPeer.incrementAndGet();
        peerEntry.setLastModified(peerEntry.getLastmodified() - senderTimeOffset); // Adjust to local time skew...
        IDCardClusterCacheEntry entry = findIDCardOrNull(peerEntry.getCacheKey());
        // Decide which to keep based on the state-pair.
        if (entry == null) {
            if (log.isDebugEnabled()) {
                log.debug("Had no entry, new entry is " + peerEntry);
            }
            addedFromPeer.incrementAndGet();
            addNewEntry(peerEntry);
        } else if (entry.getLastmodified() < peerEntry.getLastmodified()) {
            if (log.isDebugEnabled()) {
                log.debug("Replacing entry " + peerEntry + ", oldentry.time=" + entry.getLastmodified() + " newentry.time=" + peerEntry.getLastmodified() + " (skew " + senderTimeOffset + ")");
            }
            replacedFromPeer.incrementAndGet();
            replaceEntry(peerEntry);
        } else {
            // Okay, we got en entry that is older than what we have.
            if (log.isDebugEnabled()) {
                log.debug("Old replacement entry " + peerEntry + ", my.entry.time=" + entry.getLastmodified() + " incoming.entry.time=" + peerEntry.getLastmodified() + " (skew " + senderTimeOffset + ")");
            }
            long diff = entry.getLastmodified() - (peerEntry.getLastmodified());
            if (diff < 5000 && entry.getState() != IDCardState.REVOKED &&
                    peerEntry.getState().ordinal() > entry.getState().ordinal()) {
                // My entry is not revoked. The incoming entry has a "higher" state, which means "more signed" than mine. Accept it.
                if (log.isDebugEnabled()) {
                    log.debug("Replacing entry " + peerEntry + ", oldentry.time=" + entry.getLastmodified() + " newentry.time=" + peerEntry.getLastmodified() + " (skew " + senderTimeOffset + ")");
                }
                replacedFromPeer.incrementAndGet();
                replaceEntry(peerEntry);
            } else {
                ignoredFromPeer.incrementAndGet();
                if (log.isDebugEnabled()) {
                    log.debug("Ignored replacement entry " + peerEntry);
                }
            }
        }
    }

    public synchronized void entryUpdatedLocally(IDCardClusterCacheEntry entry) {
        entry.setLastModified(TimeSource.currentTimeMillis());
        obsoletionQueue.remove(entry);
        obsoletionQueue.offer(entry);
        if (notificationListener != null) {
            notificationListener.entryUpdatedLocally(entry);
        }
    }
    
    @Override
    public IDCardState updateEntryState(IDCardCacheEntry entry, IDCardState newState) {
        synchronized (entry) {
            IDCardClusterCacheEntry cacheEntry = (IDCardClusterCacheEntry)entry;
            IDCardState oldState = entry.getState();
            hash.addAndGet(-1 * cacheEntry.stateHashCode());
            cacheEntry.setState(newState);
            hash.addAndGet(cacheEntry.stateHashCode());
            return oldState;
        }
    }

    public synchronized List<String> getContentsAsList() {
        removeObsoleteEntries(true);
        List<String> contents = new ArrayList<String>(map.keySet());
        return contents;
    }
    
    public synchronized long size() {
        return map.size();
    }
    
    public long contentHash() {
        return hash.get();
    }
    
    public void setNotificationListener(IDCardCacheNotification notificationListener) {
        log.debug("notificationlistener is set to " + notificationListener);
        this.notificationListener = notificationListener;
    }
    
    private synchronized void removeObsoleteEntries(boolean all) {
        int count = 0;
        while (all || count < 32) { // Avoid doing way too much at a time. There is likely to be another chance shortly.
            IDCardClusterCacheEntry entry = obsoletionQueue.peek();
            if (log.isDebugEnabled()) {
                log.debug("Checking " + entry);
            }
            if (entry == null) {
                log.debug("Entry was null");
                break;
            }
            long age = TimeSource.currentTimeMillis() - entry.getLastmodified();
            if (age < OSOLETION_AGE) {
                log.debug("Entry not expired");
                break;
            }
            if (log.isDebugEnabled()) {
                log.debug("Obsoleting " + entry.toString() + " with age " + age);
            }
            entry = obsoletionQueue.poll();
            map.remove(entry.getCacheKey());
            hash.addAndGet(-1 * entry.hashCode());
            hash.addAndGet(-1 * entry.stateHashCode());
            count++;
        }
        if (log.isDebugEnabled()) {
            log.debug("Obsoleted " + count + " IDCard-entries from cache. Queue contains " + obsoletionQueue.size());
        }
    }

    @Override
    public long getArrivedFromPeer() {
        return arrivedFromPeer.get();
    }

    @Override
    public long getAddedFromPeer() {
        return addedFromPeer.get();
    }

    @Override
    public long getReplacedFromPeer() {
        return replacedFromPeer.get();
    }

    @Override
    public long getIgnoredFromPeer() {
        return ignoredFromPeer.get();
    }
}
