/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.log4j.Logger;

public class ClusterState {
    
    private static final Logger log = Logger.getLogger(ClusterState.class);
    
    private final Map<String, Long> syncRequestedAt = new ConcurrentHashMap<String, Long>(16);
    private Map<String, Long> lastSenderSeqno = new ConcurrentHashMap<String, Long>(16);
    private final Map<String, TimeState> timeStates = new HashMap<String, TimeState>(16);
    private Map<String, Long> lastSenderPing = new ConcurrentHashMap<String, Long>(16);

    private boolean firstSyncCompleted = false;
    private long lastReSyncAt = System.currentTimeMillis();
    private volatile boolean bootstrapped;

    private final long warnSilentPeerTime;
    private final long resyncTimeout;
    private final long maxSilentPeerTime;
    private final boolean hasAccurateTime;
    private boolean killed = false;

    public ClusterState(long warnSilentPeerTime, long maxSilentPeerTime, long resyncTimeout, boolean hasAccurateTime) {
        this.warnSilentPeerTime = warnSilentPeerTime;
        this.maxSilentPeerTime = maxSilentPeerTime;
        this.resyncTimeout = resyncTimeout;
        this.hasAccurateTime = hasAccurateTime;
    }

    public ClusterState() {
        this(5, 300, 5000, false);
    }

    public boolean isFirstSyncCompleted() {
        return firstSyncCompleted;
    }
    
    public void setFirstSyncCompleted(boolean firstSyncCompleted) {
        this.firstSyncCompleted = firstSyncCompleted;
    }
    
    public long getLastReSyncAt() {
        return lastReSyncAt;
    }
    
    public void setLastReSyncAt(long lastReSyncAt) {
        this.lastReSyncAt = lastReSyncAt;
    }

    public Map<String, Long> getSyncRequestedAt() {
        return syncRequestedAt;
    }
    
    public Map<String, Long> getLastSenderPing() {
        return lastSenderPing;
    }
    
    public Map<String, Long> getLastSenderSeqno() {
        return lastSenderSeqno;
    }
    
    public Map<String, TimeState> getTimeStates() {
        return timeStates;
    }

    public void clear(String senderUID) {
        lastSenderPing.remove(senderUID);
        syncRequestedAt.remove(senderUID);
        lastSenderSeqno.remove(senderUID);
        timeStates.remove(senderUID);
    }
    
    public void checkBootstrap() {
        if (!bootstrapped 
                && (isFirstSyncCompleted() // Completed initial sync with someone
                        || (isSingleMember() && System.currentTimeMillis() - getLastReSyncAt() > (warnSilentPeerTime))  // No one said a word.
                        || (getSyncRequestedAt().isEmpty() && System.currentTimeMillis() - getLastReSyncAt() > (maxSilentPeerTime + resyncTimeout)) // Timed out sync'ing.
                        )) {
                        
            // Either we completed a sync, or there was silence for a long while. Lets say we're in sync now.
            bootstrapped = true;
            if (isSingleMember()) {
                log.info("IntercomImpl now Bootstrapped! (No other members of cluster seen yet)");
            } else {
                log.info("IntercomImpl now Bootstrapped! (Is in contact with " + getLastSenderPing().size() + " other members of cluster)");
            }
        }
    }
    
    public boolean isSingleMember() {
        return lastSenderPing.isEmpty(); 
    }

    public boolean isBootstrapped() {
        if (!bootstrapped) {
            checkBootstrap();
        }
        return bootstrapped;
    }

    public boolean hasTimeState(String senderUID) {
        if (hasAccurateTime) {
            return true;
        }
        TimeState state = getTimeStates().get(senderUID);
        return state != null ? state.getHasState() : false; 
    }

    public boolean isKilled() {
        return killed;
    }
    
    public void kill() {
        killed = true;
    }
}
