/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.*;
import com.trifork.sosigw.config.StaticConfigApi;
import com.trifork.sosigw.config.StaticConfigImpl;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.net.*;
import java.util.Enumeration;
import java.util.Random;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class IntercomImpl implements IDCardCacheNotification, RuntimeConfigRequest, Intercom {
    
    private static final Logger log = Logger.getLogger(IntercomImpl.class); 

    private final StaticConfigApi STATIC_CONF = StaticConfigImpl.getInstance();
    
    // config:
    private static final boolean IPV4_ONLY = true;
    private final int MAX_MESSAGE_BYTESIZE = STATIC_CONF.getClusterTuningParam("max.message.size", 60000);
    private final int RESYNC_CONNECT_TIMEOUT = STATIC_CONF.getClusterTuningParam("resync.connect.timeout", 2000);  // Fairly low, as there is no point in responding to old resync requests.
    private final int RESYNC_SOTIMEOUT = STATIC_CONF.getClusterTuningParam("resync.receive.timeout", 5000);  // How long to wait for data on resync socket
    private final int PERIODIC_ALIVE_INTERVAL = STATIC_CONF.getClusterTuningParam("periodic.alive.interval", 1000); //How long to sleep between periodic pings to other members. 
    private final long WARN_SILENT_PEER_TIME = PERIODIC_ALIVE_INTERVAL * STATIC_CONF.getClusterTuningParam("warn.peer.silent.times", 5);
    private final long MAX_SILENT_PEER_TIME = PERIODIC_ALIVE_INTERVAL *  STATIC_CONF.getClusterTuningParam("remove.peer.silent.times", 300);
    private final int RECEIVER_THREAD_COUNT =  STATIC_CONF.getClusterTuningParam("receiver.thread.count", 5);
    
    private final int SENT_WINDOW_WIDTH =  STATIC_CONF.getClusterTuningParam("sent.window.width", 1000);
    private final int RESYNC_SOCKET_LISTEN_BACKLOG =  STATIC_CONF.getClusterTuningParam("resync.socket.backlog", 5);
    private final boolean HAS_ACURATE_TIME = (1 == STATIC_CONF.getClusterTuningParam("acurate.time.flow", 0));

    // state:
    private final String myUID;
    private final MulticastHandler netHandler;
    private final ServerSocket resyncSocket;
    private final SlidingWindow<String> sentWindow = new SlidingWindow<String>(SENT_WINDOW_WIDTH);
    
    private Thread[] receiverThread = new Thread[RECEIVER_THREAD_COUNT];
    private Thread[] syncRecvThread = new Thread[RECEIVER_THREAD_COUNT];
    private Thread syncReqThread;
    
    private IDCardClusterCache clusterCache;
    private Thread periodicAliveThread;
    
    private final ClusterState state = new ClusterState(WARN_SILENT_PEER_TIME, MAX_SILENT_PEER_TIME, RESYNC_SOTIMEOUT, HAS_ACURATE_TIME);
    private ConfigSender configSender;
    private ProtocolReceiver protocolReceiver;

    public IntercomImpl() throws IOException {
        netHandler = new MulticastHandler(STATIC_CONF);

        final IDCardCache idCardCache = Factory.getIDCardCache();
        
        if (idCardCache instanceof IDCardClusterCache) {
            clusterCache = (IDCardClusterCache)idCardCache;
        } else {
            String msg = "My cache must implement IDCardClusterCache";
            log.error(msg);
            throw new IllegalStateException(msg);
        }

        InetAddress resyncInetAddr = netHandler.getAddressOfLocalInterface();//selectAddressOfLocalInterface();
        this.myUID = (resyncInetAddr != null ? resyncInetAddr.getHostAddress() : "NoNet") + "." + Long.toHexString(new Random().nextLong());

        configSender = new ConfigSender(myUID, netHandler);

        if (resyncInetAddr != null) {
            resyncSocket = new ServerSocket(0, RESYNC_SOCKET_LISTEN_BACKLOG, resyncInetAddr);
        } else {
            resyncSocket = null;
        }
        
        protocolReceiver = new ProtocolReceiver(myUID, state, sentWindow, clusterCache, RESYNC_SOTIMEOUT, RESYNC_CONNECT_TIMEOUT, HAS_ACURATE_TIME, configSender);
            
        for (int i = 0; i < receiverThread.length; i++) {
            receiverThread[i] = new Thread(new Receiver(), "Intercom.Receiver");
//            receiverThread[i].setPriority(Thread.NORM_PRIORITY + 2);
            receiverThread[i].setDaemon(true);
            receiverThread[i].start();
        }
        if (resyncSocket != null) {
            PeriodicAlive pa = new PeriodicAlive(myUID, netHandler, sentWindow, state, STATIC_CONF);
            periodicAliveThread = new Thread(pa, "Intercom.PeriodicAlive");
//            periodicAliveThread.setPriority(Thread.MAX_PRIORITY); // We need as precise time as possible.
            periodicAliveThread.setDaemon(true);
            periodicAliveThread.start();
            syncReqThread = new Thread(new SyncRequester(state, resyncSocket, netHandler), "Intercom.SyncRequester");
            syncReqThread.setDaemon(true);
            syncReqThread.start();
            for (int i = 0; i < receiverThread.length; i++) {
                syncRecvThread[i] = new Thread(new SyncReceiver(myUID, resyncSocket, state, STATIC_CONF, protocolReceiver), "Intercom.SyncReceiver");
//                syncRecvThread[i].setPriority(Thread.NORM_PRIORITY + 1);
                syncRecvThread[i].setDaemon(true);
                syncRecvThread[i].start();
            }
            
            pa.sendPeriodicAlive();
            pa.sendPeriodicAlive();
        }
    }
    
    public boolean isBootstrapped() {
        return state.isBootstrapped();
    }
    
    public void entryUpdatedLocally(IDCardCacheEntry peerEntry) {
        try {
            log.debug("Entry updated locally: " + peerEntry);
            long seqNo = sentWindow.reserve();
            byte[] byteArray = Protocol.encodeVersion1IDCardEntryAsBytes(peerEntry, seqNo, myUID);
            sentWindow.add(seqNo, peerEntry.getCacheKey());
            log.debug("About to send " + byteArray.length + " bytes");
            netHandler.sendMessage(byteArray);
            log.debug("Sent " + byteArray.length + " bytes okay");
        } catch (IOException e) {
            log.error("Error encoding entry to stream", e);
        }
    }
    
    public void configUpdatedLocally() {
        configSender.sendCurrentRuntimeConfigState();
    }

    public void kill() {
        if (log.isDebugEnabled()) {
            log.debug("Being killed!");
        }
        state.kill();
        for (int i = 0; i < receiverThread.length; i++) {
            receiverThread[i].interrupt();
        }
        if (syncReqThread != null) {
            syncReqThread.interrupt();
        }
        for (int i = 0; i < syncRecvThread.length; i++) {
            if (syncRecvThread[i] != null) {
                syncRecvThread[i].interrupt();
            }
        }
        if (periodicAliveThread != null) {
            periodicAliveThread.interrupt();
        }
        netHandler.close();
        try {
            if (resyncSocket != null) {
                resyncSocket.close();
            }
        } catch (IOException e) {
            if (log.isDebugEnabled()) {
                log.debug("error while closing resyncSocket", e);
            }
        }
    }

    public String getUID() {
        return myUID;
    }
    
    private class Receiver extends Worker {
        private final byte[] buffer;

        public Receiver() {
            super(state);
            buffer = new byte[MAX_MESSAGE_BYTESIZE];
        }
        
        protected void run0() throws InterruptedException, IOException, ClassNotFoundException {
            ObjectInputStream objectInputStream = null; 
            try {
                InputStream inputStream = netHandler.recvMessage(buffer);
                objectInputStream = new ObjectInputStream(inputStream);
                protocolReceiver.receive(objectInputStream);
            } finally {
                if (objectInputStream != null) {
                    objectInputStream.close();
                }
            }
        }
    }

    public boolean isSingleMember() {
        return state.isSingleMember(); 
    }

    public void requestConfig(long timestampOfKnownConfig) {
        configSender.requestConfig(timestampOfKnownConfig);
    }

}
