/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.DatagramPacket;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.Enumeration;

import org.apache.log4j.Logger;

import com.trifork.sosigw.config.StaticConfigApi;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class MulticastHandler implements NetHandler {
    private static final Logger log = Logger.getLogger(MulticastHandler.class);
    private static final long PACKET_MIN_INTERVAL = 5; // In ms.
    private static final int PACKET_BUFFER_SIZE = 1024 * 1024; // bytes.
    private final MulticastSocket mcs;
    private final InetAddress group;
    private final int port;
    private long nextSendNotBefore = 0;
    private final StaticConfigApi config;

    /**
     * Use settings from ConfigAPI to make the instance.
     * 
     * @throws IOException
     */
    public MulticastHandler(StaticConfigApi config) throws IOException {
        this.config = config;
        this.port = config.getMulticastPort();
        this.group = config.getMulticastIpAddress();
        this.mcs = new MulticastSocket(port);

        
        if (useBindToInterfaceByProtocol()) {
            log.debug("useBindToInterfaceByProtocol");
            bindToInterfaceByProtocol();
        } else {
            log.debug("useBindToInterfaceByPrefix");
            bindToInterfaceByPrefix();
        }
        
        mcs.joinGroup(group);
        int buf = PACKET_BUFFER_SIZE;
        while (true) {
            mcs.setReceiveBufferSize(buf);
            if (mcs.getReceiveBufferSize() < buf) {
                buf = buf / 2;
            } else {
                log.info("MulticastHandler: set ReceiveBufferSize to " + buf);
                break;
            }
        }
    }
    
    public InetAddress getAddressOfLocalInterface() throws SocketException {
        return mcs.getInterface();
    }

    public static Boolean useBindToInterfaceByProtocol() {
        return Boolean.valueOf(System.getProperty("bindToInterfaceByProtocol", "false"));
    }
    
    public static String getBindToInterfaceByProtocolPreferredPrefix() {
        return System.getProperty("preferredPrefix", null);
    }

    public void close() {
        if (log.isDebugEnabled()) {
            log.debug("MulticastHandler: Closing");
        }
        mcs.close();
    }
    
    protected void bindToInterfaceByProtocol() throws SocketException {
        if(this.group == null && this.mcs == null) {
            return;
        }
        
        bindToInterfaceByProtocol(this.group, this.mcs);
    }
    
    static void bindToInterfaceByProtocol(InetAddress group, MulticastSocket mcs) throws SocketException {  
        log.debug("Trying to bind to network interface using protocol.");
        
        final String bindToInterfaceByProtocolPreferredPrefix = getBindToInterfaceByProtocolPreferredPrefix();
        boolean boundToInterfaceByProtocol = false;
        boolean boundToInterfaceByProtocolWithPreferredPrefix = false;
        
        boolean isGroupInet6 = group instanceof Inet6Address;
        
        Enumeration<NetworkInterface> networkInterfaces = NetworkInterface.getNetworkInterfaces();
        while (networkInterfaces.hasMoreElements()) {
            NetworkInterface netIf = networkInterfaces.nextElement();
            
            Enumeration<InetAddress> addresses = netIf.getInetAddresses();
            while (addresses.hasMoreElements()) {
                InetAddress addr = addresses.nextElement();
                
                boolean isAddressInet6 = addr instanceof Inet6Address;
                boolean isAddressInet4 = !isAddressInet6 && addr instanceof Inet4Address;

                boolean isPreferredPrefix = bindToInterfaceByProtocolPreferredPrefix != null ? addr.getHostAddress().startsWith(bindToInterfaceByProtocolPreferredPrefix) : false;
                if(isGroupInet6 == isAddressInet6) {
                    if(isPreferredPrefix || !boundToInterfaceByProtocol) {
                        log.info("Binding multicast to netIf " + netIf.getDisplayName() + " (addr " + addr.getHostAddress() + ")");        
                        mcs.setInterface(addr);
                        boundToInterfaceByProtocolWithPreferredPrefix = isPreferredPrefix;
                        boundToInterfaceByProtocol = true;
                    }
                } else if(!isGroupInet6 && isAddressInet4) {
                    if(isPreferredPrefix || !boundToInterfaceByProtocol) {
                        log.info("Binding multicast to netIf " + netIf.getDisplayName() + " (addr " + addr.getHostAddress() + ")");
                        mcs.setInterface(addr);
                        boundToInterfaceByProtocolWithPreferredPrefix = isPreferredPrefix;
                        boundToInterfaceByProtocol = true;
                    }
                } else {
                    if (log.isDebugEnabled()) {
                        log.debug("Not binding to : " + addr.getHostAddress() + ", not a " + (isGroupInet6 ? "Inet6" : "Inet4") + " address");
                    }
                    continue;
                }
                
                if(boundToInterfaceByProtocolWithPreferredPrefix) {
                    return;
                }
            }
        }
        
        if(!boundToInterfaceByProtocol) {
            throw new IllegalStateException("Unable to find any interface bound to an address matching protocol '" + (isGroupInet6 ? "IPv6" : "IPv4") + "'");            
        }
    }

    protected void bindToInterfaceByPrefix() throws SocketException {
        String networkPrefix = config.getClusterIpPrefix();
        if (networkPrefix == null) {
            return; // Bind to any by not binding to any specific.
        }
        
        log.debug("Trying to bind to network interface using prefix.");
        
        Enumeration<NetworkInterface> networkInterfaces = NetworkInterface.getNetworkInterfaces();
        while (networkInterfaces.hasMoreElements()) {
            NetworkInterface netIf = networkInterfaces.nextElement();
            Enumeration<InetAddress> addresses = netIf.getInetAddresses();
            while (addresses.hasMoreElements()) {
                InetAddress adr = addresses.nextElement();
                if (adr.getHostAddress().startsWith(networkPrefix)) {
                    log.info("Binding multicast to netIf " + netIf.getDisplayName() + " (addr " + adr.getHostAddress()
                            + ")");
                    mcs.setNetworkInterface(netIf);
                    return;
                } else {
                    if (log.isDebugEnabled()) {
                        log.debug("Not binding to : " + adr.getHostAddress() + ", not startsWith " + networkPrefix);
                    }
                }
            }
        }
        throw new IllegalStateException(
                "Unable to find any interface bound to an address matching '" + networkPrefix + "'");
    }

    public synchronized void sendMessage(byte[] bytes) throws IOException {
        if (mcs.isClosed()) {
            log.warn(
                    "Socket closed. Fine if we're shutting down or redeploying or the like, but should not happen during normal operation");
            return;
        }
        interPacketDelay();
        DatagramPacket p = new DatagramPacket(bytes, bytes.length, group, port);
        try {
            mcs.send(p);
        } catch (IOException e) {
            log.warn("Rethrowing IOException " + e.toString() + " caught while trying to multicast " + bytes.length
                    + " bytes to " + group + ":" + port);
        }
        if (log.isDebugEnabled()) {
            log.debug("Sent message, length=" + bytes.length);
        }
    }

    public synchronized void ensureInterPacketDelay() {
        interPacketDelay();
        nextSendNotBefore = System.currentTimeMillis();
    }

    private void interPacketDelay() {
        long now = System.currentTimeMillis();
        long delay = nextSendNotBefore - now;
        if (delay > 0) {
            nextSendNotBefore += PACKET_MIN_INTERVAL;
            try {
                Thread.sleep(delay);
            } catch (InterruptedException e) {
                // ignore.
            }
        } else {
            nextSendNotBefore = now + PACKET_MIN_INTERVAL;
        }
    }

    /**
     * Receive a single message, readable from the returned stream.
     * 
     * @param buffer
     * @return
     * @throws IOException
     */
    public InputStream recvMessage(byte[] buffer) throws IOException, InterruptedException {
        DatagramPacket recv = new DatagramPacket(buffer, buffer.length);
        mcs.receive(recv);
        if (log.isDebugEnabled()) {
            log.debug("Got message, length=" + recv.getLength());
        }
        return new ByteArrayInputStream(buffer, 0, recv.getLength());
    }
}
