/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.StaticConfigApi;

/**
 * Worker which pings for other nodes in the cluster. 
 * 
 * @author Morten Grouleff <mgr@trifork.com>
 *
 */
public class PeriodicAlive extends Worker {
    private static final Logger log = Logger.getLogger(PeriodicAlive.class);
    private final ClusterState state;
    private final NetHandler netHandler;
    private final SlidingWindow<?> sentWindow;
    private final String myUID;
    private int periodicAliveInterval;
    private int warnSilentPeerTime;
    private int maxSilentPeerTime;
    private List<String> vanishedServers = new ArrayList<String>();
    private List<String> silentServers = new ArrayList<String>();

    
    public PeriodicAlive(String myUID, NetHandler netHandler, SlidingWindow<?> sentWindow, ClusterState state, StaticConfigApi config) {
        super(state);
        this.myUID = myUID;
        this.netHandler = netHandler;
        this.sentWindow = sentWindow;
        this.state = state;
        periodicAliveInterval = config.getClusterTuningParam("periodic.alive.interval", 1000); 
        warnSilentPeerTime = periodicAliveInterval * config.getClusterTuningParam("warn.peer.silent.times", 5);
        maxSilentPeerTime = periodicAliveInterval *  config.getClusterTuningParam("remove.peer.silent.times", 300);
    }
    
    protected void run0() throws InterruptedException, IOException {
        Thread.sleep(periodicAliveInterval);
        sendPeriodicAlive();
        state.checkBootstrap();
        long now = System.currentTimeMillis();
        synchronized (state.getLastSenderPing()) {
            for (String serverUID: state.getLastSenderPing().keySet()) {
                Long lastPing = state.getLastSenderPing().get(serverUID);
                if (now - lastPing.longValue() >= warnSilentPeerTime) {
                    // We have not had a ping from this guy for a while. Consider him temporarily missing.
                    if (!silentServers.contains(serverUID)) {
                        log.warn("Server silent: " + serverUID);
                        silentServers.add(serverUID);
                    }
                    state.getSyncRequestedAt().remove(serverUID);
                } else {
                    silentServers.remove(serverUID);
                }
                if (now - lastPing.longValue() >= maxSilentPeerTime) {
                    // We have not had a ping from this guy for a while. Consider him gone.
                    vanishedServers.add(serverUID);
                    silentServers.remove(serverUID);
                }
            }
        }
        for (String serverUID : vanishedServers) {
            serverVanished(serverUID);
        }
        vanishedServers.clear();
}
    
    private void serverVanished(String senderUID) {
        log.warn("Server vanished: " + senderUID);
        state.clear(senderUID);
    }

    public void sendPeriodicAlive() {
        try {
            // This one includes the timestamp used to keep track of the time-skew of the other cluster members.
            // Therefore we grab the nethandler before we build the message to get less jitter on the timestamps.
            synchronized (netHandler) {
                netHandler.ensureInterPacketDelay(); // Make sure we do not wait when we get to the point of sending; do it now instead.
                long configTimestamp = RuntimeConfig.getConfigTimestamp();
                long seqNo = sentWindow.getTop();
                byte[] encodeVersion1AliveAsBytes = Protocol.encodeVersion1AliveAsBytes(seqNo, configTimestamp, myUID);
                netHandler.sendMessage(encodeVersion1AliveAsBytes);
            }
        } catch (IOException e) {
            log.error("Error encoding entry to stream", e);
        }
    }


}
