/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.InetSocketAddress;

import org.apache.log4j.Logger;

import com.trifork.sosigw.api.IDCardClusterCache;
import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.RuntimeConfigState;

public class ProtocolReceiver {
    private static final Logger log = Logger.getLogger(ProtocolReceiver.class);
    private final String myUID;
    private final ClusterState state;
    private final SlidingWindow<String> sentWindow;
    private final IDCardClusterCache clusterCache;
    private final int readTimeout;
    private final int connectTimeout;
    private final boolean hasAccurateTime;
    private final ConfigSender configSender;

    public ProtocolReceiver(String myUID, ClusterState state, SlidingWindow<String> sentWindow, IDCardClusterCache clusterCache, int readTimeout, int connectTimeout, boolean hasAccurateTime, ConfigSender configSender) {
        this.myUID = myUID;
        this.state = state;
        this.sentWindow = sentWindow;
        this.clusterCache = clusterCache;
        this.readTimeout = readTimeout;
        this.connectTimeout = connectTimeout;
        this.hasAccurateTime = hasAccurateTime;
        this.configSender = configSender;
        
    }
    
    public void receive(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        byte version = objectInputStream.readByte();
//        if (log.isDebugEnabled()) {
//            log.debug("receive  : Got message, version " + version);
//        }
        if (version == Protocol.PROTOCOL_VERSION_1) {
            byte command = objectInputStream.readByte();
            if (command == Protocol.PROTOCOL_V1_IDCARDENTRY) {
                receiveV1IDCardEntry(objectInputStream);
            } else if (command == Protocol.PROTOCOL_V1_ALIVE) {
                receiveV1Alive(objectInputStream);
            } else if (command == Protocol.PROTOCOL_V1_SYNCREQUEST) {
                receiveV1SyncRequest(objectInputStream);
            } else if (command == Protocol.PROTOCOL_V1_CONFIGREQUEST) {
                receiveV1ConfigRequest(objectInputStream);
            } else if (command == Protocol.PROTOCOL_V1_CONFIGSTATE) {
                receiveV1ConfigState(objectInputStream);
            } else {
                log.error("Received message with unknown command " + command);
            }
        } else {
            log.error("Received message with unknown protocol version " + version);
        }
    }

    private void receiveV1ConfigState(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        String senderUID = objectInputStream.readUTF();
        if (myUID.equals(senderUID)) { 
            return; // Ignore my own config messages...
        }
        if (log.isDebugEnabled()) {
            log.debug("Receiving RuntimeConfigState from " + senderUID);
        }
        RuntimeConfigState configState = (RuntimeConfigState)objectInputStream.readObject();
        if (log.isDebugEnabled()) {
            log.debug("Done receiving RuntimeConfigState from " + senderUID + " with timestamp " + configState.getTimestamp());
        }
        if (!myUID.equals(senderUID) && RuntimeConfig.getConfigTimestamp() < configState.getTimestamp()) {
            if (log.isInfoEnabled()) {
                log.info("Accepting a new RuntimeConfigState from " + senderUID + " with timestamp " + configState.getTimestamp());
            }
            RuntimeConfig.setRuntimeConfigState(configState);
            RuntimeConfig.saveCurrentState();
        }        
    }

    private void receiveV1SyncRequest(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        SyncRequest syncReq = (SyncRequest)objectInputStream.readObject();
        if (myUID.equals(syncReq.senderUID)) {
            InetSocketAddress peerAddress = (InetSocketAddress)objectInputStream.readObject();
            if (log.isDebugEnabled()) {
                log.debug("Intercom: Received syncrequest, starting SyncSender to " + peerAddress);
            }
            Thread thread = new Thread(new SyncSender(myUID, sentWindow, state, clusterCache, syncReq, peerAddress, readTimeout, connectTimeout), "Intercom.SyncSender");
            thread.setDaemon(true);
            thread.start();
        } else {
            if (log.isDebugEnabled()) {
                log.debug("Intercom: Ignoring syncrequest, meant for " + syncReq.senderUID);
            }
        }
    }

    private void receiveV1ConfigRequest(ObjectInputStream objectInputStream) throws IOException {
        String senderUID = objectInputStream.readUTF();
        long senderConfigTimestamp = objectInputStream.readLong();
        if (!myUID.equals(senderUID)) {
            inspectConfigTimestampFromSender(senderUID, senderConfigTimestamp);
        }
    }

    private void receiveV1Alive(ObjectInputStream objectInputStream) throws IOException {
        String senderUID = objectInputStream.readUTF();
        long senderSeqno = objectInputStream.readLong();
        long senderConfigTimestamp = objectInputStream.readLong();
        long sendertimestamp = objectInputStream.readLong();
        if (myUID.equals(senderUID)) {
            return;
        }
        state.getLastSenderPing().put(senderUID, System.currentTimeMillis());
        inspectSenderTime(senderUID, sendertimestamp);
        if (senderSeqno != 0) {
            inspectSequenceFromSender(senderUID, senderSeqno);
        }
        if (senderConfigTimestamp != 0) {
            inspectConfigTimestampFromSender(senderUID, senderConfigTimestamp);
        }
    }

    private void receiveV1IDCardEntry(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        String senderUID = objectInputStream.readUTF();
        if (myUID.equals(senderUID)) {
            return;
        }
        state.getLastSenderPing().put(senderUID, System.currentTimeMillis());
        long senderSeqno = objectInputStream.readLong();
        if (senderSeqno != 0) {
            inspectSequenceFromSender(senderUID, senderSeqno);
        }
        if (state.hasTimeState(senderUID)) {
            // We cannot accept messages from peers, if we don't know their timing state. 
            Object object = objectInputStream.readObject();
            IDCardClusterCacheEntry entry = (IDCardClusterCacheEntry)object;
            if (log.isDebugEnabled()) {
                log.debug("Received entry :" + entry.toString());
            }
            clusterCache.arrivedFromPeer(entry, getTimeOffset(senderUID));
        }
    }

    private int getTimeOffset(String senderUID) {
        if (hasAccurateTime) {
            return 0;
        }
        TimeState state = this.state.getTimeStates().get(senderUID);
        return state != null ? state.getCurOffset() : 0; 
    }

    private void inspectSenderTime(String senderUID, long sendertimestamp) {
        long currentTimeMillis = System.currentTimeMillis();
        long offset = sendertimestamp - currentTimeMillis;
        if (hasAccurateTime) {
            if (offset > 30) {
                log.warn("Configured for accurate time, but the peer " + senderUID + " is " + offset + " ahead of me!");
                return;
            }
        }
        TimeState state;
        synchronized (this.state.getTimeStates()) {
            state = this.state.getTimeStates().get(senderUID);
            if (state == null) {
                state = new TimeState();
                this.state.getTimeStates().put(senderUID, state);
                log.warn("Server discovered: " + senderUID);
            } 
            state.addSample((int)offset);
        }
        
        int curOffset = state.getCurOffset();
//        if (log.isInfoEnabled() && curOffset > ACCEPTABLE_TIME_SKEW) {
//            log.info("Time at " + senderUID + " is ahead of mine. Diff is " + curOffset);
//        } else if (log.isInfoEnabled() && curOffset < -ACCEPTABLE_TIME_SKEW) {
//            log.info("Time at " + senderUID + " is behind mine. Diff is " + curOffset);
//        }
        
        if (log.isDebugEnabled()) {
            log.debug("Time at " + senderUID + " is offset " + (sendertimestamp - currentTimeMillis) + " curOffset=" + curOffset);
        }
    }

    private void inspectConfigTimestampFromSender(String senderUID, long senderConfigTimestamp) {
        if (log.isDebugEnabled()) {
            log.debug("inspectConfigTimestampFromSender: " + senderUID + ", " + senderConfigTimestamp);
        }
        long myConfigTimestamp = RuntimeConfig.getConfigTimestamp();
        if (myConfigTimestamp == senderConfigTimestamp) {
            return; // All is well
        }
        if (myConfigTimestamp < senderConfigTimestamp) {
            //My config is too old. Request a new one.
            if (log.isDebugEnabled()) {
                log.debug("Intercom: Requesting new config from " + senderUID);
            }
            configSender.requestConfig(myConfigTimestamp);
        } else {
            // My config is the newer.
            if (log.isDebugEnabled()) {
                log.debug("Intercom: Received config-request, sendCurrentRuntimeConfigState to " + senderUID);
            }
            configSender.sendCurrentRuntimeConfigState();
        }
    }
    
    private void inspectSequenceFromSender(String senderUID, long senderSeqno) {
        if (log.isDebugEnabled()) {
            log.debug("inspectSequenceFromSender: " + senderUID + ", " + senderSeqno);
        }
        
        Long lastSeen = state.getLastSenderSeqno().get(senderUID);
            
        if (lastSeen == null) {
            // Never heard from this guy before.
            log.debug("inspectSequenceFromSender: " + senderUID + ", " + senderSeqno + " - New buddy");
            startSyncFromParty(senderUID);
        } else {
            long l = lastSeen.longValue();
            if (senderSeqno == l) {
                state.getSyncRequestedAt().remove(senderUID); // We are current with this fellow. No need for further sync with him.
            } else if (senderSeqno == l + 1) {
                state.getSyncRequestedAt().remove(senderUID); // We are current with this fellow. No need for further sync with him.
                state.getLastSenderSeqno().put(senderUID, new Long(senderSeqno));
            } else if (senderSeqno > l) {
                // We must have missed one or more messages.
                if (log.isDebugEnabled()) {
                    log.debug("inspectSequenceFromSender: " + senderUID + ", " + senderSeqno + ", missed-packet, we have him at " + l);
                }
                startSyncFromParty(senderUID);
            } else {
                // We got an "old" packet, that we're also receiving though resyncsocket concurrently..
                if (log.isDebugEnabled()) {
                    log.debug("inspectSequenceFromSender: " + senderUID + ", " + senderSeqno + ", old-packet, we have him at " + l);
                }
            }
        }
    }
    
    private void startSyncFromParty(String senderUID) {
        if (!state.getSyncRequestedAt().containsKey(senderUID)) { // Do not add sender that we already plan to sync with
            if (log.isDebugEnabled()) {
                log.debug("Enqueue sender in startSyncFromParty: " + senderUID + " (I am " + myUID + ")");
            }
            synchronized (state.getSyncRequestedAt()) {
                if (!state.getSyncRequestedAt().containsKey(senderUID)) { // Do not add sender that we already plan to sync with
                    state.getSyncRequestedAt().put(senderUID, new Long(0));
                    state.getSyncRequestedAt().notify();
                }                
            }
        }
    }
    
}
