/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class SlidingWindow<E> {
    private static final Logger log = Logger.getLogger(SlidingWindow.class); 
    private final E[] elementData;
    private final int size;
    private long topElement;
    private long nextSequenceNumber;
    
    @SuppressWarnings("unchecked")
    public SlidingWindow(int size) {
        this.size = size;
        this.elementData = (E[])new Object[size];
        this.nextSequenceNumber = System.currentTimeMillis();
        this.topElement = this.nextSequenceNumber;
    }
    
    public synchronized long reserve() {
        long seqNo = nextSequenceNumber++;
        if (seqNo - topElement > size) {
            reset();
        }
        elementData[calcIndex(seqNo)] = null;
        return seqNo;
    }

    private void reset() {
        log.warn("Overloaded! Resetting SlidingWindow");
        this.topElement = this.nextSequenceNumber;
    }

    public synchronized long getTop() {
        return topElement - 1;
    }
    
    public synchronized void add(long seqNo, E elm) {
        int idx = calcIndex(seqNo);
        elementData[idx] = elm;
        while (topElement < nextSequenceNumber && elementData[calcIndex(topElement)] != null) {
            topElement++;
        }
    }

    public synchronized List<E> getContentsSince(long lastKnownGood) {
        long count = topElement - lastKnownGood;
        if (count > size || count < 0) {
            return null; // don't know elements that old.
        }
        List<E> res = new ArrayList<E>((int)count);
        for (long seqNo = lastKnownGood + 1; seqNo < topElement; seqNo++) {
            int idx = calcIndex(seqNo);
            E e = elementData[idx];
            if (e != null) {
                res.add(e);
            }
        }
        return res;
    }

    private int calcIndex(long seqNo) {
        return (int)(seqNo % size);
    }
}
