/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import com.trifork.sosigw.config.StaticConfigApi;

public class SyncReceiver extends Worker {
    private static final Logger log = Logger.getLogger(SyncReceiver.class);
    
    private Map<String, Long> resetTheseOnError = new HashMap<String, Long>();
    private Map<String, Long> setTheseOnSuccess = new HashMap<String, Long>();

    private final byte[] buffer;
    private final ServerSocket resyncSocket;
    private final ClusterState state;
    private final String myUID;
    private final int readTimeout;

    private final ProtocolReceiver receiver;

    public SyncReceiver(String myUID, ServerSocket resyncSocket, ClusterState state, StaticConfigApi config, ProtocolReceiver receiver) {
        super(state);
        this.myUID = myUID;
        this.resyncSocket = resyncSocket;
        this.state = state;
        this.receiver = receiver;
        
        int maxMessageSize = config.getClusterTuningParam("max.message.size", 60000);
        buffer = new byte[maxMessageSize];
        
        readTimeout = config.getClusterTuningParam("resync.receive.timeout", 5000);
    }

    protected void run0() throws InterruptedException, IOException, ClassNotFoundException {
        Socket socket = resyncSocket.accept();
        socket.setSoTimeout(readTimeout); // Make sure we fail, if a read takes more than this long to complete.
        InputStream inputStream = socket.getInputStream();
        DataInputStream dataInputStream = null;
        try {
            dataInputStream = new DataInputStream(inputStream);
            while (!state.isKilled()) {
                // Get the states of the other servers known to then sync-sending party:
                byte protoversion = dataInputStream.readByte();
                if (protoversion != Protocol.PROTOCOL_VERSION_1) {
                    log.error("Unknown protocol version " + protoversion);
                    return;
                }
                int serverStateCount = dataInputStream.readInt();
                String senderUID = null;
                for (int i = 0; i < serverStateCount; i++) {
                    String serverUID = dataInputStream.readUTF();
                    long serverSeqNo = dataInputStream.readLong();
                    if (!myUID.equals(serverUID)) {
                        if (senderUID == null) {
                            senderUID = serverUID;
                        }
                        synchronized (state.getLastSenderSeqno()) {
                            Long l = state.getLastSenderSeqno().get(serverUID);
                            if (l == null) {
                                resetTheseOnError.put(serverUID, null);
                                //lastSenderSeqno.put(serverUID, new Long(serverSeqNo));
                                setTheseOnSuccess.put(serverUID, serverSeqNo);
                            } else {
                                if (l.longValue() < serverSeqNo) {
                                    resetTheseOnError.put(serverUID, l);
                                    //  lastSenderSeqno.put(serverUID, new Long(serverSeqNo));
                                    setTheseOnSuccess.put(serverUID, serverSeqNo);
                                }
                            }
                        }
                    }
                }
                // Receive the idcard states from the other end.
                long idCardMsgCount = dataInputStream.readLong();
                if (log.isInfoEnabled()) {
                    log.info("SyncReceiver: Expecting " + idCardMsgCount + " messages.");
                }
                long lastupdate = 0;
                for (int i = 0; i < idCardMsgCount; i++) {
                    long now = System.currentTimeMillis();
                    if (now - lastupdate > 100) {
                        state.setLastReSyncAt(now + readTimeout);
                        state.getSyncRequestedAt().put(senderUID, new Long(now));
                        lastupdate = now;
                    }
                    if (log.isDebugEnabled()) {
                        log.debug("SyncReceiver: Reading message #" + i + ".");
                    }
                    int length = dataInputStream.readInt();
                    if (log.isDebugEnabled()) {
                        log.debug("SyncReceiver: Reading bytes to buffer, expecting  " + length + " bytes.");
                    }
                    dataInputStream.readFully(buffer, 0, length);
                    ObjectInputStream objectInputStream = new ObjectInputStream(new ByteArrayInputStream(buffer, 0, length));
                    receiver.receive(objectInputStream);
                    objectInputStream.close();
                }

                state.getLastSenderSeqno().putAll(setTheseOnSuccess);
                setTheseOnSuccess.clear();
                resetTheseOnError.clear();
                state.setLastReSyncAt(System.currentTimeMillis());
                state.setFirstSyncCompleted(true);
                
                if (log.isInfoEnabled()) {
                    log.info("SyncReceiver: Got all " + idCardMsgCount + " messages.");
                }
                
                if (!dataInputStream.readBoolean()) {
                    if (log.isDebugEnabled()) {
                        log.debug("SyncReceiver: Got ALL messages.");
                    }
                    break;
                } else {
                    if (log.isDebugEnabled()) {
                        log.debug("SyncReceiver: Got more-messages flag.");
                    }
                }
            }
        } catch (IOException e) {
            log.warn(e.getMessage(), e);
            state.getLastSenderSeqno().putAll(resetTheseOnError);
        } finally {
            resetTheseOnError.clear();
            if (dataInputStream != null) {
                dataInputStream.close();
            }
            synchronized (state.getSyncRequestedAt()) {
                state.getSyncRequestedAt().notify();
            }
        }
    }
}
