/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import java.io.IOException;
import java.net.ServerSocket;

import org.apache.log4j.Logger;

/**
 * This class is responsible for sending out IDCard sync requests.
 * 
 * Requests are sent when the a new entry is added to {@link ClusterState#getSyncRequestedAt()}, which
 * happens when a alive request is received from an unknown node.
 * 
 * @author Morten Grouleff <mgr@trifork.com>
 * 
 * @see SyncSender which receives sync requests and responds to them.
 *
 */
public class SyncRequester extends Worker {
    private static final Logger log = Logger.getLogger(SyncRequester.class);
    
    private static final long SYNCREQUESTER_INSPECTINTERVAL = 100; 
    private static final long SYNCREQUEST_TIMEOUT = 5 * 1000; // How long from a syncrequest is sent before another is sent to the same host.

    private final ClusterState state;
    private final ServerSocket resyncSocket;
    private final NetHandler netHandler;

    public SyncRequester(ClusterState state, ServerSocket resyncSocket, NetHandler netHandler) {
        super(state);
        this.state = state;
        this.resyncSocket = resyncSocket;
        this.netHandler = netHandler;
    }
    
    protected void run0() throws InterruptedException, IOException {
        String senderUID = null;
        synchronized (state.getSyncRequestedAt()) {
            state.getSyncRequestedAt().wait(SYNCREQUESTER_INSPECTINTERVAL);
            if (state.isFirstSyncCompleted() || state.getLastReSyncAt() <= System.currentTimeMillis()) {
                Long lowest = System.currentTimeMillis() - SYNCREQUEST_TIMEOUT;
                for (String uid : state.getSyncRequestedAt().keySet()) {
                    Long when = state.getSyncRequestedAt().get(uid);
                    if (when < lowest) {
                        if (state.hasTimeState(uid)) {
                            senderUID = uid;
                            lowest = when;
                        }
                    }
                }
                if (senderUID != null) {
                    state.getSyncRequestedAt().put(senderUID, new Long(System.currentTimeMillis()));
                }
            }
        }
        if (senderUID != null) {
            Long lastKnownGood = state.getLastSenderSeqno().get(senderUID);
            if (log.isDebugEnabled()) {
                log.debug("Sending syncrequest to " + senderUID + " with lastKnownGood=" + lastKnownGood);
            }
            SyncRequest syncReq = new SyncRequest(senderUID, lastKnownGood == null ? 0 : lastKnownGood.longValue());
            byte[] byteArray = Protocol.encodeVersion1SyncRequestAsBytes(syncReq, resyncSocket.getLocalSocketAddress());
            netHandler.sendMessage(byteArray);
        }
    }

}
