/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import com.trifork.sosigw.api.IDCardCacheEntry;
import com.trifork.sosigw.api.IDCardClusterCache;
import org.apache.log4j.Logger;

import java.io.BufferedOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * This class is responsible for sending sync responses to other nodes in the cluster.
 * 
 * Requesting nodes open a new server socket and sends the hostname and port. This class then opens a new socket
 * to the listener and sends missing IDCards.
 * 
 * @author Morten Grouleff <mgr@trifork.com>
 *
 */
public class SyncSender implements Runnable {
    
    private static final Logger log = Logger.getLogger(SyncSender.class);
    
    private final SyncRequest syncReq;
    private final Socket socket;
    private final InetSocketAddress peerAddress;
    private boolean killed = false;
    private final String myUID;

    private final SlidingWindow<String> sentWindow;

    private final IDCardClusterCache clusterCache;

    private final ClusterState state;

    public SyncSender(String myUID, SlidingWindow<String> sentWindow, ClusterState state, IDCardClusterCache clusterCache, SyncRequest syncReq, InetSocketAddress peerAddress, int readTimeout, int connectTimeout) throws IOException {
        this.myUID = myUID;
        this.sentWindow = sentWindow;
        this.state = state;
        this.clusterCache = clusterCache;
        this.syncReq = syncReq;
        this.peerAddress = peerAddress;
        socket = new Socket();
        socket.setTcpNoDelay(true); // We're sending large amount, but we have a buffer, so get it out as quick as you can.
        socket.setSoTimeout(readTimeout);
        socket.connect(peerAddress, connectTimeout);
    }

    public void run() {
        DataOutputStream dataOutputStream = null;
        try {
            long lastKnownGood = syncReq.lastKnownGood;
            if (log.isDebugEnabled()) {
                log.debug("Intercom.SyncSender: Processing syncreq, lastKnownGood=" + lastKnownGood);
            }
            dataOutputStream = new DataOutputStream(new BufferedOutputStream(socket.getOutputStream(), 32768));
            long byteCount = 0;
            int idcardCount = 0;
            final long startedAt = System.currentTimeMillis();

            while (!killed) {
                byte[] otherServersstates = Protocol.encodeVersion1ServerStatesAsBytes(myUID, sentWindow, state.getLastSenderSeqno());
                dataOutputStream.write(otherServersstates);

                List<String> cacheKeys;
                synchronized (sentWindow) {
                    cacheKeys = sentWindow.getContentsSince(lastKnownGood);
                    lastKnownGood = sentWindow.getTop();
                }

                if (cacheKeys != null) {
                    if (log.isDebugEnabled()) {
                        log.debug("Intercom.SyncSender: Sending partial data from window, at most " + cacheKeys.size() + " messages");
                    }
                    if (!cacheKeys.isEmpty()) {
                        Set<String> names = new HashSet<String>(cacheKeys);
                        dataOutputStream.writeLong(names.size());
                        idcardCount += names.size();
                        for (String cacheKey: names) {
                            IDCardCacheEntry entry = clusterCache.ensureIDCard(cacheKey);
                            // Encode each entry to a byte buffer, to make sure the entry is not locked for too long. 
                            byte[] byteArray = Protocol.encodeVersion1IDCardEntryAsBytes(entry, 0, myUID);
                            dataOutputStream.writeInt(byteArray.length);
                            dataOutputStream.write(byteArray, 0, byteArray.length);
                            byteCount += 4 + byteArray.length;
                        }
                    } else {
                        dataOutputStream.writeLong(0);
                    }
                } else {
                    if (log.isDebugEnabled()) {
                        log.debug("Intercom.SyncSender: Preparing to send complete cache");
                    }
                    List<String> contents = clusterCache.getContentsAsList();
                    if (log.isInfoEnabled()) {
                        log.info("Intercom.SyncSender: Sending complete cache holding " + contents.size() + " CacheKeys");
                    }
                    dataOutputStream.writeLong(contents.size());
                    idcardCount += contents.size();
                    for (String cacheKey: contents) {
                        // Encode each entry to a byte buffer, to make sure the entry is not locked for too long. 
                        IDCardCacheEntry entry = clusterCache.ensureIDCard(cacheKey);
                        byte[] byteArray = Protocol.encodeVersion1IDCardEntryAsBytes(entry, 0, myUID);
                        dataOutputStream.writeInt(byteArray.length);
                        dataOutputStream.write(byteArray, 0, byteArray.length);
                        byteCount += 4 + byteArray.length;
                    }
                }
                long newTop = sentWindow.getTop();
                if (newTop == lastKnownGood) {
                    if (log.isDebugEnabled()) {
                        log.debug("Intercom.SyncSender: Done sending, now in sync");
                    }
                    dataOutputStream.writeBoolean(false);
                    break;
                } else {
                    if (log.isDebugEnabled()) {
                        log.debug("Intercom.SyncSender: Must send more, missing " + (newTop - lastKnownGood) + " messages.");
                    }
                    dataOutputStream.writeBoolean(true);
                }
            }
            if (log.isInfoEnabled()) {
                log.info("SyncSender completed sending state to " + peerAddress + ". Sent total bytes=" + byteCount + ",  idcards=" + idcardCount + ", took=" + (System.currentTimeMillis() - startedAt));
            }
        } catch (IOException e) {
            log.warn("Ignored exception. Did not complete syncSending", e);
        } finally {
            try {
                if (dataOutputStream != null) {
                    dataOutputStream.close();
                }
            } catch (IOException e) {
                log.warn("Ignored exception", e);
            }
            try {
                socket.close();
            } catch (IOException e) {
                log.warn("Ignored exception", e);
            }
        }
    }
    


}
