/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.config;

import dk.sosi.seal.model.UserInfo;
import org.apache.log4j.Logger;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.regex.Pattern;

/**
 * facade towards config, meant for runtime lookups of current config data, UNMODIFIABLE, except as result of receive from remote peers.
 * @author Morten Grouleff <mgr@trifork.com>
 *
 */
public class RuntimeConfig {
    private final static Logger log = Logger.getLogger(RuntimeConfig.class);
    public static final String CATEGORY_CLIENTACCESSWHITELIST = "Client Access WhiteList";
    public static final String CATEGORY_SERVICEACCESSWHITELIST = "Service Access WhiteList";
    public static final String CATEGORY_SERVICEPRESERVE_WSA = "Service Preserve WSA";
    public static final String CATEGORY_PROPERTIES = "General Properties";
    public static final String CATEGORY_CONSOLE_USERS = "Administration Console Users";
    public static final String CATEGORY_METAINFO = "MetaInfo";

    public static final String KEY_BROWSERSIGNINGURL = "browsersigning.url";
    public static final String KEY_STSURL = "sts.service.url";
    public static final String KEY_GLOBALJDBCURL = "globaldb.jdbc.url";
    public static final String KEY_GLOBALJDBCUSERNAME = "globaldb.jdbc.username";
    public static final String KEY_GLOBALJDBCPASSWORD = "globaldb.jdbc.password";
    public static final String KEY_GLOBALJDBCINTERVAL = "globaldb.transmit.interval";
    public static final String KEY_CAREPROVIDER_CVR = "careprovider.cvr";
    public static final String KEY_CAREPROVIDER_CVRNAME = "careprovider.cvr.name";
    public static final String KEY_IDCARD_SIGNING_TIMEOUT = "idcard.signing.timeout";
    public static final String KEY_BROWSER_SIGNING_TIMEOUT = "browser.signing.timeout";
    public static final String KEY_STS_NUMBER_FAILURES_BEFORE_OPENING_CIRCUITBREAKER = "sts.number.failures.before.open.cb";
    public static final String KEY_STS_MILLIS_BEFORE_ATTEMPTING_CLOSE_CIRCUITBREAKER = "sts.millis.before.attempting.close.cb";
    public static final String KEY_STS_TIMEOUT_MILLIS = "sts.timeout.millis";
    public static final String KEY_CACHE_PARTITIONING_ENABLED = "cache-partitioning.enabled";
    public static final String KEY_CACHE_PARTITIONING_HTTP_HEADER_NAME = "cache-partitioning.http-header.name";
    public static final String KEY_DECOUPLING_URL = "decoupling-component.url";
    public static final String KEY_PASS_THROUGH_ON_SIGNATURE = "pass-through-on-signature.enabled";
    public static final String KEY_NSP_PROXY_REQUESTID_HTTP_HEADER_NAME = "proxy-requestid.http-header.name";
    public static final String KEY_NSP_FORWARD_HTTP_HEADERS_NAME = "forward.http-header.names";
    public static final String KEY_FORCE_SSL_REQUESTS_ENABLED = "force-ssl-requests.enabled";

    private static RuntimeConfigState configState;
    private static final String[] DEFAULT_GLOBAL_PROPERTIES = {
        KEY_BROWSERSIGNINGURL, "http://localhost:8080/sosigw/browsersign/invoke", "url that client can reach this sosigw-cluster on for browsersigning",
        KEY_STSURL, "http://test1-cnsp.ekstern-test.nspop.dk:8080/sts/services/SecurityTokenService http://test2-cnsp.ekstern-test.nspop.dk:8080/sts/services/SecurityTokenService", "Where to find the Security Token Service (IDP). The value can be a list of urls, which will be tried in sequence. The urls should be separated by a single space character",
        KEY_GLOBALJDBCURL, "jdbc:postgresql://localhost:5432/sosigwgloballog", "The global audit log database",
        KEY_GLOBALJDBCUSERNAME, "sosigwgloballog", "Username for the global audit log database",
        KEY_GLOBALJDBCPASSWORD, "sosigwgloballog", "password for the global audit log database",
        KEY_GLOBALJDBCINTERVAL, "10000", "In milliseconds, the interval between transfers of local audit log entries from local DB to global DB.",
        KEY_CAREPROVIDER_CVR, "19343634", "CVR of organization running SOSIGW. (Use 19343634 for testing)",
        KEY_CAREPROVIDER_CVRNAME, "SOSITEST", "Name of organization running SOSIGW. (Use \"SOSITEST\" for testing)",
        KEY_IDCARD_SIGNING_TIMEOUT, "120", "Number of seconds allowed from requesting the signing digest/url to until either signign it or starting a browser signing session on the url.",
        KEY_BROWSER_SIGNING_TIMEOUT, "120", "Number of seconds allowed from starting the signing applet until the browser must have submitted the signed digest.",
        KEY_STS_NUMBER_FAILURES_BEFORE_OPENING_CIRCUITBREAKER, "5", "Number of failures allowed for an STS before the corresponding circuit breaker opens",
        KEY_STS_MILLIS_BEFORE_ATTEMPTING_CLOSE_CIRCUITBREAKER, "60000", "Milliseconds to elapse before retrying an STS whose circuit breaker has opened due to previous failures",
        KEY_STS_TIMEOUT_MILLIS, "10000", "The timeout (in milliseconds) to wait for a response to an STS request",
        KEY_CACHE_PARTITIONING_ENABLED, "false", "Whether the idcard cache should be partitioned based on an http-header",
        KEY_NSP_PROXY_REQUESTID_HTTP_HEADER_NAME, "X_NSP_PROXY_REQUESTID", "Name of proxy request id http header",
        KEY_NSP_FORWARD_HTTP_HEADERS_NAME, "X_NSP_FLOWID, X_NSP_MY_HTTP_FORWARD_HEADER1, X_NSP_MY_HTTP_FORWARD_HEADER2", "List of HTTP headers that should be forwarded"
    };


    private static synchronized void ensureDefaultProperties() {
        long timestamp = configState.getTimestamp();
        RuntimeConfigEditState editState = configState.getEditableVersion();
        boolean changed = false;
        for (int i = 0; i < DEFAULT_GLOBAL_PROPERTIES.length; i += 3) {
            if (editState.getValue(CATEGORY_PROPERTIES, DEFAULT_GLOBAL_PROPERTIES[i]) == null) {
                editState.addEntry(CATEGORY_PROPERTIES, DEFAULT_GLOBAL_PROPERTIES[i], new String[] { DEFAULT_GLOBAL_PROPERTIES[i + 1], DEFAULT_GLOBAL_PROPERTIES[i + 2] });
                changed = true;
            }
        }
        if (changed) {
            setRuntimeConfigState(new RuntimeConfigState(timestamp, editState.getResult().getState()));
        }
    }

    private static synchronized void ensureConfigState() {
        if (configState == null) {
            log.debug("Getting initial configuration");
            try {
                configState = getDatabaseVersion();
                if (configState != null) {
                    log.info("Loaded config with timestamp " + configState.getTimestamp());
                } else {
                    configState = new RuntimeConfigState(0, new RuntimeConfigItem[0]);
                }
                ensureDefaultProperties();
            } finally {
                log.debug("Getting initial configuration, done!");
            }
        }
    }

    private static RuntimeConfigState getDatabaseVersion() {
        if (!StaticConfigImpl.getInstance().isDeveloperRestrictedMode()) {
            return RuntimeConfigMapper.getInstance().loadLatestConfig();
        }

        return null;
    }

    public static RuntimeConfigState getRuntimeConfigState() {
        if (configState == null) {
            ensureConfigState();
            log.info("RuntimeConfig: Got an initial config with timestamp " + configState.getTimestamp());
        }
        return configState;
    }

    public static void setRuntimeConfigState(RuntimeConfigState configState) {
        RuntimeConfig.configState = configState;
        log.debug("Replaced config with a new one with timestamp " + configState.getTimestamp());
        if (log.isDebugEnabled()) {
            logAllEntries();
        }
        ensureDefaultProperties();
    }

    public static void saveCurrentState() {
        if (configState != null) {
            if (!StaticConfigImpl.getInstance().isDeveloperRestrictedMode()) {
                RuntimeConfigMapper.getInstance().saveConfig(configState);
            }
        }
    }

    private static void logAllEntries() {
        RuntimeConfigItem[] state = configState.getState();
        for (RuntimeConfigItem runtimeConfigItem : state) {
            log.debug(runtimeConfigItem.toString());
        }
    }

    public static long getConfigTimestamp() {
        if (configState == null) {
            return 0;
        }
        return configState.getTimestamp();
    }

    public static Map<String, String[]> getCategoryMap(String category) {
        return getRuntimeConfigState().getCategoryMap(category);
    }

    public static String[] getValue(String category, String key) {
        return getRuntimeConfigState().getValue(category, key);
    }
    
    public static boolean checkClientAccessWhiteList(String remoteAddrString, String saltInRequest, boolean requireSalt) {
        Map<String, String[]> values = getCategoryMap(CATEGORY_CLIENTACCESSWHITELIST);
        for (String[] white: values.values()) {
            String ipPrefix = white[0];
            String saltInConfig = white[1];
            if (ipPrefix == null || ipPrefix.length() == 0) {
                if (!requireSalt || saltInConfig == null || saltInConfig.equals(saltInRequest)) {
                    return true;
                }
            }
            if (ipPrefix.indexOf('/') >= 0) {
                // TODO: Extension point: Match by ip prefix with bit-width
            } else {
                if (remoteAddrString.startsWith(ipPrefix)) {
                    if (!requireSalt || saltInConfig == null || saltInConfig.equals(saltInRequest)) {
                        return true;
                    }
                }
            }
        }
        return StaticConfigImpl.getInstance().isDeveloperRestrictedMode();
    }

    public static boolean checkServiceWhiteList(String headerTo, String headerAction) {
        Map<String, String[]> values = getCategoryMap(CATEGORY_SERVICEACCESSWHITELIST);
        if (matchWSA(values, headerTo, headerAction)) {
            return true;
        }
        return StaticConfigImpl.getInstance().isDeveloperRestrictedMode();
    }

    private static boolean matchWSA(Map<String, String[]> values, String headerTo, String headerAction) {
        for (String[] white: values.values()) {
            String configTo = white[0];
            String configAction = white[1];
            if (checkRegexpMatch(configTo, headerTo) && checkRegexpMatch(configAction, headerAction)) {
                return true;
            }
        }
        return false;
    }

    // Cache compiled regex patterns here
    private static ConcurrentMap<String, Pattern> compiledPatterns = new ConcurrentHashMap<String, Pattern>();
    private static boolean checkRegexpMatch(String configRegexpString, String checkString) {

        if (configRegexpString == null || configRegexpString.length() == 0) {
            return true;
        } else if (configRegexpString.equals(checkString)) {
            return true;
        }
        Pattern pattern = compiledPatterns.get(configRegexpString);
        if (pattern == null) {
            pattern = Pattern.compile(configRegexpString);
            compiledPatterns.putIfAbsent(configRegexpString, pattern);
        }
        return pattern.matcher(checkString).matches();
    }

    public static boolean removeWSAInProxy(String headerTo, String headerAction) {
        Map<String, String[]> values = getCategoryMap(CATEGORY_SERVICEPRESERVE_WSA);
        if (matchWSA(values, headerTo, headerAction)) {
            return false;
        }
        return true; // Default: Remove all WSA while proxy'ing.
    }

    public static String getStringProperty(String key) {
        String[] values = getValue(CATEGORY_PROPERTIES, key);
        return values!= null && values.length > 0 ? values[0] : null;
    }

    public static Boolean getBooleanProperty(String key) {
        return "true".equalsIgnoreCase(getStringProperty(key));
    }

    public static long getLongProperty(String key) {
        String s = getStringProperty(key);
        try {
            return Long.parseLong(s);
        } catch (NumberFormatException nfe) {
            log.warn("Error parsing '" + s + "' as a long", nfe);
            return 0;
        }
    }

    public static int getIntProperty(String key) {
        String s = getStringProperty(key);
        try {
            return Integer.parseInt(s);
        } catch (NumberFormatException nfe) {
            log.warn("Error parsing '" + s + "' as a int", nfe);
            return 0;
        }
    }

    public static String getBrowserSigningUrlBase(String partition) {
        if (partition != null) {
            String partitionedUrl = getStringProperty(KEY_BROWSERSIGNINGURL + "." + partition);
            if (partitionedUrl != null && !partitionedUrl.trim().isEmpty()) {
                return partitionedUrl;
            }
        }
        return getStringProperty(KEY_BROWSERSIGNINGURL);
    }
    
    public static String getSecurityTokenServiceUrlList() {
        return getStringProperty(KEY_STSURL);
    }

    public static String getGlobalDatabaseJDBCUrl() {
        return getStringProperty(KEY_GLOBALJDBCURL);
    }
    public static String getGlobalDatabaseJDBCUsername() {
        return getStringProperty(KEY_GLOBALJDBCUSERNAME);
    }
    public static String getGlobalDatabaseJDBCPassword() {
        return getStringProperty(KEY_GLOBALJDBCPASSWORD);
    }
    public static long getGlobalDatabaseJDBCInterval() {
        return getLongProperty(KEY_GLOBALJDBCINTERVAL);
    }
    
    public static boolean isCPRAllowedAccessToConsole(String cpr) {
        return getValue(CATEGORY_CONSOLE_USERS, cpr) != null;
    }

    public static UserInfo getConsoleUserInfo(String cpr) {
        String[] value = getValue(CATEGORY_CONSOLE_USERS, cpr);
        if (value != null) {
            return new UserInfo(cpr, value[0], value[1], value[2], "SOSI-GW console user", "SysAdm", "SOSI-GW console user");
        }
        return null;
    }

    public static String getConsoleCVRNumber() {
        return getStringProperty(KEY_CAREPROVIDER_CVR);
    }
    public static String getConsoleCVRName() {
        return getStringProperty(KEY_CAREPROVIDER_CVRNAME);
    }

    public static long getIdCardSigningTimeout() {
        return getLongProperty(KEY_IDCARD_SIGNING_TIMEOUT) * 1000;
    }

    public static long getBrowserSigningTimeout() {
        return getLongProperty(KEY_BROWSER_SIGNING_TIMEOUT) * 1000;
    }

    public static int getSTSNumberOfFailuresBeforeOpeningCircuitBreaker() {
        return getIntProperty(KEY_STS_NUMBER_FAILURES_BEFORE_OPENING_CIRCUITBREAKER);
    }

    public static int getMillisecondsBeforeAttemptingClose() {
        return getIntProperty(KEY_STS_MILLIS_BEFORE_ATTEMPTING_CLOSE_CIRCUITBREAKER);
    }

    public static int getStsTimeoutMillis() {
        return getIntProperty(KEY_STS_TIMEOUT_MILLIS);
    }

    public static boolean isCachePartitioningEnabled() {
        return getBooleanProperty(KEY_CACHE_PARTITIONING_ENABLED);
    }

    public static String getCachePartitioningHttpHeaderName() {
        return getStringProperty(KEY_CACHE_PARTITIONING_HTTP_HEADER_NAME);
    }

    public static String getDecouplingUrl() {
        return getStringProperty(KEY_DECOUPLING_URL);
    }

    public static boolean isPassThroughOnSignatureEnabled() {
        return getBooleanProperty(KEY_PASS_THROUGH_ON_SIGNATURE);
    }

    public static String getNspProxyRequestidHttpHeaderName() {
        return getStringProperty(KEY_NSP_PROXY_REQUESTID_HTTP_HEADER_NAME);
    }

    public static String getKeyNspForwardHttpHeaderNames() {
        return getStringProperty(KEY_NSP_FORWARD_HTTP_HEADERS_NAME);
    }

    public static boolean isForceSSLRequests() {
        return getBooleanProperty(KEY_FORCE_SSL_REQUESTS_ENABLED);
    }
}
