/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.config;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.ArrayList;
import java.util.List;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.sql.DataSource;

import org.apache.log4j.Logger;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public abstract class RuntimeConfigMapperAbstractDb extends RuntimeConfigMapper {
    private static final Logger log = Logger.getLogger(RuntimeConfigMapperAbstractDb.class);

    protected static final String SOSIGWLOG_RUNTIMECONFIG = "sosigwlog.RUNTIMECONFIG";
    protected static final String SOSIGWLOG_RUNTIMECONFIG_ELM = "sosigwlog.RUNTIMECONFIG_ELM";
    protected static final String SOSIGWLOG_RUNTIMECONFIG_DATA = "sosigwlog.RUNTIMECONFIG_DATA";
    protected static final String SOSIGWLOG_RUNTIMECONFIG_PKSEQ = "sosigwlog.RUNTIMECONFIG_PKSEQ";

    private static RuntimeConfigMapperAbstractDb instance;

    protected static synchronized RuntimeConfigMapperAbstractDb getDatabaseInstance() {
        if (instance == null) {
            String jdbcDriver = StaticConfigImpl.getInstance().getJdbcDriver().toLowerCase();

            if (jdbcDriver.contains("mysql")) {
                instance = new RuntimeConfigMapperMysql();
            } else if (jdbcDriver.contains("postgresql")) {
                instance = new RuntimeConfigMapperPostgresql();
            } else if (jdbcDriver.contains("hsqldb")) {
                instance = new RuntimeConfigMapperHsqldb();
            } else {
                log.fatal("Could not determine type of database support from jdbc driver \"" + StaticConfigImpl.getInstance().getJdbcDriver() + "\"");
            }
        }
        return instance;
    }
    
    private Connection getUnitTestDBConnection() throws Exception {
        Class.forName(StaticConfigImpl.getInstance().getJdbcDriver());
        String databaseUrl = "jdbc:postgresql://localhost:5432/sosigwlog";
        String username = "sosigwlog";
        String password = "sosigwlog";
        Connection db = DriverManager.getConnection(databaseUrl, username, password);
        return db;
    }

    protected Connection getConnectionFromPool() throws Exception {
        if (StaticConfigImpl.isUnitTestMode()) {
            return getUnitTestDBConnection();
        }
        String name = StaticConfigImpl.getInstance().getDataSourceName();
        Context ctx = new InitialContext();

        DataSource ds = (DataSource)ctx.lookup(name);
        if (ds == null) {
            throw new IllegalStateException("Unable to lookup Datasource in jndi from '" + name + "'");
        }

        return ds.getConnection();
    }
    
    public void saveConfig(RuntimeConfigState state) {
        Connection connection;
        try {
            connection = getConnectionFromPool();
        } catch (Exception e) {
            log.error("Unable to get connection", e);
            throw new RuntimeException(e);
        }
        boolean success = false;
        try {
            connection.setAutoCommit(false);
            saveConfig0(connection, state);
            connection.commit();
            success = true;
        } catch (SQLException e) {
            log.error("Error saving config", e);
            throw new RuntimeException(e);
        } finally {
            if (!success) {
                try {
                    if (connection != null) {
                        connection.rollback();
                    }
                } catch (SQLException e) {
                    log.warn("Error rolling back. Ignored", e);
                }
            }
            try {
                if (connection != null) {
                    connection.close();
                }
            } catch (SQLException e) {
                log.warn("Error closing connection. Ignored", e);
            }
        }
        
    }
    
    private void saveConfig0(Connection connection, RuntimeConfigState state) throws SQLException {
        PreparedStatement topInsert = null;
        PreparedStatement topSelect = null;
        PreparedStatement elmInsert = null;
        PreparedStatement elmSelect = null;
        PreparedStatement data = null;

        try {
            topInsert = connection.prepareStatement(getTopInsertSql());
            topSelect = connection.prepareStatement(getTopSelectSql());
            elmInsert = connection.prepareStatement(getElmInsertSql());
            elmSelect = connection.prepareStatement(getElmSelectSql());       
            data = connection.prepareStatement(getDataInsertSql());

            Timestamp timestamp = new Timestamp(state.getTimestamp());
            topInsert.setTimestamp(1, timestamp);
            topInsert.executeUpdate();
            topSelect.setTimestamp(1, timestamp);
            if (topSelect.execute()) {
                ResultSet resultSet = topSelect.getResultSet();
                if (resultSet.next()) {
                    long topNodePK = resultSet.getLong(1);
                    RuntimeConfigItem[] configItems = state.getState();
                    for (RuntimeConfigItem runtimeConfigItem : configItems) {
                        elmInsert.setLong(1, topNodePK);
                        elmInsert.setString(2, runtimeConfigItem.category);
                        elmInsert.setString(3, runtimeConfigItem.key);
                        elmInsert.setInt(4, runtimeConfigItem.value.length);
                        elmInsert.executeUpdate();
                        elmSelect.setLong(1, topNodePK);
                        elmSelect.setString(2, runtimeConfigItem.category);
                        elmSelect.setString(3, runtimeConfigItem.key);

                        if (elmSelect.execute()) {
                            ResultSet elmResultSet = elmSelect.getResultSet();
                            if (elmResultSet.next()) {
                                long elmNodePK = elmResultSet.getLong(1);
                                for (int i = 0; i < runtimeConfigItem.value.length; i++) {
                                    data.setLong(1, elmNodePK);
                                    data.setInt(2, i);
                                    if (runtimeConfigItem.value[i] == null || runtimeConfigItem.value[i].length() == 0) {
                                        data.setNull(3, Types.VARCHAR);
                                    } else {
                                        data.setString(3, runtimeConfigItem.value[i]);
                                    }
                                    data.addBatch();
                                }
                            } else {
                                throw new SQLException("Did not receive PK from insert");
                            }                        

                            try { elmResultSet.close(); } catch (SQLException e) {}

                        } else {
                            throw new SQLException("Did not receive PK from insert");
                        }
                    }
                } else {
                    throw new SQLException("Did not receive PK from insert");
                }
                try { resultSet.close(); } catch (SQLException e) {}
            } else {
                throw new SQLException("Did not receive PK from insert");
            }
            data.executeBatch();
        } finally {
            try { if (topInsert != null) topInsert.close(); } catch (SQLException e) {}
            try { if (topSelect != null) topSelect.close(); } catch (SQLException e) {}
            try { if (elmInsert != null) elmInsert.close(); } catch (SQLException e) {}
            try { if (elmSelect != null) elmSelect.close(); } catch (SQLException e) {}
            try { if (data != null) data.close(); } catch (SQLException e) {}
        }
    }

    abstract protected String getDataInsertSql();
    abstract protected String getElmSelectSql();
    abstract protected String getElmInsertSql();
    abstract protected String getTopSelectSql();
    abstract protected String getTopInsertSql();

    public RuntimeConfigState loadLatestConfig() {
        Connection connection = null;
        PreparedStatement topStmt = null;
        PreparedStatement dataStmt = null;
        try {
            connection = getConnectionFromPool();
            topStmt = connection.prepareStatement("SELECT configpk, moment FROM " + SOSIGWLOG_RUNTIMECONFIG + " WHERE configpk=(SELECT MAX(configpk) FROM " + SOSIGWLOG_RUNTIMECONFIG + ")");
            if (topStmt.execute()) {
                ResultSet topSet = topStmt.getResultSet();
                if (topSet.next()) {
                    long configPK = topSet.getLong(1);
                    long timestamp = topSet.getTimestamp(2).getTime();
                    dataStmt = connection.prepareStatement("SELECT elm.elmpk, elm.category, elm.elmkey, elm.valuecount, data.idx, data.value FROM "
                            + SOSIGWLOG_RUNTIMECONFIG_ELM + " elm INNER JOIN " 
                            + SOSIGWLOG_RUNTIMECONFIG_DATA + " data ON elm.elmpk=data.elmpk " 
                            + "WHERE elm.configpk = ? "
                            + "ORDER BY elm.elmpk");
                    dataStmt.setLong(1, configPK);
                    if (dataStmt.execute()) {
                        ResultSet resultSet = dataStmt.getResultSet();
                        List<RuntimeConfigItem> items = new ArrayList<RuntimeConfigItem>();
                        String[] values = null;
                        int valuesCompleted = 0;
                        while (resultSet.next()) {
                            //long elmPK = resultSet.getLong(1);
                            String category = resultSet.getString(2);
                            String key= resultSet.getString(3);
                            int valueCount = resultSet.getInt(4);
                            int idx = resultSet.getInt(5);
                            String value= resultSet.getString(6);

                            if (values == null) {
                                values = new String[valueCount];
                                valuesCompleted = 0;
                            }
                            values[idx] = value;
                            valuesCompleted++;

                            if (valuesCompleted == valueCount) {
                                items.add(new RuntimeConfigItem(category, key, values));
                                values = null;
                            }
                        }
                        
                        try { resultSet.close(); } catch (SQLException ex) {}
                        
                        return new RuntimeConfigState(timestamp, items.toArray(new RuntimeConfigItem[items.size()]));
                    }
                }
                try { topSet.close(); } catch (SQLException e) {}
            }
        } catch (Exception e) {
            if (StaticConfigImpl.getInstance().isProductionMode()) {
                throw new ConfigurationException("Unable to load latest config from database while running in production mode. Instance will not be started. If needed, place instance in bootstrap mode by editing " + StaticConfigImpl.getInstance().getConfigfilePath() + " and set property " + StaticConfigImpl.KEY_SOSIGW_MODE, e);
            } else {
                log.error("Unable to load latest config: " + e.getMessage() + ". Starting instance anyway, as mode is not production", e);
            }
        } finally {
            try { if (topStmt != null) { topStmt.close(); } } catch (SQLException e) {}
            try { if (dataStmt != null) { dataStmt.close(); } } catch (SQLException e) {}
            try {
                if (connection != null) {
                    connection.close();
                }
            } catch (SQLException e) {
                log.warn("Error closing connection: " + e.getMessage(), e);
            }
        }
        return null;
    }

}
