/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.config;

import java.io.IOException;
import java.io.Serializable;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

/**
 * Holds all the current runtime state, meant for transmission among peers, not for editing or direct lookup: Use RuntimeConfig as facade.
 * @author Morten Grouleff <mgr@trifork.com>
 *
 */
public class RuntimeConfigState implements Serializable {
    // These are the ones to transmit across the wire among the cluster members, and store in the database. 
    private final long timestamp;
    private final RuntimeConfigItem[] state;

    // For efficient access, this is where the lookups are performed. Never serialized nor transmitted.
    protected transient Map<String, Map<String, String[]>> cacheByCategory;
    protected transient Map<String, String[]> META_INFO;
    
    private void initMetaInfo() {
        META_INFO = new HashMap<String, String[]>();
        META_INFO.put(RuntimeConfig.CATEGORY_CLIENTACCESSWHITELIST, new String[] { "IP-address", "Salt", "Comment", "Allow clients matching pattern access to sosigw. If both are set they must both match."});
        META_INFO.put(RuntimeConfig.CATEGORY_SERVICEACCESSWHITELIST, new String[] { "WS-Addressing-to regexp", "WS-Addressing-action regexp", "Comment", "Allow access to services mathing the regexps. If both are set they must both match to allow access to the service"});
        META_INFO.put(RuntimeConfig.CATEGORY_SERVICEPRESERVE_WSA, new String[] { "WS-Addressing-to regexp", "WS-Addressing-action regexp", "Comment", "Preserve WS-addressing headers in request to services mathing the regexps. If both are set they must both match."});
        META_INFO.put(RuntimeConfig.CATEGORY_PROPERTIES, new String[] { "Value", "Comment", "Global settings. See the documentation for the complete list of usefull keys."});
        META_INFO.put(RuntimeConfig.CATEGORY_CONSOLE_USERS, new String[] { "Fornavn", "Efternavn", "Email", "Users who are granted access to the administrative console. The 'key' must be the CPR of the user."});
    }
    
    public RuntimeConfigState(long timestamp, RuntimeConfigItem[] state) {
        this.timestamp = timestamp;
        this.state = state;
        initMetaInfo();
        buildCaches();
    }
    
    public long getTimestamp() {
        return timestamp;
    }

    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        initMetaInfo();
        buildCaches();
    }
    
    protected void addEntry(String category, String key, String[] values) {
        Map<String, String[]> catMap = cacheByCategory.get(category);
        if (catMap == null) {
            catMap = new HashMap<String, String[]>();
            cacheByCategory.put(category, catMap);
        }
        String[] newValues = new String[values.length];
        for (int i = 0; i < values.length; i++) {
            String val = values[i];
            if (val == null || val.length() == 0) {
                //skip
            } else {
                newValues[i] = val;
            }
        }
        catMap.put(key, newValues);
    }
    
    protected void buildCaches() {
        cacheByCategory = new HashMap<String, Map<String, String[]>>();
        for (RuntimeConfigItem item : state) {
            addEntry(item.category, item.key, item.value);
        }
    }
    
    public Map<String, String[]> getCategoryMap(String category) {
        if (RuntimeConfig.CATEGORY_METAINFO.equals(category)) {
            return META_INFO ;
        }
        Map<String, String[]> catMap = cacheByCategory.get(category);
        if (catMap != null) {
            return catMap;
        }
        return Collections.emptyMap();
    }
    
    public String[] getValue(String category, String key) {
        Map<String, String[]> catMap = getCategoryMap(category);
        return catMap.get(key);
    }
    
    public RuntimeConfigEditState getEditableVersion() {
        return new RuntimeConfigEditState(state);
    }

    public RuntimeConfigItem[] getState() {
        return state;
    }
    
}
