/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.config;

import com.trifork.sosigw.util.Util;
import org.apache.log4j.Logger;

import java.io.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Properties;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class StaticConfigImpl implements StaticConfigApi {
    private static final Logger log = Logger.getLogger(StaticConfigImpl.class); 

    public static final String KEY_CLUSTERIPPREFIX = "cluster.ip.prefix";
    public static final String KEY_MULTICASTIPADDRESS = "cluster.multicast.ip.address";
    public static final String KEY_MULTICASTPORT = "cluster.multicast.ip.port";
    public static final String KEY_CLUSTER_TUNING = "cluster.tuningparam";
    public static final String KEY_DGWS_VERSION= "dgws.version";
    public static final String KEY_KEEPALIVE_ENABLED = "keepalive.enabled";
    public static final String KEY_FMK_ISALIVE_URL = "fmk.isalive.url";
    public static final String KEY_CENTRAL_AUDITLOG_TP_KEY = "auditlog.central.tableprefix";
    public static final String KEY_JDBC_DRIVER = "jdbc.driver";
    public static final String KEY_GLOBAL_AUDITLOG_ENABLED = "global.auditlog.enabled";
    public static final String KEY_AUDITLOG_TO_DB_ENABLED = "auditlog.to.db.enabled";
    public static final String KEY_JDBC_DATASOURCE_NAME = "jdbc.datasource";

    public static final String KEY_SOSIGW_MODE = "sosigw.mode";
    public static final String KEY_SLA_LOG_ENABLED = "sosigw.slalog";

    private static final String KEY_RESTRICTED_WEBSERVICE_ENABLED = "restricted.webservice.enabled";
    private static final String KEY_STS_MBEANS_ENABLED = "sts.mbeans.enabled";
    private static final String DEFAULT_JDBC_DATASOURCE_NAME = "java:comp/env/jdbc/postgres"; // java:jdbc/sosigwlocal;
    public static final String DEFAULT_FMK_ISALIVE_URL = "http://trifork.lms.trifork.com/services/isalive";
    public static final String DEFAULT_CENTRAL_AUDITLOG_TP_KEY = null;

    public static final String DEFAULT_JDBC_DRIVER = "org.postgresql.Driver";
    private static StaticConfigApi impl;
    private static boolean unitTestMode;
    public static final String KEY_RUNTIMECONFIG_STORAGE = "runtimeconfig.storage.type";
    private Properties properties;
    private File configfile;

    public StaticConfigImpl() {
        String defaultConfigFile = Util.getConfigFiledirName() + "/sosigw-staticconf.properties";
        configfile = new File(System.getProperty("sosigw.config", defaultConfigFile));

        properties = new Properties();
        InputStream inStream = null;
        try {
            inStream = new FileInputStream(configfile);
            properties.load(inStream);
            log.info("Loaded static configuration " + configfile.getAbsolutePath());
        } catch (FileNotFoundException e) {
            String message = "SOSI-GW: Unable to load configfile.\n"
                    + "SOSI-GW:    Make sure it is available in '" + configfile.getAbsolutePath() + "'\n"
                    + "SOSI-GW:    or specify which file to load using '-Dsosigw.config=[filepath]'";
            System.err.println(message);
            log.error(message);
        } catch (IOException e) {
            log.error("Unable to load static configuration " + configfile.getAbsolutePath(), e);
        } finally {
            if (inStream != null) {
                try {
                    inStream.close();
                } catch (IOException e) {
                    log.warn("Error closing file " + configfile.getAbsolutePath(), e);
                }
            }
        }
    }

    public StaticConfigImpl(Properties properties) {
        this.properties = properties;
    }

    public static StaticConfigApi getInstance() {
        return impl;
    }
    
    public static void setInstance(StaticConfigApi impl) {
        log.warn("Setting configuration - should never happen in production");
        StaticConfigImpl.impl = impl;
    }

    public static void initSettings() {
        if (impl == null) {
            impl = new StaticConfigImpl();
        }
    }

    public String getConfigfilePath() {
        return configfile != null ? configfile.getAbsolutePath() : "no config file";
    }

    public String getMode() {
        return properties.getProperty(KEY_SOSIGW_MODE, "developer"); // Default mode is developer.
    }
    
    public boolean isDeveloperRestrictedMode() {
        return "developer".equalsIgnoreCase(getMode());
    }

    public boolean isBootstrapRestrictedMode() {
        return "bootstrap".equalsIgnoreCase(getMode());
    }

    public boolean isProductionMode() {
        return "production".equalsIgnoreCase(getMode());
    }

    public boolean isTestMode() {
        return "test".equalsIgnoreCase(getMode());
    }
    
    public boolean isSLALogEnabled() {
        return isProductionMode() || "true".equalsIgnoreCase(properties.getProperty(KEY_SLA_LOG_ENABLED, "false")); // Defaults to only SLA log in production mode.
    }

    public RuntimeConfigStorageType getRuntimeconfigStorage() {
        String storageVal = properties.getProperty(KEY_RUNTIMECONFIG_STORAGE);
        if (storageVal == null || "db".equals(storageVal)) {
            return RuntimeConfigStorageType.DATABASE;
        } else if ("properties".equals(storageVal)) {
            return RuntimeConfigStorageType.PROPERTIES;
        } else {
            throw new ConfigurationException("Unknown runtime storage type '" + storageVal + "' specified as static conf " + KEY_RUNTIMECONFIG_STORAGE);
        }
    }

    public Properties getProperties() {
        Properties copy = new Properties();
        for (Object key : properties.keySet()) {
            copy.put(key, properties.get(key));
        }
        
        return copy;
    }

    public String getClusterIpPrefix() {
        return properties.getProperty(KEY_CLUSTERIPPREFIX, null);
    }

    public InetAddress getMulticastIpAddress() throws UnknownHostException {
        return InetAddress.getByName(properties.getProperty(KEY_MULTICASTIPADDRESS, "224.0.2.7")); // Any legal multicast address, e.g. 224.1.7.13
    }

    public int getMulticastPort() {
        String s = properties.getProperty(KEY_MULTICASTPORT, "3456");
        return Integer.parseInt(s);
        // Let potential NumberFormatException flow through, aborting the start.
    }
    
    public int getClusterTuningParam(String param, int def) {
        try {
            return Integer.parseInt(properties.getProperty(KEY_CLUSTER_TUNING + "." + param, "" + def));
        } catch (Exception e) {
            log.warn(e.getMessage(), e);
            return def;
        }
    }

    public static void setUnitTestMode() {
        unitTestMode = true;
    }
    public static boolean isUnitTestMode() {
        return unitTestMode;
    }

    public String getSupportedDGWSVersion() {
        return properties.getProperty(KEY_DGWS_VERSION, "1.0.1");
    }

    public boolean useLocalTimeInSaml() {
        // Use local time with dgws 1.0, zulu time otherwise
        return "1.0".equals(getSupportedDGWSVersion());
    }

    public boolean isKeepAliveEnabled() { 
        return "true".equals(properties.getProperty(KEY_KEEPALIVE_ENABLED, "false"));
    }

    // TODO: Flyt til RuntimeConfig    
    public String getFMKIsAliveURL() { 
        return properties.getProperty(KEY_FMK_ISALIVE_URL, DEFAULT_FMK_ISALIVE_URL);
    }
    public String getCentralAuditLogTablePrefix() {
        return properties.getProperty(KEY_CENTRAL_AUDITLOG_TP_KEY, DEFAULT_CENTRAL_AUDITLOG_TP_KEY);        
    }

    public String getJdbcDriver() {
        return properties.getProperty(KEY_JDBC_DRIVER, DEFAULT_JDBC_DRIVER);
    }

    public boolean isGlobalAuditLogEnabled() {
        return "true".equals(properties.getProperty(KEY_GLOBAL_AUDITLOG_ENABLED, "true"));
    }

    public boolean isAuditLogToDbEnabled() {
        return "true".equals(properties.getProperty(KEY_AUDITLOG_TO_DB_ENABLED, "true"));
    }

    public String getDataSourceName() {
        return properties.getProperty(KEY_JDBC_DATASOURCE_NAME, DEFAULT_JDBC_DATASOURCE_NAME);
    }

    public boolean isRestrictedWebserviceEnabled() {
        return "true".equals(properties.getProperty(KEY_RESTRICTED_WEBSERVICE_ENABLED, "true"));
    }

    public boolean isSTSMBeansEnabled() {
        String strMBeanEnabled = (String) properties.get(KEY_STS_MBEANS_ENABLED);
        if (strMBeanEnabled == null || "true".equals(strMBeanEnabled)) {
            return true;
        }
        return false;
    }

}
