/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.proxy;

import com.trifork.sosigw.api.Constants;
import com.trifork.sosigw.util.timing.Timer;
import com.trifork.sosigw.ws.SosiGWFacadeBusiness;
import dk.sosi.gw._2007_09.DGWSFault;
import dk.sosi.gw._2007_09.ImplicitLoginHeader;
import dk.sosi.gw._2007_09.RequestIdCardDigestForSigningResponse;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.constants.MedComTags;
import dk.sosi.seal.model.constants.NameSpaces;
import org.apache.log4j.Logger;
import org.xmlsoap.schemas.soap.envelope.*;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.bind.*;
import javax.xml.namespace.QName;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class ImplicitLoginHandler {
    private static final Logger log = Logger.getLogger(ImplicitLoginHandler.class); 

    private final UserIDCard userIDCard;
    private final String cacheKey;

    public ImplicitLoginHandler(UserIDCard unsignedIDCard, String cacheKey) {
        this.userIDCard = unsignedIDCard;
        this.cacheKey = cacheKey;
        log.debug("ImplicitLogin for cacheKey=" + cacheKey);
    }

    public void sendResponse(HttpServletRequest req, HttpServletResponse resp) throws IOException {
        RequestIdCardDigestForSigningResponse digestForSigningImpl;
        try {
            digestForSigningImpl = SosiGWFacadeBusiness.requestIdCardDigestForSigningImpl(userIDCard, cacheKey);
        } catch (DGWSFault e) {
            sendInternalError(resp, e);
            return;
        }

        try {
            ImplicitLoginHeader implicitLoginHeader = new ImplicitLoginHeader();
            implicitLoginHeader.setRequestIdCardDigestForSigningResponse(digestForSigningImpl);
            
            Header header = new Header();
            header.getAny().add(implicitLoginHeader);

            Envelope envelope = new Envelope();
            envelope.setHeader(header);
            
            Detail detail = new Detail();
            detail.getAny().add(new JAXBElement<String>(new QName(NameSpaces.MEDCOM_SCHEMA, MedComTags.FAULT_CODE), String.class, Constants.FAULTCODE_NO_VALID_IDCARD_IN_CACHE));
            
            Fault fault = new Fault();
            fault.setDetail(detail);
            fault.setFaultcode(new QName(NameSpaces.SOAP_SCHEMA, "Server"));
            fault.setFaultstring("Login required");
            
            Body body = new Body();
            body.getAny().add(new JAXBElement<Fault>(new QName(NameSpaces.SOAP_SCHEMA, "Fault"), Fault.class, fault));
            envelope.setBody(body);
            
            ByteArrayOutputStream result = new ByteArrayOutputStream();
            Timer.start(ImplicitLoginHandler.class, "jaxb.marshal");
            try {
                Marshaller marshaller = getMarshaller();
                marshaller.marshal(envelope, result);
            } finally {
                Timer.stop(ImplicitLoginHandler.class, "jaxb.marshal");
            }
            
            // We are as sure as possible that we will not fail, so now is the time to grab the output.
            resp.setStatus(500);
            resp.setContentType("text/xml; charset=utf-8");
            resp.setContentLength(result.size());
            result.writeTo(resp.getOutputStream());
        } catch (JAXBException e) {
            sendInternalError(resp, e);
            return;
        } catch (IOException e) {
            sendInternalError(resp, e);
            return;
        }
    }

    private static final ThreadLocal<Marshaller> soapMarshaller = new ThreadLocal<Marshaller>();
    
    private Marshaller getMarshaller() throws JAXBException, PropertyException {
        Marshaller marshaller = soapMarshaller.get();
        if (marshaller == null) {
            JAXBContext jaxbContext = JAXBContext.newInstance(new Class[] {
                    ImplicitLoginHeader.class,
                    Header.class,
                    Envelope.class,
                    Body.class,
                    Detail.class,
                    Fault.class,
                    } );
            marshaller = jaxbContext.createMarshaller();
            marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, new Boolean(true));
            marshaller.setProperty(Marshaller.JAXB_ENCODING, Constants.UTF_8);
            soapMarshaller.set(marshaller);
        }
        return marshaller;
    }

    private void sendInternalError(HttpServletResponse resp, Exception e) throws IOException {
        log.error("sendInternalError", e);
        resp.setStatus(500);
        DGWSFault fault = new DGWSFault(e.getMessage(), Constants.FAULTCODE_INTERNAL_ERROR);
        
        try {            
            JAXBContext jaxbContext = JAXBContext.newInstance("org.xmlsoap.schemas.soap.envelope");
            Marshaller marshaller = jaxbContext.createMarshaller();
            marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, new Boolean(true));
            ByteArrayOutputStream result = new ByteArrayOutputStream();
            marshaller.marshal(fault, result);

            resp.setContentLength(result.size());
            result.writeTo(resp.getOutputStream());
        } catch (JAXBException e2) {
            log.error("Error sending internal error", e2);
            throw new RuntimeException(e2);
        }
    }
}
