/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.proxy;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.Constants;
import com.trifork.sosigw.api.IDCardCache;
import com.trifork.sosigw.api.IDCardCacheEntry;
import com.trifork.sosigw.api.IDCardState;
import com.trifork.sosigw.auditlog.AuditLog;
import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.config.ConfigurationException;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.StaticConfigApi;
import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.util.*;
import com.trifork.sosigw.util.timing.Timer;
import dk.sdsd.nsp.slalog.api.SLALogException;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sosi.gw._2007_09.DGWSFault;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.constants.FaultCodeValues;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.modelbuilders.IDCardModelBuilder;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.commons.httpclient.*;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.RequestEntity;
import org.apache.commons.httpclient.params.HttpConnectionManagerParams;
import org.apache.log4j.Logger;
import org.w3c.dom.*;
import org.w3c.dom.ls.LSSerializer;

import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;

/**
 *
 *
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class ProxyRequestHandler {
    private static final StaticConfigApi STATIC_CONF = StaticConfigImpl.getInstance();

    public static final Logger log = Logger.getLogger(ProxyRequestHandler.class);
    public static final Logger contentLog = Logger.getLogger(ProxyRequestHandler.class.getName() + ".content");
    private static final MultiThreadedHttpConnectionManager connectionManager = makeConnectionManager();

    private final HttpServletRequest req;
    private final HttpServletResponse resp;
    private final HttpClient httpClient;

    private String localHostAddress;

    public static void kill() {
        // Wait a while, then kill all our connections. This allows ongoing transactions to complete, within that
        // window.
        Thread thread = new Thread() {

            public void run() {
                try {
                    Thread.sleep(60 * 1000);
                } catch (InterruptedException e) {
                    // ignore
                }
                try {
                    connectionManager.shutdown();
                } catch (Throwable ignore) {
                } // Ignore everything.
                try {
                    MultiThreadedHttpConnectionManager.shutdownAll();
                } catch (Throwable t) {
                } // Ignore everything.
            }
        };
        thread.setDaemon(true);
        thread.start();
    }

    private static MultiThreadedHttpConnectionManager makeConnectionManager() {
        MultiThreadedHttpConnectionManager manager = new MultiThreadedHttpConnectionManager();
        HttpConnectionManagerParams params = new HttpConnectionManagerParams();
        params.setMaxConnectionsPerHost(HostConfiguration.ANY_HOST_CONFIGURATION,
                STATIC_CONF.getClusterTuningParam("max.connections.per.host", 400));
        params.setMaxTotalConnections(STATIC_CONF.getClusterTuningParam("max.connections.in.total", 400));
        params.setTcpNoDelay(true);
        manager.setParams(params);
        return manager;
    }

    public ProxyRequestHandler(HttpServletRequest req, HttpServletResponse resp) {

        this.req = req;
        this.resp = resp;

        httpClient = new HttpClient(connectionManager);
        try {
            localHostAddress = InetAddress.getLocalHost().getHostAddress();
        } catch (UnknownHostException e) {
            localHostAddress = "unknown";
        }
    }

    public void proxyRequest() {
        PostMethod postMethod = null;
        Document headerDocument = null;
        WsAddressingHandler addressingHandler = new WsAddressingHandler(RuntimeConfig.isForceSSLRequests());

        SLALogItem slaLogItem = Util.getSLALogItem();
        slaLogItem.setCallResultOk();
        try {
            MultipartHeaderFix multipartHeaderFix = new MultipartHeaderFix();
            AuthenticateByIP(req);
            InputStream inputStream = req.getInputStream();
            boolean multipart = isMultipart(req);
            if (multipart) {
                // Filter away attachments which are linked from the header, saves contents of the attachments for later
                inputStream = multipartHeaderFix.extractAndFilterMtomHeaderAttachents(inputStream, req.getContentType());
            }
            HeaderReadingInputStream headerReadingInputStream = new HeaderReadingInputStream(inputStream);
            String soapNamespace = headerReadingInputStream.getSoapNamespace();
            long origHeaderSize = headerReadingInputStream.getBytesRead();
            InputStream remainderStream = headerReadingInputStream.getRemainderStream();
            StringBuffer prolog = headerReadingInputStream.getProlog();

            StringBuffer soapHeader = headerReadingInputStream.getSoapHeader();
            headerDocument = SoapHeader.getSoapEnvelopeNode(soapHeader, soapNamespace);

            NodeList headerNodes = headerDocument.getElementsByTagName(soapNamespace + ":" + Constants.SOAP_HEADER);
            if (headerNodes.getLength() != 1) {
                throw new ServletException("Got " + headerNodes.getLength() + " soap-headers");
            }
            if (multipart) {
                // Replaces the MTOM links with the actual content from the attachement
                replaceMtomIncludes(headerDocument, multipartHeaderFix.getHeaderAttachmentMap());
            }
            AuthenticateBySalt(headerDocument);

            NodeList maxAgeElements = extractMaxAgeElements(headerDocument);

            IDCardCacheEntry iDCardCacheEntry = null;
            if (passThrough(headerDocument)) {
                slaLogItem.addCallParameter("sosigw-pass-through", "header-in-request");
            } else if (RuntimeConfig.isPassThroughOnSignatureEnabled() && hasSignatureElements(headerDocument)) {
                slaLogItem.addCallParameter("sosigw-pass-through", "signature-present");
            } else {
                String cacheKey = computeCacheKey(extractNameID(headerDocument));
                iDCardCacheEntry = waitForSignedIDCard(cacheKey);

                boolean replaceSucceeded = replaceIDCard(headerDocument, iDCardCacheEntry);
                boolean idCardMaxAgeHeaderExpired = false;
                boolean idCardTooOld = false;
                if (replaceSucceeded) {
                    idCardTooOld = isIDCardTooOld(iDCardCacheEntry);
                    idCardMaxAgeHeaderExpired = isMaxAgeHeaderExpired(maxAgeElements, iDCardCacheEntry);
                }
                if ((!replaceSucceeded) || idCardTooOld || idCardMaxAgeHeaderExpired) {
                    ImplicitLoginHandler implicitLoginHandler = new ImplicitLoginHandler(extractIDCard(headerDocument), cacheKey);
                    implicitLoginHandler.sendResponse(req, resp);
                    if (!replaceSucceeded) {
                        slaLogItem.setCallResultError("Login required (idcard not found)");
                    } else if (idCardTooOld) {
                        slaLogItem.setCallResultError("Login required (idcard expired)");
                    } else {
                        slaLogItem.setCallResultError("Login required (supplied max age expired)");
                    }
                    return;
                }
            }

            // Remove MaxAge whether or not passthrough is in effect
            if (maxAgeElements != null) {
                if (log.isDebugEnabled()) {
                    log.debug("Removing " + maxAgeElements.getLength() + " MaxAge elements");
                }
                removeChildNodes(maxAgeElements);
            }

            Node endpoint = addressingHandler.getWSANode(headerDocument, "EndpointReference");
            if (endpoint != null) {
                slaLogItem.setSourceEndpoint(endpoint.getTextContent());
            }

            String messageId = extractMessageId(headerDocument, slaLogItem);
            if (iDCardCacheEntry != null) {
                slaLogItem.addCallParameter("sosigw-idCard-cacheKey", iDCardCacheEntry.getCacheKey());
            }

            WSAHeaderInfo wsaHeaderInfo = addressingHandler.handleWSAddressing(headerDocument, req.getHeader("soapaction"));
            slaLogItem.setSourceSOAPAction(wsaHeaderInfo.action);
            slaLogItem.addCallParameter("sosigw-wsa-to", wsaHeaderInfo.to);
            checkTargetAccess(wsaHeaderInfo);

            if (iDCardCacheEntry != null) {
                AuditLog.logIdcardUse(iDCardCacheEntry, req, wsaHeaderInfo.to, wsaHeaderInfo.action);
            }

            String newHeaders = serializeNewHeader(headerDocument);
            if (contentLog.isDebugEnabled()) {
                contentLog.debug(newHeaders);
            }
            RequestEntity requestEntity = makeRequestEntity(req, req.getContentLength() - multipartHeaderFix.getBytesRemoved(),
                    newHeaders, origHeaderSize, remainderStream, prolog);

            postMethod = new PostMethod(wsaHeaderInfo.to);
            addNspRequestIdHttpHeader(postMethod, slaLogItem);
            addForwardedHttpHeaders(postMethod, slaLogItem);

            invokeRemoteService(postMethod, requestEntity, wsaHeaderInfo, messageId);
        } catch (DGWSFault e) {
            // Expected to be thrown while response is uncommitted.
            sendDGWSFault(headerDocument, resp, e);
            slaLogItem.setCallResultError("DGWS Fejl: " + e.getMessage());
        } catch (IOException e) {
            sendInternalErrorResponse(headerDocument, resp, e);
            slaLogItem.setCallResultError("Intern IO Fejl: " + e.getMessage());
            return;
        } catch (Exception e) {
            sendInternalErrorResponse(headerDocument, resp, e);
            slaLogItem.setCallResultError("Intern Fejl: " + e.getMessage());
            return;
        } finally {
            try {
                if (postMethod != null) {
                    postMethod.releaseConnection();
                }
            } catch (Exception e) {
                log.debug("Ignoring: " + e.getMessage(), e);
            }
        }
    }

    private void replaceMtomIncludes(Document headerDocument, Map<String, String> headerAttachmentMap) {
        if (headerAttachmentMap == null || headerAttachmentMap.size() == 0) {
            return;
        }
        Element documentElement = headerDocument.getDocumentElement();
        replaceMtomIncludesRecursive(null, documentElement, headerAttachmentMap);
    }

    private void replaceMtomIncludesRecursive(Node parent, Node current, Map<String, String> headerAttachmentMap) {
        if (current == null) {
            return;
        }
        if (parent != null && "Include".equalsIgnoreCase(current.getLocalName())) {
            Node urlAttribute = current.getAttributes().getNamedItem("href");
            if (urlAttribute != null) {
                String newContent = headerAttachmentMap.get(contentIdFromUrl(urlAttribute.getTextContent()));
                if (newContent != null) {
                    parent.removeChild(current);
                    parent.setTextContent(newContent);
                }
            }
        } else {
            NodeList childNodes = current.getChildNodes();
            for (int i=0; i<childNodes.getLength(); ++i) {
                replaceMtomIncludesRecursive(current, childNodes.item(i), headerAttachmentMap);
            }
        }
    }

    private String contentIdFromUrl(String url) {
        if (url.length() <= 4) {
            return url;
        }
        return url.substring(4); // We remove the "cid:" part
    }

    private boolean isMultipart(HttpServletRequest req) {
        String header = req.getHeader("Content-Type");
        if (header != null && header.contains("multipart/")) {
            return true;
        }
        return false;
    }

    private void addForwardedHttpHeaders(PostMethod postMethod, SLALogItem slaLogItem) {
        // Some Http headers should be read, sla-logged and passed on to the backend service being called.
        String keyNspForwardHttpHeaderNames = RuntimeConfig.getKeyNspForwardHttpHeaderNames();
        if (keyNspForwardHttpHeaderNames != null && keyNspForwardHttpHeaderNames.length() > 0) {
            String[] headerNames = keyNspForwardHttpHeaderNames.split(",");
            for (String headerName : headerNames) {
                headerName = headerName.trim();
                String reqHeader = req.getHeader(headerName);
                if (reqHeader != null) {
                    log.debug("Forwarding HTTP header named: " + headerName);
                    slaLogItem.addCallParameter("httpheader-"+headerName, reqHeader);
                    postMethod.addRequestHeader(headerName, reqHeader);
                }
            }
        }
    }

    /**
     * Generate and add a mandatory HTTP requestId header to PostMethod, and logs the value to the sla log
     * @param postMethod post method which will have the headers added
     * @param slaLogItem sla log item
     */
    private void addNspRequestIdHttpHeader(PostMethod postMethod, SLALogItem slaLogItem) {
        // Generate a proxy request id, that will be passed as HTTP header to the backend service.
        // The request id should of cause be sla-logged.
        String proxyRequestId = UUID.randomUUID().toString();
        if (proxyRequestId != null) {
            slaLogItem.addCallParameter("sosigw-proxy-requestid", proxyRequestId);
            postMethod.addRequestHeader(RuntimeConfig.getNspProxyRequestidHttpHeaderName(), proxyRequestId);
        }
    }

    /**
     * Extract message id, and add it to the sla log.
     * @param headerDocument headers
     * @param slaLogItem sla log item
     * @return extracted message id or "unknown"
     */
    private String extractMessageId(Document headerDocument, SLALogItem slaLogItem) {
        Element soapHeaderElm = (Element)headerDocument.getElementsByTagNameNS(Constants.URL_SOAP,
                Constants.SOAP_HEADER).item(0);
        String messageId = "unknown";
        NodeList medcomHeaderElmList = soapHeaderElm.getElementsByTagNameNS(Constants.URL_MEDCOM, "Header");
        if (medcomHeaderElmList != null && medcomHeaderElmList.getLength() == 1) {
            Element medcomHeaderElm = (Element)medcomHeaderElmList.item(0);
            NodeList medcomLinkingElmList = medcomHeaderElm.getElementsByTagNameNS(Constants.URL_MEDCOM, "Linking");
            Element medcomLinkingElm = null;
            if (medcomLinkingElmList != null && medcomLinkingElmList.getLength() == 1) {
                medcomLinkingElm = (Element)medcomLinkingElmList.item(0);
                try {
                    Element medcomMessageIDElm = getChild(medcomLinkingElm, Constants.URL_MEDCOM, "MessageID");
                    messageId = medcomMessageIDElm.getTextContent();
                } catch (Exception ex) {
                    log.error("Unable to extract message ID from request", ex);
                }
                slaLogItem.setMessageId(messageId);
            }
        }
        return messageId;
    }

    private boolean passThrough(Document headerDocument) {
        NodeList PassThroughElements = headerDocument.getElementsByTagNameNS(Constants.HTTP_SOSI_DK_GW_2007_09_01, Constants.PASS_THROUGH);
        boolean passThrough = PassThroughElements != null && PassThroughElements.getLength() > 0;
        removeChildNodes(PassThroughElements);
        return passThrough;
    }

    private IDCardClusterCacheEntry waitForSignedIDCard(String cacheKey) {
        IDCardCache cardCache = Factory.getIDCardCache();
        IDCardClusterCacheEntry cacheEntry = cardCache.findIDCardOrNull(cacheKey);
        if (Factory.getIntercom().isSingleMember()) {
            return cacheEntry;
        }
        if (cacheEntry == null || cacheEntry.getState() == IDCardState.REVOKED) {
            // We wait a short while, to allow the message to get through...
            log.debug("Waiting for entry for cacheKey=" + cacheKey + " to arrive.");

            try {
                int millis = STATIC_CONF.getClusterTuningParam("implicit.login.unknown.delay", 500);
                Thread.sleep(millis);
            } catch (InterruptedException e) {
                // ignore.
            }
            cacheEntry = cardCache.findIDCardOrNull(cacheKey);
        }

        if (cacheEntry != null && cacheEntry.getState() != IDCardState.SIGNED
                && cacheEntry.getState() != IDCardState.REVOKED) {
            // IDCard exists and is not signed. It is very probable that another node in
            // the cluster has done so but not propagated this to this node yet.
            // Wait a while to allow the message to get through.
            // When we finally give up, we force the user to login again, as a result of implicit-login.
            if (log.isDebugEnabled()) {
                log.debug("Waiting for entry for cacheKey=" + cacheKey + " to become signed. Current state is "
                        + (cacheEntry != null ? cacheEntry.getState() : "no-card"));
            }
            long sleepTime = STATIC_CONF.getClusterTuningParam("implicit.login.unsigned.delay", 100);
            int retryCount = STATIC_CONF.getClusterTuningParam("implicit.login.unsigned.retry", 5);
            for (int i = 0; i < retryCount; i++) {
                try {
                    Thread.sleep(sleepTime);
                } catch (InterruptedException e) {
                    // ignore.
                }
                cacheEntry = cardCache.findIDCardOrNull(cacheKey);
                if (cacheEntry != null && cacheEntry.getState() == IDCardState.SIGNED) {
                    if (log.isDebugEnabled()) {
                        log.debug("Waiting for entry for cacheKey=" + cacheKey + " to become signed. Got signed.");
                    }
                    return cacheEntry;
                }
            }
            if (log.isDebugEnabled()) {
                log.debug("Waiting for entry for cacheKey=" + cacheKey + " to become signed. Gave up. Current state is "
                        + (cacheEntry != null ? cacheEntry.getState() : "no-card"));
            }
        }
        return cacheEntry;
    }

    private String computeCacheKey(String nameID) throws DGWSFault {
        try {
            return Util.computeCacheKey(nameID);
        } catch (ConfigurationException e) {
            throw new DGWSFault(e.getMessage(), Constants.FAULTCODE_INTERNAL_ERROR);
        } catch (HttpHeaderException e) {
            throw new DGWSFault(e.getMessage(), Constants.FAULTCODE_SYNTAX_ERROR);
        }
    }

    protected void invokeRemoteService(PostMethod postMethod, RequestEntity requestEntity, WSAHeaderInfo wsaHeaderInfo,
            String messageId) throws IOException, HttpException, DGWSFault {
        postMethod.setRequestEntity(requestEntity);
        setRequestHeaders(postMethod, req, wsaHeaderInfo);

        Timer.start(ProxyRequestHandler.class, "Sending request");
        // note: do not call Util.setSLALogItem() since there may be stored one already
        SLALogItem slaLogItem = Util.getSLALogger().createLogItem("SOSIGWProxyRequestHandler.invokeRemoteService()",
                "14");
        slaLogItem.setClientIP(localHostAddress);
        slaLogItem.setMessageId(messageId);
        slaLogItem.setSourceSOAPAction(wsaHeaderInfo.action);
        slaLogItem.setTargetEndpoint(wsaHeaderInfo.to);
        try {
            try {
                // Experiment: No keep-alive..
                if (!StaticConfigImpl.getInstance().isKeepAliveEnabled()) {
                    postMethod.addRequestHeader(new Header("Connection", "close"));
                }

                httpClient.executeMethod(postMethod);
                slaLogItem.setRequestSize(requestEntity.getContentLength());
                slaLogItem.setCallResultOk();
            } catch (IOException e) {
                while (e.getCause() instanceof IOException) {
                    log.debug("Unwrapping IOException", e);
                    e = (IOException)e.getCause();
                }
                slaLogItem.setCallResultError("Caught Error invoking target service: " + e.getMessage());
                throw new DGWSFault("An error occurred while invoking action \"" + wsaHeaderInfo.action
                        + "\" on service \"" + wsaHeaderInfo.to + "\": " + e, Constants.FAULTCODE_PROXY_ERROR, e);
            } finally {
                Timer.stop(ProxyRequestHandler.class, "Sending request");
            }

            resp.setStatus(postMethod.getStatusCode());
            if (postMethod.getStatusCode() != 200) {
                slaLogItem.addCallParameter("sosigw-service-request-status", "" + postMethod.getStatusCode());
            }
            Header[] responseHeaders = postMethod.getResponseHeaders();
            for (int i = 0; i < responseHeaders.length; i++) {
                Header header = responseHeaders[i];
                if (!headerIsOnlyApplicableForConnectionToService(header)) {
                    resp.setHeader(header.getName(), header.getValue());
                }
            }

            // stream-copy result.
            Timer.start(ProxyRequestHandler.class, "Sending response");
            try {
                InputStream bodyStream = postMethod.getResponseBodyAsStream();
                ServletOutputStream servletOutputStream = resp.getOutputStream();
                long totalReplySize = Util.copyStream(bodyStream, servletOutputStream);
                slaLogItem.setReplySize(totalReplySize);
                if (Util.getSLALogItem() != null) {
                    Util.getSLALogItem().setReplySize(totalReplySize);
                }
                servletOutputStream.flush();
            } finally {
                Timer.stop(ProxyRequestHandler.class, "Sending response");
            }
        } finally {
            try {
                slaLogItem.store();
            } catch (SLALogException e) {
                log.error("Cannot store SLA Log Item", e);
            }
        }
    }

    /**
     * Some headers are only applicable to the connection between SOSIGW and the service it proxies. They should not be
     * set in the response to the original client. The Transfer-Encoding header will normally specify that the message
     * has to be chunked. The stream written to the out stream will not be chunked, and the header therefore has to be
     * removed. Note that we are currently not removing headers named in the value of the "Connection" header. The http
     * specification states that we should do that. Please see
     * http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.10 for further information.
     */
    private boolean headerIsOnlyApplicableForConnectionToService(Header header) {
        List<String> headerNames = Arrays.asList("transfer-encoding", "connection");
        return headerNames.contains(header.getName().toLowerCase());
    }

    private String extractNameID(Document headerDocument) throws DGWSFault {
        try {
            NodeList assertions = headerDocument.getElementsByTagNameNS(NameSpaces.SAML2ASSERTION_SCHEMA,
                    SAMLTags.ASSERTION);
            for (int i = 0; i < assertions.getLength(); i++) {
                Element elem = (Element)assertions.item(i);
                XmlUtil.registerElementByIdExtended(elem);
                Element idCardElement = DOMUtil.getIDCardElement(headerDocument);
                if (idCardElement != null) {
                    Element subjectNode = (Element)idCardElement.getElementsByTagNameNS(
                            NameSpaces.SAML2ASSERTION_SCHEMA, SAMLTags.SUBJECT).item(0);
                    Node nameIDNode = subjectNode.getElementsByTagNameNS(NameSpaces.SAML2ASSERTION_SCHEMA,
                            SAMLTags.NAMEID).item(0);
                    return nameIDNode.getFirstChild().getNodeValue();

                }
            }
        } catch (RuntimeException e) {
            log.info("Error parsing incoming idcard for nameid", e);
            throw new DGWSFault(Constants.MISSING_USER_IDCARD_IN_REQUEST_MSG,
                    Constants.FAULTCODE_MISSING_USER_IDCARD_IN_REQUEST, e);
        }
        throw new DGWSFault(Constants.MISSING_USER_IDCARD_IN_REQUEST_MSG,
                Constants.FAULTCODE_MISSING_USER_IDCARD_IN_REQUEST);
    }

    private boolean hasSignatureElements(Document headerDocument) throws DGWSFault {
        try {
            NodeList signatureElements = headerDocument.getElementsByTagNameNS(NameSpaces.DSIG_SCHEMA, "Signature");
            if (signatureElements.getLength() == 1) {
                Element signatureElement = (Element) signatureElements.item(0);
                NodeList digestValueElements = signatureElement.getElementsByTagNameNS(NameSpaces.DSIG_SCHEMA, "DigestValue");
                NodeList signatureValueElements = signatureElement.getElementsByTagNameNS(NameSpaces.DSIG_SCHEMA, "SignatureValue");
                NodeList x509CertificateElements = signatureElement.getElementsByTagNameNS(NameSpaces.DSIG_SCHEMA, "X509Certificate");
                if (digestValueElements.getLength() != 1 || digestValueElements.item(0).getTextContent().trim().isEmpty()) {
                    return false;
                }
                if (signatureValueElements.getLength() != 1 || signatureValueElements.item(0).getTextContent().trim().isEmpty()) {
                    return false;
                }
                if (x509CertificateElements.getLength() != 1 || x509CertificateElements.item(0).getTextContent().trim().isEmpty()) {
                    return false;
                }
                return true;
            }
        } catch (RuntimeException e) {
            log.info("Error parsing incoming idcard for existing signature", e);
            throw new DGWSFault(Constants.MISSING_USER_IDCARD_IN_REQUEST_MSG,
                    Constants.FAULTCODE_MISSING_USER_IDCARD_IN_REQUEST, e);
        }
        return false;
    }

    protected UserIDCard extractIDCard(Document doc) throws DGWSFault {
        IDCard card = new IDCardModelBuilder().buildModel(doc);
        if (card instanceof UserIDCard) {
            return SosiSeal.MakeNewIdenticalLevel4Card((UserIDCard)card);
        }
        throw new DGWSFault(Constants.MISSING_USER_IDCARD_IN_REQUEST_MSG,
                Constants.FAULTCODE_MISSING_USER_IDCARD_IN_REQUEST);
    }

    protected NodeList extractMaxAgeElements(Document headerDocument) throws DGWSFault {
        return headerDocument.getElementsByTagNameNS(Constants.HTTP_SOSI_DK_GW_2007_09_01, Constants.MAX_AGE);
    }

    protected boolean isMaxAgeHeaderExpired(NodeList maxAgeElements, IDCardCacheEntry cardCacheEntry) throws DGWSFault {
        if (maxAgeElements != null && maxAgeElements.getLength() > 0) {
            Date createdDate = cardCacheEntry.getIdCard().getCreatedDate();
            long ageInMinutes = (System.currentTimeMillis() - createdDate.getTime()) / (1000L * 60);
            for (int i = 0; i < maxAgeElements.getLength(); i++) {
                try {
                    String txt = maxAgeElements.item(i).getTextContent();
                    int maxAgeInMinutes = Integer.parseInt(txt);
                    if (ageInMinutes > maxAgeInMinutes) {
                        return true;
                    }
                } catch (NumberFormatException nfe) {
                    throw new DGWSFault("Unable to parse header " + Constants.MAX_AGE + " as integer",
                            Constants.FAULTCODE_SYNTAX_ERROR);
                }
            }
        }
        return false;
    }

    protected boolean isIDCardTooOld(IDCardCacheEntry cardCacheEntry) throws DGWSFault {
        long now = System.currentTimeMillis();
        if (cardCacheEntry.getIdCard().getExpiryDate().getTime() < now) {
            log.debug("idcard expired for " + cardCacheEntry.getCacheKey());
            return true;
        }
        return false;
    }

    protected boolean replaceIDCard(Document doc, IDCardCacheEntry signedIDCard) throws DGWSFault, IOException {
        if (signedIDCard != null && signedIDCard.getState() == IDCardState.SIGNED) {
            // Replace unsigned card with signed one from cache.
            Element unsignedIDCardElement = DOMUtil.getIDCardElement(doc);
            Node parentNode = unsignedIDCardElement.getParentNode();
            Element newElement = (Element)doc.importNode(signedIDCard.getDOMelement(), true);
            parentNode.replaceChild(newElement, unsignedIDCardElement);
            return true;
        }
        return false;
    }

    protected void setRequestHeaders(PostMethod postMethod, HttpServletRequest req, WSAHeaderInfo wsaHeaderInfo) {
        postMethod.setRequestHeader(new Header(Constants.SOAP_ACTION_HEADER, wsaHeaderInfo.action));
        if (req.getHeader("Content-Type") != null) {
            postMethod.setRequestHeader(new Header("Content-Type", req.getHeader("Content-Type")));
        } else {
            postMethod.setRequestHeader(new Header("Content-Type", Constants.HTTPHEADER_TEXT_XML_CHARSET_UTF_8));
        }
        if (req.getHeader("Accept-Encoding") != null) {
            postMethod.setRequestHeader(new Header("Accept-Encoding", req.getHeader("Accept-Encoding")));
        }
    }

    private void removeChildNodes(NodeList nodes) {
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            node.getParentNode().removeChild(node);
        }
    }

    /**
     * Build string version of soap-header, including the opening soap:envelope, stopping at the soap-header-end.
     *
     * @param soapEnvelopeNode
     * @return
     * @throws ServletException
     */
    protected static String serializeNewHeader(Node soapEnvelopeNode) throws ServletException {
        LSSerializer serializer = DOMUtil.getDomLSSerializer();
        StringBuffer sb = new StringBuffer(4000);
        sb.append(Constants.XML_VERSION_1_0_ENCODING_UTF_8);
        sb.append(serializer.writeToString(soapEnvelopeNode));
        return sb.substring(0, sb.lastIndexOf("</")); // remove ending envelope tag.
    }

    private RequestEntity makeRequestEntity(final HttpServletRequest req, final long origContentLengh, final String newHeaders,
            final long origHeaderSize, final InputStream remainderStream, final StringBuffer prolog) throws UnsupportedEncodingException {
        final byte[] newHeaderAsBytes = newHeaders.getBytes(Constants.UTF_8);
        final byte[] prologAsBytes = prolog.toString().getBytes(Constants.UTF_8);
        final long contentLength =
                origContentLengh < 0 ? -1 : origContentLengh + prologAsBytes.length + newHeaderAsBytes.length
                - origHeaderSize;

        return new RequestEntity() {

            private boolean repeatable = true;

            public long getContentLength() {
                return contentLength;
            }

            public String getContentType() {
                return req.getContentType();
            }

            public boolean isRepeatable() {
                return repeatable;
            }

            public void writeRequest(OutputStream out) throws IOException {
                if (prologAsBytes.length > 0) {
                    // MTOM requests have a prolog before the xml, if this is the case write it here.
                    out.write(prologAsBytes);
                }
                // send new headers.
                out.write(newHeaderAsBytes);

                // No longer able to repeat this request.
                repeatable = false;

                // stream-copy remainder
                Util.copyStream(remainderStream, out);
            }
        };
    }

    protected void AuthenticateByIP(HttpServletRequest req) throws DGWSFault {
        String remoteAddrString = req.getRemoteAddr();
        ClientAccessWhiteList(remoteAddrString, null, false);
    }

    protected void AuthenticateBySalt(Document headerDocument) throws DGWSFault {
        String remoteAddrString = req.getRemoteAddr();
        String salt = null; // This salt must fail.
        NodeList clientAccessElements = headerDocument.getElementsByTagNameNS(Constants.HTTP_SOSI_DK_GW_2007_09_01,
                Constants.CLIENT_ACCESS_KEY);
        if (clientAccessElements != null && clientAccessElements.getLength() > 0) {
            salt = clientAccessElements.item(0).getNodeValue();
        }
        ClientAccessWhiteList(remoteAddrString, salt, true);
    }

    protected void checkTargetAccess(WSAHeaderInfo wsaHeaderInfo) throws DGWSFault {
        if (wsaHeaderInfo.to != null && wsaHeaderInfo.to.equals(RuntimeConfig.getDecouplingUrl())) {
            return; // Decoupling url is always white listed
        } else if (!RuntimeConfig.checkServiceWhiteList(wsaHeaderInfo.to, wsaHeaderInfo.action)) {
            throw new DGWSFault("Access denied: Service not in whitelist", Constants.FAULTCODE_ACCESS_DENIED);
        }
    }

    protected static void ClientAccessWhiteList(String remoteAddrString, String saltInRequest, boolean requireSalt)
            throws DGWSFault {
        if (!RuntimeConfig.checkClientAccessWhiteList(remoteAddrString, saltInRequest, requireSalt)) {
            throw new DGWSFault("Access denied: Client " + remoteAddrString + " not in whitelist", Constants.FAULTCODE_ACCESS_DENIED);
        }
    }

    private void sendInternalErrorResponse(Document headerDocument, HttpServletResponse resp, Exception e) {
        log.error("Sending internal error: " + e.getMessage(), e);
        sendDGWSFault(headerDocument, resp, new DGWSFault(e.getMessage(), Constants.FAULTCODE_INTERNAL_ERROR, e));
    }

    private void sendDGWSFault(Document headerDocument, HttpServletResponse resp, DGWSFault e) {
        if (!resp.isCommitted()) {
            resp.setStatus(500);
            resp.setContentType("text/xml; charset=utf-8");
            String errorResponse = makeErrorResponse(headerDocument, e);
            try {
                resp.getWriter().write(errorResponse);
            } catch (IOException e1) {
                log.warn("Failed sending errorResponse: " + errorResponse, e1);
            } catch (IllegalStateException e2) {
                log.warn("Failed sending errorResponse: " + errorResponse, e2);
            }
        } else {
            // What can we do? too late to send error code...
            log.warn("Unable to send errormessage as response is committed.", e);
        }
    }

    private Element getChild(Element elm, String ns, String name) {
        NodeList list = elm.getElementsByTagNameNS(ns, name);
        if (list.getLength() != 1) {
            throw new IllegalArgumentException("Expected to find one child named \"" + name + "\" of element \""
                    + elm.getNodeName() + "\" but found " + list.getLength());
        }

        return (Element)list.item(0);
    }

    private String makeErrorResponse(Document headerDocument, DGWSFault e) {
        String flowID = "No flowID in request!";
        String messageIDFromRequest = "No messageID in request!";
        try {
            Element soapHeaderElm = (Element)headerDocument.getElementsByTagNameNS(Constants.URL_SOAP,
                    Constants.SOAP_HEADER).item(0);

            // Element securityElm = (Element)soapHeaderElm.getElementsByTagNameNS(Constants.URL_WSSE,
            // "Security").item(0);
            // Element timestampElm = (Element)securityElm.getElementsByTagNameNS(Constants.URL_WSU,
            // "Timestamp").item(0);
            // Element createdElm = (Element)timestampElm.getElementsByTagNameNS(Constants.URL_WSU, "Created").item(0);
            // String createdText = createdElm.getTextContent();

            Element medcomHeaderElm = getChild(soapHeaderElm, Constants.URL_MEDCOM, "Header");
            Element medcomLinkingElm = getChild(medcomHeaderElm, Constants.URL_MEDCOM, "Linking");

            try {
                Element medcomFlowIDElm = getChild(medcomLinkingElm, Constants.URL_MEDCOM, "FlowID");
                flowID = medcomFlowIDElm.getTextContent();

            } catch (Exception ex) {
                log.error("Unable to extract medcom flow ID from request", ex);
            }

            try {
                Element medcomMessageIDElm = getChild(medcomLinkingElm, Constants.URL_MEDCOM, "MessageID");
                messageIDFromRequest = medcomMessageIDElm.getTextContent();
            } catch (Exception ex) {
                log.error("Unable to extract medcom message ID from request", ex);
            }

        } catch (Exception ex) {
            log.error("Unable to extract medcom linking info from request", ex);
        }

        DOMImplementation impl = DOMUtil.getDOMImplementation();
        DocumentType documentType = null; // impl.createDocumentType("implicitLogin", "publicId", "systemId");
        Document document = impl.createDocument(Constants.URL_SOAP, "Envelope", documentType);
        Element root = document.getDocumentElement();
        root.setPrefix("soapEnv");

        Node soapHeader = document.createElementNS(Constants.URL_SOAP, "Header");
        soapHeader.setPrefix("soapEnv");
        root.appendChild(soapHeader);

        Node security = document.createElementNS(Constants.URL_WSSE, "Security");
        soapHeader.appendChild(security);
        Node timestamp = document.createElementNS(Constants.URL_WSU, "Timestamp");
        security.appendChild(timestamp);
        Node created = document.createElementNS(Constants.URL_WSU, "Created");
        created.setTextContent(XmlUtil.toXMLTimeStamp(new Date(), !StaticConfigImpl.getInstance().useLocalTimeInSaml()));
        timestamp.appendChild(created);

        Node medcomHeader = document.createElementNS(Constants.URL_MEDCOM, "Header");
        soapHeader.appendChild(medcomHeader);

        Node medcomLinking = document.createElementNS(Constants.URL_MEDCOM, "Linking");
        medcomHeader.appendChild(medcomLinking);

        Node medcomFlowID = document.createElementNS(Constants.URL_MEDCOM, "FlowID");
        medcomFlowID.setTextContent(flowID);
        medcomLinking.appendChild(medcomFlowID);

        Node medcomMessageID = document.createElementNS(Constants.URL_MEDCOM, "MessageID");
        medcomMessageID.setTextContent(createMessageID());
        medcomLinking.appendChild(medcomMessageID);

        Node medcomInResponseToMessageID = document.createElementNS(Constants.URL_MEDCOM, "InResponseToMessageID");
        medcomInResponseToMessageID.setTextContent(messageIDFromRequest);
        medcomLinking.appendChild(medcomInResponseToMessageID);

        Node medcomFlowStatus = document.createElementNS(Constants.URL_MEDCOM, "FlowStatus");
        medcomFlowStatus.setTextContent(FaultCodeValues.PROCESSING_PROBLEM);
        medcomHeader.appendChild(medcomFlowStatus);

        Node soapBody = document.createElementNS(Constants.URL_SOAP, "Body");
        soapBody.setPrefix("soapEnv");
        root.appendChild(soapBody);

        Node soapFault = document.createElementNS(Constants.URL_SOAP, "Fault");
        soapBody.appendChild(soapFault);

        Node soapFaultCode = document.createElementNS("", "faultcode");
        soapFaultCode.setTextContent(root.getPrefix() + ":" + "Server");
        soapFault.appendChild(soapFaultCode);

        Node soapDetail = document.createElementNS("", "detail");
        soapFault.appendChild(soapDetail);

        Node medcomFaultCode = document.createElementNS(Constants.URL_MEDCOM, "FaultCode");
        medcomFaultCode.setTextContent(e.getFaultInfo());
        soapDetail.appendChild(medcomFaultCode);

        Node soapFaultString = document.createElementNS("", "faultstring");
        soapFaultString.setTextContent(e.getMessage());
        soapFault.appendChild(soapFaultString);

        String responseAsString = DOMUtil.getDomLSSerializer().writeToString(document);
        return responseAsString;
    }

    static ThreadLocal<Integer> messageCounter = new ThreadLocal<Integer>() {
        public Integer initialValue() {
            return new Integer(0);
        }
    };

    private String createMessageID() {
        int count;
        messageCounter.set((count = messageCounter.get()) + 1);

        String id = "" + Thread.currentThread().getId() + "/" + count;

        return id;
    }

}
