/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.proxy;

import com.trifork.sosigw.api.Constants;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.util.Util;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.servlet.ServletException;

public class WsAddressingHandler {
    private boolean forceSSLRequests;

    public WsAddressingHandler(boolean forceSSLRequests) {
        this.forceSSLRequests = forceSSLRequests;
    }

    /**
     * Extract the target url and soap action from the WSAdressing part of the document.
     * If no target url could be found, then the Decoupling Component URL is used instead.
     * If no target soap action could be found, then the soap action from the HTTP header
     * is used instead.
     */
    protected WSAHeaderInfo handleWSAddressing(Document doc, String soapAction) throws ServletException {
        Node to = getWSANode(doc, "To");
        Node action = getWSANode(doc, "Action");

        WSAHeaderInfo headerInfo = handleWSAddressing(to != null ? to.getTextContent() : null, action != null ? action.getTextContent() : null, soapAction);

        removeWSANodes(doc, "ReplyTo");
        removeWSANodes(doc, "FaultTo");
        removeWSANodes(doc, "From");
        removeWSANodes(doc, "MessageID");
        removeWSANodes(doc, "RelatesTo");
        removeWSANodes(doc, "ReferenceParameters");

        if (RuntimeConfig.removeWSAInProxy(headerInfo.to, headerInfo.action)) {
            removeWSANodes(doc, "To");
            removeWSANodes(doc, "Action");
        }

        return headerInfo;
    }

    protected WSAHeaderInfo handleWSAddressing(String to, String action, String soapAction) throws ServletException {
        WSAHeaderInfo headerInfo = new WSAHeaderInfo();
        if (to != null) {
            headerInfo.to = to;

            if (forceSSLRequests) {
                if (to.toLowerCase().startsWith("http:")) {
                    headerInfo.to = "https:" + to.substring(5);
                } else if (!to.toLowerCase().contains("http")) {
                    headerInfo.to = "https://" + to;
                }
            }

            Util.getSLALogItem().addCallParameter("sosigw-wsa-decoupling", "false");
        } else {
            String decouplingUrl = RuntimeConfig.getDecouplingUrl();
            if (decouplingUrl == null || decouplingUrl.trim().isEmpty()) {
                throw new ServletException("No Decoupling Component URL configured");
            }
            headerInfo.to = decouplingUrl;
            Util.getSLALogItem().addCallParameter("sosigw-wsa-decoupling", "true");
        }

        if (action != null) {
            headerInfo.action = action;
        } else {
            if (soapAction == null || soapAction.trim().isEmpty()) {
                throw new ServletException("No SOAP Action was provided");
            }
            headerInfo.action = soapAction;
        }

        return headerInfo;
    }

    public Node getWSANode(Document doc, String localName) throws ServletException {
        NodeList nodes = doc.getElementsByTagNameNS(Constants.URN_WS_ADDRESSING_2004_08, localName);
        if (nodes.getLength() == 0) {
            nodes = doc.getElementsByTagNameNS(Constants.URN_WS_ADDRESSING_1_0, localName);
        }
        if (nodes.getLength() == 0) {
            return null;
        }
        if (nodes.getLength() == 1) {
            return nodes.item(0);
        }
        throw new ServletException("More than one WSA node of kind " + localName);
    }

    private void removeWSANodes(Document doc, String localName) {
        removeChildNodes(doc.getElementsByTagNameNS(Constants.URN_WS_ADDRESSING_2004_08, localName));
        removeChildNodes(doc.getElementsByTagNameNS(Constants.URN_WS_ADDRESSING_1_0, localName));
    }

    private void removeChildNodes(NodeList nodes) {
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            node.getParentNode().removeChild(node);
        }
    }
}
