/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.sts;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;

import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;

public class CircuitBreakerSTSProxy implements STSProxy {
    private static final Logger log = Logger.getLogger(CircuitBreakerSTSProxy.class);
    private final STSProxy sts;

    private AtomicReference<BreakerState> state;
    private STSProxyConfig config;

    private AtomicInteger attemptedCalls = new AtomicInteger(0);
    private AtomicInteger successfulCalls = new AtomicInteger(0);

    protected CircuitBreakerSTSProxy(STSProxy sts, STSProxyConfig config) {
        this.sts = sts;
        this.config = config;
        this.state = new AtomicReference<BreakerState>(new ClosedState());
        log.debug("Constructing circuit breaker for " + sts + ", numberOfFailuresBeforeOpening=" +
                getNumberOfFailuresBeforeOpening() +", millisecondsBeforeAttemptingClose=" + getMillisecondsBeforeAttemptingClose());
    }

    public CircuitBreakerSTSProxy(STSProxy sts) {
        this(sts, new STSProxyDynamicConfig());
    }

    /**
     * @return <code>true</code> is the breaker is closed, i.e. is letting calls through normally
     */
    public boolean isClosed() {
        return (state.get() instanceof ClosedState);
    }

    private BreakerState getState() {
        return state.get();
    }

    public int getNumberOfFailuresBeforeOpening() {
        return config.getNumberOfFailuresBeforeOpening();
    }

    public int getMillisecondsBeforeAttemptingClose() {
        return config.getMillisecondsBeforeAttemptingClose();
    }

    /* JMX Properties
         These are Called via Reflection from CircuitBreakerManagerMBean - do not delete
     */
    @SuppressWarnings({"UnusedDeclaration"})
    public String getStateDescription() {
        return getState().toString();
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public int getFailureCount() {
        if (isClosed()) {
            return ((ClosedState)getState()).failureCount.get();
        }
        return getNumberOfFailuresBeforeOpening();
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public int getAttemptedCalls() {
        return attemptedCalls.get();
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public int getSuccessfulCalls() {
        return successfulCalls.get();
    }

    /* END JMX Properties */


    public String callSTS(Document doc) throws STSUnreachableException {
        String response;

        try {
            getState().preInvoke(this);

            attemptedCalls.incrementAndGet();
            response = sts.callSTS(doc);
            successfulCalls.incrementAndGet();

            getState().postInvoke(this);
        } catch (STSUnreachableException e) {
            getState().onError(this);
            throw e;
        }

        return response;
    }


    public String getUrlString() {
        return sts.getUrlString();
    }

    private void tripBreaker(BreakerState currentState) {
        log.error(sts + " seems to be down, opening circuit breaker");
        changeState(currentState, new OpenState());
    }

    private void attemptReset(BreakerState currentState) {
        log.info("Circuit breaker for " + sts + " will retry once");
        changeState(currentState, new HalfOpenState());
    }

    private void reset(BreakerState currentState) {
        log.info(sts + " seems to be up again, closing circuit breaker");
        changeState(currentState, new ClosedState());
    }

    /**
     * @param oldState The expected current state. Because of threading issues, state may already have been updated by another thread; in this case, this method does nothing
     * @param newState state to change to
     */
    private void changeState(BreakerState oldState, BreakerState newState) {
        final boolean updateHappened = state.compareAndSet(oldState, newState);
        if (updateHappened) {
            log.info("State change " + oldState + " -> " + newState + " for circuit breaker for " + sts);
        }
    }


    public String toString() {
        return sts.toString() + " in CircuitBreakerState " + getState().toString();
    }


    public abstract class BreakerState {
        abstract void preInvoke(CircuitBreakerSTSProxy cb);
        abstract void postInvoke(CircuitBreakerSTSProxy cb);
        abstract void onError(CircuitBreakerSTSProxy cb);
    }

    public class HalfOpenState extends BreakerState {

        void preInvoke(CircuitBreakerSTSProxy cb) {
            // do nothing, we're just waiting to see if the coming call succeeds
        }


        void postInvoke(CircuitBreakerSTSProxy cb) {
            // call suceeded, back to closed
            cb.reset(this);
        }


        void onError(CircuitBreakerSTSProxy cb) {
            // call failed, back to open
            cb.tripBreaker(this);
        }


        public String toString() {
            return "HalfOpen";
        }
    }

    public class ClosedState extends BreakerState {
        private AtomicInteger failureCount = new AtomicInteger(0);


        void preInvoke(CircuitBreakerSTSProxy cb) {
            // do nothing, we're closed and hoping it will last
        }


        void postInvoke(CircuitBreakerSTSProxy cb) {
            resetFailureCounter();
        }


        void onError(CircuitBreakerSTSProxy cb) {
            failureCount.incrementAndGet();
            log.debug("Failure count increased to " + failureCount.get() + " for circuit breaker for " + cb.sts);
            if (failureCount.get() >= cb.getNumberOfFailuresBeforeOpening()) {
                cb.tripBreaker(this);
            }
        }

        private void resetFailureCounter() {
            failureCount.set(0);
        }


        public String toString() {
            String s = "Closed";
            if (failureCount.get() > 0) {
                s += "(failureCount=" + failureCount.get() + ")";
            }
            return s;
        }
    }

    public class OpenState extends BreakerState {
        /** the currentTimeMillis when the breaker last opened **/
        private long tripTime = System.currentTimeMillis();


        void preInvoke(CircuitBreakerSTSProxy cb) {
            long elapsed = getMillisSinceTrip();
            if(elapsed >= cb.getMillisecondsBeforeAttemptingClose()) {
                log.debug(elapsed + " ms elapsed since opening circuit breaker for " + cb.sts + ", will attempt retry");
                cb.attemptReset(this);
            } else {
                log.debug(elapsed + " ms elapsed since opening circuit breaker for " + cb.sts + ", will not attempt retry, as we're below threshold " + cb.getMillisecondsBeforeAttemptingClose());
                throw new STSUnreachableException("Circuit Breaker for " + cb.sts + " is open, did not attempt to call");
            }
        }

        private long getMillisSinceTrip() {
            return System.currentTimeMillis() - tripTime;
        }


        void postInvoke(CircuitBreakerSTSProxy cb) {
            // do nothing, we're just waiting for enough time to elapse
        }


        void onError(CircuitBreakerSTSProxy cb) {
            // do nothing, we're already open
        }


        public String toString() {
            return "Open(" + getMillisSinceTrip() + " ms elapsed)";
        }

    }
}


