/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.sts;

import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.util.HeaderReadingInputStream;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

public class STSClusterProxy implements STSProxy {
	private final static Logger log = Logger.getLogger(HeaderReadingInputStream.class);
    private final static boolean registerMBean = StaticConfigImpl.getInstance().isSTSMBeansEnabled();
    private String stsUrls;
    private List<STSProxy> proxies;

    /**
     * @param stsUrls Space-separated STS urls
     */
    public static STSClusterProxy create(String stsUrls) {
        STSClusterProxy cluster = new STSClusterProxy(createProxiesFromUrlString(stsUrls));
        cluster.stsUrls = stsUrls;
        return cluster;
    }

    protected STSClusterProxy(List<STSProxy> proxies) {
        this.proxies = proxies;
    }

    public List<STSProxy> getProxies() {
        return proxies;
    }


    private static List<STSProxy> createProxiesFromUrlString(String stsUrls) {
        List<STSProxy> createdProxies = new ArrayList<STSProxy>();

        for (String stsUrl : stsUrls.split(" ")) { // handles things like double spaces between URL's
            if (!"".equals(stsUrl.trim())) {
                createdProxies.add(new CircuitBreakerSTSProxy(new SingleSTSProxy(stsUrl)));
            }
        }
        registerCircuitBreakerProxiesAsMBeans(createdProxies);
        return createdProxies;

    }

    private static void registerCircuitBreakerProxiesAsMBeans(List<STSProxy> proxies) {
        List<CircuitBreakerSTSProxy> circuitBreakerProxies = new ArrayList<CircuitBreakerSTSProxy>(proxies.size());
        for (STSProxy proxy : proxies) {
            circuitBreakerProxies.add((CircuitBreakerSTSProxy) proxy);
        }
        // Only register mbeans if they are enabled in config
        if (registerMBean) {
            CircuitBreakerManagerMBean.registerMBeans(circuitBreakerProxies);
        }
    }


    public String callSTS(Document doc) {
        verifyStsUrlsAgainstRuntimeConfig();
        for (STSProxy proxy : proxies) {
            try {
                final String result = proxy.callSTS(doc);
                log.debug("Got response from " + proxy);
                return result;
            } catch (STSUnreachableException ignore) {
                // be it as it may, we will try the next proxy in our list
                log.debug("Unable to reach " + proxy + ": " + ignore.getMessage()!=null?ignore.getMessage():"" + ", trying next in list");
            }
        }

        // if we end up here, all our proxies failed:
        throw new STSUnreachableException("Could not contact any of the STS'es " + proxies);
    }

    /**
     * Allows changes in the RuntimeConfig to take effect without a restart of the SOSI-GW instance.
     * If the configured STS URL's have changed since this class was instantiated, we create new proxies
     */
    private void verifyStsUrlsAgainstRuntimeConfig() {
        if (stsUrls != null) { // may be null if the protected List<STSProxy> constructor was used; in that case, do nothing
            if (!stsUrls.equals(RuntimeConfig.getSecurityTokenServiceUrlList())) {
                stsUrls = RuntimeConfig.getSecurityTokenServiceUrlList();
                proxies = createProxiesFromUrlString(stsUrls);
            }
        }
    }


    public String getUrlString() {
        StringBuffer urlString = new StringBuffer();

        for (STSProxy proxy : proxies) {
            urlString.append(proxy.getUrlString());
            urlString.append(" ");
        }

        urlString.deleteCharAt(urlString.length()-1);

        return urlString.toString();
    }
}
