/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.sts;

import com.trifork.sosigw.api.Constants;
import com.trifork.sosigw.util.Util;
import com.trifork.sosigw.util.timing.Timer;
import com.trifork.sosigw.ws.IdCardSignerImpl;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HostConfiguration;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.MultiThreadedHttpConnectionManager;
import org.apache.commons.httpclient.methods.ByteArrayRequestEntity;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.params.HttpClientParams;
import org.apache.commons.httpclient.params.HttpConnectionManagerParams;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.Properties;

/**
 * STS Proxy which is able to call one STS instance
 */
public class SingleSTSProxy implements STSProxy {
    private Logger log = Logger.getLogger(SingleSTSProxy.class);
    private MultiThreadedHttpConnectionManager connectionManager;
    private final String serverUrl;
    private final Properties properties;
    private static final int HTTP_STATUS_OK = 200;
    protected STSProxyConfig config;
    /**
     * The connection timeout, we're currently using for the connection manager. This might be different from the timeout indicated by the config
     */
    private int activeConnectionTimeout;

    /**
     * @param serverUrl the address of the STS service to call
     */
    protected SingleSTSProxy(String serverUrl) {
        this.serverUrl = serverUrl;
        this.properties = SignatureUtil.setupCryptoProviderForJVM();
        this.config = new STSProxyDynamicConfig();
    }

    
    public String callSTS(Document doc) {
        // Setup WebService

        String responseFromSTS;
        try {
            String xml = getValidatedXmlString(doc);
            final byte[] xmlAsBytes = asUtf8bytes(xml);

            HttpClient httpClient = makeHttpClient();
            PostMethod postMethod = makePostMethod();
            String operationName = "invokeRemoteService";

            try {
                Timer.start(IdCardSignerImpl.class, operationName);
                postMethod.setRequestEntity(new ByteArrayRequestEntity(xmlAsBytes, Constants.HTTPHEADER_TEXT_XML_CHARSET_UTF_8));
                final String SOAP_ACTION = "";
                postMethod.setRequestHeader(new Header(Constants.SOAP_ACTION_HEADER, SOAP_ACTION));
	            postMethod.setRequestHeader(new Header(Constants.HTTPHEADER_NAME_NSI_ORIGINAL_CLIENT_IP, Util.getHttpServletRequest().getRemoteAddr()));
	            httpClient.executeMethod(postMethod);
                if (postMethod.getStatusCode() != HTTP_STATUS_OK) {
                    //noinspection ThrowCaughtLocally
                    throw new STSUnreachableException("STS Server " + serverUrl + " returned status code " + postMethod.getStatusCode());
                }

                responseFromSTS = readResponse(postMethod);
            } finally {
                postMethod.releaseConnection();
                Timer.stop(IdCardSignerImpl.class, operationName);
            }
        } catch (IOException e) {
            log.warn(this.toString() + " rethrowing exception wrapped in STSUnreachableException", e);
            throw new STSUnreachableException(e.getClass().getSimpleName() + ": " + e.getMessage());
        } catch (RuntimeException e) {
            if (e instanceof STSUnreachableException) {
                throw e; // don't wrap it once again
            } else {
                log.warn(this.toString() + " rethrowing exception wrapped in STSUnreachableException", e);
                throw new STSUnreachableException(e.getClass().getSimpleName() + ": " + e.getMessage());
            }
        }

        return responseFromSTS;
    }


    public String getUrlString() {
        return serverUrl;
    }

    private String readResponse(PostMethod postMethod) {
        final InputStream responseStream;
        try {
            responseStream = postMethod.getResponseBodyAsStream();
        } catch (IOException e) {
            log.warn(this.toString() + " could not read response body. Rethrowing exception wrapped in STSUnreachableException", e);
            throw new STSUnreachableException("Could not read response body: " + e.getMessage());
        }

        String responseFromSTS;
        try {
            responseFromSTS = Util.readAllFromStreamAsString(responseStream);
        } catch (IOException e) {
            log.warn(this.toString() + " could not read response body. Rethrowing exception wrapped in STSUnreachableException", e);
            throw new STSUnreachableException("Could not read response body: " + e.getMessage());
        }
        return responseFromSTS;
    }

    private String getValidatedXmlString(Node xmlNode) {
        return XmlUtil.node2String(XmlUtil.readXml(properties, XmlUtil.node2String(xmlNode, false, true), false), false, true);
    }

    private byte[] asUtf8bytes(String xml) {
        try {
            return xml.getBytes(Constants.UTF_8);
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException(e);
        }
    }

    private MultiThreadedHttpConnectionManager getConnectionManager() {
        if (connectionManager == null || activeConnectionTimeout != config.getHttpTimeoutMillis()) {
            // we need to recreate the connection manager if the timeout value changes (as the timeout is baked into the connections it already has pooled)
            connectionManager = new MultiThreadedHttpConnectionManager();

            HttpConnectionManagerParams params = new HttpConnectionManagerParams();
            params.setMaxConnectionsPerHost(HostConfiguration.ANY_HOST_CONFIGURATION, 10);
            params.setMaxTotalConnections(10);
            params.setTcpNoDelay(true);
            params.setConnectionTimeout(config.getHttpTimeoutMillis());
            connectionManager.setParams(params);

            activeConnectionTimeout = config.getHttpTimeoutMillis();
        }

        return connectionManager;
    }

    private HttpClient makeHttpClient() {
        final HttpClient httpClient = new HttpClient(getConnectionManager());
        HttpClientParams params = new HttpClientParams();
        params.setSoTimeout(config.getHttpTimeoutMillis());
        params.setConnectionManagerTimeout(config.getHttpTimeoutMillis());
        httpClient.setParams(params);
        return httpClient;
    }

    private PostMethod makePostMethod() {
        PostMethod postMethod = new PostMethod(serverUrl);
        postMethod.setFollowRedirects(false); // we don't want extra latency added to every call to the STS
        return postMethod;
    }


    public String toString() {
        return "STS " + serverUrl;
    }
}
