/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.util;

import org.apache.log4j.Logger;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.bootstrap.DOMImplementationRegistry;
import org.w3c.dom.ls.DOMImplementationLS;
import org.w3c.dom.ls.LSParser;
import org.w3c.dom.ls.LSSerializer;

import com.sun.org.apache.xerces.internal.impl.Constants;

import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.SAMLTags;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class DOMUtil {
    public static final Logger log = Logger.getLogger(DOMUtil.class);
    private static final boolean CACHE_DOM_HANDLERS = true;

    static { // Set default implelentation from JDK if unset. 
        if (System.getProperty(DOMImplementationRegistry.PROPERTY) == null) {
            System.setProperty(DOMImplementationRegistry.PROPERTY, "com.sun.org.apache.xerces.internal.dom.DOMImplementationSourceImpl");
        }
    }
    
    public static String getLoggableString(Node domNode) {
        try {
            return getDomLSSerializer().writeToString(domNode);
        } catch (Exception e) {
            log.error(e.getMessage(), e);
        }
        return "<Failed/>";
    }
    
    private static final ThreadLocal<LSSerializer> lsSerializer = new ThreadLocal<LSSerializer>() {
        protected LSSerializer initialValue() {
            return createLSSerializer();
        }
    };

    public static LSSerializer getDomLSSerializer() {
        if (CACHE_DOM_HANDLERS) {
            return lsSerializer.get(); 
        }
        return createLSSerializer();
    }

    private static LSSerializer createLSSerializer()  {
        LSSerializer serializer;
        DOMImplementationRegistry registry;
        try {
            registry = DOMImplementationRegistry.newInstance();
        } catch (Exception e) {
            String msg = "Unable to load dom parser";
            log.error(msg, e);
            throw new RuntimeException(msg, e);
        }
        
        DOMImplementationLS impl = (DOMImplementationLS)registry.getDOMImplementation("LS");
        serializer = impl.createLSSerializer();
        serializer.getDomConfig().setParameter(Constants.DOM_XMLDECL, Boolean.FALSE);
        serializer.getDomConfig().setParameter(Constants.DOM_NAMESPACES, Boolean.TRUE);
        return serializer;
    }

    private static final ThreadLocal<LSParser> lsParser = new ThreadLocal<LSParser>() {
        protected LSParser initialValue() {
            return createLSParser();
        }
    };

    public static LSParser getDomLSParser() {
        if (CACHE_DOM_HANDLERS) {
            return lsParser.get();
        }
        return createLSParser();
    }

    private static LSParser createLSParser() {
        DOMImplementationRegistry registry;
        try {
            registry = DOMImplementationRegistry.newInstance();
        } catch (Exception e) {
            String msg = "Unable to load dom parser";
            log.error(msg, e);
            throw new RuntimeException(msg, e);
        }
        
        DOMImplementationLS impl = (DOMImplementationLS)registry.getDOMImplementation("LS");
        LSParser builder = impl.createLSParser(DOMImplementationLS.MODE_SYNCHRONOUS, null);
        return builder;
    }

    public static DOMImplementation getDOMImplementation()  {
        DOMImplementationRegistry registry;
        try {
            registry = DOMImplementationRegistry.newInstance();
        } catch (Exception e) {
            String msg = "Unable to load dom parser";
            log.error(msg, e);
            throw new RuntimeException(msg, e);
        }
        
        DOMImplementation impl = (DOMImplementation)registry.getDOMImplementation("XML 2.0");
        return impl;
    }

    private static Element findFirstElmByNameAndId(Document doc, String namespace, String localname, String id) {
        NodeList assertions = doc.getElementsByTagNameNS(namespace, localname);
        for (int i = 0; i < assertions.getLength(); i++) {
            Element element = (Element) assertions.item(i);
            NamedNodeMap namedNodeMap = element.getAttributes();
            if (namedNodeMap != null) {
                for (int j = 0; j < namedNodeMap.getLength(); j++) {
                    Node node = namedNodeMap.item(j);
                    String name = node.getNodeName().toLowerCase();
                    if (name.endsWith("id")) {
                        String value = node.getNodeValue();
                        if (id.equals(value)) {
                            return element;
                        }
                    }
                }
            }
        }
        return null;
    }

    public static Element getIDCardElement(Document doc) {
        return findFirstElmByNameAndId(doc, NameSpaces.SAML2ASSERTION_SCHEMA, SAMLTags.ASSERTION, IDValues.IDCARD);
    }

}
