/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.util;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UTFDataFormatException;

/**
 * Read from a input stream, buffering it and making the SOAP-header available as a string, while make the body available as a stream.
 * The input stream must be an UTF-8 encoded SOAP request, as required by "medcom: Den Gode Webservice".
 * 
 * The header string includes everything in the SOAP-envelope up to the </soap-ns:Header> tag, including the start tag of the SOAP-envelope.
 * The remainder stream includes everything after the header, including the end tag of the SOAP-envelope. 
 * The two parts are thus not valid XML in them selves.
 * 
 * @author Morten Grouleff <mgr@trifork.com>
 */
public class HeaderReadingInputStream  {
//	private final static Logger log = Logger.getLogger(HeaderReadingInputStream.class); 

	private static final String HEADER = "Header";
	private static final String ENVELOPE = "Envelope";
	private static final String XMLTAG = "?xml";

	private static final long MAX_HEADER_SIZE = 128 * 1024; // Max size of header. If bigger, throws and aborts.

    private static final int BUFFER_SIZE = 8000; // Base size of stringbuffer holding header.

	enum State {
		 WAIT_FOR_XML,
		 READING_XML,
		 WAITING_FOR_ENVELOPE,
		 READING_NAMESPACE,
		 WAITING_FOR_SOAP_HEADER_END,
		 INSIDE_CDATA,
		 DONE, 
	};

	private State state = State.WAIT_FOR_XML;
	private final InputStream bufferedInputStream;
	private final LimitingInputStream limitedInputStream;
	private final StringBuffer prolog = new StringBuffer(BUFFER_SIZE);
	private final StringBuffer header = new StringBuffer(BUFFER_SIZE);
    private final int headerOrigCharLength;
    private String soapNamespace = "";

	
	public HeaderReadingInputStream(InputStream is) throws IOException {
		this.bufferedInputStream = new BufferedInputStream(is);
		this.limitedInputStream = new LimitingInputStream(is, MAX_HEADER_SIZE);
		extractSoapHeader();
		this.headerOrigCharLength = header.length();
	}
	
	public InputStream getRemainderStream() {
		return bufferedInputStream;
	}

	public StringBuffer getProlog() {
		return prolog;
	}

	public StringBuffer getSoapHeader() throws IOException {
			return header;
	}

	public int getHeaderOrigCharLength() {
        return headerOrigCharLength;
    }

	public void extractSoapHeader() throws IOException {
		int xmlIdx = 0;
		int namespaceIdx = 0;
		int endHeaderIsAfter = 0;

		while (true) {
			while (state == State.WAIT_FOR_XML) {
				char c = nextChar();
				if (c == '<') {
					state = State.READING_XML;
					xmlIdx = header.length();
				}
			}
			while (state == State.READING_XML) {
				char c = nextChar();
				if (c == '<' || c == '>') {
					int startOfNamespace = header.indexOf(ENVELOPE, xmlIdx);
					int startOfXml = header.indexOf(XMLTAG, xmlIdx);
					if ((startOfNamespace >= 0) || (startOfXml >= 0)) {
						// Someone forgot ?xml tag, so lets we go to READING_NAMESPACE
						if (xmlIdx > 0) {
							prolog.append(header.substring(0, xmlIdx-1));
							header.delete(0, xmlIdx-1);
						}
						if (startOfNamespace >= 0) {
							namespaceIdx = 0;
							soapNamespace = header.substring(1, startOfNamespace - (xmlIdx-1));
							state = State.WAITING_FOR_SOAP_HEADER_END;
						} else {
							state = State.WAITING_FOR_ENVELOPE;
						}
					} else {
						state = State.WAIT_FOR_XML;
					}
				}
			}
			while (state == State.WAITING_FOR_ENVELOPE) {
				char c = nextChar();
				if (c == '<') {
					state = State.READING_NAMESPACE;
					namespaceIdx = header.length();
				}
			}
			while (state == State.READING_NAMESPACE) {
				char c = nextChar();
				if (c == '<' || c == '>') {
					int startOfNamespace = header.indexOf(ENVELOPE, namespaceIdx);
					if (startOfNamespace >= 0) {
						soapNamespace = header.substring(namespaceIdx, startOfNamespace);
						state = State.WAITING_FOR_SOAP_HEADER_END;
//						if (log.isDebugEnabled()) {
//							log.debug("Found namespace='" + soapNamespace + "'");
//						}
					} else {
						state = State.WAITING_FOR_ENVELOPE;
					}
				}
			}
			while (state == State.WAITING_FOR_SOAP_HEADER_END) {
				char c = nextChar();
				if (c == '[') {
					// Check for <![CDATA[
					String cdata = "<![CDATA[";
					if (header.substring(header.length() - cdata.length()).equals(cdata)) {
						state = State.INSIDE_CDATA;
					}
				} else if (c == '>') {
					// search backwards for for the tag.
					String remains = header.substring(endHeaderIsAfter);
					int i = remains.lastIndexOf(soapNamespace + HEADER);
					while (--i >= 0) {
						// Search backward, looking for '< and '/' to see if this is start or end
						char c2 = remains.charAt(i);
						if (c2 == '/') {
							state = State.DONE;
							return;
						}
						if (c2 == '<') {
							endHeaderIsAfter = header.length();
							break; // start-tag. Search for next occurrence of header.
						}
					}
				}
			}
			while (state == State.INSIDE_CDATA) {
				char c = nextChar();
				if (c == '>') {
					if (header.charAt(header.length()- 2) == ']' && header.charAt(header.length()- 3) == ']') {
						endHeaderIsAfter = header.length();
						state = State.WAITING_FOR_SOAP_HEADER_END;
					}
				}
			}
		}
	}

	private char nextChar() throws IOException {
		char c = readUTF8Char();
		header.append(c);
		return c;
	}

	/**
	 * Read a UTF-8 encoded char at a time from the input stream. 
	 * Unlike the JDK, this class does not read ahead in the stream...
	 * 
	 * @param is
	 * @return a single char from the stream, 1 to 3 bytes long. 
	 * @throws IOException if stream does on reading, or if the input is not valid UTF-8
	 */
	private char readUTF8Char() throws IOException {
		InputStream is = limitedInputStream;
		int c = is.read();
		if ((c & 0x7f) == c) {
			return (char)c;	
		} 

		if (c == -1) {
			throw new IOException("End of file");
		}
		int char2, char3;
		switch (c >> 4) {
		case 12: case 13:
			/* 110X XXXX   10XX XXXX */
			char2 = is.read();
			if ((char2 & 0xC0) != 0x80)
				throw new UTFDataFormatException("malformed input"); 
			return (char)(((c & 0x1F) << 6) | (char2 & 0x3F));  
		case 14:
			/* 1110 XXXX  10XX XXXX  10XX XXXX */
			char2 = is.read();
			char3 = is.read();
			if (((char2 & 0xC0) != 0x80) || ((char3 & 0xC0) != 0x80))
				throw new UTFDataFormatException("malformed input");
			return (char)(((c & 0x0F) << 12) | ((char2 & 0x3F) << 6) | ((char3 & 0x3F) << 0));
		default:
			/* 10XX XXXX,  1111 XXXX */
			throw new UTFDataFormatException("malformed input");
		}
	}

    public long getBytesRead() {
        return limitedInputStream.size();
    }
    
    public String getSoapNamespace() {
        return soapNamespace.substring(0, soapNamespace.length() - 1);
    }
}
