/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;

/**
 * Delegate all input to the nested stream, counting the size of the response, throwing an IOException when the limit is reached.
 * 
 * @author Morten Grouleff <mgr@trifork.com>
 *
 */
public class LimitingInputStream extends InputStream {

    private long count;

    private final InputStream _in;

    private final long limit;

	private long mark;

    /**
     * @param in
     */
    public LimitingInputStream(InputStream in, long limit) {
        _in = in;
		this.limit = limit;
    }
    
    public long size() {
        return count;
    }


    public int read() throws IOException {
        int result = _in.read();
        if (result != -1) {
            count += 1;
            checkLimit();
        }
        return result;
    }


    public boolean markSupported() {
        return _in.markSupported();
    }
    
    
    public synchronized void mark(int readlimit) {
        _in.mark(readlimit);
        mark = count;
    }
    

    public synchronized void reset() throws IOException {
        _in.reset();
        count = mark;
        mark = 0;
    }

    public int read(byte[] b, int off, int len) throws IOException {
        try {
            int result = _in.read(b, off, len);
            if (result != -1) {
                count += result;
            }
            checkLimit();
            return result;
        } catch (InterruptedIOException ex) {
            count += ex.bytesTransferred;
            checkLimit();
            throw ex;
        }
    }

    protected void checkLimit() throws IOException {
    	if (count > limit) {
    		throw new IOException("Limit exceeded");
    	}
	}

	public int read(byte[] b) throws IOException {
        try {
            int result = _in.read(b);
            if (result != -1) {
                count += result;
            }
            checkLimit();
            return result;
        } catch (InterruptedIOException ex) {
            count += ex.bytesTransferred;
            checkLimit();
            throw ex;
        }
    }
    
    public int available() throws IOException {
        return _in.available();
    }

    public void close() throws IOException {
        _in.close();
    }
    
    public long skip(long n) throws IOException {
        long howmany = _in.skip(n);
        count += howmany;
        checkLimit();
        return howmany;
    }
}
