/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.util;

import com.trifork.sosigw.api.Constants;
import dk.sosi.gw._2007_09.DGWSFault;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Set;

/**
 * Class that handles filtering of MTOM attachments
 */
public class MultipartAttachmentFilter {

    private static final byte[] CONTENT_ID_BYTES ="Content-Id:".getBytes();

    public static final Logger log = Logger.getLogger(MultipartAttachmentFilter.class);

    /**
     * Extract the boundary delimiter from content-type string
     * @param contentType content-type
     * @return boundary delimiter or null if not found.
     */
    public String findBoundaryInContentType(String contentType) {
        String[] tokens = StringUtils.split(contentType, ";");
        for (String token : tokens) {
            token = token.trim();
            if (token.contains("=")) {
                int equalOffset = token.indexOf("=");
                if (equalOffset+1 < token.length()) {
                    String name = token.substring(0, equalOffset);
                    String value = token.substring(equalOffset+1);
                    if ("boundary".equalsIgnoreCase(name) && value.length() >= 2) {
                        return value.substring(1, value.length()-1); // Remove quotes
                    }
                }
            }
        }
        return null;
    }

    /**
     * Removes MTOM attachement from raw request
     * @param requestContent raw request
     * @param contentType content-type, used for reading boundary delimiter
     * @param contentIdsToRemove id's to remove from stream
     * @return content without attachements
     * @throws IOException
     */
    public byte[] filter(byte[] requestContent, String contentType,
                              Set<String> contentIdsToRemove) throws IOException, DGWSFault {
        String boundary = findBoundaryInContentType(contentType);
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        int currentBoundaryOffset, lastBoundaryOffset = 0, lastBoundaryEndOffset = 0;
        if (boundary != null) {
            boolean excludeSection = false;
            while ((currentBoundaryOffset=indexOfBoundary(requestContent, lastBoundaryEndOffset, boundary)) > -1) {
                // Se if we hit a section that should be removed
                int currentBoundaryEOL = searchBackwardUntilEOL(requestContent, currentBoundaryOffset);
                if (currentBoundaryEOL > -1 && (currentBoundaryEOL - currentBoundaryOffset > 10)) { // sanity check
                    currentBoundaryOffset = currentBoundaryEOL + 1; // We don't want the actual eol so add one
                } else {
                    // We did not find EOL - standard says 2 characters so use that
                    currentBoundaryOffset = (currentBoundaryOffset - 2 > -1) ? currentBoundaryOffset - 2 : 0;
                }
                String contentId = findContentId(requestContent, currentBoundaryOffset + boundary.length());
                if (!excludeSection) {
                    outputStream.write(requestContent, lastBoundaryOffset, currentBoundaryOffset - lastBoundaryOffset);
                } else {
                    excludeSection = false;
                }
                if (contentId != null) {
                    if (contentIdsToRemove.contains(contentId)) {
                        // Exclude until next boundary or EOF
                        excludeSection = true;
                    }
                }
                lastBoundaryOffset = currentBoundaryOffset;
                lastBoundaryEndOffset = currentBoundaryOffset + boundary.length();
            }
            if (lastBoundaryOffset < requestContent.length) {
                outputStream.write(requestContent, lastBoundaryOffset, requestContent.length-lastBoundaryOffset);
            }
        } else {
            throw new DGWSFault("Could not locate boundary information in content-type header", Constants.FAULTCODE_SYNTAX_ERROR);
        }
        return outputStream.toByteArray();
    }

    /**
     * Search backward for end of line
     * @param content to search
     * @param offset to start at
     * @return offset to end-of-line
     */
    private int searchBackwardUntilEOL(byte[] content, int offset) {
        int pos = offset;
        while (pos >= 0) {
            if (content[pos] == '\r' || content[pos] == '\n') {
                return pos;
            }
            --pos;
        }
        return -1;
    }

    /**
     * Finds content-id in raw-request
     * @param requestContent raw request
     * @param offset start at this offset
     * @return content-id or null if not found
     * @throws IOException
     */
    private String findContentId(byte[] requestContent, int offset) throws IOException {
        int contentIdOffset = indexOfContentId(requestContent, offset);
        if (contentIdOffset > -1) {
            String contentIdLine = readUntilEOL(requestContent, contentIdOffset);
            return extractContentIdFromContentIdLine(contentIdLine);
        }
        return null;
    }

    /**
     * Read a string from buffer until a newline is hit
     * @param content content to search
     * @param offset start offset
     * @return string
     * @throws IOException
     */
    private String readUntilEOL(byte[] content, int offset) throws IOException {
        int pos = offset;
        while (pos < content.length && content[pos] != '\r' && content[pos] != '\n') {
            pos++;
        }
        return new String(content, offset, (pos-offset));
    }

    /**
     * Extract content-id from the raw content-id line
     * @param contentIdLine content-id line
     * @return content-id
     */
    private String extractContentIdFromContentIdLine(String contentIdLine) {
        // example: Content-Id: <rootpart*a0fe885 f-6a70-443d-beb4-a57c2c744075@example.jaxws.sun.com>
        return StringUtils.substringBetween(contentIdLine, "<", ">");
    }

    /**
     * Find content-id line
     * @param content content to search
     * @param byteOffset offset to start at
     * @return index of content-id string or -1 if not found
     */
    private int indexOfContentId(byte[] content, int byteOffset) {
        return indexOf(content, CONTENT_ID_BYTES, byteOffset);
    }

    /**
     * Find offset to boundary
     * @param content content to search
     * @param byteOffset offset to start at
     * @param boundary delimiter to find
     * @return index of boundary string or -1 if not found
     */
    private int indexOfBoundary(byte[] content, int byteOffset, String boundary) {
        byte[] boundaryBytes = boundary.getBytes();
        return indexOf(content, boundaryBytes, byteOffset);
    }

    /**
     * Search array for another array
     * @param haystack array to search
     * @param needle array to find
     * @param offset offset to start at
     * @return offset to found array or -1 if not found
     */
    private int indexOf(byte[] haystack, byte[] needle, int offset) {
        for (int i=offset; i<haystack.length - needle.length+1; ++i) {
            boolean found = true;
            for(int j = 0; j < needle.length; ++j) {
                if (haystack[i+j] != needle[j]) {
                    found = false;
                    break;
                }
            }
            if (found) {
                return i;
            }
        }
        return -1;
    }

}
