/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.util;

import com.trifork.sosigw.api.Constants;
import dk.sosi.gw._2007_09.DGWSFault;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.io.IOUtils;
import org.apache.james.mime4j.MimeException;
import org.apache.james.mime4j.stream.BodyDescriptor;
import org.apache.james.mime4j.stream.EntityState;
import org.apache.james.mime4j.stream.Field;
import org.apache.james.mime4j.stream.MimeTokenStream;
import org.apache.log4j.Logger;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import java.io.*;
import java.util.*;

public class MultipartHeaderFix {

    public static final Logger log = Logger.getLogger(MultipartHeaderFix.class);

    private static final XMLInputFactory XML_INPUT_FACTORY = XMLInputFactory.newInstance();

    private int bytesRemoved = 0;

    // Content-Id -> base64 encoded data
    private Map<String, String> headerAttachmentMap = new HashMap<String, String>();

    public Map<String, String> getHeaderAttachmentMap() {
        return headerAttachmentMap;
    }

    public int getBytesRemoved() {
        return bytesRemoved;
    }

    /**
     * Extract header attachements and remove them from stream
     * @param inputStream content stream
     * @param contentType content-type from request
     * @return new stream without attachements
     * @throws IOException
     */
    public InputStream extractAndFilterMtomHeaderAttachents(InputStream inputStream, String contentType) throws IOException, DGWSFault {
        byte[] data = IOUtils.toByteArray(inputStream);
        try {
            // First run - extract all attachements and save them to our map.
            parseAndSaveAttachments(new ByteArrayInputStream(data), contentType);
            // Second run - inline the saved attachment and remove them from stream
            byte[] filteredData = filterAttachements(data, contentType);
            bytesRemoved = data.length - filteredData.length;
            return new ByteArrayInputStream(filteredData);
        } catch (MimeException e) {
            log.warn("Failed to extract MTOM attachements", e);
            throw new DGWSFault(e.getMessage(), Constants.FAULTCODE_SYNTAX_ERROR);
        } catch (XMLStreamException e) {
            log.warn("Failed to extract MTOM attachements", e);
            throw new DGWSFault(e.getMessage(), Constants.FAULTCODE_SYNTAX_ERROR);
        }
    }

    private byte[] filterAttachements(byte[] data, String contentType) throws IOException, DGWSFault {
        if (headerAttachmentMap.size() == 0) {
            return data;
        }
        MultipartAttachmentFilter multipartAttachmentFilter = new MultipartAttachmentFilter();
        // Filter attachements
        return multipartAttachmentFilter.filter(data, contentType, headerAttachmentMap.keySet());
    }

    /**
     * Parse and save MTOM attachments linked in header
     * @param inputStream input
     * @param contentType content type header
     * @throws IOException
     * @throws MimeException
     * @throws XMLStreamException
     */
    private void parseAndSaveAttachments(InputStream inputStream, String contentType)
            throws IOException, MimeException, XMLStreamException {
        List<String> binaryHeaderContentIDs = null;
        String currentContentId = null;
        MimeTokenStream stream = new MimeTokenStream();

        stream.parseHeadless(inputStream, contentType);
        for (EntityState state = stream.getState(); state != EntityState.T_END_OF_STREAM; state = stream.next()) {
            BodyDescriptor bodyDescriptor;
            switch (state) {
                case T_BODY:
                    bodyDescriptor = stream.getBodyDescriptor();
                    if (bodyDescriptor.getMimeType().contains("application/xop+xml")) {
                        // Save Content-Id of all attachement that should be inlined
                        binaryHeaderContentIDs = processMainBody(stream);
                    } else if (bodyDescriptor.getMimeType().contains("application/octet-stream") &&
                        currentContentId != null) {
                        String value = binaryToBase64(stream.getInputStream());
                        if (value != null) {
                            headerAttachmentMap.put(currentContentId, value);
                        }
                    }
                    break;
                case T_FIELD:
                    Field field = stream.getField();
                    if (field != null && binaryHeaderContentIDs != null) {
                        String fieldName = field.getName();
                        if ("Content-Id".equalsIgnoreCase(fieldName) && field.getBody() != null) {
                            // Save current contentId if it should be inlined
                            String body = field.getBody();
                            for (String contentId : binaryHeaderContentIDs) {
                                if (body.contains(contentId)) {
                                    currentContentId = contentId;
                                }
                            }
                        }
                    }
                    break;
                case T_START_BODYPART:
                    // New body - reset content id
                    currentContentId = null;
                    break;
                default:
                    break;
            }
        }
    }

    private String binaryToBase64(InputStream stream) {
        try {
            byte[] bytes = IOUtils.toByteArray(stream);
            byte[] encodedStream = Base64.encodeBase64(bytes);
            return new String(encodedStream);
        } catch (IOException e) {
            return null;
        }
    }

    /**
     * Extract content-id's that should be inlined
     * @param stream body stream
     * @return list of content-id's
     * @throws XMLStreamException
     */
    private List<String> processMainBody(MimeTokenStream stream) throws XMLStreamException {
        List<String> binaryHeaderContentIDs = new ArrayList<String>();
        boolean headerStarted = false;
        XMLStreamReader parser = XML_INPUT_FACTORY.createXMLStreamReader(stream.getInputStream());
        while (parser.hasNext()) {
            int next = parser.next();
            if (next == XMLStreamConstants.START_ELEMENT) {
                QName name = parser.getName();
                String localName = parser.getLocalName();
                if (isBinaryHeader(name)) {
                    headerStarted = true;
                } else {
                    if (headerStarted && "Include".equals(localName)) {
                        String link = parser.getAttributeValue(null, "href");
                        if (link != null) {
                            // Save attachment links
                            binaryHeaderContentIDs.add(extractContentIdFromLink(link));
                        }
                    }
                    headerStarted = false;
                }
            }
        }
        return binaryHeaderContentIDs;
    }

    private String extractContentIdFromLink(String link) {
        return link.replace("cid:", "");
    }

    private boolean isBinaryHeader(QName nodeName) {
        String namespaceURI = nodeName.getNamespaceURI();
        String localPart = nodeName.getLocalPart();
        /* this could be moved to config */
        if (("X509Certificate".equals(localPart) ||
                "SignatureValue".equals(localPart) ||
                "DigestValue".equals(localPart)) &&
                namespaceURI.contains("xmldsig")) {
            return true;
        }
        return false;
    }

}
