/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.util;

import com.trifork.sosigw.api.Constants;
import com.trifork.sosigw.config.ConfigurationException;
import com.trifork.sosigw.config.RuntimeConfig;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sdsd.nsp.slalog.api.SLALogger;
import dk.sdsd.nsp.slalog.impl.NSPAwareRequestWrapper;

import javax.servlet.http.HttpServletRequest;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.Enumeration;


/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class Util {
    private static final int BUFFER_SIZE = 16384;

    private static final ThreadLocal<byte[]> byteBuffer = new ThreadLocal<byte[]>();
    private static final String SOSI_MESSAGE_ID = "SOSI_MESSAGE_ID";

    private static byte[] getThreadLocalBuffer() {
        byte[] buf = byteBuffer.get();
        if (buf == null) {
            buf = new byte[BUFFER_SIZE];
            byteBuffer.set(buf);
        }
        return buf;
    }
    
    public static long copyStream(InputStream is, OutputStream os) throws IOException {
        long total=0;
        byte[] buf = getThreadLocalBuffer();
        while (true) {
            int count = is.read(buf);
            if (count < 0) {
                return total;
            }
            os.write(buf, 0, count);
            total += count;
        }
    }

    public static String readAllFromStreamAsString(InputStream in) throws UnsupportedEncodingException, IOException {
        byte[] buffer = getThreadLocalBuffer();
        int length = 0;
        while (true) {
            int i = in.read(buffer, length, buffer.length - length);
            if (i == -1) {
                break;
            }
            length += i;
            if (length == buffer.length) {
                byte[] newBuf = new byte[length * 2];
                System.arraycopy(buffer, 0, newBuf, 0, buffer.length);
                buffer = newBuf;
            }
        }
        return new String(buffer, 0, length, Constants.UTF_8);
    }

    public static final ThreadLocal<HttpServletRequest> httpRequests = new ThreadLocal<HttpServletRequest>();

    public static void setActiveHttpRequest(HttpServletRequest httpServletRequest) {
	    if (httpServletRequest == null || httpServletRequest instanceof NSPAwareRequestWrapper) {
            httpRequests.set(httpServletRequest);
	    } else {
		    httpRequests.set(new NSPAwareRequestWrapper(httpServletRequest));
	    }
    }
    public static HttpServletRequest getHttpServletRequest() {
        return httpRequests.get();
    }

    public static String getMessageID() {
        if (getHttpServletRequest() == null) {
            throw new IllegalStateException("No active HTTP request to get message id from");
        }
        return (String) getHttpServletRequest().getAttribute(SOSI_MESSAGE_ID);
    }

    public static void setMessageID(String messageId) {
        if (getHttpServletRequest() == null) {
            throw new IllegalStateException("No active HTTP request to set message id '"+messageId+"' on");
        }
        getHttpServletRequest().setAttribute(SOSI_MESSAGE_ID, messageId);
    }
    
    private static SLALogger slaLogger;
    public static void setSLALogger(SLALogger logger) {
        slaLogger = logger;
    }
    public static SLALogger getSLALogger() {
        return slaLogger;
    }
    
    public static final ThreadLocal<SLALogItem> slaLogItems = new ThreadLocal<SLALogItem>();

    public static void setSLALogItem(SLALogItem slaLogItem) {
        slaLogItems.set(slaLogItem);
    }
    public static SLALogItem getSLALogItem() {
        return slaLogItems.get();
    }
    
    public static String makeSecretKey(int secretBitlength) {
        try {
            byte[] digest = new byte[secretBitlength/8];
            SecureRandom.getInstance("SHA1PRNG").nextBytes(digest);
            StringBuffer sb = new StringBuffer();
            for (int i = 0; i < digest.length; i++) {
                byte b = digest[i];
                sb.append(Util.hexDigits.charAt((b >> 4) & 0xf));
                sb.append(Util.hexDigits.charAt(b & 0xf));
            }
            return sb.toString();
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        }
    }


    public static final String hexDigits = "0123456789abcdef";

    public static String getConfigFiledirName() {
        String dirName = ".";

        String catalinaHome = System.getProperty("catalina.home");
        String catalinaBase = System.getProperty("catalina.base");
        String jbossConfigUrl = System.getProperty("jboss.server.config.url");
        String jbossServerBase = System.getProperty("jboss.server.base.dir");
        String triforkDomainDir = System.getProperty("trifork.domain.dir");
        String triforkServerName = System.getProperty("trifork.server.name");

        if (catalinaHome != null) {
            // probably running on tomcat, make config file default location using either catalinaBase or catalinaHome
            if (catalinaBase != null && new File(catalinaBase).exists()) {
                dirName = catalinaBase + "/conf";
            } else {
                dirName = catalinaHome + "/conf";
            }
        } else if (jbossConfigUrl != null) {
            // probably running on JBoss (6)
            try {
                dirName = new URL(jbossConfigUrl).getPath();
            } catch (MalformedURLException e) {
                throw new RuntimeException("Could not parse system property jboss.server.config.url with value " + jbossConfigUrl + " as URL", e);
            }
        } else if (triforkDomainDir != null && triforkServerName != null) {
            // Probably running on trifork. Pick up config in domain config dir for this server.
            dirName = triforkDomainDir + "/config/" + triforkServerName;
        } else if (jbossServerBase != null) {
            // probably running in Wildfly (8)
            dirName = jbossServerBase + "/configuration";
        }

        return dirName;
    }

    public static String computeCacheKey(String nameID) throws ConfigurationException, HttpHeaderException {
        if (RuntimeConfig.isCachePartitioningEnabled()) {
            String httpHeaderName = RuntimeConfig.getCachePartitioningHttpHeaderName();
            if (httpHeaderName == null) {
                throw new ConfigurationException("Property '" + RuntimeConfig.KEY_CACHE_PARTITIONING_HTTP_HEADER_NAME + "' is not defined in runtime configuration");
            }
            Enumeration<?> headers = getHttpServletRequest().getHeaders(httpHeaderName);
            if (headers == null || !headers.hasMoreElements()) {
                throw new HttpHeaderException("HTTP header '" + httpHeaderName + "' is missing in request");
            }
            Object header = headers.nextElement();
            if (header == null) {
                throw new HttpHeaderException("HTTP header '" + httpHeaderName + "' is missing a value in request");
            }
            if (headers.hasMoreElements()) {
                throw new HttpHeaderException("HTTP header '" + httpHeaderName + "' present multiple times in request");
            }
            String partition = String.valueOf(header);
            if (partition.isEmpty()) {
                throw new HttpHeaderException("HTTP header '" + httpHeaderName + "' must not be empty");
            } else if (partition.indexOf("#") > 0) {
                throw new HttpHeaderException("HTTP header '" + httpHeaderName + "' must not contain '#' (value was " + partition + ")");
            }
            SLALogItem slaLogItem = getSLALogItem();
            if (slaLogItem != null) {
                slaLogItem.addCallParameter("sosigw-partition", partition);
            }
            return partition + "#" + nameID;
        } else {
            return nameID;
        }
    }

}
