/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.util.timing;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Used for timing pairs of operations. This class can either be used by
 * calling static start/stop methods, or enter/exit methods on instances
 * of the <code>Timer</code> class.
 * 
 * Operations are represented by a <code>Class</code> and <code>String</code>
 * pair that must matched, e.g.
 * <pre>
 *   Timer.start(getClass(), "operation");
 *   ...
 *   Timer.stop(getClass(), "operation");
 * </pre>
 * or 
 * <pre>
 *   Timer timer = new Timer(...);
 *   timer.enter(getClass(), "operation");
 *   ...
 *   timer.exit(getClass(), "operation");
 * </pre>
 */
@SuppressWarnings("unchecked")
public class Timer {
	
	private static Timer instance = new NopTimer();
	
	private static Log log = LogFactory.getLog(Timer.class);

	private ThreadLocal<TimerStack> tl = new ThreadLocal<TimerStack>();
	private final TimeSink sink;

	private final int maxCapacity;
	
	/**
	 * Create a Timer instance using the <code>TimeSink</code> instance for 
	 * recording timing events 
	 * @param sink the <code>TimeSink</code> to use 
	 */
	public Timer(TimeSink sink) {
		this(sink, Integer.MAX_VALUE);
	}
	
	/**
	 * Create a Timer instance using the <code>TimeSink</code> instance for 
	 * recording timing events, and  a maximum number of nested timing events 
	 * @param sink the <code>TimeSink</code> to use
	 * @param maxCapacity the max number of nested timing events
	 */
	public Timer(TimeSink sink, int maxCapacity) {
		this.sink = sink;
		this.maxCapacity = maxCapacity;
	}

	private Timer() {
		this(null);
	}

	/**
	 * Start timing an event (see enter()).
	 * @param clazz
	 * @param operation
	 */
    public static void start(Class clazz, String operation) {
		instance.enter(clazz, operation);
	}

	/**
	 * Stop timing an event (see exit()).
	 * @param clazz
	 * @param operation
	 */
	public static void stop(Class clazz, String operation) {
		instance.exit(clazz, operation);
	}
	
	/**
	 * Set the timer to use for all static start/stop calls.
	 * @param timer
	 */
	public static void setTimer(Timer timer) {
		Timer.instance = timer;		
	}
	
	/**
	 * Start timing an event. If the max number of timings is exceeded, all 
	 * non-stopped events will be discarded.
	 * @param clazz
	 * @param operation
	 */
	public void enter(Class clazz, String operation) {
		TimerStack stack = tl.get();
		if (stack == null) {
			stack = new TimerStack(maxCapacity);
			tl.set(stack);
		}
		TimerEvent e = new TimerEvent(clazz, operation, System.nanoTime());
		if (stack.push(e) == null) {
			log.error("timer stack exceeded: " + elementInfo(clazz, operation));
			discardTimerStack(stack, "timer stack exceeded");
		} else {
			sink.enter(e);
		}
	}
	
	/**
	 * Stop timing an event. Timings must be nested, i.e. the last started event
	 * must have matching class/operation. If not then all other non-stopped
	 * events will be discarded.
	 * @param clazz
	 * @param operation
	 */
	public void exit(Class clazz, String operation) {
		TimerStack stack = tl.get();
		if (stack == null) {
			log.error("empty timer.stack (" + clazz.getSimpleName() + ", " + operation + ")");
		} else {
			TimerEvent e = stack.pop();
			if (e == null) {
				log.error("missing timer.enter");
			} else if (clazz != e.clazz || operation != e.operation) {
				log.error("timer stack inconsistent: " + elementInfo(e.clazz, e.operation)
						+ elementInfo(clazz, operation));
				// the popped element must be discarded
				stack.push(e);
				discardTimerStack(stack, "timer stack inconsistent");
			} else {
				long duration = (System.nanoTime() - e.startTime) / 1000000;
				TimerEvent parent = stack.peek();
				if (parent != null) {
				    parent.childtime += duration;
				}
				sink.exit(e, duration);
			}
		}
	}
	
	private String elementInfo(Class clazz, String operation) {
		return "(" + clazz.getSimpleName() + ", " + operation + ")";
	}
	
	private void discardTimerStack(TimerStack stack, String msg) {
		TimerEvent e;
		while ((e = stack.pop()) != null) {
			sink.discard(e, msg);
		}
	}

	/**
	 * A timer that records nothing.
	 */
	public static class NopTimer extends Timer {

		public void enter(Class clazz, String operation) {
		}

		public void exit(Class clazz, String operation) {
		}
	}
}
