/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.util.timing;

import java.util.ArrayList;

/**
 * A stack of <code>TimerEvent</code>s.
 */
public class TimerStack {
	
	private ArrayList<TimerEvent> stack; 
	private int stackPointer = -1;
	private final int maxCapacity;

	/**
	 * Create a stack with no max capacity, i.e. Integer.MAX_VALUE.
	 */
	public TimerStack() {
		this(Integer.MAX_VALUE);
	}

	/**
	 * Create a stack with the given max capacity.
	 * @param maxCapacity must not be negative
	 */
	public TimerStack(int maxCapacity) {
		this(maxCapacity, 16);
	}

	/**
	 * Create a stack with the given max and initial capacity.
	 * @param maxCapacity must not be negative
	 * @param initialCapacity must not be negative
	 */
	public TimerStack(int maxCapacity, int initialCapacity) {
		stack = new ArrayList<TimerEvent>(initialCapacity);
		this.maxCapacity = maxCapacity;
	}

	/**
	 * Puts the element at the top of this tack and return it. If the maximum
	 * capacity is exceeded <code>null</code> is returned.
	 * 
	 * @param element to put on the stack
	 * @return the element put on the stack
	 */
	public TimerEvent push(TimerEvent element) {
		if (stack.size() < maxCapacity) {
			stack.add(element);
			stackPointer++;
			return element;
		} else {
			return null;
		}
	}

	/**
	 * Removes the element at the top of this stack and returns it. If this
	 * stack is empty <code>null</code> is returned.
	 *
	 * @return the top element of the stack 
	 */
	public TimerEvent pop() {
		if (stackPointer < 0) {
			return null;
		} else {
			return stack.remove(stackPointer--);
		}
	}

	/**
	 * Returns the element at the top of this stack. If this
	 * stack is empty <code>null</code> is returned.

	 * @return the top element of the stack
	 */
	public TimerEvent peek() {
		if (stackPointer < 0) {
			return null;
		} else {
			return stack.get(stackPointer);
		}
	}
}
