/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.ws;

import java.util.Date;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.xml.namespace.QName;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;

import org.apache.log4j.Logger;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.trifork.sosigw.api.Constants;
import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.util.Util;

import dk.sosi.seal.model.constants.MedComTags;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SOSIAttributes;
import dk.sosi.seal.xml.XmlUtil;

/**
 * Ensure that DGWS required headers (mainly medcom:) are included in all replies from our service. 
 * Note that "proxy" is not covered by this, as it is a plain HttpServlet. 
 * See http://www.javaworld.com/javaworld/jw-02-2007/jw-02-handler.html for a good introduction to the JAX-WS handler framework.
 *  
 * @author Morten Grouleff <mgr@trifork.com>
 *
 */
public class DGWSHandler implements SOAPHandler<SOAPMessageContext> {
    public static final Logger log = Logger.getLogger(DGWSHandler.class); 

    private static final ThreadLocal<SOAPHeader> inbound = new ThreadLocal<SOAPHeader>();
    //    private static final ThreadLocal<Map<String, List<String>>> httpHeaders = new ThreadLocal<Map<String, List<String>>>();

    public Set<QName> getHeaders() {
        // ignored.
        return null;
    }

    public void close(MessageContext context) {
        // ignored.
        
    }

    public static SOAPHeader getSOAPHeader() {
        return inbound.get();
    }
    
    public boolean handleFault(SOAPMessageContext context) {
        return handleDGWSHeaders(context, true);
    }

    public boolean handleMessage(SOAPMessageContext context) {
        Boolean isOutbound = (Boolean)context.get(SOAPMessageContext.MESSAGE_OUTBOUND_PROPERTY);
        if (!isOutbound) { // Request from client to server.
            SOAPHeader header;
            try {
                header = context.getMessage().getSOAPHeader();
                inbound.set(header);
                HttpServletRequest httpServletRequest = gethttpRequest(context);
                Util.setActiveHttpRequest(httpServletRequest);
                Util.setMessageID(findMessageId(header));
            } catch (SOAPException e) {
                inbound.set(null);
                log.error("Error handling DGWS headers", e);
            }
        } else { // Return reply to client
            handleDGWSHeaders(context, false);
            inbound.set(null);     // assist gc.
            Util.httpRequests.set(null);
        }
        return true;
    }

    private HttpServletRequest gethttpRequest(SOAPMessageContext context) {
        return (HttpServletRequest)context.get(MessageContext.SERVLET_REQUEST);
    }

    /**
     * Find IDCard.getAuthenticationLevel().getLevel() inside outbound, or return 1 if not found.
     */
    private int getIDCardSecLevelOr1(SOAPEnvelope envelope) {
        NodeList samlAssertions = envelope.getElementsByTagNameNS(NameSpaces.SAML2ASSERTION_SCHEMA, SAMLTags.ATTRIBUTE);
        for (int i = 0; i < samlAssertions.getLength(); i++) {
            Node item = samlAssertions.item(i);
            Node attr = item.getAttributes().getNamedItem("Name");
            if (item instanceof Element && attr != null && SOSIAttributes.AUTHENTICATION_LEVEL.equals(attr.getNodeValue())) {
                Element valueElm = getElm((Element)item, NameSpaces.SAML2ASSERTION_SCHEMA, SAMLTags.ATTRIBUTE_VALUE);
                try {
                    return Integer.parseInt(valueElm.getTextContent());
                } catch (NumberFormatException nfe) {
                    // ignore.
                }
            }
        }
        
        return 1;
    }

    private static Element ensureElm(Element parent, String namespaceURI, String localName) {
        Element child = getElm(parent, namespaceURI, localName);
        if (child == null) {
            child = parent.getOwnerDocument().createElementNS(namespaceURI, localName);
            parent.appendChild(child);
        }
        return child;
    }
    
    // private
    static Element getElm(Element parent, String namespaceURI, String localName) {
        if (parent == null) {
            return null;
        }
        NodeList elmByName = parent.getElementsByTagNameNS(namespaceURI, localName);
        if (elmByName != null && elmByName.getLength() == 1) {
            Node item = elmByName.item(0);
            if (item instanceof Element) {
                Element elm = (Element)item;
                return elm;
            }
        }
        return null;
    }

    

    private boolean handleDGWSHeaders(SOAPMessageContext context, boolean isFault) {
        SOAPHeader in = inbound.get();
        if (in == null) {
            return true;
        }
        try {
            SOAPEnvelope envelope = context.getMessage().getSOAPPart().getEnvelope();
            SOAPHeader outbound = envelope.getHeader();
            int idcardSecLevel = getIDCardSecLevelOr1(envelope);

            if (outbound == null) {
                outbound = envelope.addHeader();
            }

            Element outbound_wsse_security = ensureElm(outbound, NameSpaces.WSSE_SCHEMA, "Security");
            Element outbound_wsu_timestamp= ensureElm(outbound_wsse_security, NameSpaces.WSU_SCHEMA, "Timestamp");
            Element outbound_wsu_created = ensureElm(outbound_wsu_timestamp, NameSpaces.WSU_SCHEMA, "Created");
            outbound_wsu_created.setTextContent(XmlUtil.toXMLTimeStamp(new Date(), !StaticConfigImpl.getInstance().useLocalTimeInSaml()));

            Element outbound_medcomHeader = ensureElm(outbound, Constants.URL_MEDCOM, MedComTags.HEADER);

            Element out_secLevel = ensureElm(outbound_medcomHeader, Constants.URL_MEDCOM, MedComTags.SECURITY_LEVEL);
            out_secLevel.setTextContent("" + idcardSecLevel);

            if (idcardSecLevel == 4) {
                Element outbound_timeout = ensureElm(outbound_medcomHeader, Constants.URL_MEDCOM, "TimeOut");
                outbound_timeout.setTextContent("1440");
            }
            
            Element outbound_linking = ensureElm(outbound_medcomHeader, Constants.URL_MEDCOM, MedComTags.LINKING);
            
            Element in_linking = getElm(in, Constants.URL_MEDCOM, MedComTags.LINKING);
            Element outbound_msgid = ensureElm(outbound_linking, Constants.URL_MEDCOM, MedComTags.MESSAGE_ID);
            outbound_msgid.setTextContent(XmlUtil.createNonce());
            Element in_flowid = getElm(in_linking, Constants.URL_MEDCOM, MedComTags.FLOW_ID);
            if (in_flowid != null) {
                // Keep existing flowID.
                Node cloneNode = in_flowid.cloneNode(true);
                outbound_linking.getOwnerDocument().adoptNode(cloneNode);
                outbound_linking.appendChild(cloneNode);
            } else {
                // No existing flowID, make a brand new one.
                Node out_flowid = ensureElm(outbound_linking, Constants.URL_MEDCOM, MedComTags.FLOW_ID);
                out_flowid.setTextContent(XmlUtil.createNonce());
            }

            String in_msgid_value = findMessageId(in);
            if (in_msgid_value != null) {
                Element outbound_inresponsetomsgid = ensureElm(outbound_linking, Constants.URL_MEDCOM, MedComTags.IN_RESPONSE_TO_MESSAGE_ID);
                outbound_inresponsetomsgid.setTextContent(in_msgid_value);
            }

            if (!isFault) {
                // FlowStatus irrelevant for Fault / error code 500.
                Element outbound_flowstatus = ensureElm(outbound_medcomHeader, Constants.URL_MEDCOM, MedComTags.FLOW_STATUS);
                outbound_flowstatus.setTextContent("flow_finalized_succesfully");
            }

            Element in_pri = getElm(in, Constants.URL_MEDCOM, Constants.REQUEST_PRIORITY);
            if (in_pri != null) {
                Node cloneNode = in_pri.cloneNode(true);
                outbound_medcomHeader.getOwnerDocument().adoptNode(cloneNode);
                outbound_medcomHeader.appendChild(cloneNode);
            }


            // 
        } catch (SOAPException e) {
            log.error("Error handling DGWS headers", e);
        }
        return true;
    }

    private String findMessageId(SOAPHeader soapHeader) {
        Element medcomLinking = getElm(soapHeader, Constants.URL_MEDCOM, MedComTags.LINKING);
        if (medcomLinking == null) {
            return null;
        }

        Element in_msgid = getElm(medcomLinking, Constants.URL_MEDCOM, MedComTags.MESSAGE_ID);
        String in_msgid_value = null;
        if (in_msgid != null) {
            in_msgid_value = in_msgid.getTextContent();
        }
        return in_msgid_value;
    }
}
