/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.ws;

import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import javax.xml.datatype.XMLGregorianCalendar;

import com.trifork.sosigw.config.StaticConfigApi;
import com.trifork.sosigw.config.StaticConfigImpl;

import oasis.names.tc.saml._2_0.assertion.AssertionType;
import oasis.names.tc.saml._2_0.assertion.Attribute;
import oasis.names.tc.saml._2_0.assertion.AttributeStatement;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.SystemIDCard;
import dk.sosi.seal.model.SystemInfo;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.UserInfo;
import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.model.constants.MedcomAttributes;
import dk.sosi.seal.model.constants.SOSIAttributes;
import dk.sosi.seal.modelbuilders.ModelBuildException;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class IDCardFromSamlAssertion {

    private static final StaticConfigApi STATIC_CONF = StaticConfigImpl.getInstance();

    public static IDCard build(AssertionType assertion) throws ModelBuildException {

        IDCard result = null;

        if (!IDValues.IDCARD.equals(assertion.getId())) {
            return null;
        }
        

        String itSystemName = "", ocesCertHash = null, id = "", version = "",
        cpr = "", givenName = "", surName = "", email = "", occupation = "", userRole = "", 
        authorizationCode = "", careProviderID = "", careProviderIDType = "", 
        careProviderName = "", authLevel = "";
        boolean hasIDCardData = false, hasSystemLog=false, hasUserLog=false;

        Date createdDate = null, expiryDate = null;
        if (assertion.getConditions() != null) { // fix for NSP-SUPPORT 191
            // Assume that sender is using our local timezone (what else can we do?)
            int offset = Calendar.getInstance().getTimeZone().getOffset(System.currentTimeMillis());
            
            XMLGregorianCalendar notBefore = assertion.getConditions().getNotBefore();
            XMLGregorianCalendar notOnOrAfter = assertion.getConditions().getNotOnOrAfter();
    
            if (STATIC_CONF.useLocalTimeInSaml()) {
                createdDate = new Date(notBefore.toGregorianCalendar().getTimeInMillis() - offset);
                expiryDate = new Date(notOnOrAfter.toGregorianCalendar().getTimeInMillis() - offset);
            } else {
                createdDate = notBefore.toGregorianCalendar().getTime();
                expiryDate = notOnOrAfter.toGregorianCalendar().getTime();
            }
        }
        else { // bit of code duplication from SEAL to handle the case where the caller does not set a validity
            final int IDCARD_BEGIN_TIME_BUFFER_IN_MINUTES = 5;
            final int MAX_IDCARD_LIFE_IN_HOURS = 24;

            long starttime = System.currentTimeMillis() - IDCARD_BEGIN_TIME_BUFFER_IN_MINUTES * 60 * 1000;
            createdDate = new Date(starttime);
            expiryDate = new Date(starttime + MAX_IDCARD_LIFE_IN_HOURS * 60 * 60 * 1000);
        }

        String issuer = assertion.getIssuer();
        
        Boolean isUserIDCard = null;

        List<AttributeStatement> attributeStatement = assertion.getAttributeStatement();
        
        for (Iterator<AttributeStatement> iterator = attributeStatement.iterator(); iterator.hasNext();) {
            AttributeStatement stmt = iterator.next();
            List<Attribute> attributeList = stmt.getAttribute();
            for (Iterator<Attribute> iterator2 = attributeList.iterator(); iterator2.hasNext();) {
                Attribute attr = iterator2.next();
                String attributeName = attr.getName();
                if (IDValues.SYSTEM_LOG.equals(stmt.getId())) {
                    if (MedcomAttributes.IT_SYSTEM_NAME.equals(attributeName)) {
                        itSystemName = attr.getAttributeValue();
                    } else if (MedcomAttributes.CARE_PROVIDER_ID.equals(attributeName)) {
                        careProviderID = attr.getAttributeValue();
                        careProviderIDType = attr.getNameFormat();
                    } else if (MedcomAttributes.CARE_PROVIDER_NAME.equals(attributeName)) {
                        careProviderName = attr.getAttributeValue();
                    }
                    hasSystemLog = true;
                } else if (IDValues.IDCARD_DATA.equals(stmt.getId())) {
                    if (SOSIAttributes.OCES_CERT_HASH.equals(attributeName)) {
                        ocesCertHash = attr.getAttributeValue() ;
                    } else if (SOSIAttributes.IDCARD_ID.equals(attributeName)) {
                        id = attr.getAttributeValue();
                    } else if (SOSIAttributes.IDCARD_VERSION.equals(attributeName)) {
                        version = attr.getAttributeValue();
                    } else if (SOSIAttributes.IDCARD_TYPE.equals(attributeName)) {
                        if (IDCard.IDCARDTYPE_USER.equals(attr.getAttributeValue()))
                            isUserIDCard = Boolean.TRUE;
                        else if (IDCard.IDCARDTYPE_SYSTEM.equals(attr.getAttributeValue()))
                            isUserIDCard = Boolean.FALSE;
                    } else if (SOSIAttributes.AUTHENTICATION_LEVEL.equals(attributeName)) {
                        authLevel = attr.getAttributeValue();
                    }
                    hasIDCardData = true;
                } else if (IDValues.USER_LOG.equals(stmt.getId())) {
                    if (MedcomAttributes.USER_CIVIL_REGISTRATION_NUMBER.equals(attributeName)) {
                        cpr = attr.getAttributeValue();
                    } else if (MedcomAttributes.USER_GIVEN_NAME.equals(attributeName)) {
                        givenName = attr.getAttributeValue();
                    } else if (MedcomAttributes.USER_SURNAME.equals(attributeName)) {
                        surName = attr.getAttributeValue();
                    } else if (MedcomAttributes.USER_EMAIL_ADDRESS.equals(attributeName)) {
                        email = attr.getAttributeValue();
                    } else if (MedcomAttributes.USER_OCCUPATION.equals(attributeName)) {
                        occupation = attr.getAttributeValue();
                    } else if (MedcomAttributes.USER_ROLE.equals(attributeName)) {
                        userRole = attr.getAttributeValue();
                    } else if (MedcomAttributes.USER_AUTHORIZATION_CODE.equals(attributeName)) {
                        authorizationCode = attr.getAttributeValue();
                    }
                    hasUserLog = true;
                }
            }
        }

        if (version == null || !STATIC_CONF.getSupportedDGWSVersion().equals(version.trim())) {
            throw new ModelBuildException("Version mismatch on IDCard. "+version+" is not supported");
        }
        

        CareProvider careProvider = new CareProvider(careProviderIDType, careProviderID, careProviderName);
        SystemInfo systemInfo = new SystemInfo(careProvider, itSystemName);

        // All IDCard types must have a IDCardData element
        if(!hasIDCardData)  throw new ModelBuildException("IDCardData element missing for IDCard");

        // All IDCard types must have a SystemLog element
        if(!hasSystemLog)  throw new ModelBuildException("SystemLog element missing for IDCard");

        if (isUserIDCard == null)
            throw new ModelBuildException("ID Card type not found or invalid");
        else if (isUserIDCard.booleanValue()) {
            if(!hasUserLog) throw new ModelBuildException("UserLog element missing for UserIDCard");
            UserInfo userInfo = new UserInfo(cpr, givenName, surName, email, occupation, userRole, authorizationCode);
            result = new UserIDCard(STATIC_CONF.getSupportedDGWSVersion(), null, id, AuthenticationLevel.MOCES_TRUSTED_USER, ocesCertHash, issuer,
                    systemInfo, userInfo, createdDate, expiryDate, null, null, null);
        } else {
            if(hasUserLog) throw new ModelBuildException("IDCard type is 'system', but also has a UserLog element (??)");
            result = new SystemIDCard(STATIC_CONF.getSupportedDGWSVersion(), null, id, AuthenticationLevel.getEnumeratedValue(Integer.parseInt(authLevel)), ocesCertHash,
                    issuer, systemInfo, createdDate, expiryDate, null, null, null);
        }
        return result;
    
    }
}
