/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.ws;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.Constants;
import com.trifork.sosigw.api.IDCardCache;
import com.trifork.sosigw.api.IDCardState;
import com.trifork.sosigw.api.IdCardSigner;
import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.StaticConfigApi;
import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.sts.STSClusterProxy;
import com.trifork.sosigw.sts.STSProxy;
import com.trifork.sosigw.sts.STSUnreachableException;
import com.trifork.sosigw.util.SosiSeal;
import com.trifork.sosigw.util.Util;
import com.trifork.sosigw.util.timing.Timer;
import dk.sdsd.nsp.slalog.api.SLALogException;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sosi.gw._2007_09.DGWSFault;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.modelbuilders.SecurityTokenResponseModelBuilder;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;

import java.io.ByteArrayInputStream;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Properties;

/**
 * 
 * 
 * @author Jeppe Sommer <jso at trifork.com>
 * @author Morten Grouleff <mgr at trifork.com>
 */
public class IdCardSignerImpl implements IdCardSigner {
    public static final Logger log = Logger.getLogger(IdCardSignerImpl.class);
    private static final StaticConfigApi STATIC_CONF = StaticConfigImpl.getInstance();
    // We should only create the sts proxies once, so make stsService static
    private static final STSProxy stsService =
            STSClusterProxy.create(RuntimeConfig.getSecurityTokenServiceUrlList());

    public String signIdCard(String cacheKey, byte[] signedBytes, byte[] certificate) throws DGWSFault {
        // Build request ...
        final IDCardCache idCardCache = Factory.getIDCardCache();
        IDCardClusterCacheEntry cachedIDCardEntry = idCardCache.findIDCardOrNull(cacheKey);

        if (cachedIDCardEntry == null || cachedIDCardEntry.getState() == IDCardState.REVOKED) {
            // We wait a short while, to allow the message to get through...
            log.debug("Waiting for entry for cacheKey=" + cacheKey + " to arrive.");
            try {
                int millis = STATIC_CONF.getClusterTuningParam("cluster.tuningparam.unknown.sync.delay", 500);
                Thread.sleep(millis);
            } catch (InterruptedException e) {// ignore.
            }
            cachedIDCardEntry = idCardCache.findIDCardOrNull(cacheKey);
        }
        if (cachedIDCardEntry == null) {
            if (log.isDebugEnabled()) {
                log.debug("No unsigned idCard for cacheKey in cache: " + cacheKey);
            }
            throw new DGWSFault("No unsigned idCard for cacheKey in cache: " + cacheKey, Constants.FAULTCODE_NO_VALID_IDCARD_IN_CACHE);
        }
        if (cachedIDCardEntry.isTimedOutWaitingForSigning()) {
            if (log.isInfoEnabled()) {
                log.info("User tried to sign an idcard where the unsigned card has timed out: " + cacheKey);
            }
            throw new DGWSFault("Unsigned idCard for cacheKey in cache timed out: " + cacheKey, Constants.FAULTCODE_NO_VALID_IDCARD_IN_CACHE);
        }
        SOSIFactory sosiFactory = SosiSeal.getSosiSealFactory();

        UserIDCard idCard = cachedIDCardEntry.getIdCard();

        SecurityTokenRequest securityTokenRequest = sosiFactory.createNewSecurityTokenRequest();

        X509Certificate x509Cert = null;

        try {
            CertificateFactory cf = CertificateFactory.getInstance("X.509");
            x509Cert = (X509Certificate)cf.generateCertificate(new ByteArrayInputStream(certificate));
        } catch (CertificateException e1) {
            throw new DGWSFault("Exception deserializing certificate: " + e1.getMessage(), Constants.FAULTCODE_INTERNAL_ERROR, e1);
        }

        securityTokenRequest.setIDCard(idCard);

        String signature = XmlUtil.toBase64(signedBytes);
        idCard.injectSignature(signature, x509Cert);

        idCardCache.updateEntryState(cachedIDCardEntry, IDCardState.AWAITING_STS_RESPONSE);

        // We do not need to broadcast this state. It's okay to just let this cache know this.
        // The other members cannot receive the reply from the STS anyway, so they have to make the request themselves.
        //idCardCache.entryUpdatedLocally(cachedIDCardEntry);

        // Call Service
        SecurityTokenResponse reply = null;
        securityTokenRequest.setMessageID(Util.getMessageID());
        Document doc = securityTokenRequest.serialize2DOMDocument();

        SLALogItem slaLogItem = Util.getSLALogger().createLogItem("SOSIGWidCardSignerImpl.signIdCard()", "21");
        slaLogItem.setMessageId(Util.getMessageID());
        slaLogItem.setClientIP("127.0.0.1");
        try {
            reply = callSTS(sosiFactory, doc, slaLogItem);

            if (!reply.isFault()) {
                IDCard stsSignedIDCard = reply.getIDCard();
                synchronized (cachedIDCardEntry) {
                    cachedIDCardEntry.setIdCard((UserIDCard)stsSignedIDCard);
                    idCardCache.updateEntryState(cachedIDCardEntry, IDCardState.SIGNED);
                }
                idCardCache.entryUpdatedLocally(cachedIDCardEntry);
                return "ok";
            } else {
                log.warn("Got fault from STS: " + reply.getFaultCode() + ": " + reply.getFaultString());
                throw new DGWSFault(reply.getFaultString(), reply.getFaultCode());
            }
        } finally {
            try {
                slaLogItem.store();
            } catch (SLALogException e) {
                log.error("Cannot store SLA Log Item", e);
            }
        }
    }

    private SecurityTokenResponse callSTS(SOSIFactory sosiFactory, Document doc, SLALogItem slaLogItem) throws DGWSFault {
        SecurityTokenResponse reply;
        String timingOperationName = "callIdpService";
        try {
            Timer.start(IdCardSignerImpl.class, timingOperationName);
            String responseFromSTS = stsService.callSTS(doc);

            Properties properties = SignatureUtil.setupCryptoProviderForJVM();
            Document returnDoc = XmlUtil.readXml(properties, responseFromSTS, true);

            // The modelbuilder below already validates ALL signatures in the document. So we don't do any validation here

            // Build reply
            SecurityTokenResponseModelBuilder b = new SecurityTokenResponseModelBuilder(sosiFactory);
            reply = b.buildModel(returnDoc);

            SlaLogItemUtil.setSLALogValues(slaLogItem);

            slaLogItem.setCallResultOk();
        } catch (STSUnreachableException e) {
            slaLogItem.setCallResultError("Could not call STS: " + e.getMessage());
            throw new DGWSFault(e.getMessage(), Constants.FAULTCODE_INTERNAL_ERROR, e);
        } finally {
            Timer.stop(IdCardSignerImpl.class, timingOperationName);
        }
        return reply;
    }



}
