/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw;

import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Map;
import java.util.Properties;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NamingException;
import javax.sql.DataSource;

import org.apache.log4j.Logger;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.mockito.Mockito;
import org.testcontainers.containers.MySQLContainer;
import org.testcontainers.ext.ScriptUtils;
import org.testcontainers.jdbc.JdbcDatabaseDelegate;

import com.mysql.cj.jdbc.MysqlDataSource;
import com.trifork.sosigw.auditlog.AuditLog;
import com.trifork.sosigw.cluster.IDCardClusterCacheImpl;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.StaticConfigApi;
import com.trifork.sosigw.config.StaticConfigImpl;


public class BootstrapTest {
    private static final Logger log = Logger.getLogger(BootstrapTest.class);

    private static StaticConfigApi config;
    
    @BeforeClass
    public static void init() {
        config = StaticConfigImpl.getInstance();
        StaticConfigImpl.setInstance(new StaticConfigImpl());
        
        
        mysql = new MySQLContainer<>("mysql:5.7.22");
        mysql.withUsername("root").withPassword("").start();
        JdbcDatabaseDelegate jdbcDatabaseDelegate = new JdbcDatabaseDelegate(mysql, "");
        ScriptUtils.runInitScript(jdbcDatabaseDelegate, "com/trifork/sosigw/create-local-mysql-db.sql");
        ScriptUtils.runInitScript(jdbcDatabaseDelegate, "com/trifork/sosigw/create-global-mysql-db.sql");
    }
    
    @AfterClass
    public static void tearDown() {
        StaticConfigImpl.setInstance(config);
        config = null;
    }    

    
    @Before
    public void setup() {
        if(AuditLog.isWorking()) {
            AuditLog.stop();
        }
        ConfigTestUtil.init(true);
    }
    
    @After
    public void cleanup() {
        ConfigTestUtil.cleanup();
    }

    @Test
    public void testBootstrapDefaultMode() {
        doStartStop();
    }

    @Test
    public void testBootstrapIDeveloperMode() {
        ConfigTestUtil.setMode("developer");

        doStartStop();
    }

    @Test
    public void testBootstrapInBootstrapMode() {
        ConfigTestUtil.setMode("bootstrap");

        doStartStop();
    }

    @Test
    public void testBootstrapInProductionMode() throws NamingException, SQLException {
        // Create initial context
        System.setProperty(Context.INITIAL_CONTEXT_FACTORY, "org.apache.naming.java.javaURLContextFactory");
        System.setProperty(Context.URL_PKG_PREFIXES, "org.apache.naming");

        
        final Context postgresCtx = init("java:comp/env/jdbc/mock");

        final String topStmt = "SELECT configpk, moment FROM sosigwlog.RUNTIMECONFIG WHERE configpk=(SELECT MAX(configpk) FROM sosigwlog.RUNTIMECONFIG)";

        final DataSource ds = Mockito.mock(DataSource.class);
        final Connection connection = Mockito.mock(Connection.class);

        final PreparedStatement topPreparedStmt = Mockito.mock(PreparedStatement.class);

        Mockito.when(ds.getConnection()).thenReturn(connection);

        /* TOP STATEMENT */
        Mockito.when(connection.prepareStatement(topStmt)).thenReturn(topPreparedStmt);
        Mockito.when(topPreparedStmt.execute()).thenReturn(false);

        /* SETUP NAMING CONTEXT */
        postgresCtx.rebind("java:comp/env/jdbc/mock", ds);

        ConfigTestUtil.setMode("production");
        ConfigTestUtil.setStaticConfigProperty(StaticConfigImpl.KEY_JDBC_DATASOURCE_NAME, "java:comp/env/jdbc/mock");

        Factory.setTheIDCardCache(new IDCardClusterCacheImpl(null));

        doStartStop();
    }

    @Test
    public void testBootstrapInTestMode() {
        ConfigTestUtil.setMode("test");

        Factory.setTheIDCardCache(new IDCardClusterCacheImpl(null));

        doStartStop();
    }

    @Test
    public void testBootstrapInUnkownMode() {
        ConfigTestUtil.setMode(null);

        Factory.setTheIDCardCache(new IDCardClusterCacheImpl(null));

        doStartStop();
    }

    public static DataSource getDataSource(MySQLContainer<?> mysql) {
        MysqlDataSource ds = new MysqlDataSource();
        ds.setURL(mysql.getJdbcUrl());
        ds.setUser(mysql.getUsername());
        ds.setPassword(mysql.getPassword());
        return ds;
    }

    private static MySQLContainer<?> mysql;   

    @Test
    public void testBootstrapInTestModeWithCentralAuditLogTablePrefix()
            throws NamingException, SQLException, IOException {
        ConfigTestUtil.setMode("test");

        // Create initial context
        System.setProperty(Context.INITIAL_CONTEXT_FACTORY, "org.apache.naming.java.javaURLContextFactory");
        System.setProperty(Context.URL_PKG_PREFIXES, "org.apache.naming");

        final Context postgresCtx = init("java:comp/env/jdbc/postmysql");
        final Context sosigwcentralCtx = init("java:jdbc/sosigwcentral");

        final DataSource localDS = getDataSource(mysql); // SqlTests.createMemDatabase("local");//Mockito.mock(DataSource.class);
        final DataSource globalDS = getDataSource(mysql); // SqlTests.createMemDatabase("global", "sa",
                                                          // "");//Mockito.mock(DataSource.class);

        // SqlTestsUtil.executeSql(localDS, this.getClass(), "create-local-mysql-db.sql");
        // SqlTestsUtil.executeSql(globalDS, "create.sql");

        /* SETUP NAMING CONTEXT */
        postgresCtx.rebind("java:comp/env/jdbc/postmysql", localDS);
        sosigwcentralCtx.rebind("java:jdbc/sosigwcentral", globalDS);

        ConfigTestUtil.setStaticConfigProperty(StaticConfigImpl.KEY_JDBC_DATASOURCE_NAME, "java:comp/env/jdbc/postmysql");
        ConfigTestUtil.setStaticConfigProperty(StaticConfigImpl.KEY_CENTRAL_AUDITLOG_TP_KEY, "sosigwgloballog.");
        ConfigTestUtil.setStaticConfigProperty(StaticConfigImpl.KEY_JDBC_DRIVER, "com.mysql.cj.jdbc.Driver");

        ConfigTestUtil.setRuntimeConfigCategoryProperty(RuntimeConfig.KEY_GLOBALJDBCURL, mysql.getJdbcUrl());
        ConfigTestUtil.setRuntimeConfigCategoryProperty(RuntimeConfig.KEY_GLOBALJDBCUSERNAME, mysql.getUsername());
        ConfigTestUtil.setRuntimeConfigCategoryProperty(RuntimeConfig.KEY_GLOBALJDBCPASSWORD, mysql.getPassword());
        ConfigTestUtil.setRuntimeConfigCategoryProperty(RuntimeConfig.KEY_GLOBALJDBCINTERVAL, "10");

        Factory.setTheIDCardCache(new IDCardClusterCacheImpl(null));

        doStartStop(100);

    }

    private Context init(String ctx) throws NamingException {
        String[] split = ctx.split("/");

        InitialContext ic = new InitialContext();
        Context result = ic;
        for (String string : split) {
            try {
                result = result.createSubcontext(string);
            } catch (NameAlreadyBoundException ex) {
                Object lookup = result.lookup(string);
                if (lookup instanceof Context) {
                    result = (Context)result.lookup(string);
                } else {
                    break;
                }
            }
        }
        return ic;
    }

    private void doStartStop() {
        this.doStartStop(0);
    }

    private void doStartStop(long ms) {
        Bootstrap.start();
        if (ms > 0) {
            try {
                Thread.sleep(ms);
            } catch (InterruptedException e) {
                log.debug("Interrupted.");
                // Continue
            }
        }
        log.debug("Running Bootstrap. Mode:" + StaticConfigImpl.getInstance().getMode());
        //Bootstrap.stop();
    }
}
