/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.auditlog;

import static org.junit.Assert.assertEquals;
import static org.mockito.Matchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.verifyZeroInteractions;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Collection;

import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.mortbay.log.Log;

import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.config.StaticConfigApi;
import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.console.datamodel.AuditLogQuery;

import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.SystemInfo;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.UserInfo;

public class AuditLogTest {
    private static StaticConfigApi config;
    private static Object loggers, searchers;

    @BeforeClass
    public static void saveConfig() throws Exception {
        config = StaticConfigImpl.getInstance();
        StaticConfigImpl.setInstance(new StaticConfigImpl());
        loggers = getHiddenField("loggers");
        searchers = getHiddenField("searchers");

        setHiddenField("loggers", null);
        setHiddenField("searchers", null);
    }

    @AfterClass
    public static void restoreConfig() throws Exception{
        StaticConfigImpl.setInstance(config);
        
        setHiddenField("loggers", loggers);
        setHiddenField("searchers", searchers);
    }

    @Before
    public void setup() throws Exception {
        if(AuditLog.isWorking()) {
            AuditLog.stop();
        }
    }

    private static Object getHiddenField(String fieldName)
            throws NoSuchFieldException, SecurityException, IllegalArgumentException, IllegalAccessException {
        Field field = AuditLog.class.getDeclaredField(fieldName);
        field.setAccessible(true);
        Object object = field.get(null);    
        field.setAccessible(false);
        return object;
    }

    private static void setHiddenField(String fieldName, Object value) throws Exception {
        Field field = AuditLog.class.getDeclaredField(fieldName);
        field.setAccessible(true);
        if(value == null) {
            Method method = field.getType().getMethod("clear", null);
            method.invoke(field.get(null), null);
        } else {
            Method method = field.getType().getMethod("addAll", Collection.class);
            method.invoke(field.get(null), value);
        }
        field.setAccessible(false);
    }

    @Test
    public void testLog() throws Exception {
        int old = AuditLog.getQueueSize();

        IDCardClusterCacheEntry entry = generateEntry();
        AuditLog.logIdcardChange(entry, "127.0.0.1");
        AuditLog.logIdcardUse(entry, "127.0.0.1", "urn:to", "urn:action");
        
        assertEquals(old + 2, AuditLog.getQueueSize());

        AuditLogger logger = mock(AuditLogger.class);

        AuditLog.addLogger(logger);

        AuditLog.start();

        while(!AuditLog.isWorking()) {
            Log.debug("Waiting for Auditlog to start working");
            Thread.sleep(100);
        }     
        Thread.sleep(old * 50 + 100);
        
        assertEquals(0, AuditLog.getQueueSize());
        verify(logger, times(old + 2)).log(any(AuditLogItem.class));
        verifyNoMoreInteractions(logger);
    }

    @Test
    public void testSearch() throws Exception {
        AuditLogSearch searcher = mock(AuditLogSearch.class);
        AuditLogSearch otherSearcher = mock(AuditLogSearch.class);

        AuditLog.addSearcher(searcher);
        AuditLog.addSearcher(otherSearcher);

        AuditLogQuery q = new AuditLogQuery();
        AuditLog.search(q);

        verify(searcher).search(q);
        verifyNoMoreInteractions(searcher);
        verifyZeroInteractions(otherSearcher);
    }

    private IDCardClusterCacheEntry generateEntry() {
        IDCardClusterCacheEntry entry = new IDCardClusterCacheEntry("test");
        final SystemInfo sysInfo = new SystemInfo(new CareProvider("medcom:skscode", "test", "test"), "test");
        final UserInfo userInfo = new UserInfo("1111111118", "fissirul", "lohman", null, null, "role", null);
        entry.setIdCard(new UserIDCard("1.0.1", AuthenticationLevel.MOCES_TRUSTED_USER, "test", sysInfo, userInfo, "e", "2", null, null));
        return entry;
    }

}
