/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.browser;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.junit.Before;
import org.junit.Test;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.IDCardState;
import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.cluster.IDCardClusterCacheImpl;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.RuntimeConfigItem;
import com.trifork.sosigw.config.RuntimeConfigState;
import com.trifork.sosigw.config.StaticConfigImpl;


public class BrowserBasedSigningTest {
    
    private IDCardClusterCacheEntry entry;
    private Map<String, String[]> params;
    private IDCardClusterCacheImpl cache;

    @Before
    public void setup() {
        StaticConfigImpl.setInstance(new StaticConfigImpl(new Properties()));
        entry = new IDCardClusterCacheEntry("test");
        params = new HashMap<String, String[]>();
        
        cache = new IDCardClusterCacheImpl(null);
        Factory.setTheIDCardCache(cache);
    }
    
    @Test
    public void prepareBrowserSigningShouldSetBytes() {
        assertNull(entry.getBytesForSigning());
        
        BrowserBasedSigning.prepareBrowserSigning(entry, new byte[0]);
        assertNotNull(entry.getBytesForSigning());
    }
    
    @Test
    public void testMakeSigningUrl() throws Exception {
        RuntimeConfig.setRuntimeConfigState(new RuntimeConfigState(1, new RuntimeConfigItem[] { new RuntimeConfigItem(RuntimeConfig.CATEGORY_PROPERTIES, RuntimeConfig.KEY_BROWSERSIGNINGURL, new String[] { "http://localhost" })}));
        
        String url = BrowserBasedSigning.makeSigningURL(entry);
        assertTrue(url.startsWith("http://localhost"));
        assertTrue(url.contains("ID=test"));
        assertTrue(url.contains("CK="));
        
        String secret = url.substring(url.indexOf("CK=") + 3);
        assertTrue(entry.validateSecret(secret));
    }
    
    @Test
    public void findAndCheckShouldReturnNullOnMissingParameters() {
       assertNull(BrowserBasedSigning.findAndCheckIDCardCacheEntry(params));
       
       params.put("ID", new String[] { "test" });
       assertNull(BrowserBasedSigning.findAndCheckIDCardCacheEntry(params));
       
       params.clear();
       params.put("CK", new String[] { "value"});
       assertNull(BrowserBasedSigning.findAndCheckIDCardCacheEntry(params));
    }

    @Test
    public void findAndCheckShouldReturnNullOnMissingEntry()  {
        params.put("ID", new String[] {"test"});
        params.put("CK", new String[] {"value"});
        
        assertNull(BrowserBasedSigning.findAndCheckIDCardCacheEntry(params));
    }
    
    @Test
    public void findAndCheckShouldReturnEntryIfSigned() {
        params.put("ID", new String[] {"test"});
        params.put("CK", new String[] {"value"});

        IDCardClusterCacheEntry entry = cache.ensureIDCard("test");
        cache.updateEntryState(entry, IDCardState.SIGNED);
        
        assertNotNull(BrowserBasedSigning.findAndCheckIDCardCacheEntry(params));
    }
    
    @Test
    public void findAndCheckShouldReturnEntryIfSecretMatches() {
        params.put("ID", new String[] {"test"});
        params.put("CK", new String[] {"value"});

        cache.ensureIDCard("test").setSecret("value");
        assertNotNull(BrowserBasedSigning.findAndCheckIDCardCacheEntry(params));
    }
    
    @Test
    public void findAndCheckShouldReturnNullIfTimedOut() {
        params.put("ID", new String[] {"test"});
        params.put("CK", new String[] {"value"});

        IDCardClusterCacheEntry entry = cache.ensureIDCard("test");
        entry.setSecret("value");
        cache.updateEntryState(entry, IDCardState.AWAITING_BROWSER_RESPONSE);
        entry.setLastModified(0);
        assertNull(BrowserBasedSigning.findAndCheckIDCardCacheEntry(params));
    }
}
