/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.browser;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.*;

import java.io.*;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import javax.servlet.RequestDispatcher;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.sun.xml.wss.saml.internal.saml11.jaxb10.DigestValue;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.commons.io.IOUtils;
import org.junit.Before;
import org.junit.Test;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.IDCardState;
import com.trifork.sosigw.api.IdCardSigner;
import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.cluster.IDCardClusterCacheImpl;
import com.trifork.sosigw.config.StaticConfigImpl;

import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.SystemInfo;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.UserInfo;

public class BrowserSigningServletTest {
    
    private HttpServletRequest req;
    private HttpServletResponse res;
    private BrowserSigningServlet bss;
    private Map<String, String[]> params;
    private StringWriter resp;
    private IDCardClusterCacheImpl cache;

    @Before
    public void setup() throws IOException {
        req = mock(HttpServletRequest.class);
        res = mock(HttpServletResponse.class);
        params = new HashMap<String, String[]>();
        when(req.getParameterMap()).thenReturn(params);
        resp = new StringWriter();
        when(res.getWriter()).thenReturn(new PrintWriter(resp));
        
        bss = new BrowserSigningServlet();
        
        cache = new IDCardClusterCacheImpl(null);
        Factory.setTheIDCardCache(cache);
        
        StaticConfigImpl.setInstance(new StaticConfigImpl(new Properties()));
    }
    

    @Test
    public void noEntryShouldFail() throws Exception {
        bss.doGet(req, res);
        
        assertFailed();
    }
    
    @Test
    public void awaitingSignedBytesShouldShowApplet() throws Exception {
        RequestDispatcher dispatcher = mock(RequestDispatcher.class);
        when(req.getRequestDispatcher(anyString())).thenReturn(dispatcher);
        IDCardClusterCacheEntry entry = prepareEntry();
        cache.updateEntryState(entry, IDCardState.AWAITING_SIGNED_BYTES);
        
        bss.doPost(req, res);
        verify(req).setAttribute(eq("digest.to.sign"), anyString());
        verify(req).setAttribute(eq("text.for.user"), anyString());
        
        assertEquals(IDCardState.AWAITING_BROWSER_RESPONSE, entry.getState());
        verify(dispatcher).forward(req, res);
    }
    
    @Test
    public void timedOutBrowserResponseShouldFail() throws Exception {
        IDCardClusterCacheEntry entry = prepareEntry();
        cache.updateEntryState(entry, IDCardState.AWAITING_BROWSER_RESPONSE);
        entry.setLastModified(0);
        bss.doPost(req, res);
        
        assertFailed();
    }
    
    @Test
    public void awaitingBrowserResponseWithoutMessageShouldFail() throws Exception {
        IDCardClusterCacheEntry entry = prepareEntry();
        cache.updateEntryState(entry, IDCardState.AWAITING_BROWSER_RESPONSE);
        entry.setLastModified(System.currentTimeMillis());
        bss.doPost(req, res);
        
        assertFailed();
    }
    
    @Test
    public void awaitingBrowserResponseShouldSign() throws Exception {
        byte[] msg = getSignMessageFromResource("message-from-moces1-sign.xml");
        params.put("message", new String[] { XmlUtil.toBase64(msg)});
        
        IDCardClusterCacheEntry entry = prepareEntry();
        cache.updateEntryState(entry, IDCardState.AWAITING_BROWSER_RESPONSE);
        entry.setLastModified(System.currentTimeMillis());
        
        IdCardSigner signer = mock(IdCardSigner.class);
        bss.setIdCardSigner(signer);
        
        RequestDispatcher dispatcher = mock(RequestDispatcher.class);
        when(req.getRequestDispatcher(anyString())).thenReturn(dispatcher);

        bss.doPost(req, res);
        
        verify(signer).signIdCard(eq("test"), any(byte[].class), any(byte[].class));
        verify(dispatcher).forward(req, res);
    }
    
    @Test
    public void signedIsDone() throws Exception {
        IDCardClusterCacheEntry entry = prepareEntry();
        cache.updateEntryState(entry, IDCardState.SIGNED);
        entry.setLastModified(System.currentTimeMillis());
        
        RequestDispatcher dispatcher = mock(RequestDispatcher.class);
        when(req.getRequestDispatcher(anyString())).thenReturn(dispatcher);

        bss.doPost(req, res);
        
        verify(dispatcher).forward(req, res);
    }

    @Test
    public void awaitingSTSResponseShouldFail() throws Exception {
        IDCardClusterCacheEntry entry = prepareEntry();
        cache.updateEntryState(entry, IDCardState.AWAITING_STS_RESPONSE);
        entry.setLastModified(System.currentTimeMillis());
        
        bss.doPost(req, res);
        assertFailed();
    }
    
    private IDCardClusterCacheEntry prepareEntry() {
        params.put("ID", new String[] {"test"});
        params.put("CK", new String[] {"secret"});
        IDCardClusterCacheEntry entry = cache.ensureIDCard("test");
        entry.setSecret("secret");
        entry.setLastModified(System.currentTimeMillis());
        entry.setBytesForSigning(new byte[] { 'a' });
        final SystemInfo sysInfo = new SystemInfo(new CareProvider("medcom:skscode", "test", "test"), "test");
        final UserInfo userInfo = new UserInfo("1111111118", "fissirul", "lohman", null, null, "role", null);
        entry.setIdCard(new UserIDCard("1.0.1", AuthenticationLevel.MOCES_TRUSTED_USER, "test", sysInfo, userInfo, "e", "2", null, null));
        
        return entry;
    }
    
    private void assertFailed() {
        verify(res).setStatus(500);
        assertTrue(resp.getBuffer().length() > BrowserSigningServlet.extraIEPayload);
    }
    
    public byte[] getSignMessageFromResource(String name) {
        final InputStream stream = this.getClass().getResourceAsStream(name);
        final ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
        try {
            IOUtils.copy(stream, byteStream);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        return byteStream.toByteArray();
    }
}
