/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster;

import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertFalse;

import java.util.HashSet;
import java.util.Random;
import java.util.Set;

import org.junit.After;
import org.junit.Test;

import com.trifork.sosigw.api.IDCardClusterCache;
import com.trifork.sosigw.api.IDCardState;
import com.trifork.sosigw.cluster.intercom.TimeSource;

public class IDCardClusterCacheHashingTest {

    private static final Random RANDOM = new Random();

    private static final int CACHE_SIZE = 1000;

    private static class Entry {

        private final String cacheKey;

        private final IDCardState state;

        private final int hashcode;

        public Entry(String cacheKey, IDCardState state) {
            this.cacheKey = cacheKey;
            this.state = state;
            this.hashcode = RANDOM.nextInt(); // Ensure random order in a Set
        }

        @Override
        public int hashCode() {
            return hashcode;
        }
    }

    @Test
    public void testHashingSameAfterStateUpdateAndRemoval() {
        Set<Entry> cacheEntries1 = new HashSet<Entry>(CACHE_SIZE);
        Set<Entry> cacheEntries2 = new HashSet<Entry>(CACHE_SIZE);
        Set<Entry> cacheStateUpdatedEntries1 = new HashSet<Entry>(CACHE_SIZE);
        Set<Entry> cacheStateUpdatedEntries2 = new HashSet<Entry>(CACHE_SIZE);
        Set<Entry> cacheObsoleteEntries1 = new HashSet<Entry>(CACHE_SIZE);
        Set<Entry> cacheObsoleteEntries2 = new HashSet<Entry>(CACHE_SIZE);
        for (int i = 0; i < CACHE_SIZE; i++) {
            int cacheKeyLength = RANDOM.nextInt(5) + 8;
            char[] cacheKeyContent = new char[cacheKeyLength];
            for (int j = 0; j < cacheKeyLength; j++) {
                cacheKeyContent[j] = (char)('0' + RANDOM.nextInt(10));
            }
            String cacheKey = new String(cacheKeyContent);
            IDCardState state;
            switch (RANDOM.nextInt(3)) {
                case 0:
                    state = IDCardState.AWAITING_BROWSER_RESPONSE;
                    break;
                case 1:
                    state = IDCardState.AWAITING_SIGNED_BYTES;
                    break;
                case 2:
                    state = IDCardState.AWAITING_STS_RESPONSE;
                    break;
                default:
                    state = IDCardState.REVOKED;
            }
            cacheEntries1.add(new Entry(cacheKey, state));
            cacheEntries2.add(new Entry(cacheKey, state));
            if (RANDOM.nextInt(100) < 90) {
                cacheStateUpdatedEntries1.add(new Entry(cacheKey, IDCardState.SIGNED));
                cacheStateUpdatedEntries2.add(new Entry(cacheKey, IDCardState.SIGNED));
            }
            if (RANDOM.nextInt(100) < 10) {
                cacheObsoleteEntries1.add(new Entry(cacheKey, null));
                cacheObsoleteEntries2.add(new Entry(cacheKey, null));
            }
        }
        IDCardClusterCache cache1 = new IDCardClusterCacheImpl(null);
        IDCardClusterCache cache2 = new IDCardClusterCacheImpl(null);
        
        // Add the entries to both caches in different sequences
        for (Entry entry : cacheEntries1) {
            IDCardClusterCacheEntry cacheEntry = cache1.ensureIDCard(entry.cacheKey);
            cache1.updateEntryState(cacheEntry, entry.state);
        }
        for (Entry entry : cacheEntries2) {
            IDCardClusterCacheEntry cacheEntry = cache2.ensureIDCard(entry.cacheKey);
            cache2.updateEntryState(cacheEntry, entry.state);
        }
        
        // Update the state of the same entries in both caches in different sequences
        for (Entry entry : cacheStateUpdatedEntries1) {
            cache1.updateEntryState(cache1.findIDCardOrNull(entry.cacheKey), entry.state);
        }
        for (Entry entry : cacheStateUpdatedEntries2) {
            cache2.updateEntryState(cache2.findIDCardOrNull(entry.cacheKey), entry.state);
        }
        
        // Mark the same entries obsolete in both caches in different sequences.
        TimeSource.setTestTime(System.currentTimeMillis() - (25 * 60 * 60 * 1000));
        for (Entry entry : cacheObsoleteEntries1) {
            cache1.entryUpdatedLocally(cache1.findIDCardOrNull(entry.cacheKey));
        }
        for (Entry entry : cacheObsoleteEntries2) {
            cache2.entryUpdatedLocally(cache2.findIDCardOrNull(entry.cacheKey));
        }
        
        // Trigger a removal of ALL obsolete entries
        TimeSource.setTestTime(0);
        cache1.getContentsAsList();
        cache2.getContentsAsList();
        
        // Assert cache size and hash
        assertEquals("Cache does not contain the correct number of entries", CACHE_SIZE - cacheObsoleteEntries1.size(), cache1.size());
        assertEquals("Cache does not contain the same number of entries", cache1.size(), cache2.size());
        assertEquals("The fingerprint hash of the caches content are not the same", cache1.contentHash(), cache2.contentHash());
    }
    
    @Test
    public void testEntriesFromPeers() {
        IDCardClusterCache cache1 = new IDCardClusterCacheImpl(null);
        IDCardClusterCache cache2 = new IDCardClusterCacheImpl(null);
        assertEquals(cache1.contentHash(), cache2.contentHash());
        
        // Make a local cache and a cache with and initial entry form a peer
        cache1.ensureIDCard("abc");
        assertFalse("Caches should not have same hash", cache1.contentHash() == cache2.contentHash());
        long peerTime = System.currentTimeMillis() + 10000; // Ensure peer has never entries
        IDCardClusterCacheEntry peerEntry = new IDCardClusterCacheEntry("abc");
        peerEntry.setLastModified(peerTime + 1);
        cache2.arrivedFromPeer(peerEntry, 0);
        assertEquals("Caches should have the same hash", cache1.contentHash(), cache2.contentHash());
        
        // locally update one cache and let the other one get an updated version of the entry from a peer
        cache1.updateEntryState(cache1.findIDCardOrNull("abc"), IDCardState.SIGNED);
        assertFalse("Caches should not have same hash", cache1.contentHash() == cache2.contentHash());
        peerEntry = new IDCardClusterCacheEntry("abc");
        peerEntry.setLastModified(peerTime + 2);
        peerEntry.setState(IDCardState.SIGNED);
        cache2.arrivedFromPeer(peerEntry, 0);
        assertEquals("Caches should have the same hash", cache1.contentHash(), cache2.contentHash());
        
        // ensure a "lesser" updated version is ignored
        peerEntry = new IDCardClusterCacheEntry("abc");
        peerEntry.setLastModified(peerTime);
        cache2.arrivedFromPeer(peerEntry, 0);
        assertEquals("Caches should have the same hash", cache1.contentHash(), cache2.contentHash());
    }
    
    @After
    public void cleanup() {
        TimeSource.setTestTime(0);
    }
}

