/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;

import org.junit.Before;
import org.junit.Test;

import com.trifork.sosigw.api.IDCardCacheNotification;
import com.trifork.sosigw.api.IDCardState;

public class IDCardClusterCacheImplTest {
    
    private IDCardClusterCacheImpl cache;

    @Before
    public void setup() {
        cache = new IDCardClusterCacheImpl(null);
    }
    
    @Test
    public void ensureIdCardShouldAddIfNew() {
        assertNull(cache.findIDCardOrNull("test"));
        
        IDCardClusterCacheEntry e = cache.ensureIDCard("test");
        assertNotNull(cache.findIDCardOrNull("test"));
        
        assertSame(e, cache.ensureIDCard("test"));
    }

    @Test
    public void newArrivedFromPeerShouldBeAdded() {
        assertNull(cache.findIDCardOrNull("test"));
        
        IDCardClusterCacheEntry entry = new IDCardClusterCacheEntry("test");
        cache.arrivedFromPeer(entry, 0);
        
        assertSame(entry, cache.findIDCardOrNull("test"));
    }
    
    @Test
    public void newerArrivedFromPeerShouldReplace() {
        IDCardClusterCacheEntry local = cache.ensureIDCard("test");
        local.setLastModified(0);
        
        IDCardClusterCacheEntry entry = new IDCardClusterCacheEntry("test");
        entry.setLastModified(System.currentTimeMillis());
        
        cache.arrivedFromPeer(entry, 0);
        assertSame(entry, cache.findIDCardOrNull("test"));
    }
    
    @Test
    public void olderArrivedFromPeerShouldBeDiscarded() {
        IDCardClusterCacheEntry local = cache.ensureIDCard("test");
        local.setLastModified(System.currentTimeMillis());
        
        IDCardClusterCacheEntry entry = new IDCardClusterCacheEntry("test");
        entry.setLastModified(0);
        
        cache.arrivedFromPeer(entry, 0);
        assertSame(local, cache.findIDCardOrNull("test"));
    }
    
    @Test
    public void olderButHigherFromPeerShouldReplace() {
        IDCardClusterCacheEntry local = cache.ensureIDCard("test");
        local.setLastModified(System.currentTimeMillis());
        local.setState(IDCardState.NEW);
        
        IDCardClusterCacheEntry entry = new IDCardClusterCacheEntry("test");
        entry.setLastModified(System.currentTimeMillis() - 3000L);
        entry.setState(IDCardState.SIGNED);
        
        cache.arrivedFromPeer(entry, 0);
        assertSame(entry, cache.findIDCardOrNull("test"));
    }
    
    @Test
    public void getContentsAsListShouldObsoleteEntries() {
        assertTrue(cache.getContentsAsList().isEmpty());
        
        IDCardClusterCacheEntry entry = cache.ensureIDCard("test");
        entry.setLastModified(System.currentTimeMillis());
        
        assertFalse(cache.getContentsAsList().isEmpty());
        
        entry.setLastModified(0);
        assertTrue(cache.getContentsAsList().isEmpty());
    }
    
    @Test
    public void listenerShouldBeNotifiedOnUpdate() {
        IDCardCacheNotification listener = mock(IDCardCacheNotification.class);
        cache.setNotificationListener(listener);
        
        IDCardClusterCacheEntry entry = cache.ensureIDCard("test");
        
        cache.entryUpdatedLocally(entry);
        verify(listener).entryUpdatedLocally(entry);
    }
}
