/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.util.Map;
import java.util.Properties;

import org.junit.Assume;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.IDCardCache;
import com.trifork.sosigw.cluster.IDCardClusterCacheImpl;
import com.trifork.sosigw.config.StaticConfigImpl;

import static org.mockito.Mockito.*;

public class IntercomImplTest {
    
    @Before
    public void setup() {
        StaticConfigImpl config = new StaticConfigImpl(new Properties());
        StaticConfigImpl.setInstance(config);
        
        Factory.setTheIDCardCache(new IDCardClusterCacheImpl(null));
    }

    @Test
    @Ignore("IntercomImpl vil fejle, hvis ikke MulticastHandler kan forbinde til netværksinterface.")
    public void testNoNetwork() throws Exception {
        Assume.assumeTrue(!MulticastHandler.useBindToInterfaceByProtocol());
        
        IntercomImpl intercom = new IntercomImpl();
        assertTrue(intercom.getUID().startsWith("NoNet"));
    }
    
    @Test(expected=IllegalStateException.class)
    public void idCardCacheMustBeClustering() throws Exception {
        Factory.setTheIDCardCache(mock(IDCardCache.class));
        
        new IntercomImpl();
    }
    
    @Test
    //Hvad er forskellen på testNoClusteringEnabled og testClusteringEnabled?
    public void testNoClusteringEnabled() throws Exception {
        Assume.assumeTrue(!MulticastHandler.useBindToInterfaceByProtocol());
        
        int oldCount = findThreads("Intercom.Receiver");
        int oldPeriodic = findThreads("Intercom.PeriodicAlive");
        int oldSync = findThreads("Intercom.SyncRequester");
        int oldSyncRecv = findThreads("Intercom.SyncReceiver");
        
        Properties props = new Properties();
        props.setProperty("cluster.tuningparam.receiver.thread.count", "2");
        StaticConfigImpl config = new StaticConfigImpl(props);
        StaticConfigImpl.setInstance(config);
        IntercomImpl intercom = new IntercomImpl();
        
        assertEquals("Intercom.Receiver not valid", oldCount + 2, findThreads("Intercom.Receiver"));
        assertEquals("Intercom.PeriodicAlive not valid", oldPeriodic + 1, findThreads("Intercom.PeriodicAlive"));
        assertEquals("Intercom.SyncRequester not valid", oldSync + 1, findThreads("Intercom.SyncRequester"));
        assertEquals("Intercom.SyncReceiver not valid", oldSyncRecv + 2, findThreads("Intercom.SyncReceiver"));
        
        intercom.kill();
        Thread.sleep(500);
        assertEquals("Intercom.Receiver not valid", oldCount, findThreads("Intercom.Receiver"));
    }
    
    @Test
    public void testClusteringEnabled() throws Exception {
        int oldCount = findThreads("Intercom.Receiver");
        int oldPeriodic = findThreads("Intercom.PeriodicAlive");
        int oldSync = findThreads("Intercom.SyncRequester");
        int oldSyncRecv = findThreads("Intercom.SyncReceiver");
        
        Properties props = new Properties();
        props.setProperty("cluster.tuningparam.receiver.thread.count", "2");
        props.setProperty("cluster.ip.prefix", "");
        StaticConfigImpl config = new StaticConfigImpl(props);
        StaticConfigImpl.setInstance(config);
        IntercomImpl intercom = new IntercomImpl();
        
        assertEquals("Intercom.Receiver not valid", oldCount + 2, findThreads("Intercom.Receiver"));
        assertEquals("Intercom.PeriodicAlive not valid", oldPeriodic + 1, findThreads("Intercom.PeriodicAlive"));
        assertEquals("Intercom.SyncRequester not valid", oldSync + 1, findThreads("Intercom.SyncRequester"));
        assertEquals("Intercom.SyncReceiver not valid", oldSyncRecv + 2, findThreads("Intercom.SyncReceiver"));

        
        intercom.kill();
        
        Thread.sleep(500);
        
        assertEquals("Intercom.Receiver not valid", oldCount, findThreads("Intercom.Receiver"));
        assertEquals("Intercom.PeriodicAlive not valid", oldPeriodic, findThreads("Intercom.PeriodicAlive"));
        assertEquals("Intercom.SyncRequester not valid", oldSync, findThreads("Intercom.SyncRequester"));
        assertEquals("Intercom.SyncReceiver not valid", oldSyncRecv, findThreads("Intercom.SyncReceiver"));

    }
    
    private int findThreads(String name) {
        Map<Thread, StackTraceElement[]> threads = Thread.getAllStackTraces();
        int found = 0;
        for (Thread t : threads.keySet()) {
            if (name.equals(t.getName())) {
                found++;
            }
        }
        return found;
    }
}
