/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.io.InputStream;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.Enumeration;

import org.apache.log4j.Logger;
import org.junit.Assume;
import org.junit.Test;

import com.trifork.sosigw.config.StaticConfigTestImpl;

public class MulticastTest {
    private static final Logger log = Logger.getLogger(MulticastTest.class);
    
    private static final String HELLO_WORLD = "Hello World";
    private static final byte[] bytes = HELLO_WORLD.getBytes();
    
    @Test
    public void multicastShouldWork() throws Exception {
        int port = 3456;
        MulticastSocket mcs = new MulticastSocket(port);
        
        InetAddress group = InetAddress.getByName("228.0.0.7");
        
        if (MulticastHandler.useBindToInterfaceByProtocol()) {
            log.debug("Binding "+group+" to "+mcs.getLocalPort()+" using protocol");
            MulticastHandler.bindToInterfaceByProtocol(group, mcs);
        }
        
        mcs.joinGroup(group);
        for (int i = 0; i < 10; i++) {
            log.debug("Sending data");
            mcs.send(new DatagramPacket(bytes, bytes.length, group, port));
            Thread.sleep(1);
        }

        byte[] buf = new byte[1000];
        DatagramPacket recv = new DatagramPacket(buf, buf.length);
        for (int i = 0; i < 10; i++) {
            log.debug("Receinving data");
            mcs.receive(recv);
            assertEquals(HELLO_WORLD, new String(buf, 0, recv.getLength(), "UTF-8"));
        }
    }
    
    @Test
    public void testSendReceive() throws IOException, InterruptedException {
        NetHandler multicastHandler = new MulticastHandler(new StaticConfigTestImpl());
        byte[] buf = new byte[1000];

        multicastHandler.sendMessage(bytes);
        InputStream recvMessage = multicastHandler.recvMessage(buf);
        
        assertEquals(HELLO_WORLD, new String(buf, 0, recvMessage.available(), "UTF-8"));
    }
    
    @Test(expected=SocketException.class)
    public void testClosed() throws Exception {
        MulticastHandler multicastHandler = new MulticastHandler(new StaticConfigTestImpl());
        multicastHandler.close();
        multicastHandler.sendMessage(bytes);
        
        byte[] buf = new byte[1000];
        multicastHandler.recvMessage(buf);
    }
    
    @Test
    public void testBindToPrefix() throws Exception {
        StaticConfigTestImpl config = new StaticConfigTestImpl();
        config.setCluserIpPrefix("127.");
        new MulticastHandler(config);
    }
    
    @Test
    public void testBindToUnknownPrefixShouldFail() throws Exception {
        Assume.assumeTrue(!MulticastHandler.useBindToInterfaceByProtocol());
        
        try {
            StaticConfigTestImpl config = new StaticConfigTestImpl();
            config.setCluserIpPrefix("453.");
            
            new MulticastHandler(config);
        } catch(IllegalStateException e) {
            //As expected
            return;
        }
        
        fail("Expected an IllegalStateException to be thrown.");
    }
    
    @Test
    public void testInterPacketDelay() throws Exception {
        StaticConfigTestImpl config = new StaticConfigTestImpl();
        NetHandler handler = new MulticastHandler(config);
        
        handler.ensureInterPacketDelay();
        handler.ensureInterPacketDelay();
    }
}
