/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.mockito.Matchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;

import java.util.Properties;

import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;

import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.StaticConfigImpl;

public class PeriodicAliveTest {
    
    private PeriodicAlive periodicAlive;
    private NetHandler handler;
    private ClusterState state;
    private SlidingWindow<String> window;

    @Before
    public void setup() {
        handler = mock(NetHandler.class);
        state = new ClusterState();
        window = new SlidingWindow<String>(3);
        StaticConfigImpl config = new StaticConfigImpl(new Properties());
        StaticConfigImpl.setInstance(config);
        periodicAlive = new PeriodicAlive("uid", handler, window, state, config);
    }
    
    @Test
    public void testNormalRun() throws Exception {
        periodicAlive.run0();
        
        assertTrue(state.getLastSenderPing().isEmpty());
        
        verify(handler).ensureInterPacketDelay();
        verify(handler).sendMessage(any(byte[].class));
    }
    
    @Test
    public void testSendPeriodicAlive() throws Exception {
        periodicAlive.sendPeriodicAlive();
        
        verify(handler).ensureInterPacketDelay();
        ArgumentCaptor<byte[]> res = ArgumentCaptor.forClass(byte[].class);
        verify(handler).sendMessage(res.capture());
        
        ProtocolVerifier pv = new ProtocolVerifier(res.getValue());
        pv.protocolVersion1()
            .then("the operation", Protocol.PROTOCOL_V1_ALIVE)
            .then("my uid", "uid")
            .then("the seqno", window.getTop())
            .then("the timestamp", RuntimeConfig.getConfigTimestamp())
            .thenALong("the current timestamp")
            .done();
        
        verifyNoMoreInteractions(handler);
    }

    @Test
    public void testWarnPeriodShouldCancelRequests() throws Exception {
        state.getSyncRequestedAt().put("test", System.currentTimeMillis());
        state.getLastSenderPing().put("test", System.currentTimeMillis() - 6000L);
        
        periodicAlive.run0();
        
        assertFalse(state.getSyncRequestedAt().containsKey("test"));
    }
    
    @Test
    public void testServerVanshed() throws Exception {
        state.getLastSenderSeqno().put("test", 1L);
        state.getSyncRequestedAt().put("test", 1L);
        state.getTimeStates().put("test", new TimeState());
        state.getLastSenderPing().put("test", System.currentTimeMillis() - 600000L);
        
        periodicAlive.run0();
        
        assertFalse(state.getLastSenderSeqno().containsKey("test"));
        assertFalse(state.getLastSenderPing().containsKey("test"));
        assertFalse(state.getSyncRequestedAt().containsKey("test"));
        assertFalse(state.getTimeStates().containsKey("test"));
    }
}
