/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import static org.junit.Assert.assertEquals;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.net.InetSocketAddress;
import java.util.HashMap;

import org.junit.Test;

import com.trifork.sosigw.api.IDCardCacheEntry;
import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.config.RuntimeConfigItem;
import com.trifork.sosigw.config.RuntimeConfigState;

public class ProtocolTest {

    @Test
    public void testEncodeVersion1ServerStatesAsBytes() throws Exception {
        SlidingWindow<String> window = new SlidingWindow<String>(1);
        byte[] bytes = Protocol.encodeVersion1ServerStatesAsBytes("uid", window, new HashMap<String, Long>() {{ put("test", 2L); }});
        DataInputStream dis = new DataInputStream(new ByteArrayInputStream(bytes));
        
        assertEquals("First the protocol version", Protocol.PROTOCOL_VERSION_1, dis.readByte());
        assertEquals("Then the number of nodes in the message (me + the ones in lastSenderSeqno)", 2, dis.readInt());
        assertEquals("Then my own uid", "uid", dis.readUTF());
        assertEquals("Then my own seqno", window.getTop(), dis.readLong());
        assertEquals("Then the other uids", "test", dis.readUTF());
        assertEquals("Lastly the seqno for the other uid", 2, dis.readLong());
        
        assertEquals("And the no more", 0, dis.available());
    }

    @Test
    public void testEncodeVersion1RuntimeConfigStateAsBytes() throws Exception {
        RuntimeConfigState state = new RuntimeConfigState(System.currentTimeMillis(), new RuntimeConfigItem[0]);
        byte[] bytes = Protocol.encodeVersion1RuntimeConfigStateAsBytes(state, "uid");
        
        ProtocolVerifier pv = new ProtocolVerifier(bytes);
        pv.protocolVersion1()
            .then("The operation", Protocol.PROTOCOL_V1_CONFIGSTATE)
            .then("my uid", "uid")
            .then("the state", RuntimeConfigState.class)
            .done();
    }

    @Test
    public void testEncodeVersion1ConfigRequestAsBytes() throws Exception {
        byte[] bytes = Protocol.encodeVersion1ConfigRequestAsBytes(2L, "uid");
        
        ProtocolVerifier pv = new ProtocolVerifier(bytes);
        pv.protocolVersion1()
            .then("the operation", Protocol.PROTOCOL_V1_CONFIGREQUEST)
            .then("my uid", "uid")
            .then("the last timestamp", 2L)
            .done();
    }

    @Test
    public void testEncodeVersion1AliveAsBytes() throws Exception {
        byte[] bytes = Protocol.encodeVersion1AliveAsBytes(2L, 3L, "uid");
        ProtocolVerifier pv = new ProtocolVerifier(bytes);
        pv.protocolVersion1()
            .then("the operation", Protocol.PROTOCOL_V1_ALIVE)
            .then("my uid", "uid")
            .then("my seq no", 2L)
            .then("last timestamp", 3L)
            .thenALong("the current time")
            .done();
    }

    @Test
    public void testEncodeVersion1SyncRequestAsBytes() throws Exception {
        byte[] bytes = Protocol.encodeVersion1SyncRequestAsBytes(new SyncRequest("uid", 3L), new InetSocketAddress(42));
        ProtocolVerifier pv = new ProtocolVerifier(bytes);
        pv.protocolVersion1()
            .then("the operation", Protocol.PROTOCOL_V1_SYNCREQUEST)
            .then("then the request", SyncRequest.class)
            .then("then the socket address", InetSocketAddress.class)
            .done();
    }

    @Test
    public void testEncodeVersion1IDCardEntryAsBytes() throws Exception {
        byte[] bytes = Protocol.encodeVersion1IDCardEntryAsBytes(new IDCardClusterCacheEntry("test"), 2L, "uid");
        ProtocolVerifier pv = new ProtocolVerifier(bytes);
        pv.protocolVersion1()
            .then("the operation", Protocol.PROTOCOL_V1_IDCARDENTRY)
            .then("my uid", "uid")
            .then("the seqno", 2L)
            .then("the entry", IDCardCacheEntry.class)
            .done();
    }

}
