/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.cluster.intercom;

import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.cluster.IDCardClusterCacheImpl;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import java.io.DataInputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;


public class SyncSenderTest {
    
    private long seq;
    private SlidingWindow<String> sl;
    private ClusterState state;
    private ServerSocket server;
    private Thread t;
    private IDCardClusterCacheImpl clusterCache;
    private int waitTime = 0;

    @Before
    public void setup() throws Exception {
        sl = new SlidingWindow<String>(2);
        seq = sl.getTop();
        sl.add(sl.reserve(), "test");
        
        server = new ServerSocket();
        InetSocketAddress addr = new InetSocketAddress(InetAddress.getLocalHost(), 3443);
        server.bind(addr);
        
        state = new ClusterState();
        clusterCache = new IDCardClusterCacheImpl(null) {

            public synchronized IDCardClusterCacheEntry ensureIDCard(String cacheKey) {
                try {
                    Thread.sleep(waitTime);
                } catch (InterruptedException e) {}
                return super.ensureIDCard(cacheKey);
            }
        };
        SyncSender sender = new SyncSender("uid", sl, state, clusterCache, new SyncRequest("sender", seq), (InetSocketAddress)server.getLocalSocketAddress(), 2000, 2000);
        t = new Thread(sender);
        
    }
    
    @After
    public void shutdown() throws Exception {
        if (server != null) {
            server.close();
        }
    }

    @Test
    public void testSendPartial() throws Exception {
        t.start();
        
        Socket socket = server.accept();
        DataInputStream dis = new DataInputStream(socket.getInputStream());
        assertEquals("First the protocol version", Protocol.PROTOCOL_VERSION_1, dis.readByte());
        assertEquals("Then the number of entries", 1, dis.readInt());
        assertEquals("Then my uid", "uid", dis.readUTF());
        assertEquals("Then my seqno", sl.getTop(), dis.readLong());
        assertEquals("Then the number of ids to request", 1, dis.readLong());
        int length = dis.readInt();
        byte[] entry = new byte[length];
        dis.read(entry, 0, length);
        
        assertEquals("Then indicate if any more is coming", false, dis.readBoolean());
        assertEquals("And then no more", 0, dis.available());
                                
        dis.close();
    }
    
    @Test
    public void testSendFull() throws Exception {
        sl.add(sl.reserve(), "test2");
        sl.add(sl.reserve(), "test3");
        
        clusterCache.ensureIDCard("test");
        clusterCache.ensureIDCard("test2");
        clusterCache.ensureIDCard("test3");
        
        t.start();
        
        Socket socket = server.accept();
        DataInputStream dis = new DataInputStream(socket.getInputStream());
        assertEquals("First the protocol version", Protocol.PROTOCOL_VERSION_1, dis.readByte());
        assertEquals("Then the number of entries", 1, dis.readInt());
        assertEquals("Then my uid", "uid", dis.readUTF());
        assertEquals("Then my seqno", sl.getTop(), dis.readLong());
        assertEquals("Then the number of ids to request", 3, dis.readLong());
        for (int i = 0; i < 3; i++) {
            int length = dis.readInt();
            byte[] entry = new byte[length];
            dis.read(entry, 0, length);
        }
        
        assertEquals("Then indicate if any more is coming", false, dis.readBoolean());
        assertEquals("And then no more", 0, dis.available());
                                
        dis.close();
    }
    
    @Test
    public void testIncompleteResponse() throws Exception {
        clusterCache.ensureIDCard("test");
        clusterCache.ensureIDCard("test2");
        
        sl.add(sl.reserve(), "tester");
        sl.add(sl.reserve(), "tester");

        waitTime = 100;
        t.start();
        for (int i = 0; i < 10000; i++) {
            sl.add(sl.reserve(), "tester");
        }
        
        Socket socket = server.accept();
        DataInputStream dis = new DataInputStream(socket.getInputStream());
        assertEquals("First the protocol version", Protocol.PROTOCOL_VERSION_1, dis.readByte());
        assertEquals("Then the number of entries", 1, dis.readInt());
        assertEquals("Then my uid", "uid", dis.readUTF());
        dis.readLong(); // ignore the seqno
        assertEquals("Then the number of ids to request", 2, dis.readLong());
        for (int i = 0; i < 2; i++) {
            int length = dis.readInt();
            dis.read(new byte[length], 0, length);
        }
        
        assertEquals("Then indicate if any more is coming", true, dis.readBoolean());
        
        assertFalse("More data should be available", 0 == dis.available());
        
        dis.close();

    }
}
