/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.config;

import java.util.Map;
import java.util.Properties;

import org.junit.Assert;
import org.junit.Test;

public class RuntimeConfigMapperPropertiesTest {
    @Test(expected = UnsupportedOperationException.class)
    public void shouldNotAllowConfigToBeSaved() {
        RuntimeConfigMapperProperties mapper = new RuntimeConfigMapperProperties(new Properties());
        mapper.saveConfig(new RuntimeConfigState(0, new RuntimeConfigItem[] {}));
    }

    @Test
    public void shouldKnowGeneralCategory() {
        shouldKnowCategory("general", RuntimeConfig.CATEGORY_PROPERTIES);
    }

    @Test
    public void shouldKnowCategoryClientAccessWhiteList() {
        shouldKnowCategory("clientaccesswhitelist", RuntimeConfig.CATEGORY_CLIENTACCESSWHITELIST);
    }

    @Test
    public void shouldKnowCategoryServiceAccessWhiteList() {
        shouldKnowCategory("serviceaccesswhitelist", RuntimeConfig.CATEGORY_SERVICEACCESSWHITELIST);
    }

    @Test
    public void shouldKnowCategoryPreserverWsa() {
        shouldKnowCategory("servicepreservewsa", RuntimeConfig.CATEGORY_SERVICEPRESERVE_WSA);
    }

    @Test
    public void shouldReadCommaSeparatedValue() {
        RuntimeConfigState configState = makeMapperAndLoadConfig("runtimeconfig.general.commaSep", "val1,val2");

        String[] value = configState.getValue(RuntimeConfig.CATEGORY_PROPERTIES, "commaSep");
        Assert.assertEquals("val1", value[0]);
        Assert.assertEquals("val2", value[1]);
    }

    @Test
    public void shouldReadThreeTupleWithNullElement() {
        RuntimeConfigState configState = makeMapperAndLoadConfig("runtimeconfig.general.commaSep", "val1,,val2");

        String[] value = configState.getValue(RuntimeConfig.CATEGORY_PROPERTIES, "commaSep");
        Assert.assertEquals("val1", value[0]);
        Assert.assertNull(value[1]);
        Assert.assertEquals("val2", value[2]);
    }

    @Test
    public void shouldReadThreeTupleWithElementContainingSpaces() {
        RuntimeConfigState configState = makeMapperAndLoadConfig("runtimeconfig.general.commaSep", "val1,val2 is here");

        String[] value = configState.getValue(RuntimeConfig.CATEGORY_PROPERTIES, "commaSep");
        Assert.assertEquals("val1", value[0]);
        Assert.assertEquals("val2 is here", value[1]);
    }

    @Test
    public void shouldReadSingleValueAndExtractClientAccessWhiteListCategory() {
        RuntimeConfigState configState = makeMapperAndLoadConfig("runtimeconfig.clientaccesswhitelist.fromthelb",
                "9.0.5.,testSalt");

        String[] value = configState.getValue(RuntimeConfig.CATEGORY_CLIENTACCESSWHITELIST, "fromthelb");
        Assert.assertEquals("9.0.5.", value[0]);
        Assert.assertEquals("testSalt", value[1]);
    }

    @Test
    public void shouldComplainAboutUnknownCategory() {
        try {
            makeMapperAndLoadConfig("runtimeconfig.doesnotexist.key", "testVal");
            Assert.fail("Expected ConfigurationException for unknown category");
        } catch (ConfigurationException e) {
            Assert.assertTrue(e.getMessage().contains("doesnotexist"));
            Assert.assertTrue(e.getMessage().contains("runtimeconfig.doesnotexist.key"));
        }
    }

    private void shouldKnowCategory(String shortCategoryName, String longCategoryName) {
        RuntimeConfigState configState = makeMapperAndLoadConfig("runtimeconfig." + shortCategoryName + ".key",
                "testVal");

        Assert.assertEquals("testVal", configState.getValue(longCategoryName, "key")[0]);
    }

    @Test
    public void shouldIgnoreNonRuntimeProperties() {
        Properties properties = new Properties();
        properties.setProperty("static.foo.bar", "baz");
        properties.setProperty("runtimeconfig.general.foobar", "foobar");

        RuntimeConfigMapperProperties runtimeConfigMapperProperties = new RuntimeConfigMapperProperties(properties);
        RuntimeConfigState runtimeConfigState = runtimeConfigMapperProperties.loadLatestConfig();

        String[] strings = runtimeConfigState.getValue(RuntimeConfig.CATEGORY_PROPERTIES, "foobar");
        Assert.assertEquals(1, strings.length);
        Assert.assertEquals("foobar", strings[0]);
    }

    @Test
    public void shouldAcceptSeveralClientAccessWhitelistEntries() {
        Properties properties = new Properties();
        properties.setProperty("runtimeconfig.clientaccesswhitelist.client1", "9.0.5.,testSalt");
        properties.setProperty("runtimeconfig.clientaccesswhitelist.client2", "127.0.0.1,testSalt2");

        RuntimeConfigMapperProperties runtimeConfigMapperProperties = new RuntimeConfigMapperProperties(properties);
        RuntimeConfigState runtimeConfigState = runtimeConfigMapperProperties.loadLatestConfig();

        Map<String, String[]> categoryMap = runtimeConfigState
                .getCategoryMap(RuntimeConfig.CATEGORY_CLIENTACCESSWHITELIST);

        Assert.assertEquals(2, categoryMap.size());

        String[] client1Values = categoryMap.get("client1");
        Assert.assertEquals(2, client1Values.length);
        Assert.assertEquals("9.0.5.", client1Values[0]);
        Assert.assertEquals("testSalt", client1Values[1]);

        String[] client2Values = categoryMap.get("client2");
        Assert.assertEquals(2, client2Values.length);
        Assert.assertEquals("127.0.0.1", client2Values[0]);
        Assert.assertEquals("testSalt2", client2Values[1]);
    }

    @Test
    public void shouldNotBreakWhenNoDots() {
        makeMapperAndLoadConfig("foo", "bar");
    }

    @Test
    public void shouldNotBreakWhenNoDotsInRuntimeConfig() {
        try {
            makeMapperAndLoadConfig("runtimeconfig.foo", "bar");
        } catch (ConfigurationException e) {
            Assert.assertTrue(e.getMessage().contains("Must contain at least three components"));
        }
    }
    
    @Test
    public void shouldAcceptStorageTypeField()
    {
        makeMapperAndLoadConfig("runtimeconfig.storage.type", "properties");
    }
    
    private RuntimeConfigState makeMapperAndLoadConfig(String key, String val) {
        RuntimeConfigMapperProperties mapper = makeMapper(key, val);

        return mapper.loadLatestConfig();
    }

    private RuntimeConfigMapperProperties makeMapper(String key, String val) {
        Properties properties = new Properties();
        properties.setProperty(key, val);

        return new RuntimeConfigMapperProperties(properties);
    }
}
