/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.config;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

import java.util.Map;
import java.util.Properties;

import static org.junit.Assert.*;

public class RuntimeConfigStateTest {

    private static StaticConfigApi config;

    @BeforeClass
    public static void setup() {
        config = StaticConfigImpl.getInstance();
        Properties props = new Properties();
        props.setProperty("sosigw.mode", "test");
        StaticConfigImpl.setInstance(new StaticConfigImpl(props));
    }
    
    @AfterClass
    public static void end() {
        StaticConfigImpl.setInstance(config);
    }
    
    public static void makeClientAccessTestData() {
        RuntimeConfigEditState state = RuntimeConfig.getRuntimeConfigState().getEditableVersion();
        state.addEntry(RuntimeConfig.CATEGORY_CLIENTACCESSWHITELIST, "excat-with-salt", new String[] { "10.0.0.0", "mitSalt" });
        state.addEntry(RuntimeConfig.CATEGORY_CLIENTACCESSWHITELIST, "excat-without-salt", new String[] { "111.0.0.0", null });
        state.addEntry(RuntimeConfig.CATEGORY_CLIENTACCESSWHITELIST, "prefix-without-salt", new String[] { "27.0", null });
        RuntimeConfig.setRuntimeConfigState(state.getResult());
    }

    @Test
    public void test0() {
        RuntimeConfigState state = RuntimeConfig.getRuntimeConfigState();
        assertTrue(state.getCategoryMap("foo.bar").isEmpty());
        
        RuntimeConfigEditState editableVersion = state.getEditableVersion();
        editableVersion.addEntry("foo.bar", "første regel", new String[] { "bar is foo" });
        assertTrue(!editableVersion.getCategoryMap("foo.bar").isEmpty());
        
        RuntimeConfig.setRuntimeConfigState(editableVersion.getResult());
        assertTrue(!RuntimeConfig.getCategoryMap("foo.bar").isEmpty());
    }
    
    @Test
    public void testClientAccess() {    
        makeClientAccessTestData();
        
        assertTrue("10.nosalt", RuntimeConfig.checkClientAccessWhiteList("10.0.0.0", null, false));
        assertTrue("10.salt", RuntimeConfig.checkClientAccessWhiteList("10.0.0.0", "mitSalt", true));

        assertTrue("111.nosalt", RuntimeConfig.checkClientAccessWhiteList("111.0.0.0", null, false));
        assertTrue("111.salt", RuntimeConfig.checkClientAccessWhiteList("111.0.0.0", "ligemeget", true));

        assertFalse("17.nosalt", RuntimeConfig.checkClientAccessWhiteList("17.0.0.0", null, false));
        assertFalse("17.salt", RuntimeConfig.checkClientAccessWhiteList("17.0.0.0", "mitSalt", true));
        assertFalse("10.false-salt", RuntimeConfig.checkClientAccessWhiteList("10.0.0.0", "mitforkerteSalt", true));

        assertTrue("27.nosalt", RuntimeConfig.checkClientAccessWhiteList("27.0.1.2", null, false));
        assertTrue("27.salt", RuntimeConfig.checkClientAccessWhiteList("27.0.1.2", "ligemeget", true));
        assertFalse("27.false-nosalt", RuntimeConfig.checkClientAccessWhiteList("27.1.1.1", null, false));
    }

    @Test
    public void testServiceAccess() {    
        RuntimeConfigEditState state = RuntimeConfig.getRuntimeConfigState().getEditableVersion();
        state.addEntry(RuntimeConfig.CATEGORY_SERVICEACCESSWHITELIST, "regexp 1", new String[] { "^http:/pem.noget.dk.*", "^minAction.*" });
        state.addEntry(RuntimeConfig.CATEGORY_SERVICEACCESSWHITELIST, "exact ", new String[] { "http:/pem.noget.dk.exact", "minAction.exact" });
        RuntimeConfig.setRuntimeConfigState(state.getResult());
        
        assertTrue("service.exact", RuntimeConfig.checkServiceWhiteList("http:/pem.noget.dk", "minAction"));
        assertTrue("service.regexp1", RuntimeConfig.checkServiceWhiteList("http:/pem.noget.dk/et/eller/andet", "minAction?param=foo"));

        assertFalse("service.false1", RuntimeConfig.checkServiceWhiteList("http:/pem.andet.dk/et/eller/andet", "minAction?param=foo"));
        assertFalse("service.false2", RuntimeConfig.checkServiceWhiteList("http:/pem.noget.dk/et/eller/andet", "AndresAction?param=foo"));
    }
    
    @Test
    public void testMetaInfo() {
        RuntimeConfigState runtimeConfigState = RuntimeConfig.getRuntimeConfigState();
        Map<String, String[]> categoryMap = runtimeConfigState.getCategoryMap(RuntimeConfig.CATEGORY_METAINFO);
        assertNotNull(categoryMap);
        assertTrue(RuntimeConfig.CATEGORY_CLIENTACCESSWHITELIST, categoryMap.containsKey(RuntimeConfig.CATEGORY_CLIENTACCESSWHITELIST));
        assertTrue(RuntimeConfig.CATEGORY_SERVICEACCESSWHITELIST, categoryMap.containsKey(RuntimeConfig.CATEGORY_SERVICEACCESSWHITELIST));
        assertEquals(4, categoryMap.get(RuntimeConfig.CATEGORY_SERVICEACCESSWHITELIST).length);
    }
}
