/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.sts;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

public class CircuitBreakerSTSProxyTest {
    private TestProxy goodProxy1;
    private TestProxy badProxy1;
    private CircuitBreakerSTSProxy cb;

    @Before
    public void setup() {
        goodProxy1 = TestProxy.goodInstance("goodProxy1");
        badProxy1 = TestProxy.failingInstance("badProxy1");
        cb = null; // tests instantiate it, let's make sure state is not spilled over from test to test if they forget
    }

    @Test
    public void succesfulStsResponseIsCommunicatedToCallerAndBreakersStaysClosed() {
        cb = new CircuitBreakerSTSProxy(goodProxy1, new STSProxyStaticConfig(5, 30000, 10000));

        final String stsResponse = cb.callSTS(null);
        Assert.assertEquals("response from sts", "goodProxy1", stsResponse);
        Assert.assertTrue("breaker closed", cb.isClosed());
    }

    @Test
    public void urlStringIsTheSameAsTheSTSProxyWeAreProtecting() {
        cb = new CircuitBreakerSTSProxy(goodProxy1, new STSProxyStaticConfig(5, 30000, 10000));

        Assert.assertEquals("url string", goodProxy1.getUrlString(), cb.getUrlString());
    }

    @Test
    public void firstFailureThrowsExceptionToCallerButBreakerStaysClosed() {
        cb = new CircuitBreakerSTSProxy(badProxy1, new STSProxyStaticConfig(5, 30000, 10000));

        try {
            cb.callSTS(null);
            Assert.fail("Expected STSUnreacheException");
        } catch (STSUnreachableException ignore) {
            Assert.assertTrue("cb closed after first exception", cb.isClosed());
        }
    }

    @Test
    public void firstFailureOverThresholdThrowsExceptionToCallerAndBreakerOpens() {
        cb = new CircuitBreakerSTSProxy(badProxy1, new STSProxyStaticConfig(2, 30000, 10000));
        makeOneFailingCall(cb);

        try {
            cb.callSTS(null);
            Assert.fail("Expected STSUnreacheException");
        } catch (STSUnreachableException ignore) {
            Assert.assertFalse("cb open after first exception", cb.isClosed());
        }
    }

    @Test
    public void openBreakerStaysOpenWhenInsufficientTimeHasElapsed() {
        cb = new CircuitBreakerSTSProxy(badProxy1, new STSProxyStaticConfig(1, Integer.MAX_VALUE, 10000)); // make time threshold "high enough"
        makeOneFailingCall(cb); // opens the breaker

        try {
            cb.callSTS(null);
            Assert.fail("Expected STSUnreacheException");
        } catch (STSUnreachableException ignore) {
            Assert.assertFalse("cb open", cb.isClosed());
        }
    }

    @Test
    public void openBreakerEndsupClosedWhenSufficientTimeHasElapsedAndCallSucceeds() {
        cb = new CircuitBreakerSTSProxy(badProxy1, new STSProxyStaticConfig(1, 0, 10000)); // make time threshold "low enough"
        makeOneFailingCall(cb); // opens the breaker

        badProxy1.turnGood();

        cb.callSTS(null);
        Assert.assertTrue("cb closed", cb.isClosed());
    }

    @Test
    public void whenBreakerHasClosedAfterBeingOpenOneFailureIsNotEnoughToMakeItOpenAgain() {
        cb = new CircuitBreakerSTSProxy(badProxy1, new STSProxyStaticConfig(5, 0, 10000));
        makeOneFailingCall(cb); // opens the breaker

        badProxy1.turnGood();

        cb.callSTS(null); // closes the breaker from the half open state we assume it was in

        badProxy1.turnBad();
        makeOneFailingCall(cb);

        Assert.assertTrue("cb closed", cb.isClosed());
    }

    @Test
    public void openBreakerEndsupOpenAfterOnlyOneFailingCallWhenSufficientTimeHasElapsedAndCallFails() {
        cb = new CircuitBreakerSTSProxy(badProxy1, new STSProxyStaticConfig(1, 0, 10000)); // make time threshold "low enough"
        makeOneFailingCall(cb); // opens the breaker

        // let the proxy stay bad

        try {
            cb.callSTS(null);
            Assert.fail("Expected STSUnreacheException");
        } catch (STSUnreachableException ignore) {
            Assert.assertFalse("cb open", cb.isClosed());
        }
    }

    @Test
    /**
     * Regression test: this was not always the case
     */
    public void oneCircuitBreakerOpeningDoesNotOpenOtherBreakers() {
        CircuitBreakerSTSProxy shouldStayClosedCb = new CircuitBreakerSTSProxy(badProxy1, new STSProxyStaticConfig(1, Integer.MAX_VALUE, 10000));

        cb = new CircuitBreakerSTSProxy(badProxy1, new STSProxyStaticConfig(1, Integer.MAX_VALUE, 10000)); // make time threshold "high enough"
        makeOneFailingCall(cb); // opens the breaker

        try {
            cb.callSTS(null);
            Assert.fail("Expected STSUnreacheException");
        } catch (STSUnreachableException ignore) {
            Assert.assertFalse("cb open", cb.isClosed());
        }

        Assert.assertTrue("second cb closed", shouldStayClosedCb.isClosed());
    }

    @Test
    public void countsCorrectlyANumberOfSuccesfulCalls() {
        cb = new CircuitBreakerSTSProxy(goodProxy1, new STSProxyStaticConfig(5, Integer.MAX_VALUE, 10000)); // make time threshold "high enough"
        cb.callSTS(null);
        cb.callSTS(null);
        cb.callSTS(null);

        Assert.assertEquals(3, cb.getSuccessfulCalls());
        Assert.assertEquals(cb.getSuccessfulCalls(), cb.getAttemptedCalls());
    }


    @Test
    public void countsCorrectlyANumberOfFailingCalls() {
        cb = new CircuitBreakerSTSProxy(badProxy1, new STSProxyStaticConfig(5, Integer.MAX_VALUE, 10000)); // make time threshold "high enough"
        makeOneFailingCall(cb);
        makeOneFailingCall(cb);
        makeOneFailingCall(cb);

        Assert.assertEquals(0, cb.getSuccessfulCalls());
        Assert.assertEquals(3, cb.getAttemptedCalls());
    }

    @Test
    public void doesNotCountCallsMadeWhenInOpenState() {
        cb = new CircuitBreakerSTSProxy(badProxy1, new STSProxyStaticConfig(2, Integer.MAX_VALUE, 10000)); // make time threshold "high enough"
        makeOneFailingCall(cb);
        makeOneFailingCall(cb);
        makeOneFailingCall(cb);

        Assert.assertEquals(2, cb.getAttemptedCalls());
    }

    @Test
    public void toStringContainsTheToStringFromTheSTSWeAreProtecting() {
        cb = new CircuitBreakerSTSProxy(goodProxy1, new STSProxyStaticConfig(2, 0, 10000));
        Assert.assertTrue(cb.toString().contains(goodProxy1.toString()));
    }

    private void makeOneFailingCall(CircuitBreakerSTSProxy cb) {
        try {
            cb.callSTS(null);
            Assert.fail("Expected STSUnreacheException");
        } catch (STSUnreachableException ignored) {
            // expected
        }
    }
}
