/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.fail;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Random;

import org.apache.james.mime4j.MimeException;
import org.junit.Test;

public class HeaderReadingInputStreamTest {

	private static final String BASIC_SOAP_1_HEADER_PART 
		= "<?xml version=\"1.0\"?>"
		+ " <soap:Envelope xmlns:soap=\"http://www.w3.org/2001/12/soap-envelope\""
		+ " soap:encodingStyle=\"http://www.w3.org/2001/12/soap-encoding\">"
		+ " <soap:Header>"
		+ " </soap:Header>";
	
	private static final String BASIC_SOAP_1_BODY_PART 
		= "<soap:Body>"
		+ "</soap:Body>"
		+ "</soap:Envelope>";
	
	private static final String BASIC_SOAP_MAX_WHITESPACE_HEADER_PART 
	= "<?xml version=\"1.0\"?>"
	+ " < soap:Envelope xmlns:soap = \"http://www.w3.org/2001/12/soap-envelope\" "
	+ " soap:encodingStyle = \"http://www.w3.org/2001/12/soap-encoding\" > "
	+ " < soap:Header > "
	+ " < / soap:Header >";

	private static final String BASIC_SOAP_CDATA_HEADER_PART 
	= "<?xml version=\"1.0\"?>"
	+ " < soap:Envelope xmlns:soap = \"http://www.w3.org/2001/12/soap-envelope\" "
	+ " soap:encodingStyle = \"http://www.w3.org/2001/12/soap-encoding\" > "
	+ " < soap:Header > "
	+ " <![CDATA["
	+ BASIC_SOAP_MAX_WHITESPACE_HEADER_PART // Nest a complete header inside CDATA. Ugly stuff...
	+ " ]]> " 
	+ "< / soap:Header >";

	private static final String BASIC_SOAP_MTOM_HEADER
	= "<?xml version=\"1.0\" encoding=\"UTF-8\"?><S:Envelope xmlns:S=\"http://schemas.xmlsoap.org/soap/envelope/\">\n"
	+ "\t<S:Header>\n"
	+ "\t</S:Header>";

	private static final String BASIC_SOAP_MTOM_CONTENT_TYPE=
			"Content-Type: multipart/related; type=\"application/xop+xml\"; boundary=\"uuid:6fb8dea2-01d1-40d0-8906-15450983d322\"; start=\"<root.message@cxf.apache.org>\"; start-info=\"text/xml\"";
	private static final String BASIC_SOAP_MTOM_REQ
	= "--uuid:6fb8dea2-01d1-40d0-8906-15450983d322\n"
	+ "Content-Type: application/xop+xml; charset=UTF-8; type=\"text/xml\";\n"
	+ "Content-Transfer-Encoding: binary\n"
	+ "Content-ID: <root.message@cxf.apache.org>\n"
	+ "\n"
	+ "<?xml version=\"1.0\" encoding=\"UTF-8\"?><S:Envelope xmlns:S=\"http://schemas.xmlsoap.org/soap/envelope/\">\n"
	+ "\t<S:Header>\n"
	+ "\t</S:Header>\n"
	+ "</S:Envelope>\n"
	+ "--uuid:6fb8dea2-01d1-40d0-8906-15450983d322\n"
	+ "Content-Type: application/octet-stream\n"
	+ "Content-Transfer-Encoding: binary\n"
	+ "Content-ID: <9a29dc17-eac4-4b58-b4ee-d6ead68879a8-1@cxf.apache.org>\n"
	+ "\n"
	+ "asdfasdfsadfsdfsadfasdf\n"
	+ "--uuid:6fb8dea2-01d1-40d0-8906-15450983d322--\n";

	private static final String MISSING_XML_TAG_MTOM_HEADER =
			"<soap:Envelope xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" +
					"\t<soap:Header>\n" +
					"\t</soap:Header>";

	private static final String MISSING_XML_TAG_MTOM_CONTENT_TYPE =
			"Content-Type: multipart/related; type=\"application/xop+xml\"; boundary=\"uuid:ea0668b8-3197-48fd-9bdb-adbf9d8dbc58\"; start=\"<root.message@cxf.apache.org>\"; start-info=\"text/xml\"";
	private static final String MISSING_XML_TAG_MTOM_REQ
		= "--uuid:ea0668b8-3197-48fd-9bdb-adbf9d8dbc58\n" +
		"Content-Type: application/xop+xml; charset=UTF-8; type=\"text/xml\"\n" +
		"Content-Transfer-Encoding: binary\n" +
		"Content-ID: <root.message@cxf.apache.org>\n" +
		"\n" +
		"<soap:Envelope xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" +
		"\t<soap:Header>\n" +
		"\t</soap:Header>\n" +
		"\t<soap:Body>\n" +
		"\t</soap:Body>\n" +
		"</soap:Envelope>\n" +
		"--uuid:ea0668b8-3197-48fd-9bdb-adbf9d8dbc58\n" +
		"Content-Type: application/octet-stream\n" +
		"Content-Transfer-Encoding: binary\n" +
		"Content-ID: <ed9d3154-3b3d-45ca-9465-d094a3dfdd58-1@cxf.apache.org>\n" +
		"Content-Disposition: attachment;name=\"fusesource_logo.png\"\n";

	private static final String BASIC_SOAP_1 = BASIC_SOAP_1_HEADER_PART + BASIC_SOAP_1_BODY_PART;
	private static final String BASIC_SOAP_WHITESPACE = BASIC_SOAP_MAX_WHITESPACE_HEADER_PART + BASIC_SOAP_1_BODY_PART;
	private static final String BASIC_SOAP_CDATA = BASIC_SOAP_CDATA_HEADER_PART + BASIC_SOAP_1_BODY_PART;

	@Test
	public void testMtomRequest() throws IOException, MimeException {
		String testString = BASIC_SOAP_MTOM_REQ;
		byte[] bytes = testString.getBytes("UTF-8");
		InputStream is = new ByteArrayInputStream(bytes);
		HeaderReadingInputStream hris = new HeaderReadingInputStream(is);
		StringBuffer soapHeader = hris.getSoapHeader();
		InputStream remainderStream = hris.getRemainderStream();
		assertNotNull(remainderStream);
		assertEquals(BASIC_SOAP_MTOM_HEADER, soapHeader.toString());
	}

	@Test
	public void testMtomMissingXmlTag() throws IOException, MimeException {
		String testString = MISSING_XML_TAG_MTOM_REQ;
		byte[] bytes = testString.getBytes("UTF-8");
		InputStream is = new ByteArrayInputStream(bytes);
		HeaderReadingInputStream hris = new HeaderReadingInputStream(is);
		StringBuffer soapHeader = hris.getSoapHeader();
		InputStream remainderStream = hris.getRemainderStream();
		assertNotNull(remainderStream);
		assertEquals(MISSING_XML_TAG_MTOM_HEADER, soapHeader.toString());
	}

	@Test
	public void testBasic1() throws IOException, MimeException {
		String testString = BASIC_SOAP_1;
		byte[] bytes = testString.getBytes("UTF-8");
		InputStream is = new ByteArrayInputStream(bytes);
		HeaderReadingInputStream hris = new HeaderReadingInputStream(is);
		StringBuffer soapHeader = hris.getSoapHeader();
		InputStream remainderStream = hris.getRemainderStream();
		assertNotNull(remainderStream);
		assertEquals(BASIC_SOAP_1_HEADER_PART, soapHeader.toString());
	}
	
    @Test
	public void testBasicWhitespace() throws IOException, MimeException {
		String testString = BASIC_SOAP_WHITESPACE;
		byte[] bytes = testString.getBytes("UTF-8");
		InputStream is = new ByteArrayInputStream(bytes);
		HeaderReadingInputStream hris = new HeaderReadingInputStream(is);
		StringBuffer soapHeader = hris.getSoapHeader();
		InputStream remainderStream = hris.getRemainderStream();
		assertNotNull(remainderStream);
		assertEquals(BASIC_SOAP_MAX_WHITESPACE_HEADER_PART, soapHeader.toString());
	}

    @Test
	public void testBasicCdata() throws IOException, MimeException {
		String testString = BASIC_SOAP_CDATA;
		byte[] bytes = testString.getBytes("UTF-8");
		InputStream is = new ByteArrayInputStream(bytes);
		HeaderReadingInputStream hris = new HeaderReadingInputStream(is);
		StringBuffer soapHeader = hris.getSoapHeader();
		InputStream remainderStream = hris.getRemainderStream();
		assertNotNull(remainderStream);
		assertEquals(BASIC_SOAP_CDATA_HEADER_PART, soapHeader.toString());
	}

    @Test
	public void testOverFlow() throws IOException {
		InputStream inputStream = new InputStream() {

			public int read() throws IOException {
				return ' ';
			}
		};
		try {
	        HeaderReadingInputStream hris = new HeaderReadingInputStream(inputStream);
	        assertNotNull(hris);
		} catch (IOException e) {
			// Fine. Detected too much data in header.
			return;
		}
		fail();
	}

    @Test
	public void testRandomBytes() throws IOException, MimeException {
		final Random rand = new Random();
		InputStream inputStream = new InputStream() {

			public int read() throws IOException {
				return rand.nextInt(256);
			}
		};
		try {
	        HeaderReadingInputStream hris = new HeaderReadingInputStream(inputStream);
	        assertNotNull(hris);
		} catch (IOException e) {
			// Fine. Detected non-utf8 stuff.
			return;
		}
		fail();
	}
}
