/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.util;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.cluster.IDCardClusterCacheImpl;
import com.trifork.sosigw.config.*;
import org.apache.catalina.core.DummyRequest;
import org.junit.Before;
import org.junit.Test;

import java.util.Arrays;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;


/**
 * @author $LastChangedBy: kpn $ $LastChangedDate: 2013-09-12 11:08:05 +0200 (Tor, 12 Sep 2013) $
 * @version $Revision: 1145 $
 */
public class UtilTest {

    private static final String PARTITION_HTTP_HEADER = "SOSIGW_PARTITION";

    private DummyRequest httpServletRequest;
    private Map<String, List<String>> headers = new HashMap<String, List<String>>();

    @Before
    public void before() {
        StaticConfigImpl.initSettings();
        httpServletRequest = new DummyRequest() {
            
            @Override
            public Enumeration<?> getHeaders(String name) {
                if (headers.containsKey(name)) {
                    return Collections.enumeration(headers.get(name));
                } else {
                    return Collections.enumeration(Collections.emptyList());
                }
            }
        };
        Util.setActiveHttpRequest(httpServletRequest);
        Factory.setTheIDCardCache(new IDCardClusterCacheImpl(null));
    }

    @Test
    public void testComputeNameIDNoPartitioning() {
        assertEquals("foo", Util.computeCacheKey("foo"));
        assertEquals("foo#bar", Util.computeCacheKey("foo#bar"));
    }

    @Test
    public void testComputeNameIDWithPartitioningEnabled() {
        try {

            RuntimeConfigEditState editableVersion = RuntimeConfig.getRuntimeConfigState().getEditableVersion();
            editableVersion.addEntry(RuntimeConfig.CATEGORY_PROPERTIES, RuntimeConfig.KEY_CACHE_PARTITIONING_ENABLED, new String[]{"true"});
            RuntimeConfigState result = editableVersion.getResult();
            RuntimeConfig.setRuntimeConfigState(result);

            try {
                Util.computeCacheKey("foo");
                fail("Expected ConfigurationException");
            } catch (ConfigurationException e) {
                assertEquals("Property 'cache-partitioning.http-header.name' is not defined in runtime configuration", e.getMessage());
            }

            editableVersion = RuntimeConfig.getRuntimeConfigState().getEditableVersion();
            editableVersion.addEntry(RuntimeConfig.CATEGORY_PROPERTIES, RuntimeConfig.KEY_CACHE_PARTITIONING_ENABLED, new String[]{"true"});
            editableVersion.addEntry(RuntimeConfig.CATEGORY_PROPERTIES, RuntimeConfig.KEY_CACHE_PARTITIONING_HTTP_HEADER_NAME, new String[]{PARTITION_HTTP_HEADER});
            result = editableVersion.getResult();
            RuntimeConfig.setRuntimeConfigState(result);

            try {
                Util.computeCacheKey("foo");
                fail("Expected HttpHeaderException");
            } catch (HttpHeaderException e) {
                assertEquals("HTTP header 'SOSIGW_PARTITION' is missing in request", e.getMessage());
            }

            headers.put(PARTITION_HTTP_HEADER, Collections.singletonList((String)null));
            try {
                Util.computeCacheKey("foo");
                fail("Expected HttpHeaderException");
            } catch (HttpHeaderException e) {
                assertEquals("HTTP header 'SOSIGW_PARTITION' is missing a value in request", e.getMessage());
            }

            headers.put(PARTITION_HTTP_HEADER, Arrays.asList("OnePartition", "AnotherPartition"));
            try {
                Util.computeCacheKey("foo");
                fail("Expected HttpHeaderException");
            } catch (HttpHeaderException e) {
                assertEquals("HTTP header 'SOSIGW_PARTITION' present multiple times in request", e.getMessage());
            }

            headers.put(PARTITION_HTTP_HEADER, Collections.singletonList(""));
            try {
                Util.computeCacheKey("foo");
                fail("Expected HttpHeaderException");
            } catch (HttpHeaderException e) {
                assertEquals("HTTP header 'SOSIGW_PARTITION' must not be empty", e.getMessage());
            }

            headers.put(PARTITION_HTTP_HEADER, Collections.singletonList("SomePartitionID#thatIsIllegal"));
            try {
                Util.computeCacheKey("foo");
                fail("Expected HttpHeaderException");
            } catch (HttpHeaderException e) {
                assertEquals("HTTP header 'SOSIGW_PARTITION' must not contain '#' (value was SomePartitionID#thatIsIllegal)", e.getMessage());
            }

            headers.put(PARTITION_HTTP_HEADER, Collections.singletonList("SomePartitionID"));
            assertEquals("SomePartitionID#foo", Util.computeCacheKey("foo"));
            
        } finally {
            RuntimeConfigEditState editableVersion = RuntimeConfig.getRuntimeConfigState().getEditableVersion();
            editableVersion.addEntry(RuntimeConfig.CATEGORY_PROPERTIES, RuntimeConfig.KEY_CACHE_PARTITIONING_ENABLED, new String[]{"false"});
            RuntimeConfigState result = editableVersion.getResult();
            RuntimeConfig.setRuntimeConfigState(result);
        }

    }
}
