/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.ws;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.cluster.IDCardClusterCacheImpl;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.RuntimeConfigEditState;
import com.trifork.sosigw.config.RuntimeConfigState;
import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.testclient.ServiceStubClient;
import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.sosi.gw._2007_09.DGWSFault;
import dk.sosi.gw._2007_09.RequestIdCardDigestForSigningResponse;
import dk.sosi.gw._2007_09.SosiGWFacade;
import junit.framework.Assert;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;

import javax.xml.ws.BindingProvider;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.soap.SOAPFaultException;
import java.util.Collections;

import static junit.framework.Assert.assertNull;
import static junit.framework.Assert.assertTrue;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

/**
 * @author $LastChangedBy: kpn $ $LastChangedDate: 2014-06-10 11:11:52 +0200 (Tir, 10 Jun 2014) $
 * @version $Revision: 1690 $
 */
public class PartitionedCacheTest extends AbstractIntegrationTests {

    private static final String PARTITION_HTTP_HEADER = "SOSIGW_PARTITION";
    private static final String PARTITION = "MyPartition";
    private static final String NAME_ID = "MitBrugerNavnSomErMereEnd10tegnLangt";
    private RuntimeConfigState previous;

    @Before
    public void enablePartitioningInServer() {
        RuntimeConfigEditState editableVersion = RuntimeConfig.getRuntimeConfigState().getEditableVersion();
        previous = editableVersion.getResult();
        editableVersion.addEntry(RuntimeConfig.CATEGORY_PROPERTIES, RuntimeConfig.KEY_CACHE_PARTITIONING_ENABLED, new String[] { "true"} );
        editableVersion.addEntry(RuntimeConfig.CATEGORY_PROPERTIES, RuntimeConfig.KEY_CACHE_PARTITIONING_HTTP_HEADER_NAME, new String[] { PARTITION_HTTP_HEADER} );
        editableVersion.addEntry(RuntimeConfig.CATEGORY_PROPERTIES, RuntimeConfig.KEY_BROWSERSIGNINGURL + "." + PARTITION, new String[] { "http://partitionhost:8080/sosigw/browsersign/invoke" } );
        RuntimeConfigState result = editableVersion.getResult();
        RuntimeConfig.setRuntimeConfigState(result);
    }
    
    @Before
    public void cleanCache() {
        Factory.setTheIDCardCache(new IDCardClusterCacheImpl(null));        
    }

    @After
    public void disablePartitioningInServer() {
        RuntimeConfig.setRuntimeConfigState(previous);
    }

    @Override
    protected SosiGWFacade getFacade() {
        SosiGWFacade facade = super.getFacade();
        BindingProvider bindingProvider = (BindingProvider) facade;
        bindingProvider.getRequestContext().put(MessageContext.HTTP_REQUEST_HEADERS, Collections.singletonMap(PARTITION_HTTP_HEADER, Collections.singletonList(PARTITION)));
        return facade;
    }

    @Test(timeout = 60000)
    public void testMisingPartitionHeaderOnFacadeCall() throws Exception {
        SosiGWFacade sosiGWFacade = super.getFacade();

        try {
            Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
            sosiGWFacade.requestIdCardDigestForSigning(securityHeader, "");
        } catch (DGWSFault e) {
            assertEquals("HTTP header 'SOSIGW_PARTITION' is missing in request", e.getMessage());
        }

        assertNull(Factory.getIDCardCache().findIDCardOrNull(NAME_ID));
        assertNull(Factory.getIDCardCache().findIDCardOrNull(PARTITION + "#" + NAME_ID));
    }

    @Test(timeout = 60000)
    public void testPartitionKeysNotUsedInCacheAfterRequest() throws Exception {
        disablePartitioningInServer();
        assertNull(Factory.getIDCardCache().findIDCardOrNull(NAME_ID));
        assertNull(Factory.getIDCardCache().findIDCardOrNull(PARTITION + "#" + NAME_ID));

        SosiGWFacade sosiGWFacade = getFacade();
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        sosiGWFacade.requestIdCardDigestForSigning(securityHeader, "");

        assertNotNull(Factory.getIDCardCache().findIDCardOrNull(NAME_ID));
        assertNull(Factory.getIDCardCache().findIDCardOrNull(PARTITION + "#" + NAME_ID));
    }

    @Test(timeout = 60000)
    public void testPartitionKeysUsedInCacheAfterRequest() throws Exception {
        assertNull(Factory.getIDCardCache().findIDCardOrNull(NAME_ID));
        assertNull(Factory.getIDCardCache().findIDCardOrNull(PARTITION + "#" + NAME_ID));

        SosiGWFacade sosiGWFacade = getFacade();
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        sosiGWFacade.requestIdCardDigestForSigning(securityHeader, "");

        assertNull(Factory.getIDCardCache().findIDCardOrNull(NAME_ID));
        assertNotNull(Factory.getIDCardCache().findIDCardOrNull(PARTITION + "#" + NAME_ID));
    }
    
    @Test(timeout = 60000)
    public void testPartitionKeyInBrowserUrl() throws Exception {
        SosiGWFacade sosiGWFacade = getFacade();
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        RequestIdCardDigestForSigningResponse response = sosiGWFacade.requestIdCardDigestForSigning(securityHeader, "");

        assertNotNull(response.getBrowserUrl());
        assertTrue(response.getBrowserUrl().contains(PARTITION + "%23" + NAME_ID)); // URL escaped
    }
    
    @Test(timeout = 60000)
    public void testPartitionHostInBrowserUrl() throws Exception {
        SosiGWFacade sosiGWFacade = getFacade();
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        RequestIdCardDigestForSigningResponse response = sosiGWFacade.requestIdCardDigestForSigning(securityHeader, "");

        assertNotNull(response.getBrowserUrl());
        assertTrue(response.getBrowserUrl().startsWith("http://partitionhost"));
    }
    
    @Test(timeout = 60000)
    public void testDefaultHostInBrowserUrl() throws Exception {
        RuntimeConfigEditState editableVersion = RuntimeConfig.getRuntimeConfigState().getEditableVersion();
        editableVersion.addEntry(RuntimeConfig.CATEGORY_PROPERTIES, RuntimeConfig.KEY_BROWSERSIGNINGURL + "." + PARTITION, new String[] { "" } );
        RuntimeConfig.setRuntimeConfigState(editableVersion.getResult());
        
        SosiGWFacade sosiGWFacade = getFacade();
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        RequestIdCardDigestForSigningResponse response = sosiGWFacade.requestIdCardDigestForSigning(securityHeader, "");

        assertNotNull(response.getBrowserUrl());
        assertTrue(response.getBrowserUrl().startsWith("http://localhost"));
    }
    
    @Test(timeout = 60000)
    public void testPartitionKeysUsedInCacheAfterSigning() throws Exception {
        assertNull(Factory.getIDCardCache().findIDCardOrNull(NAME_ID));
        assertNull(Factory.getIDCardCache().findIDCardOrNull(PARTITION + "#" + NAME_ID));
        createSignedIdCard();
        assertNull(Factory.getIDCardCache().findIDCardOrNull(NAME_ID));
        assertNotNull(Factory.getIDCardCache().findIDCardOrNull(PARTITION + "#" + NAME_ID));
    }
    
    @Test(timeout = 60000)
    public void testMisingPartitionHeaderOnProxyCall() throws Exception {
        createSignedIdCard();
        
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        int responsePayloadSize = 10;
        String requestPayload = "small payload";
        ServiceStubClient client = new ServiceStubClient("localhost:8586");
        Header medcomHeader = null;
        try {
            client.doInvoke(medcomHeader, securityHeader, requestPayload, responsePayloadSize, 0, SERVER_URL + "/proxy/soap-request", false);
        } catch (SOAPFaultException e) {
            assertEquals("HTTP header 'SOSIGW_PARTITION' is missing in request", e.getMessage());
        }
    }
    
    @Test(timeout = 60000)
    public void testIncorrectPartitionHeaderOnProxyCall() throws Exception {
        createSignedIdCard();
        
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        ServiceStubClient client = new ServiceStubClient("localhost:8586");
        client.addHttpHeader(PARTITION_HTTP_HEADER, "Incorrect" + PARTITION);
        
        try {
            client.doInvoke(null, securityHeader, "", 0, 0, SERVER_URL + "/proxy/soap-request", false);
        } catch (SOAPFaultException e) {
            assertEquals("Login required", e.getMessage());
        }
    }
    
    @Test(timeout = 60000)
    public void testCorrectPartitionHeaderOnProxyCall() throws Exception {
        createSignedIdCard();
        
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        ServiceStubClient client = new ServiceStubClient("localhost:8586");
        client.addHttpHeader(PARTITION_HTTP_HEADER, PARTITION);
        
        String response = client.doInvoke(null, securityHeader, "", 0, 0, SERVER_URL + "/proxy/soap-request", false);
        assertEquals("Incorrect response from service stub", "#", response);
    }

    @Test(timeout = 60000)
    public void testCorrectPartitionHeaderAfterLogoutOnProxyCall() throws Exception {
        createSignedIdCard();
        
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        ServiceStubClient client = new ServiceStubClient("localhost:8586");
        client.addHttpHeader(PARTITION_HTTP_HEADER, PARTITION);

        String response = getFacade().logoutWithResponse(securityHeader, null);
        Assert.assertEquals("ok", response);

        try {
            client.doInvoke(null, securityHeader, "", 0, 0, SERVER_URL + "/proxy/soap-request", false);
        } catch (SOAPFaultException e) {
            assertEquals("Login required", e.getMessage());
        }
    }
}
