/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.ws;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.Constants;
import com.trifork.sosigw.cluster.IDCardClusterCacheImpl;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.RuntimeConfigEditState;
import com.trifork.sosigw.config.RuntimeConfigState;
import com.trifork.sosigw.testclient.ServiceStubClient;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.w3._2000._09.xmldsig_.*;

import static junit.framework.Assert.assertTrue;
import static org.junit.Assert.assertEquals;

/**
 * @author $LastChangedBy: jso $ $LastChangedDate: 2015-12-10 14:30:38 +0100 (Tor, 10 Dec 2015) $
 * @version $Revision: 2036 $
 */
public class PassThroughProxyTest extends AbstractIntegrationTests {

    private RuntimeConfigState previous;

    @Before
    public void enablePassThroughOnSignatureInServer() {
        RuntimeConfigEditState editableVersion = RuntimeConfig.getRuntimeConfigState().getEditableVersion();
        previous = editableVersion.getResult();
        editableVersion.addEntry(RuntimeConfig.CATEGORY_PROPERTIES, RuntimeConfig.KEY_PASS_THROUGH_ON_SIGNATURE,
                new String[] { "true" });
        RuntimeConfigState result = editableVersion.getResult();
        RuntimeConfig.setRuntimeConfigState(result);
    }

    @Before
    public void cleanCache() {
        Factory.setTheIDCardCache(new IDCardClusterCacheImpl(null));
    }

    @After
    public void disablePassThroughOnSignatureInServer() {
        RuntimeConfig.setRuntimeConfigState(previous);
    }

    @Test
    public void testPassThroughOnRequest() throws Exception {
        createSignedIdCard();

        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        ServiceStubClient client = new ServiceStubClient("localhost:8586", true, true, true, true, true);

        String response = client.doInvoke(null, securityHeader, "", 0, 0, SERVER_URL + "/proxy/soap-request", false);
        assertEquals("Incorrect response from service stub", "#", response);
        assertTrue("ID card was exchanged", lastRequest.toString().indexOf("Issuer>sosigwtester<") != -1);
        assertTrue("PassThrough header was not removed", lastRequest.toString().indexOf(Constants.PASS_THROUGH) == -1);
        assertTrue("MaxAge header was not removed", lastRequest.toString().indexOf(Constants.MAX_AGE) == -1);
    }

    @Test
    public void testNoPassThroughWithSignatureElementsWhenNotEnabled() throws Exception {
        disablePassThroughOnSignatureInServer();
        createSignedIdCard();

        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        addSignatureElements(securityHeader);

        ServiceStubClient client = new ServiceStubClient("localhost:8586", true, true, true, false, false);

        String response = client.doInvoke(null, securityHeader, "", 0, 0, SERVER_URL + "/proxy/soap-request", false);
        assertEquals("Incorrect response from service stub", "#", response);
        assertTrue("ID card was not exchanged", lastRequest.toString().indexOf("Issuer>TEST1-NSP-STS<") != -1);
    }

    @Test
    public void testNoPassThroughWithoutSignature() throws Exception {
        createSignedIdCard();
        
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        
        ServiceStubClient client = new ServiceStubClient("localhost:8586", true, true, true, false, false);
        
        String response = client.doInvoke(null, securityHeader, "", 0, 0, SERVER_URL + "/proxy/soap-request", false);
        assertEquals("Incorrect response from service stub", "#", response);
        assertTrue("ID card was not exchanged", lastRequest.toString().indexOf("Issuer>TEST1-NSP-STS<") != -1);
    }

    @Test
    public void testNoPassThroughWithoutDigestValue() throws Exception {
        createSignedIdCard();
        
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        addSignatureElements(securityHeader);
        securityHeader.getAssertion().getSignature().setSignedInfo(null);
        
        ServiceStubClient client = new ServiceStubClient("localhost:8586", true, true, true, false, false);
        
        String response = client.doInvoke(null, securityHeader, "", 0, 0, SERVER_URL + "/proxy/soap-request", false);
        assertEquals("Incorrect response from service stub", "#", response);
        assertTrue("ID card was not exchanged", lastRequest.toString().indexOf("Issuer>TEST1-NSP-STS<") != -1);
    }

    @Test
    public void testNoPassThroughWithoutSignatureValue() throws Exception {
        createSignedIdCard();
        
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        addSignatureElements(securityHeader);
        securityHeader.getAssertion().getSignature().setSignatureValue(null);
        
        ServiceStubClient client = new ServiceStubClient("localhost:8586", true, true, true, false, false);
        
        String response = client.doInvoke(null, securityHeader, "", 0, 0, SERVER_URL + "/proxy/soap-request", false);
        assertEquals("Incorrect response from service stub", "#", response);
        assertTrue("ID card was not exchanged", lastRequest.toString().indexOf("Issuer>TEST1-NSP-STS<") != -1);
    }

    @Test
    public void testNoPassThroughWithoutX509Certificate() throws Exception {
        createSignedIdCard();
        
        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        addSignatureElements(securityHeader);
        securityHeader.getAssertion().getSignature().setKeyInfo(null);
        
        ServiceStubClient client = new ServiceStubClient("localhost:8586", true, true, true, false, false);
        
        String response = client.doInvoke(null, securityHeader, "", 0, 0, SERVER_URL + "/proxy/soap-request", false);
        assertEquals("Incorrect response from service stub", "#", response);
        assertTrue("ID card was not exchanged", lastRequest.toString().indexOf("Issuer>TEST1-NSP-STS<") != -1);
    }
    
    @Test
    public void testPassThroughWithAllSignatureElements() throws Exception {
        createSignedIdCard();

        Security securityHeader = IdCardTestUtils.makeIDCardInSecurity();
        addSignatureElements(securityHeader);

        ServiceStubClient client = new ServiceStubClient("localhost:8586", true, true, true, false, false);

        String response = client.doInvoke(null, securityHeader, "", 0, 0, SERVER_URL + "/proxy/soap-request", false);
        assertEquals("Incorrect response from service stub", "#", response);
        assertTrue("ID card was exchanged", lastRequest.toString().indexOf("Issuer>sosigwtester<") != -1);
    }

    private void addSignatureElements(Security securityHeader) {
        Signature signature = new Signature();
        SignedInfo signedInfo = new SignedInfo();
        Reference reference = new Reference();
        KeyInfo keyInfo = new KeyInfo();
        X509Data x509Data = new X509Data();

        securityHeader.getAssertion().setSignature(signature);
        signature.setSignedInfo(signedInfo);
        signedInfo.getReference().add(reference);
        reference.setDigestValue(new byte[] { 1, 2, 3, 4 });
        signature.setSignatureValue(new byte[] { 5, 6, 7, 8 });
        signature.setKeyInfo(keyInfo);
        keyInfo.setX509Data(x509Data);
        x509Data.setX509Certificate(new byte[] { 1, 2, 3, 4, 5, 6, 7, 8 });
    }

}
